#include <algorithm>
#include <cstdint>
#include <iomanip>
#include <iostream>
#include <string>
#include <sstream>
#include "Serial.h"

#ifndef HV_CAEN_N1471_H
#define HV_CAEN_N1471_H
class HV_CAEN_N1471
{
  public:
    //Class and communication setup
    HV_CAEN_N1471(const std::string nPort="caen_hv", const int nBaud=9600, const double timeout=HV_CAEN_N1471::TIMEOUT);//Constructor: serial port name, Baud rate, timeout
    ~HV_CAEN_N1471(){delete communicator;};//Destructor
    bool ResetCommunication(const std::string nPort="caen_hv", const int nBaud=9600, const double timeout=HV_CAEN_N1471::TIMEOUT);
        //Reset the communication parameters: serial port, transmission rate, timeout
    //-------------Channel related monitoring commands: contains module address, and channel numbers (0..3: each, 4:all) as inputs. In the case of error it returns with false and error information is written to standard error output otherwise as true--------------------
    bool getVoltageSet(const int module_address, std::vector<double> &voltages, const int channel_no=4){return chvalue<double>(module_address,"VSET",voltages,channel_no);};
        //Get the value of set voltage
    bool getAboutVoltageSet(const int module_address, std::vector<double> &minvoltages, std::vector<double> &maxvoltages, std::vector<int> &dec_dig, const int channel_no=4) {return chabout(module_address,"V",minvoltages,maxvoltages,dec_dig,channel_no,""); };
        //Provides information about minimal, maximal voltages able to set and decimal digits.
    bool getVoltageActual(const int module_address, std::vector<double> &voltages, const int channel_no=4){return chvalue<double>(module_address,"VMON",voltages,channel_no);};
        //Get the actual value of voltage
    bool getCurrentSet(const int module_address, std::vector<double> &micamps, const int channel_no=4){return chvalue<double>(module_address,"ISET",micamps,channel_no);};
        //Get the value of set current in microampers
    bool getAboutCurrentSet(const int module_address, std::vector<double> &minmicamps, std::vector<double> &maxmicamps, std::vector<int> &dec_dig, const int channel_no=4){return chabout(module_address,"I",minmicamps,maxmicamps,dec_dig,channel_no,"S"); };
        //Provides information about minimal, maximal microampers able to set and decimal digits.
    bool getCurrentActual(const int module_address, std::vector<double> &micamps, const int channel_no=4){return chvalue<double>(module_address,"IMON",micamps,channel_no);};
        //Get the value of the actual current in microampers
    bool getAboutCurrentActual(const int module_address, std::vector<int> &dec_dig, const int channel_no=4){return chvalue<int>(module_address,"IMDEC",dec_dig,channel_no);};
        //Get the properties of the actual current: number of decimal digits (the IMRANGE value is absent)
    bool getMaxVoltageSet(const int module_address, std::vector<double> &voltages, const int channel_no=4){return chvalue<double>(module_address,"MAXV",voltages,channel_no);};
        //Get the value of maximal set voltage
    bool getAboutMaxVoltageSet(const int module_address, std::vector<double> &minvoltages, std::vector<double> &maxvoltages, std::vector<int> &dec_dig, const int channel_no=4){return chabout(module_address,"MV",minvoltages,maxvoltages,dec_dig,channel_no,""); };
        //Provides information about the minimal, maximal possible value(s) of maximal allowed limit voltages and decimal digits.
    bool getMaxVoltageUpramp(const int module_address, std::vector<double> &voltage_s, const int channel_no=4){return chvalue<double>(module_address,"RUP",voltage_s,channel_no);};
        //Get the value of maximal ramp up in voltage/s
    bool getAboutMaxVoltageUpramp(const int module_address, std::vector<double> &minvoltage_s, std::vector<double> &maxvoltage_s, std::vector<int> &dec_dig, const int channel_no=4){return chabout(module_address,"RUP",minvoltage_s,maxvoltage_s,dec_dig,channel_no,""); };
        //Provides information about minimal, maximal up ramp speed in volt/s able to set and decimal digits.
    bool getMaxVoltageDownramp(const int module_address, std::vector<double> &voltage_s, const int channel_no=4){return chvalue<double>(module_address,"RDW",voltage_s,channel_no);};
        //Get the value of maximal ramp down in voltage/s
    bool getAboutMaxVoltageDownramp(const int module_address, std::vector<double> &minvoltage_s, std::vector<double> &maxvoltage_s, std::vector<int> &dec_dig, const int channel_no=4){return chabout(module_address,"RDW",minvoltage_s,maxvoltage_s,dec_dig,channel_no,""); };
        //Provides information about minimal, maximal down ramp speed in volt/s able to set and decimal digits.
    bool getMaxOvercurrentTime(const int module_address, std::vector<double> &seconds, const int channel_no=4){return chvalue<double>(module_address,"TRIP",seconds,channel_no);};
        //Get the value of the maximal overcurrent time in seconds
    bool getAboutMaxOvercurrentTime(const int module_address, std::vector<double> &minseconds, std::vector<double> &maxseconds, std::vector<int> &dec_dig, const int channel_no=4) {return chabout(module_address,"TRIP",minseconds,maxseconds,dec_dig,channel_no,""); };
        //Provides information about minimal, maximal time allowed to spend in overcurrent state and decimal digits.
    bool getPwrDownSetup(const int module_address, std::vector<bool> &rampt_killf, const int channel_no=4){return chstr2bool(module_address,"PDWN","RAMP","KILL",rampt_killf, channel_no);};
        //Get the powerdown value RAMP(t)/KILL(f)
    bool getPolarity(const int module_address, std::vector<bool> &plust_negf, const int channel_no=4){return chstr2bool(module_address,"POL","+","-",plust_negf, channel_no);};
        //Get the HV polarity +(t)/-(f)
    bool getStatus(const int module_address, std::vector<bool> &on, std::vector<bool> &ramp_up, std::vector<bool> &ramp_dwn, std::vector<bool> &curr_ne_set, std::vector<bool> &volt_overshot, std::vector<bool> &volt_undershot, std::vector<bool> &maxv_protect, std::vector<bool> &off_bytrip, std::vector<bool> &over_pwr, std::vector<bool> &overheat, std::vector<bool> &disabled_remote_and_off, std::vector<bool> &kill_bypanel, std::vector<bool> &interlck_bypanel, std::vector<bool> &cal_error, const int channel_no=4);
        //Provides status information: over/undershot set+/-250V, over power: >1.7W, overheat: >105C
    //--------------------------Module related monitoring commands: contains module address as input. In the case of error it returns with false and error information is written to standard error output otherwise as true--------------------
    bool getAbout(const int module_address, std::string &name, int &no_channels, std::string &firmware_release, int &serial_no);//Get general information about the module as type(name), number of channels, firmware release and serial number
    bool getInterlock(const int module_address, bool &interlockmode_opent_closef, bool &interlockstatus_tyes_fno);
        //Get interlock related parameters: mode=OPEN(t)/CLOSE(f) and current status=YES(1)/NO(0)
    bool getCommunication(const int module_address, bool &controlmode_remotet_localf, bool &localtermination_ont_offf);
        //Get communication information, like control mode(remote(t)/local(f)) and the status of termination status in daisy chaned modules (on(t)/off(f))
    bool getAlarms(const int module_address, std::vector<bool> &ch_alarm, bool &pwr_fail, bool &over_pwr, bool &hvclk_fail);//Returns with the value of alarms: of each channel, POWER_FAIL, OVER_POWER, HV Clock FAIL (outside 200+/-10 kHz range)
    //--------------------------Channel related set commands
    bool setVoltageSet(const int module_address, const double &voltage, const int channel_no){return setvaldouble(module_address,"VSET",voltage,&HV_CAEN_N1471::getAboutVoltageSet,channel_no);};//Sets HV to reach in V
    bool setCurrentSet(const int module_address, const double &micamp, const int channel_no){return setvaldouble(module_address,"ISET",micamp,&HV_CAEN_N1471::getAboutCurrentSet,channel_no);};//Sets the current in microampers
    bool setMaxVoltageSet(const int module_address, const double &voltage, const int channel_no){return setvaldouble(module_address,"MAXV",voltage,&HV_CAEN_N1471::getAboutMaxVoltageSet,channel_no);};//Sets the maximal allowed HV to set V
    bool setMaxVoltageUpramp(const int module_address, const double &voltage_s, const int channel_no){return setvaldouble(module_address,"RUP",voltage_s,&HV_CAEN_N1471::getAboutMaxVoltageUpramp,channel_no);};//Sets the maximal allowed up ramp value in V/s
    bool setMaxVoltageDownramp(const int module_address, const double &voltage_s, const int channel_no){return setvaldouble(module_address,"RDW",voltage_s,&HV_CAEN_N1471::getAboutMaxVoltageDownramp,channel_no);};//Sets the maximal allowed down ramp value in V/s
    bool setMaxOvercurrentTime(const int module_address, const double &second, const int channel_no){return setvaldouble(module_address,"TRIP",second,&HV_CAEN_N1471::getAboutMaxOvercurrentTime,channel_no);};//Sets the maximal allowed overcurrent time in s
    bool setPwrDownSetup(const int module_address, const bool &rampt_killf, const int channel_no){return bool2strch(module_address,"PDWN","RAMP","KILL",rampt_killf);};//Sets the the powerdown mode to ramp(t)/kill(f)
    //bool setAboutCurrentActual(const int module_address, const bool &range_hit_lof, const int channel_no){return bool2strch(module_address,"IMRANGE","HIGH","LOW",range_hit_lof);};//Sets the resolution of the current to high(t)/low(f) resolution -- this parameter is not available
    bool setHVOnOff(const int module_address, const bool &ont_offf, const int channel_no){std::string val=""; return communicate_errorcheck(module_address,false,(ont_offf ? "ON" : "OFF"),val,channel_no);};//Switch HV on(t)/off(f)
    //Module related set commands: modul address should be provided. Returns with true in the case of success, otherwise reason of error is printed on std error output
    bool setInterlock(const int module_address, const bool &interlockmode_opent_closef){return bool2strch(module_address,"BDILKM","OPEN","CLOSED",interlockmode_opent_closef);};//Set the interlock modo to OPEN(t)/CLOSED(0)
    bool setClrAlarm(const int module_address){std::string val=""; return communicate_errorcheck(module_address,false,"BDCLR",val);};//Clears alarm signal
  private:
    //static members
    static const uint8_t LF=0x0a;
    static const uint8_t CR=0x0d;
    static const double TIMEOUT;//Timeout
    static const double WAITSUCCESS;//Waiting some time for the answer/next trial in the communication
    //members
    Serial *communicator;//Communication object
    //methods
    //--------------------intermediate level communication methods
    bool chabout(const int module_address, const std::string &parameter, std::vector<double> &minimum_value, std::vector<double> &maximal_value, std::vector<int> &no_digits, const int channel_no=4, const std::string ext_dec="");
        //Get the "about" information by setting the module address, the first common letter(s) for parameter (this is the "parameter"), the channel number (0..3 for each, 4=all) and the extension letters after the "parameter" for decimal digits. Returning values are minimum_value, maximal_value and number of digits vectors, whose content is cleared. Returns with true for successful communication, otherwise error information is asserted.
    
    bool chstr2bool(const int module_address, const std::string &parameter, const std::string &trueans, const std::string &faseans, std::vector<bool> &vals, const int channel_no=4);
        //Get the actual value of a parameter and transforming it to boolean by setting the module address, the command letter, the true and false answers and the channel number (0..3 for each, 4=all). Returning values are the requested value. Returns with true for successful communication, otherwise error information is asserted.
    
    bool bool2strch(const int module_address, const std::string &parameter, const std::string &trueans, const std::string &falseans, const bool &vals, const int channel_no=4);
        //Set the requested value of a parameter by transforming it from boolean by setting the module address, the command letter, the true and false answer strings and the channel number (0..3 for each, 4=all). Returns with true for successful communication, otherwise error information is asserted.
    
    bool setvaldouble(const int module_address, const std::string &parameter, const double &val,bool (HV_CAEN_N1471::*about)(const int, std::vector<double> &, std::vector<double>&, std::vector<int>&, const int), const int channel_no=4);//Set the value of the parameter of module related to channel (0..3 for each, 4=all) by val. The about function is required as it provides decimal digits. Returns with true for successful communication, otherwise error information is asserted.
    
    template <typename T> bool chvalue(const int module_address, const std::string &parameter, std::vector<T> &vals, const int channel_no=4);
        //Get the actual value of a parameter by setting the module address, the command letter and the channel number (0..3 for each, 4=all). Returning values are the requested value. Returns with true for successful communication, otherwise error information is asserted.
    
    template <typename T> bool valuech(const int module_address, const std::string &parameter, const T &val, const int &decdigits, const int channel_no=4);
        //Set the requested value of a parameter by transforming it to communication format by the help of the module address, the command letter, number of decimal digits and the channel number (0..3 for each, 4= for all). Returns with true for successful communication, otherwise error information is asserted.
    
    //---------------------low level communication methods
    bool communicate(const int module_address, const bool mont_setf, const std::string &parameter, std::string &value, const int channel_no=4);
        //Communicate with serial interface: it sets the module_address, the command: type MONitor(t)/SET(f), the parameter, the requested value (only for set input) and channel number (0..3 specific channel, 4: all)
        //It returns with true in the case of successful end when the requested value(s) will be set to value (only for monitor).
        //Otherwise with false and the reason of the error is indicated in the value part as:
        // CMD (wrong command format or command not recognized); CH (channel field not present or wrong channel value);
        // PAR (parameter field not present or not recognized); VAL (wrong set value <min or >max); LOC (command set with module in LOCAL mode)
        // COE (communication error); WRA (wrong address)
    bool communicate_errorcheck(const int module_address, const bool mont_setf, const std::string &parameter, std::string &value, const int channel_no=4);
        //Similar to the previous one, but writes the error messages into standard error device
};

template <typename T> bool HV_CAEN_N1471::chvalue(const int module_address, const std::string &parameter, std::vector<T> &vals, const int channel_no)
//Get the actual value of a parameter by setting the module address, the command letter and the channel number (0..3 for each, 4=all). Returning values are the requested value. Returns with true for successful communication, otherwise error information is asserted.
{
    std::string reply;
    vals.clear();
    if(communicate_errorcheck(module_address,true,parameter,reply,channel_no))
    {
        std::stringstream ss(reply);
        T dval;
        while(ss>>dval) vals.push_back(dval);
        return true;
    }
    else return false;
}

template <typename T> bool HV_CAEN_N1471::valuech(const int module_address, const std::string &parameter, const T &val, const int &decdigits, const int channel_no)
        //Set the requested value of a parameter by transforming it to communication format by the help of the module address, the command letter, number of decimal digits and the channel number (0..3 for each, 4=all). Returns with true for successful communication, otherwise error information is asserted.
{
    std::stringstream ss;
    ss<<std::fixed<<std::setprecision(decdigits)<<val;
    std::string sr(ss.str());
    return communicate_errorcheck(module_address,false,parameter,sr,channel_no);
}
#endif

