#include "IOOperation.h"

namespace IOOperation
{
  InterruptHandle::InterruptHandle(bool exactbit)
  //Constructor: if exactbit is true, then it notifies only those, whose value exactly the requested number. If it is false (bit), then it notifies those, whoose bit has been changed 
  {
    number_or_bit_mode=exactbit;//Setting interrupt storage mode
    receiver_list=new std::multimap<uint32_t,InterruptReceiver*> ();//Allocate memory
    interrupt_attention=0x0;//Set interesting bits/numbers
    changed=false;
  }
  
  InterruptHandle::~InterruptHandle()
  //Destructor: it calls the DisableInterrupt method of each associated InterruptReceivers
  {
    //Removing each Interrupt receivers
    while(!(receiver_list->empty()))
    {
      std::multimap<uint32_t,InterruptReceiver*>::iterator i= receiver_list->begin();//Set to the start
      ((*i).second)->DisableInterrupt(this,0x0,true);//Disable all interrupts corresponding to the InterruptReceiver
    }
    delete receiver_list;//Delete the list
  }
  
  void InterruptHandle::Register(InterruptReceiver *ir, uint32_t mask)
  //Registers an interrupt receiver related to mask events.
  {
    changed=false;//Initialize changed indicator
    if(number_or_bit_mode)
    {
    	this->addremove_element(ir,mask,false);//Add the element in the case of exact match
    	if(mask<32) interrupt_attention|=0x1<<(mask); 
    }
    else
    {
    	//Define up and down bit (down for AND to find the bit 1 and up to check the remaining bits)
    	uint32_t down=0x1;
    	uint32_t up=0xffffffff;
    	//We should add it bitwise
    	while((mask&up)!=0)
    	{
    	  if((mask&down)==down)
    	  {
    	    this->addremove_element(ir,down,false);//Add element
    	    interrupt_attention|=down;//Add element to the attention register
    	  }
    	  up-=down;//Decrease up value by the actual down 
    	  down<<=1;//Move one bit higher
    	}
    }
    return;
  }
  
  bool InterruptHandle::UnRegister(InterruptReceiver *ir, uint32_t mask)
 //Unregisters the entries kept by ir for the mask events. Returns with true, if all of the entries related to ir are cleared
  {
    changed=false;//Initialize changed indicator
    if(number_or_bit_mode)
    {
      this->addremove_element(ir,mask,true);//Remove the element in the case of exact match
      if((mask<32)&&(receiver_list->count(mask) == 0)) interrupt_attention&=~(0x1<<mask);//Remove interrupt attention for the specified value
    }
    else
    {
    	//Define up and down bit (down for AND to find the bit 1 and up to check the remaining bits)
    	uint32_t down=0x1;
    	uint32_t up=0xffffffff;
    	//We should add it bitwise
    	while((mask&up)!=0)
    	{
    	  if((mask&down)==down)
    	  {
    	    this->addremove_element(ir,down,true);
    	    if(receiver_list->count(down) == 0) interrupt_attention&=~down;//Remove attention bits
    	  }
    	  up-=down;//Decrease up value by the actual down 
    	  down<<=1;//Move one bit higher
    	}
    }
    //Looking for any remaining entries of ir
    bool found=false;
    for(std::multimap<uint32_t,InterruptReceiver*>::iterator i=receiver_list->begin(); (i!=receiver_list->end())&&(!found); i++) found=((i->second)==ir);
    return (!found);//If there has not been found any, it means, that they are cleared
  }
      
  bool InterruptHandle::UnRegister(InterruptReceiver *ir)
 //Unregisters the entries kept by ir for all of the events.  Returns always with true
  {
    //Lookinf for each occurences of ir
    changed=false;//Initialize changed indicator
    std::multimap<uint32_t,InterruptReceiver*>::iterator temporary,i=receiver_list->begin();//Temporary iterator and ordinary
    while(i!=receiver_list->end())
    {
      if((i->second)==ir)
      {
        temporary=i;
        uint32_t mask=i->first;
        i--;
        this->addremove_element(ir,temporary->first,true);
        if ((!((number_or_bit_mode)&&(mask>=32)))&&(receiver_list->count(mask) == 0)) interrupt_attention&=~ (number_or_bit_mode ? (0x1<<mask) : mask);//Remove entry/bit from the attention
      };//One has been found: remove the element from the list its index will be updated
      i++;//Go to the next one
    }
    return true;
  }
  
  void InterruptHandle::Notify(uint32_t status, uint32_t mask)
  //Notifies the mask related InterruptReceivers with status: this method should be called if any change has been happened
  {
    if(number_or_bit_mode)//Exact mode
    {
    	std::pair<std::multimap<uint32_t,InterruptReceiver*>::iterator, std::multimap<uint32_t,InterruptReceiver*>::iterator>  range=receiver_list->equal_range(mask);//Get the range of dataset
    	for(std::multimap<uint32_t,InterruptReceiver*>::iterator i= range.first; i!=range.second; i++) i->second->Action(this,status);
    }
    else
    {
      //Bitwise mode: define up and down bit (down for AND to find the bit 1 and up to check the remaining bits)
      uint32_t down=0x1;
      uint32_t up=0xffffffff;
      //Creating a list of interruptreceivers to be notified
      std::list<InterruptReceiver*> *irs=new std::list<InterruptReceiver*>;
      //We should add them bitwise
      while((mask&up)!=0)
      {
        if((mask&down)==down)
        {
          //This bit has been set -- looking for the related entries
          std::pair<std::multimap<uint32_t,InterruptReceiver*>::iterator, std::multimap<uint32_t,InterruptReceiver*>::iterator>  range=receiver_list->equal_range(down);//Get the range of dataset
    	  for(std::multimap<uint32_t,InterruptReceiver*>::iterator i= range.first; i!=range.second; i++) irs->push_back(i->second);//Add items to the list
        }
    	up-=down;//Decrease up value by the actual down 
    	down<<=1;//Move one bit higher
      }
      irs->unique();//Remove multiple elements from the list
      for(std::list<InterruptReceiver*>::iterator i=irs->begin(); i!=irs->end(); i++) (*i)->Action(this,status);//Notifies the Interrupt receivers
      delete irs;//Delete elements
    }
    return;
  }
  
  void InterruptHandle::addremove_element(InterruptReceiver *ir, uint32_t mask, bool remove)
  //It adds(f)/remove(t) an element of InterruptReceiver* for the list to the mask key (it avoids multiple entry of the same mask and ir pair for add operation)
  {
  	//We are checking, have it been registered yet
    	std::pair<std::multimap<uint32_t,InterruptReceiver*>::iterator, std::multimap<uint32_t,InterruptReceiver*>::iterator>  range=receiver_list->equal_range(mask);
    	bool found=false;
    	std::multimap<uint32_t,InterruptReceiver*>::iterator i= range.first;
    	while((i!=range.second)&&(!found)){if((i->second)==ir) found=true; else i++;} //Matching entry has been found
    	if((!found)&&(!remove)) 
    	{
    	  if(receiver_list->count(mask)==0) changed=true;//If this is the first entry, set the changed flag
    	  receiver_list->insert(std::pair<uint32_t,InterruptReceiver*>(mask,ir));//Add the new entry
    	}
    	else if ((found)&&(remove))
    	{
    	  receiver_list->erase(i);//Remove the entry
    	  if(receiver_list->count(mask)==0) changed=true;//This was the last entry for this mask, set the changed flag
    	}
    	return;
  }
  
  //InterruptReceiver CLASS
  
  InterruptReceiver::InterruptReceiver()
  //Constructor
  {
    handlers= new std::list<InterruptHandle*> ();//Allocate memory for the interrupt handler
  }
  
  InterruptReceiver::~InterruptReceiver()
  //Destructor
  {
    //Remove references from all of the interrupt Handlers
    while(!(handlers->empty()))
    {
    	std::list<InterruptHandle*>::iterator i=handlers->begin();
     	this->DisableInterrupt(*i,0x0,true);
     }
    delete handlers;//Delete the handlers
  }
  
  void InterruptReceiver::EnableInterrupt(InterruptHandle *ih, uint32_t mask)
  //Enable interrupt service at ih for mask dataset: returns with success or fail
  {
    //Sending request to Interrupt handler
    ih->Register(this,mask);//Try to register new entry: if OK it returns with the original handler
    //Looking for any entries
    std::list<InterruptHandle*>::iterator i=handlers->begin();
    bool found=false;
    while((i!=handlers->end())&&(!found)){if((*i++)==ih) found=true;}
    //Insert new velue, if it is needed
    if(!found) handlers->push_back(ih);
    return;
  }
  
   void InterruptReceiver::DisableInterrupt(InterruptHandle *ih, uint32_t mask, bool all)
   //Disables the interrupt service at if for mask dataset(all=false) or each (all=true): returns with success or fail
   {
     bool cleared= (all ? ih->UnRegister(this) : ih->UnRegister(this, mask));//Unregister all occurances
     if(cleared) handlers->remove(ih);//Remove the interrupt handler from the list
     return;
   }
   
   //InterruptRegister CLASS
   
   bool InterruptRegister::reset(uint32_t mask)
   //This method resets the LAM service and set mask, which contains the variables to be attended. If LAM service not started yet it starts and mask=0, it disable LAM service
  {
    bool retval=true;//Returning value
    if(lservice)
    {
      //LAM service is on - disable it
      retval=resetlamff();//Resets LAM flip-flop 
      if(retval) retval=!testlam(false);//LAM f-f should be not set
      if(retval) retval=enabledisable_lam(false);//Disable LAM
      if(retval) retval=!testlam(true);//Test LAM disabled status
      if(retval) lservice=false;
     }
     if (!(retval && (mask==0x0))) //LAM disabled and mask 0 => return (disable LAM), otherwise the following
     {
	//Reading the input
	uint32_t input=0x0;//Value of the input
	if(retval) retval=read(input,false);//Read the pure input
	if(retval) retval=write((uint16_t) input, false);//Set it on the storage register
	if(retval) retval=write((uint16_t) ~mask, true);//Sets the mask
	if(retval) retval=read(input,true);//Read changed values
	if(retval) retval=resetlamff();//Resets LAM flip-flop 
	if(retval) retval=!testlam(false);//LAM f-f should be not set
	if(retval) retval=enabledisable_lam(true);//Enable LAM
	if(retval) retval=testlam(true);//LAM should be enabled
	if(retval)
	{
	  lservice=true;//Set LAM service indicator
	  Notify(previous^input,input);//Notify the classes about the current value
	  previous^=input;
	}
      }
      return retval;//Return with the corresponding message
  }
  
   //Output Register
  
  OutputRegister::OutputRegister(Camac &camac, uint8_t station_number, bool jumper_execute)
  //Constructor: camac controller and station number and execute jumper setup (1-true, 2-false)
  {
    cam = &camac;//Set camac controller
    station=station_number;//Set station number
    for(int i=0; i<2; i++) message[i]=0;//Set the software images
    jumper=jumper_execute;//Set execute switch
    //Initializing the device: clear each channel
    if((!clearbuffer(true))||(!clearbuffer(false)))
    {
      std::cerr<<"ERROR: IOOperation::IOOperation: clearing the buffer was unsuccessful"<<std::endl;//Clear the content of the registers
      //station=0;//Set station number to zero: it will give errors any cases
    }
  }
  
  OutputRegister::~OutputRegister()
  //Destructor
  {
    //Clearing the buffer fully
    if((!clearbuffer(true))||(!clearbuffer(false)))
    {
      std::cerr<<"ERROR: IOOperation::~IOOperation: clearing the buffer was unsuccessful"<<std::endl;//Clear the content of the registers
      station=0;//Set station number to zero: it will give errors any cases
    }
    cam=0;//Reset camac
  }
  
  bool OutputRegister::Set(uint32_t pattern, bool direct)
  //This method set bits of pattern to on, if direct is false. Otherwise, it writes the pattern directly.
  {
    uint32_t msg;//Message to be sent to buffer
    bool retval=true;//Return value
    //Create message to be sent
    for(int i=0; i<2; i++)
    {
      if (retval)
      {
	msg=(pattern>>(16*i))&0xffff;//Cutting the message into parts
	if(!direct) msg|=message[i];//Determine the message to do not disturb previously set bits
	if(((message[i]|msg)&0xffff)!=message[i])
	//It should be done something, if the already set pattern differs from the intended one
	{
	  retval=overwritebuffer(msg,(i+1)%2);//Overwrite buffer
	  if(retval&&(!jumper)) retval=execute((i%2==0) ? LO_A : HI_A);//Execute, if it is necessary
	  if(retval) message[i]=0xffff&msg;//Overwrite software representation
	}
      }
    }
    return retval;
  }
  
  bool OutputRegister::Read(uint32_t *pattern, bool software)
  //Read the currently set pattern, if software is true, does not ask camac module; returns with the success of communication
  {
    bool retval=true;//Returning value
    //The following block commented out, bacause this function is not available on the real device (only software is available) 
    retval=software;
    //Ask module
    /*if(!software)
    {
      for(int i=0; i<2; i++) 
      {
	if(retval) 
	{
	  uint32_t buff;//Buffer
	  retval=readbuffer(buff, (i+1)%2);//Ask the module
	  if(retval) message[i]=buff&0xffff;//Update software storage
	}
      }
    }*/
    //Get the answer from the software storage
    *pattern=(uint32_t)(((uint32_t) message[1])<<16)+((uint32_t)message[0]);
    return retval;
  }
  
  bool OutputRegister::Reset(uint32_t pattern)
  //This method set bits of pattern to off.
  {
    uint32_t msg;//Message to be sent to buffer
    bool retval=true;//Return value
    for(int i=0; i<2; i++)
    {
      if (retval)
      {
	//Create message to be sent
	msg=(~((pattern>>(16*i))&0xffff))&0xffff;//Cutting the message into parts and negate them
	msg&=message[i];//Determine the requested message to be written
	if(((message[i]&msg)&0xffff)!=message[i])
	//It should be done something, if the already set pattern differs from the intended one
	{
	  retval=overwritebuffer(msg,(i+1)%2);//Overwrite buffer
	  if(retval&&(!jumper)) retval=execute((i%2==0) ? LO_A : HI_A);//Execute, if it is necessary
	  if(retval) message[i]=0xffff&msg;//Overwrite software representation
	}
      }
    }
    return retval;
  }
  
  
}
