#include "Logic.h"

namespace Logic
{

  DualGate::DualGate(Camac &camac, uint8_t station_number)
  //Constructor: selects the camac comminicator and station number. Program retrieve the actual control register 
  {
    cam=&camac;//Set CAMAC controller
    station=station_number;//Set station number
    actual_control_register=0x0;//Initialize the copy of the control register
    if(!read_control_register()) std::cerr<<"ERROR: logic::DualGate reading the control register has not been succeeded!"<<std::endl;//Read the control register
  }
  
  //All of the following string configuration is the following: "1".."8" set as direct AND, "N1"/"n1"..."N8"/"n8" sets as negative input AND
  std::string DualGate::GetConfiguration(bool direct)
  //Retrieve the actual configuration of the dual gate. direct: read from module(t)/return with the stored value (f)
  {
    if (direct&&(!read_control_register())) return std::string("ERROR");//Bad communication
    return bit2str(actual_control_register);//Returning with local copy, if everything is OK
  }
  
  bool DualGate::SetConfiguration(std::string config)
  //Sets the "config" configuration
  {
    uint16_t message=str2bit(config);//Transforming to bits
    return ((message==actual_control_register) ? true : set_control_register(message));//Set message configuration
  }
  
  bool DualGate::SelectiveConfiguration(std::string config, bool set)
  //Selectively sets(t)/resets(f) the configuration (for reset, all configuration will be direct)
  {
    uint16_t mask= (set ? str2bit(config) : ~((str2bit(config)<<8)|str2bit(config)));//Transforming to bits and creating masks: for set the actual one, for reset negative of the selected
    uint16_t message=(set ? actual_control_register|mask : actual_control_register&mask) ;//Create message to send: set comparing by the mask and reset by switch off the requested bits
    return ((message==actual_control_register) ? true : set_control_register(message));//Sometimes does not need to do any operation...
  }
  
  bool DualGate::read_control_register()
  //This program reads the actual value of the control register
  {
    uint32_t rcv;//stores the received message
    bool retval=cam->getData(READ_A,READ_F,station, rcv, Camac::STATUS_BLANK);//read from the device
    if(retval) actual_control_register=rcv&0xffff;//update local copy
    return retval;
  }
  
  bool DualGate::set_control_register(uint16_t value)
  //Sets the control register and updates its local representation
  {
    bool retval=cam->setData(WRITE_A_SELECT_1_4|WRITE_A_SELECT_5_8|WRITE_A_NEG_1_4|WRITE_A_NEG_5_8, WRITE_F, station, (const uint32_t)value, Camac::STATUS_BLANK);//Send the configuration
    if(retval) actual_control_register=value;//update local copy
    return retval;//Return with success or fail
  }
  
  uint16_t DualGate::str2bit(const std::string input)
  //Converts string logics into bits
  {
    uint16_t result=0x0;//Start initial value
    bool next_item_negate=false;//Shows, that the next number should be negative
    for(int i=0; i<(int)input.size(); i++)
    {
      if((input[i]=='N')||(input[i]=='n')) next_item_negate=true;//it should be a negative
      else if((input[i]>'0')&&(input[i]<='9'))
      {
        //This is a valid number: set the corresponding bits of result
        int number=(int)input[i]-(int)'0';
        result|=0x1<<(number-1);//Set as it contributes
	if(next_item_negate) result|=0x1<<(number+7);//Set the negative sign
	next_item_negate=false;//Set it to default
      }
      else next_item_negate=false;//Cancel negative, if it is not followed by a number
    }
    return result;
  }
  
  std::string DualGate::bit2str(const uint16_t input)
  //Converts string logics into bits
  {
    std::string answer="";//Create an empty string
    uint8_t present=input&0xff;
    uint8_t negative=(input>>8)&0xff;
    for(int i=0; i<8; i++)
    {
      if((present&0x1)==0x1) answer+= std::string(((negative&0x1)==0x1) ? "N" : "")+std::to_string(i+1);
      present>>=1;
      negative>>=1;
    }
    if((int)answer.size()==0) answer="0";
    return answer;
  }
}