/******************************************
 * Modified by Laszlo Temleitner: 14.03.2018: new structure format allows to read and write more than one bulky dataset as a new column. TODO: test
 */

#include "MTESTRawData.h"

MTESTRawData::MTESTRawData()
//Initialization of the class variables
{
  //Initialize comment variables
  clear_comments();
  clear_motors();
  clear_counters();
}

bool MTESTRawData::WriteNewSnapshot(bool clr_prev)
//This function creates a new snapshot from the last saved datasets, if clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared
{
  if(runs.size()==0) return false;
  //Figure out: does it need bulky writing?
  std::vector<counter_data>::iterator max_counter=std::max_element(counters.begin(),counters.end(),[](const counter_data &c1, const counter_data &c2){return c1.no_channels<c2.no_channels;});
  bool bulky=(max_counter->no_channels >  MTESTRawData::BULK_LIMIT);//Qualify as bulky, if the number of channels is over 10
  //Create proper filename and open file
  runs.back().snapshots++;
  std::string filename(runs.back().filename+"_"+std::to_string(runs.back().run_no));
  if (bulky) filename+="_"+std::to_string(runs.back().item_no)+"_"+std::to_string(runs.back().snapshots)+".dat"; else filename+=".dat";
  std::ofstream os(filename.c_str(),std::ofstream::out|std::ofstream::app);
  if(!os.is_open()) return false;
  bool retval=write(os,bulky,(runs.back().item_no==1)&&(runs.back().snapshots==1));
  if (retval&&bulky)
  {
    std::string symfile_name = runs.back().filename+"_"+std::to_string(runs.back().run_no)+"_"+std::to_string(runs.back().item_no);
    if(runs.back().snapshots!=1) remove(symfile_name.c_str());
    symlink(filename.c_str(),symfile_name.c_str());
  }
  if(retval&&clr_prev)
  {
    clear_motors(true,true);//Clear positions of the scanning motors
    clear_time();//Clear time records
    clear_counters(true);//Clear the counters
  }
  return retval;
}

bool MTESTRawData::WriteNewItem(bool clr_prev)
//This function writes a new item from the last saved datasets, if clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared
{
  //In the case of new item, we should add a new item to the runs
  if(runs.size()==0) return false;
  run_data r=runs.back();
  r.item_no++;
  r.snapshots=0;
  runs.push_back(r);
  return WriteNewSnapshot(clr_prev);
}

bool MTESTRawData::WriteNewRun(bool clr_prev)
//This function writes the start of a new run from the last saved datasets, if clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared
{
  //In the case of new run, we should add a new item to the runs
  if(runs.size()==0) return false;
  run_data r=runs.back();
  r.run_no++;
  r.item_no=1;
  r.snapshots=0;
  runs.push_back(r);
  return WriteNewSnapshot(clr_prev);
}

bool MTESTRawData::WriteNewSample(std::string filename, bool clr_prev, int start_run_no)
//This function writes new sample data to the requested main file at the first run, first item and first snapshot, if clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared. There is a possibility set up the start run
{
  //In the case of new sample, we should add new items to the runs
  run_data r;
  r.filename=filename;
  r.run_no=start_run_no;
  r.item_no=1;
  r.snapshots=0;
  //Avoid append situation for "bulky" datasets
  if(std::any_of(counters.begin(),counters.end(),[](const counter_data &c1){return c1.no_channels> MTESTRawData::BULK_LIMIT;}))
  {
    bool found=true;
    while(found)
    {
	std::ifstream ifs((filename+"_"+std::to_string(r.run_no)+"_1_1.dat").c_str());
	found=ifs.is_open();
	if(found) r.run_no++;//To avoid overwrite situation
    }
  }
  runs.push_back(r);
  return WriteNewSnapshot(clr_prev);
}


bool MTESTRawData::write(std::ofstream &os, bool bulky, bool head)
//This method writes the current datasets to file in bulky(t) or in single(f). In single, this can be just an update(f) or write header(t) part, as well.
{
  bool retval=true;
  //Bulky mode
  if(bulky) return writeHead(os, true)&&writeStandingMotors(os)&&writeBulkyScanningMotors(os)&&writeBulkyTimes(os)&&writeBulkyCounters(os);
  else if (head)
  {
    retval=writeHead(os,false)&&writeStandingMotors(os)&&writeSingleCountersHead(os)&&writeSingleScanningMotors(true,os)&&writeSingleTimes(os,true);
    os<<std::endl;
  }
  if(retval)
  {
    retval=writeSingleCounters(os)&&writeSingleScanningMotors(false, os)&&writeSingleTimes(os);
    os<<std::endl;
  }
  return retval;
}

bool MTESTRawData::read(std::ifstream &is)
//Reads the input stream and store the dataset read in
{
  bool bulky=false;
  bool retval=readHead(is,bulky)&&readStandingMotors(is);
  if(bulky&&retval) return readBulkyScanningMotors(is) && readBulkyTimes(is) && readBulkyCounters(is);
  else if (retval)
  {
    //Get header
    std::string line;
    std::getline(is,line);
    size_t first_motor=line.find("[pul",0);
    first_motor=line.rfind("]",first_motor)+1;
    std::stringstream ss(line.substr(0,first_motor));
    readSingleCounters(ss);
    std::stringstream sas(line.substr(first_motor,std::string::npos));
    readSingleScanningMotors(sas);
    //Get the counts
    if(!is.eof())
    {
      std::getline(is,line);
      while(!is.eof())
      {
	
	std::stringstream sa(line);
	readSingleCounters(sa);
	readSingleScanningMotors(sa);
	readSingleTimes(sa);
	std::getline(is,line);
      }
    }
    return true;
  }
  else return false;
}


bool MTESTRawData::writeHead(std::ofstream &os, bool bulky)
//Writes the head information to the requested filestream
{
  if(!(os.is_open())) return false;//Avoid the case of non-opened channel
  os<<"# MTEST RAW datafile version "<<actual_version<<" "<<(bulky ? "BULK" : "SINGLE")<<std::endl;
  os<<"# Sample: "<<comment_sample<<std::endl;//Comments about the sample
  os<<"# Setup information (M-monochromator, S-sample, D-detector): "<<std::endl;
  os<<"# \tMS collimator: "<<comment_ms_collimator<<std::endl;
  os<<"# \tMS slits: "<<comment_ms_slits<<std::endl;
  os<<"# \tS environment: "<<comment_s_environment<<std::endl;
  os<<"# \tS holder: "<<comment_s_holder<<std::endl;
  os<<"# \tSD slits: "<<comment_sd_slits<<std::endl;
  os<<"# \tD position: "<<comment_d_position<<std::endl;
  return true;
}

bool MTESTRawData::readHead(std::ifstream &is,bool &bulky)
//Reads the head information from the requested filestream
{
  if(!is.is_open()) return false;//Avoid the case of non-opened channel
  std::string line;
  std::getline(is,line);//Get the first line
  std::stringstream ss(line);
  for(int i=0; i<5; i++) ss>>line;
  int version;
  ss>>version;//Get the actual version: it can be any role in afterwards...
  std::string b;
  ss>>b;
  bulky=(b=="BULK");
  comment_sample=get_comment(is);
  std::getline(is,line);//Skip the next line
  comment_ms_collimator=get_comment(is);
  comment_ms_slits=get_comment(is);
  comment_s_environment=get_comment(is);
  comment_s_holder=get_comment(is);
  comment_sd_slits=get_comment(is);
  comment_d_position=get_comment(is);
  return true;
}

bool MTESTRawData::writeStandingMotors(std::ofstream &os)
//Writes the list of standing motors into filestream
{
  if(!os.is_open()) return false;
  //Get the number of stationary motors
  int stmotors=std::count_if(motors.begin(),motors.end(),[](const motor_data& md){ return md.standing; });
  os<<"# Number of stationary motors: "<<stmotors<<std::endl;
  if(stmotors!=0)
  {
    os<<"#\tname position_in_pulses position_in_angles"<<std::endl;
    std::vector<motor_data>::iterator it=motors.begin();
    while((it=find_if(it,motors.end(),[](const motor_data& md) { return md.standing; }))!=motors.end())
    {
      os<<"#\t"<<it->name<<" "<<it->pos_pulses[0];
      if(! it->only_pulses) os<<" "<<it->pos_angles[0];
      os<<std::endl;
      it++;
    };
  }
  return true;
}

bool MTESTRawData::readStandingMotors(std::ifstream &is)
//Reads the list of standing motors from the filestream
{
  if(!is.is_open()) return false;
  //Get the number of stationary motors
  std::string line;
  std::getline(is,line);//Get the first line
  std::stringstream ss(line);
  for(int i=0; i<5; i++) ss>>line;
  int stmotors;
  ss>>stmotors;//Get the number of static motors
  if(stmotors!=0)
  {
    std::getline(is,line);//Skip the comment line
    for(int i=0; i<stmotors; i++)
    {
      std::stringstream sb(get_comment(is));
      std::string name;
      int pulse;
      
      sb>>name>>pulse;
      if(sb.rdbuf()->in_avail() ==0) addMotor(name,pulse, true, 0);
      else
      {
	double degrees;
	sb>>degrees;
	addMotor(name,pulse, true, &degrees);
      }
      
    }
  }
  return true;
}

bool MTESTRawData::writeBulkyScanningMotors(std::ofstream &os)
//Writes the list and position of scanning motors to a file in bulky mode
{
  if(!os.is_open()) return false;
  os<<"# Number of scanning motors: "<<std::count_if(motors.begin(),motors.end(),[](const motor_data& md){ return !(md.standing); })<<std::endl;
  os<<"#\tname position_in_pulse position_in_degrees"<<std::endl;
  std::vector<motor_data>::iterator it=motors.begin();
  while((it=find_if(it,motors.end(),[](const motor_data& md){ return !(md.standing); }))!=motors.end()) 
  {
    os<<"#\t"<<it->name<<" "<<it->pos_pulses.back();
    if(!it->only_pulses) os<<" "<<it->pos_angles.back();
    os<<std::endl;
    it++;
  };
  return true;
}

bool MTESTRawData::readBulkyScanningMotors(std::ifstream &is)
//Set the list and position of scanning motors from a file in bulky mode
{
  if(!is.is_open()) return false;
  //Get the number of scanning motors
  std::string line;
  std::getline(is,line);//Get the first line
  std::stringstream ss(line);
  for(int i=0; i<5; i++) ss>>line;
  int scmotors;
  ss>>scmotors;//Get the number of scanning motors
  std::getline(is,line);//Skip comment line
  for(int i=0; i<scmotors; i++)
  {
    std::stringstream sb(get_comment(is));
    std::string name;
    int pulse;
    sb>>name>>pulse;
    if(sb.rdbuf()->in_avail() ==0) addMotor(name,pulse, false, 0);
    else
    {
      double degrees;
      sb>>degrees;
      addMotor(name,pulse, false, &degrees);
    }
  }
  return true;
}

bool MTESTRawData::writeSingleScanningMotors(bool head, std::ofstream &os)
//Writes the list and position of scanning motors to a file in single mode. If the head is true, only head information will be written, otherwise the dataset
{
  if(!os.is_open()) return false;
  std::vector<motor_data>::iterator it=motors.begin();
  while((it=find_if(it,motors.end(),[](const motor_data& md){ return !(md.standing); }))!=motors.end())
  {
    if(head) os<<it->name<<" [pulse"<<((it->only_pulses) ? "] " : ",degrees] ");
    else
    {
      os<<it->pos_pulses.back()<<" ";
      if(!it->only_pulses) os<<" "<<it->pos_angles.back()<<" ";
    }
    it++;
  }
  return true;
}

void MTESTRawData::readSingleScanningMotors(std::stringstream &ss)
//Reads the list and position of scanning motors from a stringstream single mode.: we assume, it contains only motor related parts
{
  //We assume, this will be a header -- try to get the number of motors
  size_t pos=0;
  std::string line=ss.str();
  int no_motors=0;
  while((pos=line.find("[pulse",pos+1))!=std::string::npos) no_motors++;
  if(no_motors>0)
  {
    for(int i=0; i<no_motors; i++)
    {
      std::string name, dimensions;
      ss>>name>>dimensions;
      //Create new item to store
      motor_data m;
      m.name=name;
      m.standing=false;
      m.only_pulses=(dimensions=="[pulse]") ;//Indicate, that this motor does not have angular representation
      std::vector<motor_data>::iterator it=std::find_if(motors.begin(),motors.end(),[name](const motor_data& md){ return md.name == name; });//Look for the name
      if (it==motors.end()) motors.push_back(m);//Otherwise, we do not have to do anything...
    }
  }
  else
  {
    //We should read the datasets in
    std::vector<motor_data>::iterator it=motors.begin();
    while((it=find_if(it,motors.end(),[](const motor_data& md){ return !(md.standing); }))!=motors.end())
    {
      int pulse;
      ss>>pulse;
      it->pos_pulses.push_back(pulse);
      //Read angular value, if it is associated
      if(!it->only_pulses)
      {
	double angle;
	ss>>angle;
	it->pos_angles.push_back(angle);
      }
      it++;
    }
  }
}

bool MTESTRawData::writeBulkyCounters(std::ofstream &os)
//This method writes the datasets of counters into an output file
{
  if(!os.is_open()) return false;
  os<<"# Number of counters: "<<counters.size()<<std::endl;
  std::vector<counter_data>::iterator max_counter=std::max_element(counters.begin(),counters.end(),[](const counter_data &c1, const counter_data &c2){return c1.no_channels<c2.no_channels;});
  //Creating output for auxiliary datasets
  os<<"#\t counter_name number_of_channels number_of_additionals [channel_couts] [additional_counts] [captions]"<<std::endl;
  if(counters.size()>1)
  {
    std::vector<counter_data>::iterator it=counters.begin();
    
    while(it!=counters.end())
    {
      if (it->no_channels <=  MTESTRawData::BULK_LIMIT)
      {
        os<<"#\t "<<it->name<<" "<<it->no_channels<<" "<<it->no_additionals<<" ";
        for(int i=(int)it->count_data.size()-(it->no_channels); i<(int)it->count_data.size(); i++) os<<it->count_data[i]<<" ";
        for(int i=(int)it->additionals.size()-(it->no_additionals); i<(int)it->additionals.size(); i++) os<<it->additionals[i]<<" ";
        for(int i=0; i<(int)it->caption.size(); i++) os<<it->caption[i]<<" ";
        os<<std::endl;
      }
      it++;
    }
  }
  {
    //Creating dataset for the bulky counters
    std::vector<counter_data>::iterator it=counters.begin();
    
    //Writing header
    while(it!=counters.end())
    {
      if (it->no_channels >  MTESTRawData::BULK_LIMIT)
      {
        os<<"#\t "<< it->name <<" "<<it->no_channels<<" "<<it->no_additionals<<" * ";
        for(int i=(int)it->additionals.size()-(it->no_additionals); i<(int)it->additionals.size(); i++) os<<it->additionals[i]<<" ";
        for(int i=(int)it->caption.size()-(it->no_additionals); i<(int)it->caption.size(); i++) os<<it->caption[i]<<" ";
        os<<std::endl;
      }
      it++;
    }
  }
  //Writing out the counts
  os<<"# caption_of_maximum [counts]"<<std::endl;
  for(int i=0; i<max_counter->no_channels; i++)
  {
      os<<max_counter->caption[i]<<" ";
      for(std::vector<counter_data>::iterator it=counters.begin(); it!=counters.end(); it++)
      {
          if(it->no_channels > MTESTRawData::BULK_LIMIT)
          {
              if(i < it->no_channels) os<<it->count_data[(int)it->count_data.size()-it->no_channels+i]<<" ";
              else os<<"-1 ";//Indicating out of valid range
          }
      }
      os<<std::endl;
  }
  return true;
}

bool MTESTRawData::readBulkyCounters(std::ifstream &is)
//This method creates datasets of counters from input file
{
  if(!is.is_open()) return false;
  std::string line;
  std::getline(is,line);//Get the first line
  std::stringstream ss(line);
  for(int i=0; i<4; i++) ss>>line;
  int no_cnt;
  ss>>no_cnt;//Get the number of counters
  std::getline(is,line);//Skip the comment line
  std::vector<std::string> bulky_name;//Registers bulky names
  std::vector<int> bulky_no_ch, bulky_no_add;// Registers bulky number of channels and number of additional datasets
  std::vector<std::vector<double>> bulky_add_counts;// Registers bulky additional counts
  std::vector<std::vector<std::string>> bulky_acapt;// Registers captions of bulky additional counts
  //Read the head of counters
  for(int i=0; i<no_cnt; i++)
  {
    std::string kamu,name;
    int no_ch,no_add;
    std::getline(is,line);
    std::stringstream sb(line);
    sb>>kamu>>name>>no_ch>>no_add;
    std::vector<long> cnt;
    std::vector<std::string> capt;
    std::vector<double> add_counts;
    if(no_ch <= MTESTRawData::BULK_LIMIT)
    {
      //Set the "Single counter"
      //Get channel counts, additional data and captions
      for(int j=0; j<no_ch; j++)
      {
        long data;
        sb>>data;
        cnt.push_back(data);
      }
      for(int j=0; j<no_add; j++)
      {
        double data;
        sb>>data;
        add_counts.push_back(data);
      }
      for(int j=0; j<no_ch+no_add; j++)
      {
        std::string caption;
        sb>>caption;
        capt.push_back(caption);
      }
      addCounterHead(name,no_ch,no_add,capt);
      addCounts(name,cnt, no_add ? &add_counts : 0);
    }
    else
    {
      //This is a bulky dataset -- set the new names
      bulky_name.push_back(name);
      bulky_no_ch.push_back(no_ch);
      bulky_no_add.push_back(no_add);
      sb>>kamu;
      if (no_add > 0)
      {
        std::vector<std::string> acapt;//Additional caption
        for(int j=0; j<no_add; j++)
        {
            double data;
            sb>>data;
            add_counts.push_back(data);
        }
        for(int j=0; j<no_add; j++)
        {
            std::string caption;
            sb>>caption;
            acapt.push_back(caption);
        }
        bulky_add_counts.push_back(add_counts);
        bulky_acapt.push_back(acapt);
      }
    }
  }
  std::getline(is,line);//Skip the comment line of bulky comment line
  int max_bulky_channels = *std::max_element( bulky_no_ch.begin(), bulky_no_ch.end());
  int no_bulky_columns = bulky_name.size();
  std::vector<std::string> bulky_label;
  std::vector< std::vector<long> > bulky_counts( no_bulky_columns, std::vector<long> (0) );
  for(int i=0; i<max_bulky_channels; i++)
  {
    std::string labl;
    is>>labl;
    bulky_label.push_back(labl);
    for(int j=0; j<no_bulky_columns; j++)
    {
        long data;
        is>>data;
        if (data >= 0) bulky_counts[j].push_back(data);
    }
  }
  //Creating proper structures
  int add_idx=0;
  for(int j=0; j<no_bulky_columns; j++)
  {
      std::vector<std::string> bcapt = bulky_label;
      if ( bulky_no_add[j] > 0) bcapt.insert(bcapt.end(), bulky_acapt[j].begin(), bulky_acapt[j].end());//Unify with addional data, if there is any
      addCounterHead(bulky_name[j], bulky_no_ch[j], bulky_no_add[j], bcapt);
      addCounts(bulky_name[j], bulky_counts[j], bulky_no_add[j] ? &bulky_add_counts[add_idx++] : 0);
  }
  return true;
}


bool MTESTRawData::writeSingleCountersHead(std::ofstream &os)
//This method writes head information into file in the case of single counters
{
  if(!os.is_open()) return false;
  for(std::vector<counter_data>::iterator it=counters.begin(); it<counters.end(); it++)
  {
   os<<it->name<<" [";
   for(int i=0; i<it->no_channels; i++) os<<it->caption[i]<<((i<it->no_channels-1) ? "," :"");
   if(it->no_additionals)
   {
     os<<";";
     for(int i=it->no_channels; i<(int)it->caption.size(); i++) os<<it->caption[i]<<((i<(int)it->caption.size()-1) ? "," :"");
   }
   os<<"] ";
  }
  return true;
}

void MTESTRawData::readSingleCounters(std::stringstream &ss)
//This method read head/line information from stringstream in the case of single counters: we assume, that it contains only the Counter related parts for head line. For counts, this is not necessary.
{
  std::string whole(ss.str());
  if(whole.find("[",0)!=std::string::npos)
  {
    //Contains head information
    while(ss.rdbuf()->in_avail() !=0)
    {
      std::string name;
      ss>>name;
      //Get the arguments: []
      std::string argument;
      while(argument.find("]")==std::string::npos)
      {
	std::string aux;
	ss>>aux;
	argument+=aux;
      }
      //Chop off [ ] characters
      size_t start=argument.find("[")+1;
      size_t stop=argument.find("]");
      argument=argument.substr(start,stop-start);
      //Separate it into channel and additional parts
      stop=argument.find(";");
      std::string cnt_part=argument.substr(0,stop);
      start=0;
      int no_channels=0;
      int no_additionals=0;
      std::vector<std::string> caption;
      //Determine the counters
      size_t aux=std::string::npos;
      do{
	aux=cnt_part.find(",",start);
	caption.push_back(cnt_part.substr(start,aux-start));
	no_channels++;
	start=aux+1;
      }while(aux!=std::string::npos);
      if(stop!=std::string::npos)
      {
	//There are auxiliary datasets, too
	std::string aux_part=argument.substr(stop+1,std::string::npos);
	start=0;
	do{
	  aux=aux_part.find(",",start);
	  caption.push_back(aux_part.substr(start,aux-start));
	  no_additionals++;
	  start=aux+1;
	}while(aux!=std::string::npos);
      }
      addCounterHead(name,no_channels,no_additionals,caption);
    }
  }
  else
  {
    //Contains line
    for(std::vector<counter_data>::iterator it=counters.begin(); it!=counters.end(); it++)
    {
      std::vector<long> cnt;
      for(int i=0; i<it->no_channels; i++)
      {
	long data;
	ss>>data;
	cnt.push_back(data);
      }
      if(it->no_additionals > 0)
      {
	std::vector<double> add;
	for(int i=0; i<it->no_additionals; i++)
	{
	  double datad;
	  ss>>datad;
	  add.push_back(datad);
	}
	addCounts(it->name, cnt, &add);
      }
      else addCounts(it->name, cnt, 0);
    }
  }
  return;
}

bool MTESTRawData::writeSingleCounters(std::ofstream &os)
//This method writes counter datasets into file in the case of single counters
{
  if(!os.is_open()) return false;
  for(std::vector<counter_data>::iterator it=counters.begin(); it<counters.end(); it++)
  {
   for(int i=(int)it->count_data.size()-it->no_channels; i<(int)it->count_data.size(); i++) os<<it->count_data[i]<<" ";
   for(int i=(int)it->additionals.size()-it->no_additionals; i<(int)it->additionals.size(); i++) os<<it->additionals[i]<<" ";
  }
  return true;
}

bool MTESTRawData::writeBulkyTimes(std::ofstream &os)
//Writes the times in the case of bulky mode
{
  if(!os.is_open()) return false;
  os<<"# Measurement started(LOCAL): "<<Write_Time(start_time.back())<<" snapshot taken(LOCAL): "<<Write_Time(end_time.back())<<std::endl;
  return true;
}

bool MTESTRawData::readBulkyTimes(std::ifstream &is)
//Read the times in the case of bulky mode
{
  if(!is.is_open()) return false;
  std::string line, kamu;
  std::getline(is,line);
  std::stringstream ss(line);
  for(int i=0; i<3; i++) ss>>kamu;
  start_time.push_back(Read_Time(ss));
  for(int i=0; i<2; i++) ss>>kamu;
  end_time.push_back(Read_Time(ss));
  return true;
}

bool MTESTRawData::writeSingleTimes(std::ofstream &os, bool head)
//Writes the times in the case of single mode. if head is true, head will be written, otherwise data
{
  if(!os.is_open()) return false;
  if(head) os<<" start_date(LOCAL) end_date(LOCAL)";
  else os<<" "<<Write_Time(start_time.back())<<" "<<Write_Time(end_time.back());
  return true;  
}

void MTESTRawData::readSingleTimes(std::stringstream &ss)
//Read the times in the case of single mode
{
  start_time.push_back(Read_Time(ss));
  end_time.push_back(Read_Time(ss));
}

void MTESTRawData::addMotor(std::string name, int pos_pulses, bool standing, double *pos_angles)
//This method adds a motor into the list, which can be standing (T) or that kind of motor, whoose position can be changed after this count. The position of the motors would be written in pulses and angles, if it is possible (pos_angles=0 means: only pulses)
{
  motor_data m;
  m.name=name;
  m.standing=standing;
  m.only_pulses=(pos_angles==0);
  m.pos_pulses.push_back(pos_pulses);
  if(!m.only_pulses) m.pos_angles.push_back(*pos_angles);
  std::vector<motor_data>::iterator it=std::find_if(motors.begin(),motors.end(),[name](const motor_data& md){ return md.name == name; });//Look for the name
  if (it==motors.end()) motors.push_back(m);
  else if (standing) *it=m;//Update the position of standing motors
  else //Update the position of scanning motors
  {
    it->pos_pulses.push_back(pos_pulses);
    if(! (it->only_pulses)) it->pos_angles.push_back(*pos_angles);
  }
  return;
}

bool MTESTRawData::getMotor(std::string name, std::vector<int> &pos_pulses, bool &standing, bool &only_pulses, std::vector<double> &pos_angles)
//This method gets the motor data from list, which can be standing (T) or that kind of motor, whoose position can be changed after this count. The position of the motors would be written in pulses and angles(only_pulses: not possible), if it is possible. Returns with true, if the name is agrees with one among the available motors
{
  std::vector<motor_data>::iterator it=std::find_if(motors.begin(),motors.end(),[name](const motor_data& md){ return md.name == name; });//Look for the name
  if (it==motors.end()) return false;
  else
  {
    pos_pulses=it->pos_pulses;
    standing=it->standing;
    only_pulses=it->only_pulses;
    if(! (it->only_pulses)) pos_angles=it->pos_angles;
    return true;
  }
}

void MTESTRawData::addCounterHead(std::string name, int no_channels, int no_additionals, std::vector<std::string> &ch_names)
//This method adds a counter to the list of counters: the name of the counter, number of channels, number of additional channels and the names of the channels and additionals
{
  counter_data c;
  c.name=name;
  c.no_channels=no_channels;
  c.no_additionals=no_additionals;//Number of additional datasets
  c.caption=ch_names;
  std::vector<counter_data>::iterator it=std::find_if(counters.begin(),counters.end(),[name](const counter_data& cd){ return cd.name == name; });//Look for the name
  if (it==counters.end()) counters.push_back(c);
  else *it=c;//Update
  return;
}

bool MTESTRawData::getCounterHead(std::string name, int &no_channels, int &no_additionals, std::vector<std::string> &ch_add_names)
//This method provides basic information about a counter appeared on the list of counters: the name of the counter, number of channels, number of additional channels and the names of the channels and additionals
{
  std::vector<counter_data>::iterator it=std::find_if(counters.begin(),counters.end(),[name](const counter_data& cd){ return cd.name == name; });//Look for the name
  if (it==counters.end()) return false;
  else
  {
    no_channels=it->no_channels;
    no_additionals=it->no_additionals;
    ch_add_names=it->caption;
    return true;
  }
}

bool MTESTRawData::addCounts(std::string name, std::vector<long> &count_data, std::vector<double> *additionals)
//This method adds the counts into the right counter. Parameters are the name of the counter, the dataset and additional datasets
{
  std::vector<counter_data>::iterator it=std::find_if(counters.begin(),counters.end(),[name](const counter_data& cd){ return cd.name == name; });//Look for the name
  if (it==counters.end()) return false;//Avoid uninitialized case
  else
  {
    if((int)count_data.size()!=it->no_channels) return false;//Number of counters are not conform
    if((it->no_additionals)&&(additionals!=0))
    {
      if((int)additionals->size()!=it->no_additionals) return false;//Number of additional counters are not conform
      it->additionals.insert(it->additionals.end(),additionals->begin(),additionals->end());
    }
    it->count_data.insert(it->count_data.end(),count_data.begin(),count_data.end());
    return true;
  }
}

bool  MTESTRawData::getCounts(std::string name, std::vector<long> &count_data, std::vector<double> &additionals)
//This method get the data belonging to a counter. Parameters are the name of the counter, the dataset and additional datasets
{
  std::vector<counter_data>::iterator it=std::find_if(counters.begin(),counters.end(),[name](const counter_data& cd){ return cd.name == name; });//Look for the name
  if (it==counters.end()) return false;//Avoid uninitialized case
  else
  {
    count_data=it->count_data;
    if(it->no_additionals) additionals=it->additionals;
    return true;
  }
}

std::string MTESTRawData::get_comment(std::ifstream &is)
//This method cuts the comment from a line started wits some information
{
  std::string line;
  std::getline(is,line);
  return line.replace(0,line.find(":")+2,"");
}

void MTESTRawData::clear_comments()
//This routine clears the comments
{
  comment_sample="";
  comment_ms_collimator="";
  comment_ms_slits="";
  comment_s_holder="";
  comment_s_environment="";
  comment_sd_slits="";
  comment_d_position="";
}

void MTESTRawData::clear_motors(bool empty, bool onlyscanning)
//Clears the list of motors: if empty is true, it clears only the positional records, onlyscanning does this task only at scanning motors
{
  if(!empty) motors.clear();
  else std::for_each(
    motors.begin(),motors.end(),
     [onlyscanning](motor_data &md)
     { 
       if(!(md.standing && onlyscanning))
       {
	 md.pos_pulses.clear();
	 md.pos_angles.clear();
       }
     }
    );
  return;
}

void MTESTRawData::clear_counters(bool empty)
//Clears the list of counters: if empty is true, it clears only the positional records
{
  if(!empty) counters.clear();
  else std::for_each(counters.begin(),counters.end(),[](counter_data &cd){ cd.count_data.clear(); cd.additionals.clear();});
}

std::time_t Read_Time(std::stringstream &ss)
//Reads time data and converts it to standard format
{
  //Remove - and : signs
  std::string date,time;
  ss>>date>>time;
  std::size_t pos;
  while((pos=date.find("-",0))!=std::string::npos) date.replace(pos,1," ");
  while((pos=time.find(":",0))!=std::string::npos) time.replace(pos,1," ");
  std::stringstream sb(date+" "+time);
  //Get into tm structure
  tm timed;
  int data;
  sb>>data;
  timed.tm_year=data-1900;
  sb>>data;
  timed.tm_mon=data-1;
  sb>>timed.tm_mday;
  sb>>timed.tm_hour;
  sb>>timed.tm_min;
  sb>>timed.tm_sec;
  timed.tm_isdst=-1;
  return mktime(&timed);
}

std::string Write_Time(std::time_t &time)
//This function converts time to output
{
  char buffer[80];
  strftime(buffer,80,"%F %T",std::localtime(&time));
  return std::string(buffer);
}
