/******************************************
 * Modified by Laszlo Temleitner: 14.03.2018: new structure format allows to read and write more than one bulky dataset as a new column. TODO: test
 */
#include <algorithm>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <ctime>

#include <unistd.h>

class MTESTRawData
/*This class describes and provides an interface to write and read raw data files measured on MTEST diffractometer. This class designed
 to cooperate both data acquisition side and for data evaluation, as well*/
{
  private:
      //Private members
      static const int BULK_LIMIT = 10;//Maximal limit of counter items to classify it as non-BULK -> this case a line is appended insted of column 
  public:
    struct motor_data
    {
      std::string name;
      bool standing;//t: motor is not moving, f: motor would be a moving one
      bool only_pulses;//If it is true, only pulses will be provided
      std::vector<int> pos_pulses;//position in pulses
      std::vector<double> pos_angles;//position in angles
    };
    struct counter_data
    {
      std::string name;//Name of the counter
      int no_channels;//Number of channels
      int no_additionals;//Number of additional datasets
      std::vector<std::string> caption;//Caption of the counts, if 2<=no_channels<=10, or additional, if no_channels>=10
      std::vector<long> count_data;//Stores the counts belonging to the channels
      std::vector<double> additionals;//The value of additional counts, if any
    };
    struct run_data
    {
      std::string filename;//Main filename
      int run_no;//Run number
      int item_no;//Item number
      int snapshots;//Number of snapshots
    };
    const static int actual_version=1;//Actual version of the raw file format
    std::vector<motor_data> motors;//Stores the names of the motors
    std::vector<counter_data> counters;//Stores the names of counters
    std::vector<run_data> runs;//Stores the run datasets
    
    //Private methods
    
    void clear_comments();//This routine clears the comments
    void clear_motors(bool empty=true, bool onlyscanning=true);//Clears the list of motors: if empty is true, it clears only the positional records, onlyscanning does this task only at scanning motors
    void clear_counters(bool empty=true);//Clears the list of counters: if empty is true, it clears only the positional records
    void clear_time(){start_time.clear(); end_time.clear();}//Clears all of the times
    void clear_runs(){runs.clear();}//Clears runs data
  private:
    //Private methods
    std::string get_comment(std::ifstream &is);//This method cuts the comment from a line started wits some information
    bool writeHead(std::ofstream &os, bool bulky);//Writes the head information to the requested filestream (bulk:t or single mode:f)
    bool readHead(std::ifstream &is, bool &bulky);//Reads the head information from the requested filestream, bulky or single mode
    bool writeStandingMotors(std::ofstream &os);//Writes the list of standing motors into filestream
    bool readStandingMotors(std::ifstream &is);//Reads the list of standing motors from the filestream
    bool writeBulkyScanningMotors(std::ofstream &os);//Writes the list and position of scanning motors to a file in bulky mode
    bool readBulkyScanningMotors(std::ifstream &is);//Set the list and position of scanning motors from a file in bulky mode
    bool writeSingleScanningMotors(bool head, std::ofstream &os);//Writes the list and position of scanning motors to a file in single mode. If the head is true, only head information will be written, otherwise the dataset
    void readSingleScanningMotors(std::stringstream &ss);//Reads the list and position of scanning motors from a stringstream single mode.: we assume, it contains only motor related parts
    bool writeBulkyCounters(std::ofstream &os);//This method writes the datasets of counters into an output file
    bool readBulkyCounters(std::ifstream &is);//This method creates datasets of counters from input file
    bool writeSingleCountersHead(std::ofstream &os);//This method writes head information into file in the case of single counters
    void readSingleCounters(std::stringstream &ss);//This method read head/line information from stringstream in the case of single counters: we assume, that it contains only the Counter related parts for head line. For counts, this is not necessary.
    bool writeSingleCounters(std::ofstream &os);//This method writes counter datasets into file in the case of single counters
    bool writeBulkyTimes(std::ofstream &os);//Writes the times in the case of bulky mode
    bool readBulkyTimes(std::ifstream &is);//Read the times in the case of bulky mode
    bool writeSingleTimes(std::ofstream &os, bool head=false);//Writes the times in the case of single mode. if head is true, head will be written, otherwise data
    void readSingleTimes(std::stringstream &ss);//Read the times in the case of single mode
  
  public:
    //Public members
    std::string comment_sample;//Comments about the sample
    std::string comment_ms_collimator;//Comment on collimator setup applied between the monochromator and sample
    std::string comment_ms_slits;//Comment on the setup of slits applied between the monochromator and sample
    std::string comment_s_holder;//Comment on the sample holder
    std::string comment_s_environment;//Comment on the applied sample environment
    std::string comment_sd_slits;//Comment on the setup of slits applied between the sample and the detector
    std::string comment_d_position;//Comment on the position of the detector
    std::vector<std::time_t> start_time, end_time;//Start and end time of the measurement
    
    
    MTESTRawData();//Initialization of the class variables
    ~MTESTRawData(){clear_motors(); clear_counters(); clear_time(); clear_runs();};//Destructor
    
    void addMotor(std::string name, int pos_pulses, bool standing=true, double *pos_angles=0);//This method adds a motor into the list, which can be standing (T) or that kind of motor, whoose position can be changed after this count. The position of the motors would be written in pulses and angles, if it is possible (pos_angles=0 means: only pulses)
    bool getMotor(std::string name, std::vector<int> &pos_pulses, bool &standing, bool &only_pulses, std::vector<double> &pos_angles);//This method gets the motor data from list, which can be standing (T) or that kind of motor, whoose position can be changed after this count. The position of the motors would be written in pulses and angles(only_pulses: not possible), if it is possible. Returns with true, if the name is agrees with one among the available motors
    void addCounterHead(std::string name, int no_channels, int no_additionals, std::vector<std::string> &ch_names);//This method adds a counter to the list of counters: the name of the counter, number of channels, number of additional channels and the names of the channels and additionals
    bool getCounterHead(std::string name, int &no_channels, int &no_additionals, std::vector<std::string> &ch_add_names);//This method provides basic information about a counter appeared on the list of counters: the name of the counter, number of channels, number of additional channels and the names of the channels and additionals
    bool addCounts(std::string name, std::vector<long> &count_data, std::vector<double> *additionals); //This method adds the counts into the right counter. Parameters are the name of the counter, the dataset and additional datasets
    bool getCounts(std::string name, std::vector<long> &count_data, std::vector<double> &additionals);//This method get the data belonging to a counter. Parameters are the name of the counter, the dataset and additional datasets
    void setTimes(std::time_t &start, std::time_t &end){start_time.push_back(start); end_time.push_back(end);};//Set start and end times
    void getTimes(std::vector<std::time_t> &start, std::vector<std::time_t> &end){start=start_time; end=end_time;};//Get start and end times
    bool WriteNewSnapshot(bool clr_prev=true);//This function creates a new snapshot from the last saved datasets, if clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared
    bool WriteNewItem(bool clr_prev=true);//This function writes a new item from the last saved datasets, if clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared
    bool WriteNewRun(bool clr_prev=true);//This function writes the start of a new run from the last saved datasets, if clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared
    bool WriteNewSample(std::string filename, bool clr_prev=true, int start_run_no=1);//This function writes new sample data to the requested main file at the first run, first item and first snapshot.  If clr_prev is true, then the last saved counters, times, and scanning motors positions will be cleared. The start_run_no specifies the start run number
    bool write(std::ofstream &os, bool bulky, bool head);//This method writes the current datasets to file in bulky(t) or in single(f). In single, this can be just an update(f) or write header(t) part, as well.
    bool read(std::ifstream &is);//Reads the input stream and store the dataset read in
};

std::time_t Read_Time(std::stringstream &ss);//Reads time data and converts it to standard format
std::string Write_Time(std::time_t &time);//This function converts time to output
