#include "Mesy.h"


uint16_t Mesy::MesyCommand::Commandbuffernumber=0;//Initialization of commandbuffer number
uint8_t Mesy::MesyCommand::Master_id=0;//Sets master id to zero
uint64_t Mesy::MesyCommand::Lastcommandtime;//Timestamp of the last sent command

//CellEvents

Mesy::CellEvents::CellEvents()
//Constructor: initializes the structure
{
    //Initialize the structure
    std::for_each(auxtimers.begin(), auxtimers.end(),[](double &a){a=0.0;});
    std::for_each(trigger_source.begin(), trigger_source.end(),[](Mesy::TRIGGER_SRC &a){a=Mesy::TRG_NO;});
    std::for_each(compare_register.begin(), compare_register.end(),[](uint16_t &a){a=0;});
}

bool Mesy::CellEvents::SetPeriodicalCheck(const Mesy::INPUT_TYP input_cell, const double period_ms)
//Check the number of counts on input_cell periodically in each 0.01..655.36ms
{
    //Check conformity
    if(!input_cell_conformity(input_cell)) return false;
    double period = floor(period_ms*100.0)/100.0;//Cutting the result to 2 decimal digits
    if((period_ms<0.01)||(period_ms>655.36))
    {
        std::cerr<<"ERROR (CellEvents::SetPeriodicalCheck): The period ("<<period_ms<<") is out from valid range (0.01..655.36ms)!"<<std::endl;
        return false;
    }
    //Looking for already set timer
    int timer=0;
    while ((timer < Mesy::NAUXTIMERS)&&((auxtimers[timer]!=period)||(auxtimers[timer]!=0.00))) timer++;
    if (timer == Mesy::NAUXTIMERS)
    {
        std::cerr<<"ERROR (CellEvents::SetPeriodicalCheck): The present period ("<<period_ms<<") does not conform any of the allocated timers! Skipping the present one!"<<std::endl;
        return false;
    }
    auxtimers[timer] = period;//Registering the time
    trigger_source[input_cell] = static_cast<Mesy::TRIGGER_SRC>(timer+static_cast<int>(Mesy::TRG_ATIMER_0));
    compare_register[input_cell] = 0;
    return true;
}

bool Mesy::CellEvents::SetEachEdge(const Mesy::INPUT_TYP input_cell)
//Check on its own rising edge
{
    if(!input_cell_conformity(input_cell)) return false;
    trigger_source[input_cell] = Mesy::TRG_COMPREG;
    compare_register[input_cell] = Mesy::CMP_TRIG_RISEDGE;
    return true;
}

bool Mesy::CellEvents::SetEachRearDigitEdge(const Mesy::INPUT_TYP input_cell, bool digital1t_digital2f)
//Check on by the rising edge of digital input 1(t) or 2(f)
{
    if(!input_cell_conformity(input_cell)) return false;
    trigger_source[input_cell] = (digital1t_digital2f ? Mesy::TRG_DIGIN_1 : Mesy::TRG_DIGIN_2);
    compare_register[input_cell] = 0;
    return true;
}

bool Mesy::CellEvents::SetEachTwoPowerCnts(const Mesy::INPUT_TYP input_cell, int power)
//Create events at each x mod 2^power = 1 counts
{
    //Check conformity
    if(!input_cell_conformity(input_cell)) return false;
    if((power>Mesy::CMP_TRIG_COVERFLOW)||(power<0))
    {
        std::cerr<<"ERROR (CellEvents::SetEachTwoPowerCnts): The power of 2 ("<<power<<") is out from valid range (0.."<<Mesy::CMP_TRIG_COVERFLOW<<")!"<<std::endl;
        return false;
    }
    trigger_source[input_cell] = Mesy::TRG_COMPREG;
    compare_register[input_cell] = power;
    return true;
}


bool Mesy::CellEvents::input_cell_conformity(const INPUT_TYP input_cell)
//Check is within input range
{
    if(input_cell>=Mesy::INP_NO)
    {
        std::cerr<<"ERROR (CellEvents::input_cell_conformity): The input cell no ("<<static_cast<int>(input_cell)<<") is out from valid range (0.."<<Mesy::INP_NO-1<<")!"<<std::endl;
        return false;
    }
    else return true;
}

//MesyDAQ

Mesy::MesyDAQ::MesyDAQ(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes,\
 const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events, std::string raw_out_name):\
 allow_parameter_consideration(header_parameters), cell_events(events), raw_name(raw_out_name)
//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices and the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped)
{
    //Initialize the allowed variables tables
    no_active_tubes=0;
    int indx_header=0;
    int indx_event=0;
    for(size_t i=0; i<active_tubes.size(); i++)
    {
        //Neutron events related initialization
        std::array<bool, Mesy::NMCPDTUBES> trscp;
        std::array<int, Mesy::NMCPDTUBES> dindx;
        for(size_t j=0; j<(active_tubes[i]).size(); j++)
        {
            for(size_t k=0; k<(active_tubes[i][j]).size(); k++)
            {
                trscp[twoD2oneD(j,k)] = active_tubes[i][j][k];
                dindx[twoD2oneD(j,k)] = (active_tubes[i][j][k] ? no_active_tubes++ : -1);
            }
        }
        allow_neutron_collection.push_back(trscp);
        neutron_index.push_back(dindx);
        //Initializing header parameters to watch
        std::array<int, Mesy::NPARAMS> pars;
        for(size_t j=0; j<Mesy::NPARAMS; j++) pars[j] = ((allow_parameter_consideration[i][j]==Mesy::PAS_BLANK) ? -1 : indx_header++);
        parameter_index.push_back(pars);
        //Initialize events
        std::array<int, Mesy::INP_NO> cevent;
        for(size_t j=0; j<Mesy::INP_NO; j++) cevent[j] = (((cell_events[i]).trigger_source[j]==Mesy::TRG_NO) ? -1 : indx_event++);
        event_index.push_back(cevent);
    }
    //Initialize acceptance of data, header parameters and events
    accept_neutron = (no_active_tubes>0);
    accept_parameters = (indx_header>0);
    accept_events = (indx_event>0);
    return;
}

//registering events
void Mesy::MesyDAQ::RegisterParams(const std::array<uint64_t,NPARAMS> params, const uint64_t time, const int mcpd_id)//Registers the changed header parameters happened in time at mcpd_id interface
{
    if(accept_parameters)
    {
        for(int i=0; i<NPARAMS; i++)
        {
            if (allow_parameter_consideration[mcpd_id][i]!=Mesy::PAS_BLANK) check_parameter_and_store( parameter_index[mcpd_id][i], params[i],time);
        }
    }
    return;
}

void Mesy::MesyDAQ::RegisterTriggerEvent(const Mesy::TRIGGER_SRC trigger_id, const INPUT_TYP data_id, const uint32_t data_value, const uint64_t time, const int mcpd_id)//Registers a trigger event on data_id by data_value triggered by trigger_id at time and at the mcpd_id interface
{
    if(accept_events && (event_index[mcpd_id][static_cast<int>(data_id)]!=-1)) check_event_and_store(event_index[mcpd_id][static_cast<int>(data_id)], trigger_id, data_value, time);
    return;
}

void Mesy::MesyDAQ::RegisterNeutronEvent(const int tube_no, const int amplitude, const int position, const uint64_t time, const int mcpd_id)//Registers a neutron event happened at tube [3 bit of module| 3bit of slot of module] of mcpd_id interface at time 
{
    if(accept_neutron && allow_neutron_collection[mcpd_id][tube_no]) check_neutron_and_store(neutron_index[mcpd_id][tube_no], amplitude, position, time);
    return;
}

std::vector< std::array<std::array<bool, Mesy::NTUBE>, Mesy::NMPSD8>> Mesy::MesyDAQ::GetActiveDetectors() const
//Returns with the arrays containing active detectors by define them as true [mcpd_id][mpsd_no][tube]
{
    std::vector< std::array<std::array<bool, Mesy::NTUBE>, Mesy::NMPSD8>> result;
    for(int i=0; i<=GetMaxMCPDId(); i++)
    {
        std::array<std::array<bool, Mesy::NTUBE>, Mesy::NMPSD8> one_mcpd;
        for(int j=0; j<Mesy::NMPSD8; j++)
        {
            for(int k=0; k<Mesy::NTUBE; k++) one_mcpd[j][k] = allow_neutron_collection[i][twoD2oneD(j,k)];
        }
        result.push_back(one_mcpd);
    }
    return result;
}

bool  Mesy::MesyDAQ::write_raw()
//Writes raw datasets into file
{
    if(raw_name!="")
    {
        for(int i=0; i<GetNoActiveDetectors(); i++)
        {
            std::string fn = raw_name+"_raw"+std::to_string(i+1)+".dat";
            std::ofstream os(fn.c_str());
            if(!os.is_open())
            {
                std::cerr<<"ERROR(MesyDAQ::write_raw) at open the \""<<fn<<"\" to write!"<<std::endl;
                return false;
            }
            else write_rawfile(i, os);
        }
    }
    return true;
}


//MesyDAQFixedDiscriminators

Mesy::MesyDAQFixedDiscriminators::MesyDAQFixedDiscriminators(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes,\
const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events,\
const std::vector< std::array<int,2> > &discr, const std::string raw_out_name):\
MesyDAQ(active_tubes, header_parameters, events, raw_out_name), discriminator(discr)
//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices, amplitude discriminators (LOW,HIGH) for each active tubes, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped)
{
    if(GetNoActiveDetectors()!=static_cast<int>(discriminator.size()))
    {
        std::cerr<<"ERROR (MesyDAQFixedDiscriminators::MesyDAQFixedDiscriminators): the size of discriminator array ("<<discriminator.size()<<") does not correspond to the number of active detector tubes ("<<GetNoActiveDetectors()<<")!"<<std::endl;
    }
    return;
}

void Mesy::MesyDAQFixedDiscriminators::ModifyDiscriminator(const int mcpd_id, const int mpsd_no, const int tube_no, const int low, const int high)
//Changes the software discriminator setup of a given tube
{
    if((mcpd_id>GetMaxMCPDId())||(mcpd_id<0))
    {
        std::cerr<<"ERROR (MesyDAQFixedDiscriminators::ModifyDiscriminator): The current mcpd_id ("<<mcpd_id<<") is out from valid range (0.."<<GetMaxMCPDId()<<")!"<<std::endl;
        return;
    }
    if((mpsd_no>=Mesy::NMPSD8)||(mpsd_no<0))
    {
        std::cerr<<"ERROR (MesyDAQFixedDiscriminators::ModifyDiscriminator): The current mpsd_no ("<<mpsd_no<<") is out from valid range (0.."<<Mesy::NMPSD8-1<<")!"<<std::endl;
        return;
    }
    if((tube_no>=Mesy::NTUBE)||(tube_no<0))
    {
        std::cerr<<"ERROR (MesyDAQFixedDiscriminators::ModifyDiscriminator): The current tube_no ("<<tube_no<<") is out from valid range (0.."<<Mesy::NTUBE-1<<")!"<<std::endl;
        return;
    }
    if((low<0)||(high<0)||(low>=Mesy::ASIZE)||(high>=Mesy::ASIZE))
    {
        std::cerr<<"ERROR (MesyDAQFixedDiscriminators::ModifyDiscriminator): The current low ("<<low<<") or high ("<<high<<") is out from valid range (0.."<<Mesy::ASIZE-1<<")!"<<std::endl;
        return;
    }
    int internal_indx = getInternalIndex(mcpd_id, mpsd_no, tube_no);
    discriminator[internal_indx][LOW]  = (high>low) ? low : high;
    discriminator[internal_indx][HIGH] = (high>low) ? high : low;
    return;
}

void Mesy::MesyDAQFixedDiscriminators::check_neutron_and_store(const int internal_indx, const int amplitude, const int position, const uint64_t time)
//Apply discriminators and store a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
{
    if((discriminator[internal_indx][LOW]<=amplitude)&&(discriminator[internal_indx][HIGH]>=amplitude)) store_neutron(internal_indx, amplitude, position, time);//Real data
    else store_neutron_out_of_discriminator(internal_indx, amplitude, position, time);//Storing rejected data for statistical reason
    return;
}

//MesyDAQCWSteadySample
        
Mesy::MesyDAQCWSteadySample::MesyDAQCWSteadySample(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events, const std::vector< std::array<int,2> > &discr, const std::string raw_out_name): MesyDAQFixedDiscriminators(active_tubes, header_parameters, events, discr, raw_out_name)
//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices, amplitude discriminators (LOW,HIGH) for each active tubes, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped)
{
    //Initialize data collector arrays
    for(int i=0; i<GetNoActiveDetectors(); i++)
    {
        std::array<Mesy::MESY_CNT,Mesy::POSAMPSIZE> arr;
        std::array<Mesy::MESY_CNT,Mesy::ASIZE> arra;
        std::array<Mesy::MESY_CNT,Mesy::PSIZE> arrp;
        arr.fill(0);
        arra.fill(0);
        arrp.fill(0);
        recorded_data.push_back(arr);
        amplitudes_data.push_back(arra);
        position_data.push_back(arrp);
    }
    return;
}

bool Mesy::MesyDAQCWSteadySample::GetRecordedData(Mesy::MESY_CNT *data, const int tube_no)
//Returns with the actually recorded dataset for a specific tube or all of the tubes -- it is assumed, that the required data is already allocated the order is all position data then all of the amplitudes
{
    if(tube_no==-1)
    {
        int position_shift = GetNoActiveDetectors() * Mesy::PSIZE;
        for(int i=0; i<GetNoActiveDetectors(); i++)
        {
            std::copy(position_data[i].begin(),position_data[i].end(), data+i*Mesy::PSIZE);
            std::copy(amplitudes_data[i].begin(),amplitudes_data[i].end(), data+position_shift+i*Mesy::ASIZE);
        }
    }
    else if ((tube_no > -1) && ( tube_no < GetNoActiveDetectors()))
    {
        std::copy(position_data[tube_no].begin(),position_data[tube_no].end(), data);
        std::copy(amplitudes_data[tube_no].begin(),amplitudes_data[tube_no].end(), data+Mesy::PSIZE);
    }
    else return false;
    return true;
}

void Mesy::MesyDAQCWSteadySample::store_neutron_out_of_discriminator(const int internal_indx, const int amplitude, const int position, const uint64_t time)
//Stores a rejected neutron event by discriminator happened on internal_indx tube index with signal amplitude at position and at a given time (position data will not be stored)
{
    recorded_data[internal_indx][Mesy::ASIZE*position+amplitude]++;
    amplitudes_data[internal_indx][amplitude]++;
    return;
}

void Mesy::MesyDAQCWSteadySample::store_neutron(const int internal_indx, const int amplitude, const int position, const uint64_t time)
//Stores a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
{
    store_neutron_out_of_discriminator(internal_indx, amplitude, position, time);//Register the data (except position information)
    position_data[internal_indx][position]++;
    return;
}

void Mesy::MesyDAQCWSteadySample::write_rawfile(int internal_tube_no, std::ofstream &os)
//Writes the raw dataset of internal_tube_no to os stream
{
    os<<"#down: position, left: amplitudes"<<std::endl;
    for(int i=0; i<Mesy::PSIZE; i++)
    {
        for(int j=0; j<Mesy::ASIZE; j++) os<<" "<<recorded_data[internal_tube_no][i*Mesy::ASIZE+j];
        os<<std::endl;
    }
    return;
}

void Mesy::MesyDAQCWSteadySample::reset()
//This method resets the recorded array
{
    std::for_each(recorded_data.begin(),recorded_data.end(),[](std::array<Mesy::MESY_CNT,POSAMPSIZE> &a){a.fill(0);});
    std::for_each(amplitudes_data.begin(),amplitudes_data.end(),[](std::array<Mesy::MESY_CNT,ASIZE> &a){a.fill(0);});
    std::for_each(position_data.begin(),position_data.end(),[](std::array<Mesy::MESY_CNT,PSIZE> &a){a.fill(0);});
    return;
}

//MesyDAQMTESTPowderGatedByMonitorPulse
Mesy::MesyDAQMTESTPowderGatedByMonitorPulse::MesyDAQMTESTPowderGatedByMonitorPulse(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array<int,2> > &software_discriminators, const Mesy::PARAMETER_SRC pulse_input_line, const int mcpd_id_no, const std::string raw_out_name, const bool paramt_eventf, const int max_mcpd): MesyDAQCWSteadySample(active_tubes, {{ {paramt_eventf ? pulse_input_line : PAS_BLANK, PAS_BLANK, PAS_BLANK, PAS_BLANK} }}, {Mesy::CellEvents()}, software_discriminators, raw_out_name), monitor_parametert_eventf(paramt_eventf)
//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD][tube:0..<NTUBE] indices, amplitude discriminators (LOW,HIGH) for each active tubes, input line of the monitor start/stop signal, the mcpd module id, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped), the parameter/event setup, finally the highest mcpd id
{
    if((max_mcpd!=0)||(!paramt_eventf))
    {
        std::cerr<<"ERROR (constructor of MesyDAQMTESTPowderGatedByMonitorPulse): only max_mcpd=0 and paramt_eventf=true setup is supported!"<<std::endl;
        return;
    }
    if(monitor_parametert_eventf) monitor_par_event_no = getParameterIndex(mcpd_id_no, pulse_input_line);
    else monitor_par_event_no = getCellInternalIndex(mcpd_id_no, static_cast<int>(pulse_input_line));
    accept_neutron=false;//Default: monitor counter is in OFF
    return;
}

//MesyDAQCWConstRotation

Mesy::MesyDAQCWConstRotation::MesyDAQCWConstRotation(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events, const std::vector< std::array<int,2> > &discr, const int rot_axis_resolution, const std::array<int,2> mcpdno_cellno_of_fullrot_indicator, const std::string raw_out_name, const int buffersize): MesyDAQFixedDiscriminators(active_tubes, header_parameters, events, discr, raw_out_name), tube_no(buffersize,0), buffer_position(buffersize, 0), buffer_time_since_event(buffersize,0)
//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices, amplitude discriminators (LOW,HIGH) for each active tubes, resolution of the rotation axis, {mcpd,cellno} of event counter indicating 360 deg rotation, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped), size of the temporary buffer
{
    rotation_resolution = rot_axis_resolution;
    rotation_event_index = getCellInternalIndex(mcpdno_cellno_of_fullrot_indicator[0], mcpdno_cellno_of_fullrot_indicator[1]);
    if(rotation_event_index<0)
    {
        std::cerr<<"ERROR (MesyDAQCWConstRotation::MesyDAQCWConstRotation): the cell in mcpd no. "<<mcpdno_cellno_of_fullrot_indicator[0]<<" and cell no. "<<mcpdno_cellno_of_fullrot_indicator[1]<<" does not have got a registered event!"<<std::endl;
        return;
    }
    //Initialize the data storage arrays
    for(int i=0; i<GetNoActiveDetectors(); i++)
    {
        std::vector<std::array<Mesy::MESY_CNT,Mesy::PSIZE>> twod;//Matrix of one tube
        for(int j=0; j<rotation_resolution; j++)
        {
            std::array<Mesy::MESY_CNT,Mesy::PSIZE> pos;//Position dataset for one rotation angular segment
            pos.fill(0);
            twod.push_back(pos);
        }
        position_data.push_back(twod);
    }
    //Initialization of local variables
    first_not_full_rotation = true;
    last_event_date = 0;
    actual_no_data = 0;
    return;
}

void Mesy::MesyDAQCWConstRotation::check_event_and_store(const int internal_indx, const Mesy::TRIGGER_SRC trigger_id, const uint32_t data_value, const uint64_t time)
//Apply discriminators and store a counter event happened on internal_indx for the effect of trigger_id by data_value at time -- this function implements only the rotation event part
{
    if(internal_indx == rotation_event_index)
    {
        if(!first_not_full_rotation)
        {
            //Discretize and update
            discretize_buffer(time);
            last_event_date = time;
        }
        else first_not_full_rotation = false;
    }
    return;
}

void Mesy::MesyDAQCWConstRotation::discretize_buffer(uint64_t event_time)
//This method discretizes the buffered datasets based on elapsed time and rotation resolution and resets the buffer
{
    uint64_t elapsed_time = event_time - last_event_date;
    double delta = static_cast<double>(elapsed_time)/rotation_resolution;
    for (int i=0; i < actual_no_data; i++) position_data [tube_no[i]] [static_cast<Mesy::MESY_CNT>(buffer_time_since_event[i]/delta)%rotation_resolution] [buffer_position[i]]++;
    actual_no_data = 0;
    return;
}

void Mesy::MesyDAQCWConstRotation::store_neutron(const int internal_indx, const int amplitude, const int position, const uint64_t time)
//Stores a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
{
    if(!first_not_full_rotation)
    {
        tube_no[actual_no_data]                 = internal_indx;
        buffer_time_since_event[actual_no_data] = time-last_event_date;
        buffer_position[actual_no_data++]       = position;
    }
    return;
}

//MesyCommand

Mesy::MesyCommand::MesyCommand(UDPClient &command, UDPServerTimeOut &data, MesyDAQ &msdaq, std::vector<uint8_t> &thresholds, std::vector<uint8_t> &gains, const uint8_t master)//Constructor: sets the UDP communication objects for command (might be the Mesytec device or data collector computer, which act as relay) and for data channels (to the data collector computer), the data acquisition object (doing the high-level acquisition), thresholds (discriminator) for each MPSD modules,  gains for each active tubes and the master channel number.
{
    
    Master_id=master;
    Udpcommand=&command;
    Udpdata=&data;
    Mesydaq=&msdaq;
    //std::atomic_init(&Daq,false);
    Daq=false;
    module_thresholds = thresholds;
    active_tube_gain = gains;
  
    //Set the non-changing (and initial values) of the sendbuffer
    Sendbuffer[MB_BUFTYP]=(0x1<<15);//Command buffer
    Sendbuffer[MB_LENHED]=10;//Length of the header
    Sendbuffer[MB_TIMSLO]=0;//Timestamp low
    Sendbuffer[MB_TIMSMI]=0;//Timestamp mid
    Sendbuffer[MB_TIMSHI]=0;//Timestamp hi
}

bool Mesy::MesyCommand::UpdateSetup()
//Sets up the DAQ system based on Mesydaq
{
    std::vector<std::array<Mesy::PARAMETER_SRC, Mesy::NPARAMS>> parameter_list = Mesydaq->GetHeaderParameterSetup();//Contains the parameter source in [mcpd:0..hiindex][param_no:0..<NPARAMS] indices
    std::vector<int> ids = Udpcommand->GetFarIDs();//ID's
    std::vector<Mesy::CellEvents> events = Mesydaq->GetEventSetup();
    std::vector< std::array<std::array<bool, Mesy::NTUBE>, Mesy::NMPSD8>> active_tubes = Mesydaq->GetActiveDetectors();
    bool retval =true;
    int active_module_indx = 0;
    
    for (int id : Udpcommand->GetFarIDs())
    {
      //MCPD
      //Setting up IP and communication
      std::string mcpd_ip = Udpcommand -> GetFarIP(id);//("0");
      std::string command_ip = Udpcommand -> GetLocalIP();//("00")
      std::string data_ip    = Udpdata    -> GetLocalIP();//("00");
      uint16_t cmd_port = Udpcommand -> GetLocalPort(), data_port = Udpdata ->GetLocalPort();
      if(retval) retval = MCPDCommIPPortSetup(mcpd_ip, command_ip, data_ip, cmd_port, data_port, id);
      if(Mesy::debug&&retval) std::cout<<"#MCPD: "<<mcpd_ip<<", Command: "<<command_ip<<"/"<<cmd_port<<",  Data: "<<data_ip<<"/"<<data_port<<std::endl;
      if(retval) retval=MCPDDAQStop();//Stop the acquisition (if the DAQ program stopped earlier accidentally)
      if(retval) retval=MCPDCommSetTX(CAP_TPA, id);//Set up transmission to time-position-amplitude
      if(Mesy::debug&&retval) std::cout<<"#MCPD: "<<id<<" has been set to TIME-POSITION-AMPLITUDE transmission"<<std::endl;
      //General setup
      uint64_t time=0;//master time
      uint16_t run_no=0;//Run number
      if(retval) retval=MCPDGenSetTiming(id == Master_id, id, id != Master_id);//Sets up the synchronization
      if(Mesy::debug&&retval) std::cout<<"#MCPD: "<<id<<" has been set to "<<((id==Master_id) ? "MASTER" : "SLAVE")<<" mode"<<std::endl;
      if(retval) retval=MCPDGenSetClock(time);//Sets the master MCPD clock
      if(Mesy::debug&&retval) std::cout<<"#Master clock has been setting to zero"<<std::endl;
      if(retval) retval=MCPDGenSetRunNo(run_no);//Sets the run number
      if(Mesy::debug&&retval) std::cout<<"#Master: run number has been set to 0"<<std::endl;
      //Set the parameters
      for(int j=0; j<Mesy::NPARAMS; j++)
      {
          if(retval&&(parameter_list[id][j] != PAS_BLANK))
          {
             retval=MCPDCATParameterSetSource(j, parameter_list[id][j], id);
             if(Mesy::debug&&retval) std::cout<<"#MCPD: "<<id<<", parameter no "<<j<<" has been set to parameter source "<<static_cast<int>(parameter_list[id][j])<<std::endl;
          }
      }
      for(int j=0; j<Mesy::NAUXTIMERS; j++)
      {
          if(retval) retval = MCPDCATSetTimer(j, events[id].auxtimers[j], id);//Set up the auxiliary timers
          if(Mesy::debug&&retval) std::cout<<"#MCPD: "<<id<<", timer no "<<j<<" has been setting to "<<events[id].auxtimers[j]<<"ms"<<std::endl;
      }
      for(int j=0; j<static_cast<int>(Mesy::INP_NO); j++)
      {
          if(retval) retval = MCPDCATEventSetup(static_cast<Mesy::INPUT_TYP>(j), events[id].trigger_source[j], events[id].compare_register[j], id);//set up events
          if(Mesy::debug&&retval) std::cout<<"#MCPD: "<<id<<", counter no "<<j<<" has been setting to trigger source no. "<<static_cast<int>(events[id].trigger_source[j])<<" with compare register: "<<static_cast<int>(events[id].compare_register[j])<<std::endl;
      }
      //Get number of connected modules
      std::string micro, fpga;
      std::array<bool,Mesy::NMPSD8> connected_modules;
      if(retval) retval = MCPDGenGetAbout(micro, fpga, connected_modules, id);
      if(Mesy::debug&&retval)
      {
          std::cout<<"#MCPD: "<<id<<", connected modules: ";
          std::for_each( connected_modules.begin(), connected_modules.end(), [](bool n){std::cout<<" "<<n;});
          std::cout<<std::endl;
      }
      for(int i=0; i<Mesy::NMPSD8; i++)
      {
          if((retval)&&(std::any_of(active_tubes[id][i].begin(),active_tubes[id][i].end(),[](bool active){return active;})))
	  {
              if(!connected_modules[i])
              {
                 std::cerr<<"ERROR: There is a discrepancy at mcpd_id: "<<id<<", module no "<<i<<": according to scan, the module is not present, although it has been requested!"<<std::endl;
                 return false;
              }
              else
              {
                retval = MPSDDiscriminator(module_thresholds[active_module_indx], i, id);//Sets the low level discriminator for the id(MCPD module).mpsd_no(MPSD-8+ module: 0..7)
                if (Mesy::debug&&retval) std::cout<<"#MCPD: "<<id<<", module no: "<<i<<" hard LLD has been set to "<<static_cast<int>(module_thresholds[active_module_indx])<<std::endl;
                active_module_indx++;
                for(int j=0; j< Mesy::NTUBE; j++)
                {
                  if(retval&&active_tubes[id][i][j])
                  {
                      retval = MPSDGain(active_tube_gain[Mesydaq->getInternalIndex(id, i, j)], j, i, id);
                      if(Mesy::debug&&retval) std::cout<<"#MCPD: "<<id<<", module no: "<<i<<", tube: "<<j<<", the gain has been set to "<<static_cast<int>(active_tube_gain[Mesydaq->getInternalIndex(id, i, j)])<<std::endl;
                  }
                }
              }
          }
      }
      Mesydaq->ResetData(false);//Reset the datasets
      //bool MPSDPulser(const bool ont_offf=false, const uint8_t amplitude=100, const PULSER_POS pos=PUL_MID, const uint8_t tube_no=0, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Switches on(t)/off(f) (ont_offf) the pulser of id(MCPD module).mpsd_no(MPSD-8+ module:0..7). If it switches to on, it sets  to tube_no(each: 0..7) by pos position and amplitude
    }
    return retval;
}

//"MCPDComm" ------------------------------------------------------Communication commands
bool Mesy::MesyCommand::MCPDCommSetID(const uint8_t new_id, const uint8_t current_id)
//Sets the current to the new ID of the MCPD device
{
    std::vector<uint16_t> par={new_id};
    bool retval=send_command(4,current_id, par,true,false);//4 new_id;  ans: new_id should be the same
    if(retval&&(par.size()==1)) retval=(par[0]==new_id);//Check that new_id is equal to the control parameter
    else return false;
    if(retval)
    {
        //Update the ID's
        if(current_id==Master_id) Master_id=new_id;
        Udpcommand->ModifyFar(current_id,new_id);
    }
    return retval;
}


bool Mesy::MesyCommand::MCPDCommIPPortSetup(std::string &mcpd_ip, std::string &command_ip, std::string &data_rcv_ip, uint16_t &cmd_port, uint16_t &data_port, const uint8_t id)
//Sets the MCPD protocol parameters of id: the IP addresses in dotted format of the MCPD ("0" means no change), the command sending computer ("0" means no change,"00" means this computer), the data receiving computer ("0" means no change,"00" means this computer), command and data UDP port numbers (0 means no change). I overwrites the original variables showing the present setup.
{
    //NOTE:the returning array is 15 word long instead of 14. The meaning of the last word is not known
    //TODO: Update the effect related to Udpcommand and Udpdata
    std::vector<uint16_t> pars(14,0);
    if(mcpd_ip!="0")
    {
        std::string temp;
        std::transform(mcpd_ip.begin(),mcpd_ip.end(),std::back_inserter(temp),[](char ch){return ((ch=='.') ? ' ' : ch);});
        std::stringstream ss(temp);
        for(int i=0;i<4; i++) ss>>pars[i];
    }
    if(command_ip=="0") pars[11]=1;
    else if(command_ip!="00")
    {
        std::string temp;
        std::transform(command_ip.begin(),command_ip.end(),std::back_inserter(temp),[](char ch){return ((ch=='.') ? ' ' : ch);});
        std::stringstream ss(temp);
        for(int i=10;i<14; i++) ss>>pars[i];
    }
    if(data_rcv_ip=="0") pars[5]=1;
    else if(data_rcv_ip!="00")
    {
        std::string temp;
        std::transform(data_rcv_ip.begin(),data_rcv_ip.end(),std::back_inserter(temp),[](char ch){return ((ch=='.') ? ' ' : ch);});
        std::stringstream ss(temp);
        for(int i=4;i<8; i++) ss>>pars[i];
    }
    pars[8]=cmd_port;
    pars[9]=data_port;
    bool retval=send_command(5,id, pars,false,true);// 5 mcpd_ip1/../4 /data_rcv_ip1/../4 /cmd_port/data_port/command_ip1/../4    in the case of "0": ip1=0; "00" ip1=ip2=0
    if(retval&&(pars.size()>13))
    {
        mcpd_ip.clear();
        data_rcv_ip.clear();
        command_ip.clear();
        for(int i=0; i<4; i++) mcpd_ip+=std::to_string(pars[i])+".";
        for(int i=4; i<8; i++) data_rcv_ip+=std::to_string(pars[i])+".";
        cmd_port=pars[8];
        data_port=pars[9];
        for(int i=10; i<14; i++) command_ip+=std::to_string(pars[i])+".";
        mcpd_ip.pop_back();
        data_rcv_ip.pop_back();
        command_ip.pop_back();
        int moduleport=Udpcommand->GetFarPort(static_cast<int>(id));
        Udpcommand->ModifyFar(static_cast<int>(id),mcpd_ip,moduleport);
        //for(int i=0; i<pars.size(); i++) std::cout<<pars[i]<<" ";
        //std::cout<<std::endl;
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MCPDCommGetTX(uint16_t &capabilities, uint16_t &current_setup, const uint8_t id)
//Reports on the Tx capabilities and the current setup of the MCPD unit addressed by id. Capabilities bitmap is used.
{
    std::vector<uint16_t> pars;
    bool retval=send_command(22,id, pars);//22 no params; ans: capabilities/current_setup
    if(retval&&(pars.size()==2))
    {
        capabilities=pars[0];
        current_setup=pars[1];
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MCPDCommSetTX(const uint16_t capability, const uint8_t id)
//Sets the Tx capability of the MCPD unit. One of the capabilities bitmap should be used
{
    std::vector<uint16_t> par={capability};
    bool retval=send_command(23,id, par);//23 capability; ans: current setup should be the same
    if(retval&&(par.size()==1)&&(par[0]==capability)) return true;
    else return false;
}

//"MCPDGen" -------------------------------------------------------General purpose commands
bool Mesy::MesyCommand::MCPDGenSetTiming(const bool mastert_slavef, const uint8_t id, const bool last_slave)
//Sets up the synchronization of the id MCPD module as master(t)/slave(f). At the last slave, the "last_slave" should be set to true, which sets the correct termination.
{
    std::vector<uint16_t> pars={static_cast<uint16_t>(mastert_slavef ? 1 : 0), static_cast<uint16_t>((mastert_slavef||last_slave) ? 0 :1)};
    bool retval=send_command(6,id, pars);//6 slave0master1/termination_on0_off1
    if(retval&&(pars.size()==2)&&(pars[0]==static_cast<uint16_t>(mastert_slavef ? 1 : 0))&&(pars[1]==static_cast<uint16_t>((mastert_slavef||last_slave) ? 0 :1)))
    {
        if(mastert_slavef) Master_id=id;//set the private master to the new value if master is selected
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MCPDGenSetClock(uint64_t &time)
//Sets the master MCPD clock to a given value, which is overwritten by the older value of the timer (resolution: 6.5535ms, because the lowest 16 bits are not transmitted). Only the lowest 48 bits are used.
{
    //NOTE:it returns with only 16..48 bits of the clock, as bits16..31/bits32..48/something(it was 0)
    std::vector<uint16_t> pars={static_cast<uint16_t>(time&0xffff),static_cast<uint16_t>((time>>16)&0xffff),static_cast<uint16_t>((time>>32)&0xffff)};
    bool retval=send_command(7,Master_id, pars);//proceed to the master 7 bits0..15/bits16..31/bits32..48
    if(retval&&(pars.size()==3))
    {
        time=static_cast<uint64_t>((static_cast<uint64_t>(pars[0])<<16)|(static_cast<uint64_t>(pars[1])<<32));
        return true;
    }
    return false;
}

bool Mesy::MesyCommand::MCPDGenSetRunNo(uint16_t &run_no)
//Sets the run number as new and overwrites it by the old value 
{
    std::vector<uint16_t> par={run_no};
    bool retval=send_command(8,Master_id, par);//proceed to the master according to private member and implement: 8 run_no
    if(retval&&(par.size()==1))
    {
        run_no=par[0];
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MCPDGenGetAbout(std::string &microcontroller_version, std::string &fpga_version, std::array<bool, Mesy::NMPSD8> &connected_modules, const uint8_t id)
//Reports on the microcontroller, FPGA versions and about the connected modules
{
    std::vector<uint16_t> pars, pars2;
    bool retval=send_command(51,id, pars);//51 no param; ans: uc_major/uc_minor/(upper: FPGA major, lower: FPGA minor)
    if (retval) retval = send_command(36,id, pars2);//36 no param; ans: codes of the modules connected to it
    if(retval&&(pars.size()==3)&&(pars2.size()==9))
    {
        microcontroller_version=std::to_string(pars[0])+"."+std::to_string(pars[1]);
        fpga_version=std::to_string(pars[2]>>8)+"."+std::to_string(pars[2]&0xff);
        for(int i=0; i<NMPSD8; i++) connected_modules[i] = (pars2[i]==105); 
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MCPDGenReadRegister(std::vector<uint16_t> &register_value, const uint16_t start_register_addr, const uint16_t no_regs, const uint8_t id)
//Retrieve no_regs register values starting from start_register_addr of id MCPD module and putting it as register_value. start_register_no: 102=tx capabilities; 103=tx mode
{
    //NOTE: It is working only in no_regs=1, start_register_addr=102 case; Also answer 1st number: no_regs, 2+2*i nd: 0x9/0x7306, which followed by the value
    if(11+2*no_regs > Replybuffersize)
    {
        std::cerr<<"ERROR(MesyCommand::MCPDGenReadRegister): the size of the Replybuffer ("<<Replybuffersize<<") is less than required by the parameters to receive ("<<11+2*no_regs<<")!!"<<std::endl;
        std::cerr<<"\tHint: increase the MesyCommand::Replybuffersize at the class declaration!"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={no_regs,start_register_addr};
    bool retval=send_command(32,id, pars);//32 no_regs/start_register_addr; no_regs/{????/register_value}
    if(retval&&(pars.size()>2)&&(pars[0]=no_regs))
    {
        register_value.clear();
        for(int i=2; i<static_cast<int>(pars.size()); i+=2) register_value.push_back(pars[i]);
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MCPDGenWriteRegister(const std::vector<uint16_t> register_value, const uint16_t start_register_addr, const uint16_t no_regs, const uint8_t id)
//Writes no_regs pieces of registers by register_value starting from start_register_addr of id MCPD module. start_register_no: 103=tx mode
{
    //NOTE: it returns with number of registers, then {something,new_value}, where something might be the previous setup value.
    std::vector<uint16_t> pars={no_regs,start_register_addr};
    for(int i=0; i<no_regs; i++) pars.push_back(register_value[i]);
    bool retval=send_command(31,id, pars);//31 no_regs/start_register_addr/register_value
    if(retval&&(pars.size()>2)&&(pars[0]=no_regs))
    {
        for(int i=2; i<static_cast<int>(pars.size()); i+=2) if(retval) retval=(pars[i]==register_value[i/2-1]);
        return retval;
    }
    else return false;
}

//"MCPDCAT" -------------------------------------------------------Counter/timer/ADC parameter related commands of the MCPD module
bool Mesy::MesyCommand::MCPDCATEventSetup(const Mesy::INPUT_TYP input_source, const Mesy::TRIGGER_SRC trigger, const uint16_t compare_register, const uint8_t id)
//Setup the event related to the "input_source" at the id MCPD module. Event can be emitted on a regular basis in time when aux timer is set (compare register will not used) or by the rising edge of incoming counts with the help of compare_register. If the compare register is x=0..20 then it triggers at 2^x=1 counts, otherwise use CMP_TRIG_COVERFLOW or CMP_TRIG_RISEDGE constants
{
    if(((input_source==Mesy::INP_ANALOG_1)||(input_source==Mesy::INP_ANALOG_2))&&(trigger==Mesy::TRG_COMPREG))
    {
            std::cerr<<"ERROR(MesyCommand::MCPDCATEventSetup) Using of compare register is not allowed for Analog inputs!"<<std::endl;
            return false;
    }
    std::vector<uint16_t> pars={input_source,trigger,compare_register};
    if((input_source==Mesy::INP_ANALOG_1)||(input_source==Mesy::INP_ANALOG_2)) pars[2]=0;
    uint16_t p=pars[2];
    return send_command(9,id, pars)&&(pars.size()==3)&&(pars[0]==input_source)&&(pars[1]==trigger)&&(p==pars[2]);// 9 input_source/trigger/compare_register
}

bool Mesy::MesyCommand::MCPDCATSetTimer(const uint16_t timer_no, const double msec, const uint8_t id)
//Sets the timing interval of the "timer_no"(0..3) auxiliary timer. The timing interval in ms will be rounded to 0.01ms. Available interval is 0.01..655.36ms
{
    if(msec>655.36)
    {
            std::cerr<<"ERROR(MesyCommand::MCPDCATSetTimer) The timing interval you provided ("<<msec<<") should be in the range of 0.01..655.36ms !"<<std::endl;
            return false;
    }
    std::vector<uint16_t> pars={timer_no,static_cast<uint16_t>(msec*100)};
    return send_command(10,id, pars)&&(pars.size()==2)&&(pars[0]==timer_no)&&(pars[1]==static_cast<uint16_t>(msec*100));// 10 timer_no/capture_register
}

bool Mesy::MesyCommand::MCPDCATParameterSetSource(const uint16_t requested_parameter_no, const Mesy::PARAMETER_SRC source, const uint8_t id)
//Selects a source from the available ones to appear in the head of each data frame as parameter no. "requested_parameter_no"(0..3: 0,1: stop at acquisition; 2,3: runs continously) shipped by the id MCPD module
{
    if(requested_parameter_no>3)
    {
        std::cerr<<"ERROR(MesyCommand::MCPDCATParameterSetSource) The requested parameter value ("<<(int)requested_parameter_no<<") (to appear in the header of each event frame)"<<std::endl<<"\t is outside the valid range (0..3)";
        return false;
    }
    std::vector<uint16_t> pars={requested_parameter_no,source};
    return send_command(11,id, pars)&&(pars.size()==2)&&(pars[0]==requested_parameter_no)&&(source==pars[1]);//11 requested_parameter_no/source
}

bool Mesy::MesyCommand::MCPDCATGetIOCounters(std::vector<uint16_t> &analog_inputs, std::vector<uint16_t> &analog_outputs, std::vector<bool> &digital_inputs, std::vector<bool> &digital_outputs, uint64_t &event_counter, std::vector<uint64_t> &parameter, const uint8_t id)
//Retrieve the value of the actual 2 analog input/output, 6 digital input/2output, the event counter and all of the parameters belonging to id MCPD module
{
    std::vector<uint16_t> pars;
    bool retval=send_command(12,id, pars);//12 w/o parameters; ans: ADC1(12bit)/ADC2(12bit)/DAC1(12bit)/DAC2(12bit)/TTLout(2bit)/TTLinp(6bit)/event_counter_lo/_mid/_hi/ parameter0_lo/_mid/_hi... parameter3_hi
    //NOTE:This command returns with 2 extra words in contrast to written in the documentation; then nor digital inputs/outputs, parameters are changing by the effect of related commands
    if(retval&&(pars.size()>=21))
    {
        analog_inputs.clear();
        analog_outputs.clear();
        digital_inputs.clear();
        digital_outputs.clear();
        parameter.clear();
        for(int i=1; i>=0; i--) analog_inputs.push_back(pars[i]);//Reverse order to be in correspondence with the labels on the back of the panel
        for(int i=2; i<4; i++) analog_outputs.push_back(pars[i]);
        for(int j=0; j<6; j++) digital_inputs.push_back((pars[5]>>j)&0x1);
        for(int j=0; j<2; j++) digital_outputs.push_back((pars[4]>>j)&0x1);
        event_counter=static_cast<uint64_t>(pars[6])|(static_cast<uint64_t>(pars[7])<<16)|(static_cast<uint64_t>(pars[8])<<32);
        for(int i=0; i<4; i++) parameter.push_back(static_cast<uint64_t>(pars[9+3*i])|(static_cast<uint64_t>(pars[10+3*i])<<16)|(static_cast<uint64_t>(pars[11+3*i])<<32));
        return true;
    }
    else return false;
}

//"MCPDPort" ------------------------------------------------------Hardware port related commands of the MCPD module
bool Mesy::MesyCommand::MCPDPortAnalogOutSet(const uint16_t out1, const uint16_t out2, const uint8_t id)
//Controls the rear Analog Out 1 and 2 (+/-3V or 0..6V related to jumper) by out1 and out2 (by the lowest 12 bits) of id MCPD module
{
    std::vector<uint16_t> pars={static_cast<uint16_t>(out1&0x0fff),static_cast<uint16_t>(out2&0x0fff)};
    return send_command(17,id, pars)&&(pars.size()==0);//17 out1/out2
}

bool Mesy::MesyCommand::MCPDPortRS232Send(const std::string message, const uint8_t id)
//Sends the message by the rear RS232 port (9600Baud/sec,8N1) of id MCPD module
{
    std::vector<uint16_t> pars={static_cast<uint16_t>(message.size()+3)};
    std::for_each(message.begin(),message.end(),[&pars](char n){pars.push_back(static_cast<uint16_t>(n));});
    pars.push_back(0xd);
    pars.push_back(0xa);
    pars.push_back('\0');
    return send_command(18,id, pars)&&(pars.size()==0);//18 length_of_string/character1/../character_(length_of_string)
}

/*bool Mesy::MesyCommand::MCPDPortRS232Receive(std::string &message, const uint8_t id)
//Receives the message from the rear RS232 port (9600Baud/sec,8N1) of id MCPD module
{
    //NOTE: this method is not available, or access method through port RS232 is not specified
    std::vector<uint16_t> pars;
    bool retval=send_command(19,id, pars);//19 no parameter; answer: length_of_string/character1/../character_(length_of_string)
    if(retval&&(pars.size()>0)&&(static_cast<uint16_t>(pars.size())==pars[0]+1))
    {
        message.clear();
        std::for_each(pars.begin()+1,pars.end(),[&message](char n){message+=n;});
        return true;
    }
    else return false;
}*/

bool Mesy::MesyCommand::MCPDPortDigitalOutSet(bool out1, bool out2, const uint8_t id)
//Controls the rear Digital(TTL) Out 1 and 2 by out1 and out2 of id MCPD module
{
    std::vector<uint16_t> par={static_cast<uint16_t>((out1 ? 0x0001 : 0x0000)|(out2 ? 0x0002 : 0x0000))};
    return send_command(21,id, par)&&(par.size()==0);;//21 out1|(out2<<1)
}

//"MPSD" ----------------------------------------------------------MPSD-8+ related commands
bool Mesy::MesyCommand::MPSDGain(const uint8_t gain, const uint8_t tube_no, const uint8_t mpsd_no, const uint8_t id)
//Sets the gain for the detector tube at id(MCPD module).mpsd_no(MPSD-8+ module:0..7).tube_no(each: 0..7, all:8)
{
    if(mpsd_no>7)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDGain): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..7"<<std::endl;
        return false;
    }
    if(tube_no>8)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDGain): the tube_no parameter ("<<(int)tube_no<<") should be in range 0..8"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no,tube_no,gain};
    bool retval=send_command(13,id, pars);//13 mpsd_no/tube_no/gain
    if((retval)&&(pars.size()>=2+((tube_no==8) ? 8 : 1))&&(pars[0]==mpsd_no)&&(pars[1]==tube_no))
    {
        for(int i=0; i<((tube_no==8) ? 8 : 1); i++) {if(retval) retval=(pars[i+2]==gain);}
        //NOTE: for tube_no=8 it returns with the transmitted value for all of the tubes and with an unknown parameter
    }
    return retval;
}

bool Mesy::MesyCommand::MPSDDiscriminator(const uint8_t threshold, const uint8_t mpsd_no, const uint8_t id)
//Sets the low level discriminator for the id(MCPD module).mpsd_no(MPSD-8+ module: 0..7)
{
    //NOTE: it returns with the sent+the gain parameter(not used in this implementation)
    if(mpsd_no>7)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDDiscriminator): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..7"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no,threshold};
    return send_command(14,id, pars)&&(pars.size()>1)&&(pars[0]==mpsd_no)&&(pars[1]==threshold);//14 mpsd_no/threshold
}

bool Mesy::MesyCommand::MPSDPulser(const bool ont_offf, const uint8_t amplitude, const Mesy::MesyCommand::PULSER_POS pos, const uint8_t tube_no, const uint8_t mpsd_no, const uint8_t id)
//Switches on(t)/off(f) (ont_offf) the pulser of id(MCPD module).mpsd_no(MPSD-8+ module:0..7). If it switches to on, it sets  to tube_no(each: 0..7) by pos position and amplitude
{
    //NOTE: it replies the command and returns with 3 WORDS. The one part of the 2nd is amplitude<<3
    if(mpsd_no>7)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDPulser): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..7"<<std::endl;
        return false;
    }
    if((tube_no>8)&&ont_offf)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDPulser): the tube_no parameter ("<<(int)tube_no<<") should be in range 0..8"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no,tube_no,pos,amplitude,static_cast<uint16_t>( ont_offf ? 1 : 0)};
    return send_command(15,id, pars)&&(pars.size()>4)&&(pars[0]==mpsd_no)&&(pars[1]==tube_no)&&(pars[2]==pos)&&(pars[3]==amplitude)&&(pars[4]==static_cast<uint16_t>( ont_offf ? 1 : 0));//15 mpsd_no/tube_no/pos/amplitude/ont(1)_offf(0)
}

bool Mesy::MesyCommand::MPSDSetMode(const bool post_amplitf, const uint8_t mpsd_no, const uint8_t id)
//Sets the mode of detector tube at id(MCPD module).mpsd_no(MPSD-8+ module:0..7, 8: all) to transmit only position(t)/amplitudes(f). Only takes effect, if the MCPD capability is set to CAP_POS_AMP or CAP_TPOS_TAMP
{
    if(mpsd_no>8)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDSetMode): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..8"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no,static_cast<uint16_t>( post_amplitf ? 0 : 1)};
    return send_command(16,id, pars)&&(pars.size()==2)&&(pars[0]==mpsd_no)&&(pars[1]==static_cast<uint16_t>( post_amplitf ? 0 : 1));//16 mpsd_no/post(0)_amplitf(1)
}

bool Mesy::MesyCommand::MPSDAbout(uint16_t &evbus_capabilities, uint16_t &evbus_current, uint16_t &firmware, const uint8_t mpsd_no, const uint8_t id)
//Returns with the eventbus transmit capability bitmap, its current setup and the firmware version number of the id(MCPD module).mpsd_no(MPSD-8+ module: 0..7) module
{
    if(mpsd_no>7)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDAbout): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..7"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no};
    bool retval=send_command(24,id, pars);//24 mpsd_no; ans: mpsd_no(should be the same)/evbus_capabilities/evbus_current/firmware
    if(retval&&(pars.size()==4)&&(pars[0]==mpsd_no))
    {
        evbus_capabilities=pars[1];
        evbus_current=pars[2];
        firmware=pars[3];
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MPSDSetTX(const uint16_t evbus_set, const uint8_t mpsd_no, const uint8_t id)
 //NOTE: lack of reliable documentation
//Sets the Tx capability(CAP_POS_AMP or CAP_TPOS_TAMP or CAP_TPA) of the MPSD unit at id(MCPD module).mpsd_no(MPSD-8+ module:0..7)
{
    //TODO:test
    if(mpsd_no>7)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDSetTX): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..7"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no,evbus_set};
    bool retval=send_command(53,id, pars);//25 mpsd_no/evbus_set???
    std::cout<<"MPSDWriteRegister: remove pars.size()==0"<<std::endl;
    return retval&&(pars.size()==0);
}

bool Mesy::MesyCommand::MPSDReadRegister(uint16_t &register_value, const uint16_t register_no, const uint8_t mpsd_no, const uint8_t id)
//Retrieve the content of register_no of id(MCPD module).mpsd_no(MPSD-8+ module:0..7) module putting it as register_value. register_no: 0=tx capabilities
{
    if(mpsd_no>7)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDAbout): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..7"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no,register_no};
    bool retval=send_command(52,id, pars);//52 mpsd_no/register_no; ans: mpsd_no(should be the same)/register_no(should be the same)/register_value
    if(retval&&(pars.size()==3)&&(pars[0]==mpsd_no)&&(pars[1]==register_no))
    {
        register_value=pars[2];
        return true;
    }
    else return false;
}
      
bool Mesy::MesyCommand::MPSDWriteRegister(const uint16_t register_value, const uint16_t register_no, const uint8_t mpsd_no, const uint8_t id)
//Writes register_value of register_no of id(MCPD module).mpsd_no(MPSD-8+ module:0..7) module. register_no: 0=tx capabilities (value: CAP_)
{
    if(mpsd_no>7)
    {
        std::cerr<<"ERROR(MesyCommand::MPSDWriteRegister): the mpsd_no parameter ("<<(int)mpsd_no<<") should be in range 0..7"<<std::endl;
        return false;
    }
    std::vector<uint16_t> pars={mpsd_no,register_no,register_value};
    return send_command(53,id, pars)&&(pars.size()==3)&&(pars[0]==mpsd_no)&&(pars[1]==register_no)&&(pars[2]==register_value);//53 mpsd_no/register_no/register_value
}

//"MCPDDAQ" -------------------------------------------------------Data acquisiton commands

bool Mesy::MesyCommand::MCPDDAQStart()
//Starts data acquisition, timers starting to run
{
    std::vector<uint16_t> pars;
    Daq=true;
    //Start data acquisition thread
    Mesydaq->ResetData();
    std::thread t(server_loop,std::ref(*this));
    t.detach();
    Daq=send_command(1,Master_id, pars)&&(pars.size()==0);//1 no param; neutron and trigger events are send to data buffer
    return Daq;
}

bool Mesy::MesyCommand::MCPDDAQStop()
//Stops data acquisition, timers stops to run
{
    std::vector<uint16_t> pars;
    bool retval=send_command(2,Master_id, pars)&&(pars.size()==0);//2 no param
    if(retval)
    {
        //Stop data acquisition 
        Daq=false;
        return true;
    }
    else return false;
}

bool Mesy::MesyCommand::MCPDDAQContinue()
//Continue data acquisition and timers continue to run
{
    //TODO:test
    std::vector<uint16_t> pars;
    return send_command(3,Master_id, pars)&&(pars.size()==0);//3 no param
}

bool Mesy::MesyCommand::MCPDDAQReset()
//Stops data acquisition and resets all timers to zero
{
    //TODO:test
    std::vector<uint16_t> pars;
    return send_command(0,Master_id, pars)&&(pars.size()==0);//0 no param
}


bool Mesy::MesyCommand::send_command(const uint16_t command_no, const uint8_t target_mcpd_id, std::vector<uint16_t> &parameters, const bool change_id, const bool change_ip)
//Sends command_no on the command channel to the target_mcpd_id MCPD device with parameters. The answer will overwrite the parameter array. It returns with true in the case of successful transmission. If command sent to change the id of MCPD set the change_id to true, also if you would like to change the IP of the module, set the change_ip to true
{
    if(11+static_cast<int>(parameters.size()) > Sendbuffersize)
    {
        std::cerr<<"ERROR: (MesyCommand::send_command) the size of the Sendbuffer ("<<Sendbuffersize<<") is less than required by the parameters ("<<11+parameters.size()<<")!!"<<std::endl;
        std::cerr<<"\tHint: increase the MesyCommand::Sendbuffersize at the class declaration!"<<std::endl;
        return false;
    }
    //Create message to send
    Sendbuffer[MB_LENBUF]=10+parameters.size();//Buffer length
    Sendbuffer[MB_BUFNUM]=Commandbuffernumber;//Command buffer number
    Sendbuffer[MB_CMDRID]=command_no;//Command number
    Sendbuffer[MB_MCPSTA]=target_mcpd_id<<8;//Target MCPD ID | Status(blank)
    
    uint16_t crc=Sendbuffernonvarycrc^Sendbuffer[MB_LENBUF]^Sendbuffer[MB_BUFNUM]^Sendbuffer[MB_CMDRID]^Sendbuffer[MB_MCPSTA];//Setting the CRC for the header part
    for(size_t i=0; i<parameters.size();i++)
    {
        Sendbuffer[10+i]=parameters[i];
        crc^=Sendbuffer[10+i];
    }
    Sendbuffer[MB_CRCSUM]=crc;//Set the command crc
    Sendbuffer[Sendbuffer[MB_LENBUF]]=0xffff;//Set the end of the message
#ifdef DEBUG_MESYCMD
    std::cout<<"Message to send contains "<<std::dec<<Sendbuffer[0]+1<<" of words:"<<std::endl;
    for(int i=0; i<=Sendbuffer[0]; i++) std::cout<<"\t"<<i+1<<": 0x"<<std::hex<<std::setw(4)<<Sendbuffer[i]<<std::dec<<std::endl;
#endif
    //Communicate
    bool second=false;//We do a second trial
    bool get=false;//Indicator, that we have got the right data
    do
    {
        int reply=Udpcommand->SendAndReceive(reinterpret_cast<const char*>(Sendbuffer),sizeof(crc)*(Sendbuffer[MB_LENBUF]+1),reinterpret_cast<char*>(Replybuffer),sizeof(Replybuffer),static_cast<int>(target_mcpd_id),second,!change_ip);
        if(reply>=0)
        {
#ifdef DEBUG_MESYCMD
            std::cout<<"Message received contains "<<std::dec<<reply/2<<" of words:"<<std::endl;
            for(int i=0; i<reply/2; i++) std::cout<<"\t"<<i+1<<": 0x"<<std::hex<<std::setw(4)<<int(Replybuffer[i])<<std::dec<<std::endl;
#endif           
            if(Replybuffer[MB_LENBUF]>Replybuffersize)
            {
                std::cerr<<"ERROR(MesyCommand::send_command): the size of the Replybuffer ("<<Replybuffersize<<") is less than required by the parameters ("<<Replybuffer[MB_LENBUF]<<")!!"<<std::endl;
                std::cerr<<"\tHint: increase the MesyCommand::Replybuffersize at the class declaration!"<<std::endl;
                return false;
            }
            //Checking, is this a command frame and the Commandbuffernumber, MCPD ID (except if ID is changed, when the answer comes from the new ID) and command number is corresponded if not, take a new turn
            if(((Replybuffer[MB_BUFTYP]>>15)&0x1)&&(Replybuffer[MB_BUFNUM]==Commandbuffernumber)&&(((Replybuffer[MB_MCPSTA]>>8)==target_mcpd_id)||change_id)&&((Replybuffer[MB_CMDRID]&0x7fff)==command_no))
            {
                //communication is OK, update command buffer and indicate as OK
                Commandbuffernumber++;
                get=true;
                if(Replybuffer[MB_CMDRID]>>15)
                {
                    std::cerr<<"ERROR(MesyCommand::send_command): The command no. "<<command_no<<" is transmitted, but the module were not able to interpret it"<<std::endl;
                    return false;
                }
                Lastcommandtime=static_cast<uint64_t>(static_cast<uint64_t>(Replybuffer[MB_TIMSLO])|(static_cast<uint64_t>(Replybuffer[MB_TIMSMI])<<16)|(static_cast<uint64_t>(Replybuffer[MB_TIMSHI])<<32));
#ifdef DEBUG_MESYCMD
                std::cout<<"\tTime: : 0x"<<std::hex<<std::setw(12)<<long(Lastcommandtime)<<std::dec<<" "<<Lastcommandtime<<std::endl;
#endif 
                parameters.clear();
                for(uint16_t i=Replybuffer[MB_LENHED]; i<Replybuffer[MB_LENBUF]; i++) parameters.push_back(Replybuffer[i]);
                return true;
            }
        }
        else return false;//Error happened at receiving
        second=-second;
    }while(second&&(!get));
    return false;
}



void Mesy::MesyCommand::server_loop(Mesy::MesyCommand &cmd)
//This method starts the data server loop, which process incoming datasets
{
  const int bsize=32754;//Size of the buffer in words
  uint16_t buffer[bsize];//Receive buffer size
  const int buffersize=sizeof(buffer);//Size in bytes
  //We accepts all incoming calls..
  bool tolerate_other_sender=true;
  int expected_sender_id=0;
  int len=0;
  int max_mcpd_value = cmd.Mesydaq->GetMaxMCPDId();
  std::vector<bool> first(max_mcpd_value+1,true);
  bool overall_first=true;
  uint16_t prev_pckg=0;
  std::vector<std::array<uint64_t,Mesy::NPARAMS>> pars;
  for(int i=0; i <= max_mcpd_value; i++)
  {
    std::array<uint64_t, Mesy::NPARAMS> par;
    par.fill(0);
    pars.push_back(par);
  }
  const double max_wait_time = 10.0;//Maximal allowed time before sending communication warning
  int max_disconnect_times = static_cast<int>((max_wait_time*1000)/cmd.Udpdata->GetTimeoutMS());//Number of maximal allowed trials before qualify as disconnected
  int counts = max_disconnect_times;
  while(cmd.Daq)
  {
    if((len=cmd.Udpdata->ReceiveMessage(reinterpret_cast<char*>(buffer),buffersize,tolerate_other_sender,expected_sender_id))>0)
    {
//	if (len>0) std::cout<<"Message received!"<<std::endl;
	counts = max_disconnect_times;
        len=(sizeof(char)*len)/sizeof(uint16_t);
        //Check package identity
        if((!overall_first)&&buffer[MB_BUFNUM]!=static_cast<uint16_t>(prev_pckg+1)) std::cout<<"Missing data package between "<<prev_pckg<<" and "<<buffer[MB_BUFNUM]<<std::endl;
        prev_pckg=buffer[MB_BUFNUM];
        int mcpd_id=(buffer[MB_MCPSTA]>>8)&0xff;
        if(mcpd_id <= max_mcpd_value)
        {
            //Check changes of the parameters
            bool change=false;
            if(first[mcpd_id])
            {
                for(int i=0; i<4; i++) pars[mcpd_id][i]=static_cast<uint64_t>(buffer[9+3*i])|(static_cast<uint64_t>(buffer[10+3*i])<<16)|(static_cast<uint64_t>(buffer[11+3*i])<<32);
                change=true;
                first[mcpd_id]=false;
                overall_first = false;
            }
            else
            {
                for(int i=0; i<4; i++)
                {
                    uint64_t par=static_cast<uint64_t>(buffer[9+3*i])|(static_cast<uint64_t>(buffer[10+3*i])<<16)|(static_cast<uint64_t>(buffer[11+3*i])<<32);
                    if(par!=pars[mcpd_id][i])
                    {
                        pars[mcpd_id][i]=par;
                        change=true;
			if(Mesy::debug) std::cout<<"#Parameter("<<mcpd_id<<") "<<i<<" "<<par<<" "<<std::endl;
                    }
                }
            }
            if(change||(len>21))
            {
                //New message received from module, which contains events
                uint64_t main_time = static_cast<uint64_t>(buffer[MB_TIMSLO])|(static_cast<uint64_t>(buffer[MB_TIMSMI])<<16)|(static_cast<uint64_t>(buffer[MB_TIMSHI])<<32);
                if(change) cmd.Mesydaq->RegisterParams(pars[mcpd_id], main_time, mcpd_id);
                for(int i=21; i<len; i+=3)
                {
                    uint64_t event=static_cast<uint64_t>(buffer[i])|(static_cast<uint64_t>(buffer[i+1])<<16)|(static_cast<uint64_t>(buffer[i+2])<<32);
                    uint32_t time = (event&0x7ffff);
                    if((event>>47)&0x1)
                    {
                        //This is a trigger event
                        Mesy::TRIGGER_SRC trigger_id = static_cast<Mesy::TRIGGER_SRC>((event>>44)&0x7);//Trigger source
                        Mesy::INPUT_TYP data_id = static_cast<Mesy::INPUT_TYP>((event>>40)&0xf);//Counter id
                        uint32_t data_value = (event>>19)&0x1fffff;//Counter value
			if(Mesy::debug) std::cout<<"#Trigger("<<mcpd_id<<"):"<<static_cast<int>(trigger_id)<<" "<<static_cast<int>(data_id)<<" "<<data_value<<" "<<static_cast<int>(main_time+time)<<std::endl;
                        cmd.Mesydaq->RegisterTriggerEvent(trigger_id, data_id, data_value, main_time+time, mcpd_id);
                    }
                    else
                    {
                        //This is neutron event
                        int tube=((event>>39)&0x7)|((event>>41)&0x38);//Tube index is created by [module_no,(tube_no|0x7)]
                        int amplitude=((event>>29)&0x3ff);
                        int position=((event>>19)&0x3ff);
			if(Mesy::debug) std::cout<<"#Neutron("<<mcpd_id<<"): "<<tube<<" "<<amplitude<<" "<<position<<" "<<static_cast<long long>(main_time+time)<<std::endl;
                        cmd.Mesydaq->RegisterNeutronEvent(tube, amplitude, position, main_time+time, mcpd_id);
                    }
                }
            }
        }
    }
    else
    {
	//Timeout happened
	if (--counts == 0)
	{
	    //Number of repetitions reached the maximal allowed value
	    std::cerr<<"WARNING(Mesy::MesyCommand::server_loop): "<<max_wait_time<<"s has been elapsed without receiving any packets from MCPD8 module!"<<std::endl;
	    counts = max_disconnect_times;
	}
    }
  }
  return;
}
