#include <algorithm>
#include <atomic>
#include <cmath>
#include <fstream>
#include <iomanip>
#include <numeric>
#include <sstream>
#include <string>
#include <thread>
#include <vector>
#include "UDP.h"


#ifndef MESY_H
#define MESY_H

enum MESY_BUFFER:int {MB_LENBUF=0, MB_BUFTYP=1, MB_LENHED=2, MB_BUFNUM=3, MB_CMDRID=4, MB_MCPSTA=5, MB_TIMSLO=6, MB_TIMSMI=7, MB_TIMSHI=8, MB_CRCSUM=9};//Description of Mesytec buffer structure: buffer length, buffer type, header length, buffer number, command/run id, MCPD ID | status, timestamp lo,mid,hi, crc (only for commands)

enum PSD_REGISTERS:uint16_t {PSD_TXCAPABILITY=0,PSD_TXCURRENT=1, PSD_FIRMWARE=2};//MPSD-8+ registers 
enum MCP_REGISTERS:uint16_t {MCP_TXCAPABILITY=102,MCP_TXCURRENT=103};//MCPD-8 registers

namespace Mesy
{

typedef uint32_t  MESY_CNT;//Single channel type of the counter
static const bool debug=false;//Provides debugging message of the module

static const int PSIZE=1024;//Size of position information
static const int ASIZE=1024;//Size of amplitude information
static const int POSAMPSIZE=PSIZE*ASIZE;//Multiplicity of position and amplitude size
static const int NTUBE=8;//Number of tubes might be connected to one mpsd-8+ module
static const int NMPSD8=8;//Number of mpsd-8+ modules can be connected to MCPD-8
static const int NMCPDTUBES=NTUBE*NMPSD8;//Maximum number of tubes might be connected to one MCPD-8 module
static const int NPARAMS=4;//Maximum number of parameters might be transmitted by one MCPD-8 module
static const int NAUXTIMERS=4;//Number of auxiliary timers
//special values of the compare register
static const uint16_t CMP_TRIG_COVERFLOW=21;//trigger on counter overflow
static const uint16_t CMP_TRIG_RISEDGE=22;//trigger on rising edge of the counter
//Constants
enum PARAMETER_SRC:uint16_t {PAS_MONCHO_0=0, PAS_MONCHO_1=1, PAS_MONCHO_2=2, PAS_MONCHO_3=3, PAS_DIGITAL_1=4, PAS_DIGITAL_2=5, PAS_ALLINPUT=6, PAS_EVENTCNT=7, PAS_CLOCK=8, PAS_BLANK=9};
//parameter sources: 4 monitor/chopper input; 2 rear digital(TTL) inputs; all monitor/chopper+rear digital+2 adc; event counter, master clock, blank
enum TRIGGER_SRC:uint16_t {TRG_NO=0, TRG_ATIMER_0=1, TRG_ATIMER_1=2, TRG_ATIMER_2=3, TRG_ATIMER_3=4, TRG_DIGIN_1=5, TRG_DIGIN_2=6, TRG_COMPREG=7};
//trigger source: no trigger; 4 aux register; rising edge of 2 rear digital inputs; compare register
enum INPUT_TYP:uint16_t {INP_MONCHO_0=0, INP_MONCHO_1=1, INP_MONCHO_2=2, INP_MONCHO_3=3, INP_DIGITAL_1=4, INP_DIGITAL_2=5, INP_ANALOG_1=6, INP_ANALOG_2=7, INP_NO=8};
//input numbers: 4 monitor/chopper input; 2 rear digital(TTL) inputs; 2 rear analog(ADC) inputs

class CellEvents
//Describes a structures connected to one MCPD module 
{
    private:
        //methods
        bool input_cell_conformity(const INPUT_TYP input_cell);//Check is within input range

    public:
        //members
        std::array<double, Mesy::NAUXTIMERS> auxtimers;//Auxiliary timers
        std::array<Mesy::TRIGGER_SRC, Mesy::INP_NO> trigger_source;//Trigger sources
        std::array<uint16_t, Mesy::INP_NO> compare_register;//Compare register values
        //methods
        CellEvents();//Constructor: initializes the structure
        ~CellEvents()=default;//Destructor
        //Adding new members -- it returns with true in the case of success
        bool SetPeriodicalCheck(const Mesy::INPUT_TYP input_cell, const double period_ms);//Check the number of counts on input_cell periodically in each 0.01..655.36ms
        bool SetEachEdge(const Mesy::INPUT_TYP input_cell);//Check on its own rising edge
        bool SetEachRearDigitEdge(const Mesy::INPUT_TYP input_cell, bool digital1t_digital2f=true);//Check on by the rising edge of digital input 1(t) or 2(f)
        bool SetEachTwoPowerCnts(const Mesy::INPUT_TYP input_cell, int power);//Create events at each x mod 2^power = 1 counts
};

class MesyDAQ
//This class describes a medium-level data acquisition object cooperating with low-level routines of MesyCommand
{
    private:
        std::vector<std::array<Mesy::PARAMETER_SRC, Mesy::NPARAMS>> allow_parameter_consideration;//Array to allow considering/collecting parameters
        std::vector<std::array<int, Mesy::NPARAMS>> parameter_index;//Parameter index of the tube used internally
        std::vector<std::array<bool, Mesy::NMCPDTUBES>> allow_neutron_collection;//Array to allow neutron data collection related to a specific tube
        std::vector<std::array<int, Mesy::NMCPDTUBES>> neutron_index;//Neutron data index of the tube used internally
        std::vector<Mesy::CellEvents> cell_events;//Contains information about all of the events
        std::vector<std::array<int, Mesy::INP_NO>> event_index;//Event data index of the cell used internally
        int no_active_tubes;//Number of active tubes 
        //methods
        inline int twoD2oneD(const int mpsd_no, const int tube) const {return mpsd_no*Mesy::NTUBE+tube;};//Returns with the internal representation of the tube and mpsd number
    protected:
        bool accept_parameters;//Indicates, that it accepts changed parameters to process
        bool accept_events;//Indicates, that it accepts trigger events to process
        bool accept_neutron;//Indicates, that it accepts neutron events to process
        std::string raw_name;//Main name of the raw output file. If it were blank, raw output is not created
        //methods
        //inline int getMaxMCPDId() const {return static_cast<int>(allow_neutron_collection.size())-1;}//Returns with the highest mcpd id 
        inline int getParameterIndex(const int mcpd_id, Mesy::PARAMETER_SRC param) {return parameter_index[mcpd_id][std::distance(allow_parameter_consideration[mcpd_id].begin(), std::find(allow_parameter_consideration[mcpd_id].begin(), allow_parameter_consideration[mcpd_id].end(), param))];}//Returns with the internal index of the mcpd_id/param
        inline int getCellInternalIndex(const int mcpd_id, const int cell_no) const {return event_index[mcpd_id][cell_no];};//Returns with the internal index of the cell event
        virtual void check_parameter_and_store(const int internal_indx, const uint64_t value, const uint64_t time){return;};//Apply discriminators and store the value of a parameter having internal_indx at a given time
        virtual void check_neutron_and_store(const int internal_indx, const int amplitude, const int position, const uint64_t time){return;};//Apply discriminators and store a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
        virtual void check_event_and_store(const int internal_indx, const Mesy::TRIGGER_SRC trigger_id, const uint32_t data_value, const uint64_t time){return;};//Apply discriminators and store a counter event happened on internal_indx for the effect of trigger_id by data_value at time
        bool write_raw();//Writes raw datasets into file
        virtual void write_rawfile(int internal_tube_no, std::ofstream &os)=0;//Writes the raw dataset of internal_tube_no to os stream
        virtual void reset()=0;//Resets the recorded dataset
    public:
        //members
        //methods
        MesyDAQ(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events, const std::string raw_out_name="");//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD8][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices and the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped)
        ~MesyDAQ()=default;//Destructor
        //registering events
        void RegisterParams(const std::array<uint64_t,NPARAMS> params, const uint64_t time, const int mcpd_id);//Registers the changed header parameters happened in time at mcpd_id interface
        void RegisterTriggerEvent(const Mesy::TRIGGER_SRC trigger_id, const INPUT_TYP data_id, const uint32_t data_value, const uint64_t time, const int mcpd_id);//Registers a trigger event on data_id by data_value triggered by trigger_id at time and at the mcpd_id interface
        void RegisterNeutronEvent(const int tube_no, const int amplitude, const int position, const uint64_t time, const int mcpd_id);//Registers a neutron event happened at tube [3 bit of module| 3bit of slot of module] of mcpd_id interface at time 
        inline int getInternalIndex(const int mcpd_id, const int mpsd_no, const int tube) const {return neutron_index[mcpd_id][mpsd_no*Mesy::NTUBE+tube];}//Returns with the internal index of a given tube
        //Infos about the stored datasets
        inline int GetNoActiveDetectors() const {return no_active_tubes;};//Returns with the number of active tubes
        inline int GetMaxMCPDId() const {return static_cast<int>(allow_neutron_collection.size())-1;}//Returns with the highest mcpd id 
        std::vector< std::array<std::array<bool, Mesy::NTUBE>, Mesy::NMPSD8>> GetActiveDetectors() const;//Returns with the arrays containing active detectors by define them as true [mcpd_id][mpsd_no][tube]
        std::vector<std::array<Mesy::PARAMETER_SRC, Mesy::NPARAMS>> GetHeaderParameterSetup() const {return allow_parameter_consideration;};//Returns with a vector containing the parameter source in [mcpd:0..hiindex][param_no:0..<NPARAMS] indices
        std::vector<Mesy::CellEvents> GetEventSetup() const{return cell_events;};//Returns cell event information in [mcpd_id] order
        std::string GetFilename(){return raw_name;};//Returns with the actually set up getfilename
        void SetFilename(const std::string filename){raw_name=filename; return;};//Sets the raw filename
        //recorded data manipulation
        void ResetData(bool raw_write=false){if(raw_write) write_raw(); reset(); return;};//This method resets the recorded array. If raw_write is true, then it will writes recorded datasets to file before doing it
        void WriteRaws(std::string fn=""){if(fn!="") SetFilename(fn); write_raw(); return;};//Writes raw datasets into fn/into already set file
        virtual bool GetRecordedData(Mesy::MESY_CNT *data, const int tube_no=-1)=0;//Returns with the actually recorded dataset for a specific tube or all of the tubes -- it is assumed, that the required data is already allocated
        virtual std::vector<int> GetDatasetDimensions()=0;//Returns with the dimensions of the dataset [tube_no][]...
        int GetNoDatasets(){std::vector<int> dims=GetDatasetDimensions(); return std::accumulate(dims.begin(),dims.end(), 1, std::multiplies<int>());}//Returns with the number of datasets
};

class MesyDAQFixedDiscriminators: public MesyDAQ
//This class describes a DAQ, where discriminators belonging to each tubes do not depend on the time variable 
{
    private:
        //members
        enum INDX:int {LOW=0, HIGH=1};//Declares low and high indices
        std::vector< std::array<int,2> > discriminator;//Discriminator for each of active tubes
        //methods
        void check_neutron_and_store(const int internal_indx, const int amplitude, const int position, const uint64_t time);//Apply discriminators and store a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
    protected:
        virtual void store_neutron(const int internal_indx, const int amplitude, const int position, const uint64_t time)=0;//Stores a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
        virtual void store_neutron_out_of_discriminator(const int internal_indx, const int amplitude, const int position, const uint64_t time){return;};//Stores a rejected neutron event by discriminator happened on internal_indx tube index with signal amplitude at position and at a given time. Default: do not take care about
    public:
        MesyDAQFixedDiscriminators(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events, const std::vector< std::array<int,2> > &discr, const std::string raw_out_name="");//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD8][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices, amplitude discriminators (LOW,HIGH) for each active tubes, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped)
        ~MesyDAQFixedDiscriminators()=default;
        void ModifyDiscriminator(const int mcpd_id, const int mpsd_no, const int tube_no, const int low, const int high);//Changes the software discriminator setup of a given tube
        std::array<int,2> GetDiscriminators(int tube_no){return discriminator[tube_no];};//Returns with the discriminators related to a specific tube
        std::vector<std::array<int,2>> GetDiscriminators(){return discriminator;};//Returns with all of the discriminators
};

class MesyDAQCWSteadySample: public MesyDAQFixedDiscriminators
//This class describes a DAQ using constant wavelength neutrons and supposing stable sample (time coordinates will not be registered)
{
    private:
        //members
        std::vector<std::array<Mesy::MESY_CNT,Mesy::POSAMPSIZE>> recorded_data;//Stores the recorded counts for each active tubes in [POSITION*ASIZE+AMPLITUDE]
        std::vector<std::array<Mesy::MESY_CNT,Mesy::ASIZE>> amplitudes_data;//Stores the recorded amplitudes for each active tubes
        std::vector<std::array<Mesy::MESY_CNT,Mesy::PSIZE>> position_data;//Stores the positions filtered by low and high level of discriminators belonging to each active tubes 
        //methods
        void store_neutron(const int internal_indx, const int amplitude, const int position, const uint64_t time);//Stores a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
        void store_neutron_out_of_discriminator(const int internal_indx, const int amplitude, const int position, const uint64_t time);//Stores a rejected neutron event by discriminator happened on internal_indx tube index with signal amplitude at position and at a given time (position data will not be stored)
        void write_rawfile(int internal_tube_no, std::ofstream &os);//Writes the raw dataset of internal_tube_no to os stream
        void reset();//Resets the recorded dataset
    public:
        MesyDAQCWSteadySample(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events, const std::vector< std::array<int,2> > &discr, const std::string raw_out_name="");//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD8][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices, amplitude discriminators (LOW,HIGH) for each active tubes, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped)
        ~MesyDAQCWSteadySample(){if(raw_name!="") write_raw();};//Destructor: writing the dataset into file before disappearing
        bool GetRecordedData(Mesy::MESY_CNT *data, const int tube_no=-1);//Returns with the actually recorded dataset for a specific tube or all of the tubes -- it is assumed, that the required data is already allocated the order is all position data then all of the amplitudes
        std::vector<int> GetDatasetDimensions(){return {{GetNoActiveDetectors(), 2, Mesy::ASIZE}}; };//Returns with the dimensions of the dataset [tube_no][]...
};

class MesyDAQMTESTPowderCatchAll: public MesyDAQCWSteadySample
//This class describes the general powder sample setup used at MTEST diffractometer, supposing that monitor counter is out of order
{
    public:
        MesyDAQMTESTPowderCatchAll(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array<int,2> > &software_discriminators, const std::string raw_out_name="", const int max_mcpd=0): MesyDAQCWSteadySample(active_tubes, {{ {PAS_BLANK, PAS_BLANK, PAS_BLANK, PAS_BLANK} }}, {Mesy::CellEvents()}, software_discriminators, raw_out_name){};//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD8][tube:0..<NTUBE] indices, amplitude discriminators (LOW,HIGH) for each active tubes, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped), finally the highest mcpd id
};

class MesyDAQMTESTPowderGatedByMonitorPulse: public MesyDAQCWSteadySample
//This class describes the general powder sample setup used at MTEST diffractometer, where data collection is initiated and stopped by TTL pulses of the monitor counter
{
    private:
        int monitor_par_event_no;//Number of the parameter or event belonging to the monitor input
        bool monitor_parametert_eventf;//Parameter or event registration of the monitor signals
    public:
        MesyDAQMTESTPowderGatedByMonitorPulse(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array<int,2> > &software_discriminators, const Mesy::PARAMETER_SRC pulse_input_line, const int mcpd_id_no, const std::string raw_out_name="", const bool paramt_eventf=true, const int max_mcpd=0);//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD8][tube:0..<NTUBE] indices, amplitude discriminators (LOW,HIGH) for each active tubes, input line of the monitor start/stop signal, the mcpd module id, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped), the parameter/event setup, finally the highest mcpd id
        ~MesyDAQMTESTPowderGatedByMonitorPulse()=default;
        inline void check_parameter_and_store(const int internal_indx, const uint64_t value, const uint64_t time) {if((monitor_parametert_eventf)&&(internal_indx==monitor_par_event_no)) accept_neutron=(value==1); return;};//If the correct parameter no has been catched, it allows neutron event detection, if it is in 1 value
        inline void check_event_and_store(const int internal_indx, const Mesy::TRIGGER_SRC trigger_id, const uint32_t data_value, const uint64_t time){if((!monitor_parametert_eventf)&&(internal_indx==monitor_par_event_no)) accept_neutron=(data_value==1); return;};//If the correct cell event has been catched, it allows neutron event detection, if it is in 1 value
};

class MesyDAQCWConstRotation: public MesyDAQFixedDiscriminators
//This class describes a DAQ using constant wavelength neutrons and supposing constant rotation on a given axis: between 2 events, the datasets are stored in internal buffer, then put into discrate bins
{
    private:
        //members
        std::vector<std::vector<std::array<Mesy::MESY_CNT,Mesy::PSIZE>>> position_data;//Stores discretized amplitude datasets in [tube][time_resolution][position] form
        std::vector<int> tube_no;//Temporary storage for tube number of neutron data
        std::vector<int> buffer_position;//Temporary storage for positions of neutron data
        std::vector<uint64_t> buffer_time_since_event;//Temporary storage for time elapsed since the last event
        int rotation_resolution;//Resolution on rotation axis
        int actual_no_data;//Actual number of data in the buffer
        int rotation_event_index;//Index of the rotation event among the other events
        uint64_t last_event_date;//Date of the last event
        bool first_not_full_rotation;//Indicates the first partial rotation, whose data should be expelled from the data collection
        //methods
        void discretize_buffer(uint64_t event_time);//This method discretizes the buffered datasets based on elapsed time and rotation resolution and resets the buffer
        void store_neutron(const int internal_indx, const int amplitude, const int position, const uint64_t time);//Stores a neutron event happened on internal_indx tube index with signal amplitude at position and at a given time
        void check_event_and_store(const int internal_indx, const Mesy::TRIGGER_SRC trigger_id, const uint32_t data_value, const uint64_t time);//Apply discriminators and store a counter event happened on internal_indx for the effect of trigger_id by data_value at time -- this function implements only the rotation event part
    public:
        MesyDAQCWConstRotation(const std::vector<std::array< std::array<bool, Mesy::NTUBE> , Mesy::NMPSD8>> &active_tubes, const std::vector< std::array< Mesy::PARAMETER_SRC, Mesy::NPARAMS > > &header_parameters, const std::vector<Mesy::CellEvents> &events, const std::vector< std::array<int,2> > &discr, const int rot_axis_resolution, const std::array<int,2> mcpdno_cellno_of_fullrot_indicator, const std::string raw_out_name="", const int buffersize=10000);//Constructor: active_tubes (t: active, f:passive) for each possible [mcpd:0..hiindex][mpsd:0..<NMPSD8][tube:0..<NTUBE] indices, header_parameters (containing the parameter source) for each [mcpd:0..hiindex][param_no:0..<NPARAMS] indices, cell events structure for each [mcpd:0..hiindex] indices, amplitude discriminators (LOW,HIGH) for each active tubes, resolution of the rotation axis, {mcpd,cellno} of event counter indicating 360 deg rotation, the name of raw_out_name used when the destructor called to write raw data to file (if it were blank, writing is skipped), size of the temporary buffer
        ~MesyDAQCWConstRotation()=default;
};

class MesyCommand
{
  public:
    //static members
    //capabilities bitmap
    static const uint16_t CAP_POS_AMP=0x1;//position or amplitudes
    static const uint16_t CAP_TPOS_TAMP=0x2;//time+ position/amplitudes
    static const uint16_t CAP_TPA=0x4;//time+position+amplitudes
    //pulser positions
    enum PULSER_POS:uint16_t {PUL_LEFT=0,PUL_MID=2,PUL_RIGHT=1};
    //methods
    //MesyCommand(UDPPassive &command, UDPPassive &data, const uint8_t master=Master_id);//Constructor: sets the UDP communication objects for command (might be the Mesytec device or data collector computer, which act as relay) and for data channels (to the data collector computer) and the master channel number.
    MesyCommand(UDPClient &command, UDPServerTimeOut &data, MesyDAQ &msdaq, std::vector<uint8_t> &thresholds, std::vector<uint8_t> &gains, const uint8_t master);//Constructor: sets the UDP communication objects for command (might be the Mesytec device or data collector computer, which act as relay) and for data channels (to the data collector computer), the data acquisition object (doing the high-level acquisition), thresholds (discriminator) for each MPSD modules,  gains for each active tubes and the master channel number.
    ~MesyCommand(){Udpcommand=NULL;Udpdata=NULL; Mesydaq=NULL; };//Sets the udp and acquisition to default
    //General methods
    bool UpdateSetup();//Sets up the DAQ system based on Mesydaq and local variables
    bool IsMeasurementON(){return Daq;};//Returns with true, if actually measurement is performed
    bool GetRecordedData(Mesy::MESY_CNT *data, const int tube_no=-1){return Mesydaq->GetRecordedData(data, tube_no);};//Returns with the actually recorded dataset for a specific tube or all of the tubes -- it is assumed, that the required data is already allocated
    void WriteRaws(std::string fn=""){Mesydaq->WriteRaws(fn); return;};//Writes raw datasets into fn/into already set file
    //Commands: return with true, if communication has been done successfully (and values are set or reliable)
    //"MCPDComm" ------------------------------------------------------Communication commands
      bool MCPDCommSetID(const uint8_t new_id, const uint8_t current_id=Master_id);//Sets the current to the new ID of the MCPD device
      bool MCPDCommIPPortSetup(std::string &mcpd_ip, std::string &command_ip, std::string &data_rcv_ip, uint16_t &cmd_port, uint16_t &data_port, const uint8_t id=Master_id);//Sets the MCPD protocol parameters of id: the IP addresses in dotted format of the MCPD ("0" means no change), the command sending computer ("0" means no change,"00" means this computer), the data receiving computer ("0" means no change,"00" means this computer), command and data UDP port numbers (0 means no change). I overwrites the original variables showing the present setup.
      //TODO: create "read version of the IP setup"
      bool MCPDCommGetTX(uint16_t &capabilities, uint16_t &current_setup, const uint8_t id=Master_id);//Reports on the Tx capabilities and the current setup of the MCPD unit addressed by id. Capabilities bitmap is used.
      bool MCPDCommSetTX(const uint16_t capability=CAP_TPA, const uint8_t id=Master_id);//Sets the Tx capability of the MCPD unit. One of the capabilities bitmap should be used
    //"MCPDGen" -------------------------------------------------------General purpose commands
      bool MCPDGenSetTiming(const bool mastert_slavef, const uint8_t id=Master_id, const bool last_slave=false);//Sets up the synchronization of the id MCPD module as master(t)/slave(f). At the last slave, the "last_slave" should be set to true, which sets the correct termination.
      bool MCPDGenSetClock(uint64_t &time);//Sets the master MCPD clock to a given value, which is overwritten by the older value of the timer(resolution: 6.5535ms: because the lowest 16 bits are not transmitted). Only the lowest 48 bits are used.
      bool MCPDGenSetRunNo(uint16_t &run_no);//Sets the run number as new and overwrites it by the old value 
      bool MCPDGenGetAbout(std::string &microcontroller_version, std::string &fpga_version, std::array<bool,NMPSD8> &connected_modules, const uint8_t id=Master_id);//Reports on the microcontroller, FPGA versions and about the connected modules
      bool MCPDGenReadRegister(std::vector<uint16_t> &register_value, const uint16_t start_register_addr=MCP_TXCAPABILITY, const uint16_t no_regs=1, const uint8_t id=Master_id);//Retrieve no_regs register values starting from start_register_addr of id MCPD module and putting it as register_value. start_register_no: 102=tx capabilities; 103=tx mode
      bool MCPDGenWriteRegister(const std::vector<uint16_t> register_value, const uint16_t start_register_addr=MCP_TXCURRENT, const uint16_t no_regs=1, const uint8_t id=Master_id);//Writes no_regs pieces of registers by register_value starting from start_register_addr of id MCPD module. start_register_no: 103=tx mode
    //"MCPDCAT" -------------------------------------------------------Counter/timer/ADC parameter related commands of the MCPD module
      bool MCPDCATEventSetup(const Mesy::INPUT_TYP input_source, const Mesy::TRIGGER_SRC trigger=Mesy::TRG_NO, const uint16_t compare_register=0, const uint8_t id=Master_id);//Setup the event related to the "input_source" at the id MCPD module. Event can be emitted on a regular basis in time when aux timer is set (compare register will not used) or by the rising edge of incoming counts with the help of compare_register. If the compare register is x=0..20 then it triggers at 2^x=1 counts, otherwise use CMP_TRIG_COVERFLOW or CMP_TRIG_RISEDGE constants
      bool MCPDCATSetTimer(const uint16_t timer_no, const double msec, const uint8_t id=Master_id);//Sets the timing interval of the "timer_no"(0..3) auxiliary timer. The timing interval in ms will be rounded to 0.01ms. Available interval is 0.01..655.36ms
      bool MCPDCATParameterSetSource(const uint16_t requested_parameter_no=0, const Mesy::PARAMETER_SRC source=Mesy::PAS_MONCHO_1, const uint8_t id=Master_id);//Selects a source from the available ones to appear in the head of each data frame as parameter no. "requested_parameter_no"(0..3: 0,1: stop at acquisition; 2,3: runs continously) shipped by the id MCPD module
      bool MCPDCATGetIOCounters(std::vector<uint16_t> &analog_inputs, std::vector<uint16_t> &analog_outputs, std::vector<bool> &digital_inputs, std::vector<bool> &digital_outputs, uint64_t &event_counter, std::vector<uint64_t> &parameter, const uint8_t id=Master_id);//Retrieve the value of the actual 2 analog input/output, 6 digital input/2output, the event counter and all of the parameters belonging to id MCPD module
    //"MCPDPort" ------------------------------------------------------Hardware port related commands of the MCPD module
      bool MCPDPortAnalogOutSet(const uint16_t out1=0, const uint16_t out2=0, const uint8_t id=Master_id);//Controls the rear Analog Out 1 and 2 (+/-3V or 0..6V related to jumper) by out1 and out2 (by the lowest 12 bits) of id MCPD module
      bool MCPDPortRS232Send(const std::string message, const uint8_t id=Master_id);//Sends the "(MESSSAGE_LENGTH+1)+message+CR+LF" by the rear RS232 port (9600Baud/sec,8N1) of id MCPD module
      //bool MCPDPortRS232Receive(std::string &message, const uint8_t id=Master_id);//Receives the message from the rear RS232 port (9600Baud/sec,8N1) of id MCPD module
      //NOTE: this method is not available, or access method through port RS232 is not specified 
      bool MCPDPortDigitalOutSet(bool out1=false, bool out2=false, const uint8_t id=Master_id);//Controls the rear Digital(TTL) Out 1 and 2 by out1 and out2 of id MCPD module
    //"MPSD" ----------------------------------------------------------MPSD-8+ related commands
      bool MPSDGain(const uint8_t gain=92, const uint8_t tube_no=8, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Sets the gain for the detector tube at id(MCPD module).mpsd_no(MPSD-8+ module:0..7).tube_no(each: 0..7, all:8)
      bool MPSDDiscriminator(const uint8_t threshold=10, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Sets the low level discriminator for the id(MCPD module).mpsd_no(MPSD-8+ module: 0..7)
      bool MPSDPulser(const bool ont_offf=false, const uint8_t amplitude=100, const PULSER_POS pos=PUL_MID, const uint8_t tube_no=0, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Switches on(t)/off(f) (ont_offf) the pulser of id(MCPD module).mpsd_no(MPSD-8+ module:0..7). If it switches to on, it sets  to tube_no(each: 0..7) by pos position and amplitude
      bool MPSDSetMode(const bool post_amplitf=true, const uint8_t mpsd_no=8, const uint8_t id=Master_id);//Sets the mode of detector tube at id(MCPD module).mpsd_no(MPSD-8+ module:0..7, 8: all) to transmit only position(t)/amplitudes(f). Only takes effect, if the MCPD capability is set to CAP_POS_AMP or CAP_TPOS_TAMP
      bool MPSDAbout(uint16_t &evbus_capabilities, uint16_t &evbus_current, uint16_t &firmware, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Returns with the eventbus transmit capability bitmap, its current setup and the firmware version number of the id(MCPD module).mpsd_no(MPSD-8+ module: 0..7) module
      bool MPSDSetTX(const uint16_t tx_capability, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Sets the Tx capability of the MPSD unit at id(MCPD module).mpsd_no(MPSD-8+ module:0..7) NOTE: lack of reliable documentation
      bool MPSDReadRegister(uint16_t &register_value, const uint16_t register_no=PSD_TXCAPABILITY, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Retrieve the content of register_no of id(MCPD module).mpsd_no(MPSD-8+ module:0..7) module putting it as register_value. register_no: 0=tx capabilities
      bool MPSDWriteRegister(const uint16_t register_value, const uint16_t register_no=PSD_TXCURRENT, const uint8_t mpsd_no=0, const uint8_t id=Master_id);//Writes register_value of register_no of id(MCPD module).mpsd_no(MPSD-8+ module:0..7) module. register_no: 0=tx capabilities (value: CAP_)
    //"MCPDDAQ" -------------------------------------------------------Data acquisiton commands
      bool MCPDDAQStart();//Starts data acquisition, timers starting to run
      bool MCPDDAQStop();//Stops data acquisition, timers stops to run
      bool MCPDDAQContinue();//Continue data acquisition and timers continue to run
      bool MCPDDAQReset();//Stops data acquisition and resets all timers to zero
    //Collected data
      //bool GetCollectedData(const int mcpd_id, const int mpsd_no, const int tube, std::array<int,Mesy::ASIZE> &data, const bool positiont_amplitudesf=true) const;//Returns with the collected amplitudes(f)/positions(t) data for the tube of mpsd_no of mcpd id. Return value is true if indexing errors have been avoided
      //bool GetCollectedData(const int mcpd_id, const int mpsd_no, const int tube, std::ofstream &os) const;//Writes the collected positions-amplitudes data [POSITION*ASIZE+AMPLITUDE] to os for the tube of mpsd_no of mcpd id. Return value is true if indexing errors have been avoided
    //Conditions
//      bool AddConditionLimit(uint64_t limit_value, std::function<bool(uint64_t,uint64_t)> op, const uint16_t requested_parameter_no=0, const uint8_t id=Master_id){return cpar.AddCondition(requested_parameter_no, limit_value, op, id);}//Adds a condition for a parameter. The op can be any comparison operator. If in the package from id the [requested_parameter_no] [op] [limit_value] is true, then the dataset is recorded, otherwise not.
//      void ClearConditions(){cpar.ClearConditions();};//Clears all of the conditions
  private:
    //static members and constants
    static const uint16_t Sendbuffernonvarycrc=0x800a;//CRC of the non-vary part of the Sbuffer
    static const int Sendbuffersize=500;//Size of the sendbuffer
    static const int Replybuffersize=500;//Size of the sendbuffer
    static uint16_t Commandbuffernumber;//Number of the command buffer
    static uint8_t Master_id;//Id of the master module
    static uint64_t Lastcommandtime;//Timestamp of the last sent command
    //Shared resources
    std::atomic<bool> Daq;//Data acquisition is on(t)/off(f) shared resource: it is set by MCPDDAQStart() and MCPDDAQStop()
//    ConditionParameter cpar;//Condition for parameters
    Mesy::MesyDAQ *Mesydaq;//Mesytec Data AcQusition object: does it on middle level
    std::vector<uint8_t> module_thresholds;//Array to store hardware low level discriminators of each active modules
    std::vector<uint8_t> active_tube_gain;//Gain of each active tubes
    //members
    UDPClient *Udpcommand;//UDP communicator object reserved for commands
    UDPServerTimeOut *Udpdata;//UDP communicator object reserved for communicating with data collector device
    uint16_t Sendbuffer[Sendbuffersize];//Communication buffer to make the communication fast
    uint16_t Replybuffer[Replybuffersize];//Communication buffer to make the communication fast
    //methods
    bool send_command(const uint16_t command_no, const uint8_t target_mcpd_id, std::vector<uint16_t> &parameters, const bool change_id=false, const bool change_ip=false);//Sends command_no on the command channel to the target_mcpd_id MCPD device with parameters. The answer will overwrite the parameter array. It returns with true in the case of successful transmission. If command sent to change the id of MCPD set the change_id to true, also if you would like to change the IP of the module, set the change_ip to true
    
    void reset_recorded_data();//This method resets the recorder array
    static void server_loop(MesyCommand &cmd);//This method starts the data server loop, which process incoming datasets
};
}
#endif
