#include "Motor.h"

namespace Positioner
{  
  const uint16_t StepMotorController::minspeed[16]={ 12, 24, 36, 60, 96, 144, 180, 228, 276, 336, 396, 456, 516, 588, 684, 792};//Minimal speed array (step/sec)
  const uint16_t StepMotorController::maxspeed[16]={ 180, 365, 550, 740, 930, 1120, 1300, 1490, 1680, 1870, 2060, 2250, 2440, 2620, 2810, 3000};//Maximal speed array
  const double   StepMotorController::ramptime[8]={0.0, 0.2, 0.5, 1.0, 1.4, 2.0, 2.8, 4.0};//Ramp in seconds
  
  const uint32_t ParallelMotor::mot_pow_arr[12]={IOOperation::OutputRegister::MOT_MON_TLT, IOOperation::OutputRegister::MOT2, IOOperation::OutputRegister::MOT_MON_SHFT, IOOperation::OutputRegister::MOT4, IOOperation::OutputRegister::MOT_MON_ROT, IOOperation::OutputRegister::MOT_DIFFM_ARM, IOOperation::OutputRegister::MOT7, IOOperation::OutputRegister::MOT8, IOOperation::OutputRegister::MOT9, IOOperation::OutputRegister::MOT10, IOOperation::OutputRegister::MOT11, IOOperation::OutputRegister::MOT12};//Commands for the motor power controller
  
  StepMotorController::StepMotorController(Camac &camcomm, uint8_t slotno)
  //Constructor: camac communicator object and slot number are the parameters 
  {
    slot=slotno;
    comm=&camcomm;
    //Initialize auxiliary running parameter variables
    for(int i=0; i<6; i++)
    {
      auxrun[i]=false;
      auxconnected[i]=false;
      //Get status of motors
      int trials=0;//Number of unsuccessful trials
      lam_status[i]=COMM_ERROR;
      while((lam_status[i]==COMM_ERROR)&&(trials++<5)) lam_status[i]=ReadStatusRegister(i+1);//Read status register
    }
  }
  
  uint32_t StepMotorController::ReadStatusRegister(uint8_t motor_no)
  //Read the status register for the requested motor
  {
    if(!check_motors(motor_no)) return BAD_MOTOR;//Check motor
    uint32_t status1,status2=0;
    bool retval=comm->getData(motor_no-1+BASE_A, STREG_F, slot, status1, Camac::STATUS_X| Camac::STATUS_Q);//Get low 16 bit part
    if(retval) retval=comm->getData((motor_no-1)/2+AUX_STREGISTER_A, STREG_F, slot, status2, Camac::STATUS_X| Camac::STATUS_Q);//Get upper 16 bit part
    if(retval) return (status1 | (((status2>>((motor_no%2)*8) )&0xff)<<16));//Reconstruct the 24 bit status register
    else return COMM_ERROR;
  }
  
  bool StepMotorController::SetRunningParameters(uint8_t motor_no, double ramp_sec, int min_speed_step_sec, int max_speed_step_sec, bool exact_position, bool over_zero_run_enable)
  //Set the running parameters for motor: the ramp in seconds, the minimal and maximal speed in step/seconds, the exact(t)/or +/-1 position tolerance (f), enabling overrun trough zero point if the requested position is closer that way
  {
    if(!check_motors(motor_no)) return false;//Check motor
    //Create running parameter word
    uint16_t code=over_zero_run_enable ? 0x8000 :0x0;//Set over zero part
    code+=((uint16_t) ramp2code(ramp_sec) <<12)+(((uint16_t) speed2code(min_speed_step_sec, false))<<8)+(uint16_t) speed2code(max_speed_step_sec, true);//Set ramp minimal and maximal speeds
    bool retval=comm->setData(motor_no-1+RUNPAR_A, WRITE_F, slot,(uint32_t) code, Camac::STATUS_X | Camac::STATUS_Q);
    if(retval)
    {
      usleep(CNTR_WAIT);//Wait some time...
      if(!((auxconnected[motor_no-1] && (auxrun[motor_no-1]!=exact_position))))
      {
	auxrun[motor_no-1]=!exact_position;//Set the requested change
	code=(auxrun[motor_no-(motor_no-1)%2] ? 0x10 : 0x0) +(auxrun[motor_no-(motor_no-1)%2-1] ? 0x1000 : 0x0);//Set the code word
	retval=comm->setData((motor_no-1)/2+AUX_RUNPAR_A, WRITE_F, slot,(uint32_t) code, Camac::STATUS_X | Camac::STATUS_Q);
	if (retval)
	{
	  usleep(CNTR_WAIT);//Wait the command to be processed
	  auxconnected[motor_no-1]=true;//Set the connected status to true
	}
      }
    }
    return retval;//Return
  }
  
  bool StepMotorController::EnableDisableLAM(uint8_t motor_no, uint32_t lam, bool enable)
  //Enables (t) or disables(f) the set of LAMs (separated by "|") for motor.
  {
    if(!check_motors(motor_no)) return false;//Check motor
    //Update internal pattern
    uint32_t new_pattern=lam_status[motor_no-1];
    if (enable) new_pattern&=(~lam)&LAM_ALL;
    else new_pattern|=lam | ((((new_pattern|lam)&LAM_ALL) == LAM_ALL) ? ST_LAM_ANY_DISABLED : 0x0); //Disable: add the modified pattern
    
    if (new_pattern!=lam_status[motor_no-1])
    {
      //Update LAM
      bool retval=comm->setData(BASE_A+motor_no-1, LAM_ED_F, slot, new_pattern, Camac::STATUS_X | Camac::STATUS_Q);
      if (retval)
      {
	usleep(CNTR_WAIT);//Wait the command to be processed
	lam_status[motor_no-1]=new_pattern;
      }
      return retval;
    }
    else return true;//Not need to do anything
  }
  
  uint8_t StepMotorController::ramp2code(double ramp)
  //This function converts the ramp in sec units into coded output.
  {
    uint8_t i=0;//Index
    while((ramptime[i++]<ramp)&&(i<8));
    if((i>1)&&((ramp-ramptime[i-2])<0.001)) i--;//Correcting small differences
    return --i;//Returns with correct setup
  }
  
  uint8_t StepMotorController::speed2code(int speed, bool max)
  //This function converts the speed in step/sec units into coded output. It does for minimal (false) or maximal (true)
  {
    uint8_t i=0;//Loop variable
    const uint16_t *poi= (max ? maxspeed : minspeed);
    while((poi[i++]<(uint16_t) speed)&&(i<16));
    //Evaluate the results
    i--;
    if (i==0)
    {
      //It is less than the minimal speed or equal
      if(!max) return 0;
      else if (speed==0) return 0;
      else return (uint8_t) ((speed*16)/poi[0] - ((((speed*16)%poi[0])==0) ? 1 : 0));//Set the speed
    }
    else if ((i==15)&&(speed>=poi[i])) return (max ? 0xff : 0xf);//It is higher, than the maximal speed
    else
    {
      //Other cases: "poi[i]" is >=speed
      if(max)
      {
        //Maximal speed
        int whole=((speed-poi[i-1])*16)/(poi[i]-poi[i-1])+(((((speed-poi[i-1])*16)%(poi[i]-poi[i-1]))==0) ? 0 : 1);//Set speed
        return (uint8_t) (((i-1)<<4)+0xf+whole);//Set the speed
      }
      //Minimal speed case
      else if (poi[i]==speed) return i;//Equal case
      else return i-1;//For safety reason, we will set lower speed
    }
  }
  
  Motor* Motor::MOVING_motor=0;//Motor, which actually moves
  bool Motor::stop_indicator=false;//Stop indicator value 
  std::mutex Motor::mut;//Mutex to save MOVING_motor operations
  
  Motor::Motor(Camac &camcomm, std::string motor_name, uint8_t slotno, uint32_t hilimp, uint32_t lowlimp, int bit, ORDER ord, int acc_diff)
  //Constructor: the camac communicator object; the name of the motor, slot number of the motor controller card; low and high limit of the motor in uniform pulses; order of the encoded value (in the case of STRAIGHT, the pulses are the same as uniform pulses, otherwise this class communicates the motor with the corresponding pulses and with program objects via uniform pulses: -limit...+limit); the acceptable smallest difference in positioning; the encoder range in bits
  {
    communicator=&camcomm;//Communicates with camac device
    addr=slotno;//Slot number
    highlimit=(hilimp>lowlimp) ? hilimp : lowlimp;//Higher limit in uniform pulses
    lowlimit=(hilimp>lowlimp) ? lowlimp : hilimp;//Low limit in uniform pulses
    order=ord;//Ordering of the variables
    bits=bit;//Available halfbits
    half=1<<(bits-1);//Half of the available range in numbers
    acceptable_difference=acc_diff;//Set the acceptable difference
    setName(motor_name);
    last_position=0xffffffff;//Start value of the position
    idir=true;//Indicate incresing direction to move to
  }
  
  bool Motor::moveTo(uint32_t target)
  //Moves the motor to the target uniform pulse position, returns with true, if not any troubles happened
  {
    //Before anything, checking the actual position
    if(!critical_check_allocate(this)) return false;//Try to get access to exclusive use of motors
    stop_indicator=false;//Reset stop indicator
    //Before anything, checking the actual position
    uint32_t actual=getposition();
    bool move_accept=false;
    //Check communication error
    if (actual==COMMUNICATION_ERROR)
    {
      stop_indicator=false;//Reset stop indicator
      critical_deallocate(this);
      return false;
    }
    //Checking limit conditions
    idir= (target>actual);
    if ((actual<lowlimit)&&idir&&(target<=highlimit)) move_accept=true;
    else if ((actual>highlimit)&&(!idir)&&(target>=lowlimit)) move_accept=true;
    else if ((target>=lowlimit)&&(target<=highlimit)) move_accept=true;
    //Check acceptable_difference condition
    if (move_accept && (abs((int)(target-actual))<=acceptable_difference)) move_accept=false;
    //Check non-stop indicator condition
    if(!move_accept || !check_not_stop())
    {
      stop_indicator=false;//Reset stop indicator
      critical_deallocate(this);
      return false;
    }
    //Try to prepare to move and start it
    int i=0;//Loop variable
    while((!setposHW_preparetomove())&&(i++<5)) sleep(1u);
    if(i<5) { i=0; while((!setposHW(uniform2machine(target)))&&(i++<5)) sleep(1u);}
    if(i>=5)
    {
      setposHW_postmove();
      stop_indicator=false;//Reset stop indicator
      critical_deallocate(this);
      return false;
    }
    //Check again the stop flag
    if(!check_not_stop())
    {
      stop_indicator=false;//Reset stop indicator
      critical_deallocate(this);//The stopHW already do the task of postmove
      return false;
    }
    //Start monitored move
    std::thread move_control_thread(monitored_move,target,this,0);
    move_control_thread.detach();//Allow the tread to run in background
    return true;//New thread has been successfully started
  }
  
  void* Motor::monitored_move(const uint32_t target_pos, Motor *pThis, int depth)
  //void* Motor::monitored_move(void *arguments)
  //This function supervises the move to get into the required position. Inputs are the target position, the pointer and the number of trials made before
  {
    uint32_t act_pos;//Actual position
    uint32_t prev_pos;//Previous position
    int counter=0;//Counts the number of position checks
    int change_counter=0;//This variable contains the last counter value, where any move has been happened
    //int equality_counter=0;//Counts the number of consecutive act_pos=prev_pos appearence
    int oscillation_counter=0;//This counter counts the going through on target position counts
    bool regular_end=false;//Indicates, that the motor reached the requested position
    bool stop_end=false;//Indicates, that move stopped due to user stop request
    bool start_new_stacked=false;//Indicates, that the actual move has been stacked and a new one is required
    bool start_new_oscillate=false;//Indicates, that the actual move started to oscillate
    do
    {
      sleep((unsigned int) pThis->CHECK_TIME);//wait some time until the next checking: CHECK_TIME is in seconds
      counter++;
      if(!(pThis->check_not_stop())) stop_end=true;//---------- CHECK STOP----------
      else
      {
	prev_pos=pThis->last_position;//Update position
	act_pos=pThis->getposition();//Check current position
	if(act_pos!=COMMUNICATION_ERROR)
	{
	  if(target_pos!=act_pos) pThis->idir= (target_pos>act_pos);//Update correct direction (have an importance for limit switches in the case of overrun)
	  //Using the LAM functionality
	  if(pThis->isFinished() && ((act_pos=pThis->getposition())==target_pos)) regular_end=true;
	  else if(!(pThis->check_not_stop())) stop_end=true;//---------- CHECK STOP----------
	  else if(counter>1)
	  {
	    //actual!=previous, skip first count, where
	    if(prev_pos!=act_pos) change_counter=counter;//Set the last change index
	    else if (((counter-change_counter)>=pThis->MAX_TIMES)|| (pThis->isFinished()))
	    {
	      //The move of the motor probably stacked
#ifdef DEBUG_MOTOR
              std::cout<<"Seems to be stacked..."<<std::endl;
#endif
	      if( ( act_pos>target_pos ? act_pos-target_pos : target_pos-act_pos ) <=static_cast<uint32_t>(pThis->acceptable_difference)) regular_end=true;//But this difference looks acceptable
	      else start_new_stacked=true;//We should start a new cycle....
	    }
	    if (((act_pos>target_pos)&&(prev_pos<target_pos))||((act_pos<target_pos)&&(prev_pos>target_pos)))
	    {
	      oscillation_counter++;
#ifdef DEBUG_MOTOR
              std::cout<<"Oscillation detected: "<<oscillation_counter<<std::endl;
#endif
	      if(oscillation_counter==3) start_new_oscillate=true;//The motor started to oscillate around the final position, so we should start a new cycle
	    }
	    if( !regular_end && !start_new_stacked && !start_new_oscillate && !(pThis->check_not_stop())) stop_end=true;//---------- CHECK STOP----------
	  }
	}
      }
    }while( !regular_end && !stop_end && !start_new_oscillate && !start_new_stacked);
#ifdef DEBUG_MOTOR
    if(regular_end) std::cout<<"LAM asserted"<<std::endl;
    if(stop_end) std::cout<<"STOP asserted"<<std::endl;
    if(start_new_oscillate) std::cout<<"Oscillating"<<std::endl;
    if(start_new_stacked) std::cout<<"Stacked"<<std::endl;
#endif
    
    if((start_new_stacked || start_new_oscillate)&&(depth>=MAX_DEPTH))
    {
      //perform Stop() operation
      pThis->Stop();
      pThis->check_not_stop();
      stop_end=true;
    }
    
    if(regular_end||stop_end)
    {
      //The move task has been finished properly or by user stop
      int i=0;
      if(!stop_end) while((!(pThis->setposHW_postmove()))&&(i++<5)) sleep(1u);//Stop the enable status
      pThis->stop_indicator=false;//Reset indicator
      critical_deallocate(pThis);
    }
    else
    {
      //We should start a new cycle by sending a setposHW command
      int i=0;
      if(start_new_oscillate)
      {
	//Oscillation: try to stop, then prepare a new run
	pThis->Stop();
	pThis->check_not_stop();
	pThis->stop_indicator=false;//Reset indicator
      }
      else while((!(pThis->setposHW_postmove()))&&(i++<2)) usleep(100000u);//Stop the enable status of stacked motor  
      act_pos=pThis->getposition();//Check current position
      if(act_pos!=COMMUNICATION_ERROR)
      {
	while((!(pThis->setposHW_preparetomove()))&&(i++<4)) usleep(100000u);
        if(i<4)
        {
	  i=0;
	  while((!(pThis->setposHW(pThis->uniform2machine(target_pos))))&&(i++<5)) sleep(1u);
	  Motor::monitored_move(target_pos, pThis, depth+1);
        }
        else
	{
	  //If it were not start in any case: stop 
	  pThis->Stop();
	  pThis->stop_indicator=false;//Reset indicator
          critical_deallocate(pThis);
	}
      }
      else
      {
	//If it were trouble with the communication on actual position
	pThis->Stop();
	pThis->stop_indicator=false;//Reset indicator
        critical_deallocate(pThis);
      }
    }
    return 0;
  }
  
  bool Motor::setLimit(uint32_t lim, bool hilow)
  //Set the position limit of the motor in uniform pulses. It will be the high (t) or the low (f) limit. It returns, if not any troubles happened
  {
    if (hilow && (lim>lowlimit))
    {
      highlimit=lim;
      return true;
    }
    else if (!hilow && (lim<highlimit))
    {
      lowlimit=lim;
      return true;
    }
    else return false;
  }
  
  uint32_t Motor::getposition()
  //Returns with the actual position of the motor in uniform pulses. 
  {
    //Getting the hardware dependent position
    uint32_t result;
    int i=0;
    while(!getposHW(result) && (i++<5)) sleep(1u);//Transfer settings
    if (i>=5) return COMMUNICATION_ERROR;
    else
    {
      last_position=machine2uniform(result);//Otherwise converting the result into uniform pulses
      return last_position;
    }
  }
  
  bool Motor::critical_check_allocate(Motor *mot)
  //This method tries to allocate the motor as moving motor, returns with true, if it has been succeeded, otherwise (if other motor is on move) returns with false
  {
    std::lock_guard<std::mutex> lock(mut);//Lock mutex
    bool retval=(MOVING_motor==0);
    if(retval) MOVING_motor=mot;
    return retval;//Return and release mutex
  }
  
  bool Motor::critical_deallocate(Motor *mot)
  //This method deallocates the motor from moving list
  {
    std::lock_guard<std::mutex> lock(mut);//Lock mutex
    bool retval=(mot==MOVING_motor);
    if(retval) MOVING_motor=0;
    return retval;
  }
  
  bool Motor::check_not_stop()
  //This method checks the indicator set by the Stop method. If it has been set, it performs emergencyStop operation (but does not null the MOVING_motor pointer) and returns with FALSE! If the indicator has not been asserted, it returns with TRUE
  {
    if(stop_indicator)
    {
      int i=0;
      while(!(MOVING_motor->emergencyStop()) && (i++<5)) sleep(1u);//Set emergency stop
      return (i>=5);//Set successful or unsuccessful stop (successful transmit: false, faled: false)
    }
    else return true;
  }
  
  uint32_t Motor::uniform2machine(uint32_t uniform)
  //Do the conversion from uniform to machine pulses
  {
    switch(order)
    {
      case STRAIGHT: return uniform; break;
      case SHIFT: return (uniform+half)&((1<<bits)-1); break;
      case INVERSE: return half&uniform ? uniform-half : (half+(half-1))^uniform; break;
      default: return half&uniform ? uniform-half : ((half+(half-1))^uniform)+1;
    }
  }
  
  uint32_t Motor::machine2uniform(uint32_t machine)
  //Do the conversion from machine to uniform pulses
  {
    switch(order)
    {
      case STRAIGHT: return machine; break;
      case SHIFT: return (machine+half)&((1<<bits)-1); break;
      case INVERSE: return half&machine ? (half+(half-1))^machine : half|machine; break;
      default: return half&machine ? ((half+(half-1))^machine)+1 : half|machine;
    }
  }
 
  //ParallelMotor -----
  ParallelMotor* ParallelMotor::ACTUAL_display=0;//Initialization of motor display pointer
  bool ParallelMotor::HILIM_off=false;//High limit switch is off 
  bool ParallelMotor::LOLIM_off=false;//Low limit switch is off 
 
  bool ParallelMotor::getposHW(uint32_t &result)
  //Returns with the actual position of the motor.
  {
    if(!change_display()) return false;//It is not allowed to modify the actual display, except this is the moving motor
    if(MOVING_motor==this) interreg->Check();
    return communicator->getData(addr, DISP_VAL_F, displayslot, result, Camac::STATUS_Q);//Ask actual position
  }
  
  bool ParallelMotor::setposHW_postmove()
  //Switch off the motors, disable and clear LAMS 
  {
    bool retval= powerOnOff(false) && controller->EnableDisableLAM(1, Positioner::StepMotorController::LAM_STOP|Positioner::StepMotorController::LAM_OUT|Positioner::StepMotorController::LAM_STACK, false);//power off and disable LAMS
    if (!retval) return false;
    //Get the status register
    uint32_t streg=controller->ReadStatusRegister(MOT_CNTRL_NO);
    if(streg==StepMotorController::COMM_ERROR) return false; 
    else
    {
      if((streg&StepMotorController::ST_LAM_STACK_SET)==StepMotorController::ST_LAM_STACK_SET)
      {
        retval=controller->ClearLAM(MOT_CNTRL_NO,6);
#ifdef DEBUG_MOTOR
        std::cout<<"LAM 6 detected! Wrong direction (6 alone) or motor stop due to wrong running parameters (1+6)"<<std::endl;
#endif
      }

      if(retval&((streg&StepMotorController::ST_LAM_STOP_SET)==StepMotorController::ST_LAM_STOP_SET))
      {
        retval=controller->ClearLAM(MOT_CNTRL_NO,1);
#ifdef DEBUG_MOTOR
        std::cout<<"LAM 1 detected! Motor stopped: at reqested position or \"Stop\" event (1 alone) or positioning interrupted (1+5/6)"<<std::endl;
#endif
      }

      if(retval&((streg&StepMotorController::ST_LAM_OUT_SET)==StepMotorController::ST_LAM_OUT_SET)) 
      {
        retval=controller->ClearLAM(MOT_CNTRL_NO,5);
#ifdef DEBUG_MOTOR
        std::cout<<"LAM 5 detected! The allowance interval has not been reached (5+1)"<<std::endl;
#endif
      }
    }
    return retval;
  }
  
  void  ParallelMotor::Action(IOOperation::InterruptHandle *ih, uint32_t mask)
  //Performs the action related to the mask and interrupt register 
  {
    //First of all, check, that this is the correct receiver (in the case of motor moving: that one, other case: others)
    if((ih==interreg)&&((MOVING_motor==this)||(MOVING_motor==0)))
    {
      HILIM_off=((mask&IOOperation::InterruptRegister::LIMIT_HI_OFF)==IOOperation::InterruptRegister::LIMIT_HI_OFF);
      LOLIM_off=((mask&IOOperation::InterruptRegister::LIMIT_LO_OFF)==IOOperation::InterruptRegister::LIMIT_LO_OFF);
    }
    return;
  }
  
  bool ParallelMotor::powerOnOff(bool on_off)
  //Switch on (t) the power for the current motor or switch off the motor
  {
    if(MOVING_motor==this)       //Avoid situation of power off or on other motors
    {
      if(on_off) change_display();//Lock the display
      bool retval=on_off ? power_control->Set(IOOperation::OutputRegister::MOTOR_ON|mot_pow_arr[(int) output_number],false) : power_control->Reset(IOOperation::OutputRegister::MOTOR_ON|mot_pow_arr[(int) output_number]);//Switch off/on the motor
      //Check the status
      int i=0;
      interreg->Check();
      if((retval)&&(on_off))
      {
	//Start to move
	while((!((idir&&HILIM_off)||((!idir)&&LOLIM_off))) && (i++<3)) {usleep(200000u); interreg->Check();};//Waiting for one of the two limit switch off becomes true
	return (i<3);
      }
      if((retval)&&(!on_off))
      {
	//Stop the move
	while((HILIM_off||LOLIM_off) && (i++<6)) {usleep(200000u); interreg->Check();};//Waiting for both limit switch off becomes false
	return (i<6);
      }
      else return false;//Bad communication
    }
    else return false;//This is not the right motor
  }
  
  bool ParallelMotor::change_display()
  //this method try to change the display: true: success, false: motor is moving, changing to other is not allowed
  {
    if((MOVING_motor!=0)&&(MOVING_motor!=this)) return false;//It is not allowed to modify the actual display, except this is the moving motor
    if(ACTUAL_display!=this) //Does it different from the actual one?
    {
      if (!( ((ACTUAL_display!=0) ? ACTUAL_display->displayConnect(false) : this->displayConnect(false)) && this->displayConnect(true) )) return false;//Stop reading, switch to the present motor - return in the case of error
      else ACTUAL_display=this;
    }
    return true;
  }
  
  bool CushionAPMotor::COM_air=false;//Compressed air is on indicator
  
  void CushionAPMotor::Action(IOOperation::InterruptHandle *ih, uint32_t mask)
  //Changes the status of the "comair" compressed air on indicator
  {
    if(ih==interreg) COM_air=((mask&IOOperation::InterruptRegister::COMPRESSEDAIR_ON)==IOOperation::InterruptRegister::COMPRESSEDAIR_ON);
    ParallelMotor::Action(ih, mask);//Call parent class routine
  }
  
  bool CushionAPMotor::CushionOnOff(bool on, bool evacuate)
  //Switch on/off air cushions, or evacuate the air from the system fully
  {
    bool retval=true;
    if(on)
    {
      //Should be switched it on
      retval=power_control->Set(IOOperation::OutputRegister::VLV_CA_R | IOOperation::OutputRegister::VLV_CA_H,false);//Switch on air valves
      if(retval)
      {
	if(!evacuate)
	{
	    int counter=0;
	    do
	    {
	      interreg->Check();
	      if(!COM_air) usleep(200000u);
	      
	    }while((!COM_air)&&(counter++<10));//Check compressed air on status each 0.2 seconds 
	    if (counter>9)
	    {
	      CushionOnOff(false,false);//If not success: switch off the air-cusions
	      retval=false;
	    }
	}
	else
	{
	    sleep(2);//Wait 2 seconds
	    retval=CushionOnOff(false,false);//Switch off motor controller board
	}
      }
    }
    else
    {
      //Should be switched it off
      retval=power_control->Reset(IOOperation::OutputRegister::VLV_CA_R|IOOperation::OutputRegister::VLV_CA_H);//Switch off air valves
      while(COM_air){interreg->Check();usleep(200000u);};//Check switch off status
    }
    return retval;
  }
}
