#include <cstddef>
#include <iostream>
#include <mutex>
#include <string>
#include <thread>
#include <unistd.h>
#include "Camac.h"
#include "IOOperation.h"

#ifndef MOTOR_H
#define MOTOR_H

namespace Positioner
{
  enum ORDER {STRAIGHT,SHIFT,INVERSE,INVERSE_SKIP};
  //Describes the order of motor encoders: straight (only positive range), shift (after the highest positive, the most highest modulus negative follows), inverse (after the highest positive, the most smallest modulus negative follows), inverse_skip (same as inverse, but it avoid double zero trouble: after the highest positive, a zero(should skip), then the most smallest modulus negative follows)

  class StepMotorController
  //This class describes the functions of a EL458 kind step motor controller
  {
     public:
       //static members-status double word
       static const uint32_t ST_LAM_STOP_SET=0x1;//LAM 1 set
       static const uint32_t ST_LAM_START_SET=0x2;//LAM 2 set
       static const uint32_t ST_LAM_LIM_SW_SET=0x4;//LAM 3 set
       static const uint32_t ST_LAM_ZERO_SET=0x8;//LAM 4 set
       static const uint32_t ST_LAM_STOP_DISABLED=0x10;//LAM 1 disable
       static const uint32_t ST_LAM_START_DISABLED=0x20;//LAM 2 disable
       static const uint32_t ST_LAM_LIM_SW_DISABLED=0x40;//LAM 3 disable
       static const uint32_t ST_LAM_ZERO_DISABLED=0x80;//LAM 4 disable
       static const uint32_t ST_MOTOR_RUNNING=0x100;//Motor running
       static const uint32_t ST_START_PRESSED=0x200;//Start botton pressed
       static const uint32_t ST_LIMIT_BW=0x400;//Backward limit reached
       static const uint32_t ST_LIMIT_FW=0x800;//Forward limit reached
       static const uint32_t ST_AT_DESTINATION=0x1000;//Destination reached
       static const uint32_t ST_INCR_LAMP_DEFECT=0x2000;//Incr. lamp defect 
       static const uint32_t ST_PRELIMIT_BW=0x4000;//Pre-Backward limit reached
       static const uint32_t ST_PRELIMIT_FW=0x8000;//Pre-Forward limit reached
       static const uint32_t ST_LAM_OUT_SET=0x10000;//LAM 5 set
       static const uint32_t ST_LAM_STACK_SET=0x20000;//LAM 6 set
       static const uint32_t ST_LAM_OUT_DISABLED=0x40000;//LAM 5 disabled
       static const uint32_t ST_LAM_STACK_DISABLED=0x80000;//LAM 6 disabled
       static const uint32_t ST_LAM_ANY_SET=0x100000;//LAM 1-7 set (in the case of motor 1: for all motors)
       static const uint32_t ST_LAM_CHECKSUM_SET=0x200000;//LAM 7 set
       static const uint32_t ST_LAM_ANY_DISABLED=0x400000;//LAM 1-7 disabled (in the case of motor 1: for all motors)
       static const uint32_t ST_LAM_CHECKSUM_DISABLED=0x800000;//LAM 7 disabled
       //static members - LAMS
       static const uint32_t LAM_STOP=ST_LAM_STOP_DISABLED;//1: position reached, stopped, positioning error
       static const uint32_t LAM_START=ST_LAM_START_DISABLED;//2: command from manual, start button pressed
       static const uint32_t LAM_LIM_SW=ST_LAM_LIM_SW_DISABLED;//3: limit switch or weaker limit switch reached
       static const uint32_t LAM_ZERO=ST_LAM_ZERO_DISABLED;//4: zero position reached
       static const uint32_t LAM_OUT=ST_LAM_OUT_DISABLED;//5: motor stopped out from the allowed range
       static const uint32_t LAM_STACK=ST_LAM_STACK_DISABLED;//6: motor stack or went to wrong direction
       static const uint32_t LAM_CHECKSUM=ST_LAM_CHECKSUM_DISABLED;//7: checksum error
       static const uint32_t LAM_ALL=LAM_STOP|LAM_START|LAM_LIM_SW|LAM_ZERO|LAM_OUT|LAM_STACK|LAM_CHECKSUM;//set/unset all LAMs
       //static members: timing, errors, etc.
       static const uint32_t BAD_MOTOR=0x1<<30;//Motor number does not exist
       static const uint32_t COMM_ERROR=0x1<<31;//Communication error
       static const useconds_t CNTR_WAIT=1000u;//Wait 200 ms before the next command
       //methods
       StepMotorController(Camac &camcomm, uint8_t slotno);//Constructor: camac communicator object and slot number are the parameters 
       ~StepMotorController(){comm=0;};//Destructor
       std::string getCAMName(){return comm->getName();}//Returns with the ID of the CAMAC controller
       uint8_t getStation(){return slot;};//Returns with the station number
       uint32_t ReadPosition(uint8_t motor_no){if(!check_motors(motor_no)) return BAD_MOTOR; uint32_t position=0; if(comm->getData(motor_no-1+BASE_A, READ_F, slot, position, Camac::STATUS_X| Camac::STATUS_Q)) return position; else return COMM_ERROR;};//Read the current position of the motor
       uint32_t ReadModulCharacteristic(){uint32_t characteristic=0; if(comm->getData(MOD_CHAR_A, READ_F, slot, characteristic, Camac::STATUS_X| Camac::STATUS_Q)) return characteristic; else return COMM_ERROR;};//Reads the module characteristics
       uint32_t ReadStatusRegister(uint8_t motor_no);//Read the status register for the requested motor
       bool TestAnyLAMAnyMotor(){return comm->Control(BASE_A, TEST_F, slot, Camac::STATUS_X| Camac::STATUS_Q); };//Test LAM 1-7 for all motors 1-6
       bool TestAnyLAMMotor(uint8_t motor_no){if(!check_motors(motor_no)) return false; return comm->Control(BASE_A+(((motor_no+1)%2)*8), TEST_F+(motor_no>2 ? 4 : 0), slot+ (motor_no>4 ? 1 : 0), Camac::STATUS_X| Camac::STATUS_Q);};//Test LAM 1-7 for the requested motor
       bool TestLAM(uint8_t motor_no, uint8_t lam){if(!check_motors(motor_no)) return false; return comm->Control(BASE_A+(((motor_no+1)%2)*8)+lam, TEST_F+(motor_no>2 ? 4 : 0), slot+ (motor_no>4 ? 1 : 0), Camac::STATUS_X | Camac::STATUS_Q);};//Test for motor and specific LAM
       bool ClearLAM(uint8_t motor_no, uint8_t lam){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+(((motor_no+1)%2)*7)+lam, CLEAR_F+(motor_no>2 ? 4 : 0), slot+ (motor_no>4 ? 1 : 0), Camac::STATUS_X | Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Clears the specific LAM for the given motor
       bool MoveTo(uint8_t motor_no, uint32_t destination){if(!check_motors(motor_no)) return false; bool retval=comm->setData(BASE_A+motor_no-1, WRITE_F, slot, destination, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Move the motor to destination
       bool SetRunningParameters(uint8_t motor_no, double ramp_sec, int min_speed_step_sec, int max_speed_step_sec, bool exact_position=true, bool over_zero_run_enable=false);//Set the running parameters for motor: the ramp in seconds, the minimal and maximal speed in step/seconds, the exact(t)/or +/-1 position tolerance (f), enabling overrun trough zero point if the requested position is closer that way
       bool EnableDisableLAM(uint8_t motor_no, uint32_t lam, bool enable);//Enables (t) or disables(f) the set of LAMs (separated by "|"), or LAM_ENABLE_ALL/LAM_DISABLE_ALL for motor.
       bool Stop(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+motor_no-1, STOP_F, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Stops the motor (set LAM1)
       bool StartSlowFW(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+((motor_no-1)%2)*7+1, MANUAL_MOVE_F+((motor_no-1)/2)*2, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Starts slow forward for the given motor
       bool StartSlowBW(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+((motor_no-1)%2)*7+2, MANUAL_MOVE_F+((motor_no-1)/2)*2, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Starts slow backward for the given motor
       bool StartFastFW(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+((motor_no-1)%2)*7+3, MANUAL_MOVE_F+((motor_no-1)/2)*2, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Starts fast forward for the given motor
       bool StartFastBW(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+((motor_no-1)%2)*7+4, MANUAL_MOVE_F+((motor_no-1)/2)*2, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Starts fast backward for the given motor
       bool OneStepFW(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+((motor_no-1)%2)*7+5, MANUAL_MOVE_F+((motor_no-1)/2)*2, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//One step forward for the given motor
       bool OneStepBW(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+((motor_no-1)%2)*7+6, MANUAL_MOVE_F+((motor_no-1)/2)*2, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//One step backward for the given motor
       bool GoToReference(uint8_t motor_no){if(!check_motors(motor_no)) return false; bool retval=comm->Control(BASE_A+((motor_no-1)%2)*7+7, MANUAL_MOVE_F+((motor_no-1)/2)*2, slot, Camac::STATUS_X| Camac::STATUS_Q); if(retval) {usleep(CNTR_WAIT); return true;} else return false;};//Goes to the reference point with the given motor
       bool TestStartButtonPressed(uint8_t motor_no){if(!check_motors(motor_no)) return false; return comm->Control(BASE_A+(motor_no-1)*2+1, MANUAL_TEST_F, slot, Camac::STATUS_X| Camac::STATUS_Q);};//Check the start button pressed for the given motor
       bool CheckDestinationReached(uint8_t motor_no){if(!check_motors(motor_no)) return false; return comm->Control(BASE_A+(motor_no-1)*2+2, MANUAL_TEST_F, slot, Camac::STATUS_X| Camac::STATUS_Q);};//Checks destination = current position for the given motor
     private:
       //static members
       static const uint8_t READ_F=0;//Read function code
       static const uint8_t STREG_F=1;//Status register function code
       static const uint8_t TEST_F=8;//Test LAM function code
       static const uint8_t CLEAR_F=10;//Clear LAM function code
       static const uint8_t WRITE_F=16;//Write function code
       static const uint8_t LAM_ED_F=17;//LAM Enable-disable function code
       static const uint8_t STOP_F=24;//Stop function code
       static const uint8_t MANUAL_MOVE_F=26;//Manual move function code
       static const uint8_t MANUAL_TEST_F=27;//Manual test function code
       static const uint8_t BASE_A=0;//Basic subcode
       static const uint8_t MOD_CHAR_A=14;//Module characteristic subcode
       static const uint8_t AUX_STREGISTER_A=6;//Auxiliary status register (base) subcode
       static const uint8_t RUNPAR_A=6;//Running parameters (base) subcode
       static const uint8_t AUX_RUNPAR_A=12;//Auxiliary running parameters (base) subcode
       static const uint8_t SECOND_MANUAL_A=7;//Second manual move parameters (base) subcode
       static const uint16_t minspeed[16];//Minimal speed array (step/sec)
       static const uint16_t maxspeed[16];//Maximal speed array
       static const double   ramptime[8];//Ramp in seconds
       //private members
       Camac* comm;//Communicates with CAMAC
       uint8_t slot;//Slot number
       bool auxrun[6];//Auxiliary running parameters: true: +/-1 difference is allowed
       bool auxconnected[6];//Auxiliary running parameters: it is true, if it has been connected earlier ([i] true: auxrun[i] represents the real situation )
       uint32_t lam_status[6];//Stores the LAM 1..7 status on each motors (enabled:1, disabled:0)
       //private methods
       bool check_motors(uint8_t motor_no){return ((motor_no>0)&&(motor_no<=6));}//Checking the conformity of the input parameter
       uint8_t ramp2code(double ramp); //This function converts the ramp in sec units into coded output.
       uint8_t speed2code(int speed, bool max);//This function converts the speed in step/sec units into coded output. It does for minimal (false) or maximal (true)
  };

  class Motor
  //This class describes a motor object, which has a communication channel and slot, able to give information about its position and goes to the required position -- routines tested by 2theta controller (16 bits, INVERSE_SKIP motor)
  //The basic concept: only one motor can be move at the same time, it is not possible to reach CAMAC during that time, except the Move routine
  {
    public:
      //Public routines
      Motor(Camac &camcomm, std::string motor_name, uint8_t slotno, uint32_t hilimp, uint32_t lowlimp, int bits, ORDER ord, int acc_diff=0);
      //Constructor: the camac communicator object; name of the motor; slot number of the motor controller card; low and high limit of the motor in uniform pulses; the encoder range in bits; order of the encoded value (in the case of STRAIGHT, the pulses are the same as uniform pulses, otherwise this class communicates the motor with the corresponding pulses and with program objects via uniform pulses: -limit...+limit); the acceptable smallest difference in positioning
      ~Motor(){communicator=NULL;};//Destructor
      //Routines
      static bool isAllStopped(){return (MOVING_motor==0);}//Returns true, if all of the move operation has been finished
      uint32_t getPosition(){return ((MOVING_motor!=0) ? last_position: getposition());};//Returns the actual position of the motor in uniform pulses or the last checked position, if any motor is MOVING
      bool moveTo(uint32_t target);//Moves the motor to the target uniform pulse position, returns with true, if not any troubles happened
      static bool Stop(){if (MOVING_motor==0) return false; else {stop_indicator=true; return true;}}//Stops the current move task
      bool setLimit(uint32_t lim, bool hilow=true);//Set the position limit of the motor in uniform pulses. It will be the high (t) or the low (f) limit. It returns, if not any troubles happened
      uint32_t getLimit(bool hilow=true){return hilow ? highlimit : lowlimit;};//Returns with the high (t), or low (f) limit in uniform pulses
      std::string getName(){return name;};//Returns with the name of the motor
      void setName(std::string motor_name){name=motor_name; return;};//Sets the name of the motor
      std::string getCAMName(){return communicator->getName();}//Returns with the ID of the CAMAC controller
      uint8_t getAddress(){return addr;};//Returns with the station number
      int getEncoderBits(){return bits;};//Returns with the number of bits of the encoder
      int getAcceptedDifference(){return acceptable_difference;};//Returns with the number of acceptable differences
      //static variables
      static const uint32_t COMMUNICATION_ERROR=0xffffffff;//Communication error
    protected:
      Camac *communicator;//Communicates with camac device
      uint8_t addr;//Slot number
      static Motor *MOVING_motor;//Motor, which actually moves
      bool idir;//Increasing(t) direction at motor move or decreasing(f)
      //Protected method
      bool acceptedzero(){return (acceptable_difference==0);}//If accepted difference is zero, it returns with true
    private:
      //private members
      uint32_t highlimit;//Higher limit in uniform pulses
      uint32_t lowlimit;//Low limit in uniform pulses
      ORDER order;//Ordering of the variables
      uint32_t half;//Half of the available range in numbers
      int bits;//Available number of bits
      int acceptable_difference;//Smallest acceptable difference at motor positioning
      std::string name;//Stores the name of the motor
      uint32_t last_position;//Last position of the motor
      static std::mutex mut;//Mutex for moving motor access
      static bool stop_indicator;//Stop indicator 
      //constants
      const static int CHECK_TIME=1;//Checking time seconds of the position of the moving motor
      const static int MAX_TIMES=16/CHECK_TIME;//The maximum allowed time in the unit of CHECK_TIME before the motor stop, if encoder shows the same position
      const static int MAX_DEPTH=2;//Maximal number of trials in move
      //private methods
      uint32_t getposition();//Returns the actual position of the motor in uniform pulses
      //static methods
      static bool critical_check_allocate(Motor *mot);//This method tries to allocate the motor as moving motor, returns with true, if it has been succeeded, otherwise (if other motor is on move) returns with false
      static bool critical_deallocate(Motor *mot);//This method deallocates the motor from moving list
      static bool check_not_stop();//This method checks the indicator set by the Stop method. If it has been set, it performs emergencyStop operation (but does not null the MOVING_motor pointer) and returns with FALSE! If the indicator has not been asserted, it returns with TRUE
      //conversion, recursive run
      static void *monitored_move(const uint32_t target, Motor *pThis, int depth);//This function supervises the move to get into the required position. Inputs are the target position, the number of attemption to move before and a pointer to the present class
      //conversion, recursive run
      uint32_t uniform2machine(uint32_t uniform);//Do the conversion from uniform to machine pulses
      uint32_t machine2uniform(uint32_t machine);//Do the conversion from machine to uniform pulses
      //virtual methods
      virtual bool getposHW(uint32_t &result)=0;//Implements the hardware dependent reading of the position. It returns with true, if the preparations are finished, with false if error happened or hw limit 
      virtual bool setposHW_preparetomove(){return true;};//Implements the preparation to move before starting the motor move. switch does not allows the move
      virtual bool setposHW(uint32_t target)=0;//Implements the hardware dependent moving of the motor. It returns with true, if the machine understand the command
      virtual bool setposHW_postmove(){return true;};//Implements the task to do after successful motor move. It returns with true, if it was successful
      virtual bool isFinished(){return false;};//Does the move has been finished (means: is the motor stopped?)
      virtual bool emergencyStop()=0;//Stops the motor move/returns to that position, which was checked just after this function call
  };

 
  class AngleMotor: virtual public Motor
  //This class add postion<-> angle transformation properties -- except constructor 2nd, it has been tested by 2theta controller
  {
    public:
      //Public routines
      AngleMotor(Camac &camcomm, std::string motor_name, uint8_t slotno, uint32_t hilimp, uint32_t lowlimp, int bits, ORDER ord, double shift, double slope, int acc_diff=0):Motor(camcomm, motor_name, slotno, hilimp, lowlimp, bits, ord, acc_diff) { trsf_shift=shift; trsf_slope=slope;};
      //Constructor: the camac communicator object; name of the motor; slot number of the motor controller card; low and high limit of the motor in uniform pulses; the encoder range in bits; order of the encoded value (in the case of STRAIGHT, the pulses are the same as uniform pulses, otherwise this class communicates the motor with the corresponding pulses and with program objects via uniform pulses: -limit...+limit); shift and slope are corresponds the coeffitients of the angle=(position-shift)*slope expression; finally the acceptable smallest difference in positioning
      AngleMotor(Camac &camcomm, std::string motor_name, uint8_t slotno, double hilim, double lowlim, int bits, ORDER ord, double shift, double slope, int acc_diff=0):Motor(camcomm, motor_name, slotno, 0x1, 0x0, bits, ord, acc_diff) { trsf_shift=shift; trsf_slope=slope; if(slope>0.0) {setALimit(hilim,true); setALimit(lowlim,false);} else {setALimit(hilim,false); setALimit(lowlim,true);} };
      //Constructor: the camac communicator object; name of the motor; slot number of the motor controller card; low and high limit of the motor in degrees; the encoder range in bits; order of the encoded value (in the case of STRAIGHT, the pulses are the same as uniform pulses, otherwise this class communicates the motor with the corresponding pulses and with program objects via uniform pulses: -limit...+limit); shift and slope are corresponds the coeffitients of the angle=(position-shift)*slope expression; finally the acceptable smallest difference in positioning - NOT TESTED
      ~AngleMotor(){while(!(this->isAllStopped())) usleep(1000u);};//Destructor
      //Positions
      double getAPosition(){return pos2angle(Motor::getPosition());};//Returns the actual position of the motor in angles
      bool moveATo(double target){return Motor::moveTo(angle2pos(target));};//Moves the motor to the target angular position, returns with true, if not any troubles happened
      //Limits
      bool setALimit(double lim, bool hilow=true){return Motor::setLimit(angle2pos(lim),((trsf_slope<0.0) ? (!hilow) : hilow));};//Set the position limit of the motor in angles. It will be the high (t) or the low (f) limit. It returns, if not any troubles happened
      double getALimit(bool hilow=true){return pos2angle(Motor::getLimit((trsf_slope<0.0) ? (!hilow) : hilow));};//Returns with the high (t), or low (f) limit in angles
      //Conversion
      double getConversion(bool shift=true){ return  (shift ? trsf_shift : trsf_slope); };//Returns with the conversion parameter: shift(t), slope(f)
      void setConversion(double value, bool shift=true) { if (shift) trsf_shift=value; else trsf_slope=value; return;};//Set the conversion parameter: shift(t), slope(f)
      void setZero(double value){trsf_shift+=value/trsf_slope; return;};//Sets the zero point shift, which was found at the present "value" degrees
    private:
      //members
      double trsf_shift;//degrees=slope*(position-shift)
      double trsf_slope;
      //methods - conversion
      double pos2angle(uint32_t pos) { return trsf_slope*(pos-trsf_shift);};
      uint32_t angle2pos(double angle) { return ((uint32_t) (angle/trsf_slope+trsf_shift));};
  };
  
  class HarwellMotor: public AngleMotor
  //This class describes a Harwell serial motor (contains the full available implementation of Harwell Digital Positioning Unit) - all routines tested by 2 theta controller
  {
    public:
      //Constructor-destructor
      HarwellMotor(Camac &camcomm, std::string motor_name, uint8_t slotno, uint32_t hilimp, uint32_t lowlimp, IOOperation::OutputRegister &power_cntrl, int acc_diff=0): Motor(camcomm, motor_name, slotno, hilimp, lowlimp, 16, INVERSE_SKIP, acc_diff), AngleMotor(camcomm, "", (uint8_t) 0x0, (uint32_t) 0x0, (uint32_t) 0x0, 0, (ORDER) 0, (double) (0x1<<15), (double) 0.01, 0){remote_enable=&power_cntrl; this->getPosition();last_status=0;};//Constructor: the camac communicator object; name of the motor; slot number of the motor controller card; low and high limit of the motor in uniform pulses; OutputRegister object to enable motor operation; number of bits (including sign bit: maximal allowed is 16) finally the acceptable smallest difference in positioning
      ~HarwellMotor(){while(!(this->isAllStopped())) usleep(1000u); remote_enable=0;};//Destructor
      IOOperation::OutputRegister *getOutReg(){return remote_enable;};//Returns with the Output register, which allow this motor to run
      //Other public methods
      bool setManual(){ if(MOVING_motor!=0) return false; else return (communicator->setData(POS,MANUAL_F,addr,0x0,Camac::STATUS_X) && remote_enable->Reset(IOOperation::OutputRegister::HW_MOTORS_ON));};//Sets the manual control mode
      uint8_t getStatus() { if((MOVING_motor==0)&&(getPosition()==COMMUNICATION_ERROR)) return HRW_ST_COMM_ERR; else return last_status;};//Returns with status information
      //constants: public status registers
      static const uint8_t HRW_ST_HIGH_LIMIT=0x1;//Forward
      static const uint8_t HRW_ST_LOW_LIMIT=0x2;//Backward
      static const uint8_t HRW_ST_SPEC_FUNCTION=0x4;
      static const uint8_t HRW_ST_MANUAL_MODE=0x8;
      static const uint8_t HRW_ST_COMM_ERR=0x10;
    private:
      //static members
      static const uint8_t POS=0;//position related subcode
      static const uint8_t DISABLE_A=1;//Disable LAM subcode
      static const uint8_t ENABLE_A=2;//Enable LAM subcode
      static const uint8_t POS_F=0;//position function code
      static const uint8_t TEST_F=8;//Test LAM flip-flop function code 
      static const uint8_t LAM_F=10;//LAM related function code
      static const uint8_t STOP_F=18;//Immediate stop function
      static const uint8_t MANUAL_F=20;//Manual mode function code
      static const uint8_t MOVE_F=22;//motor move function code (Load+Run)
      uint8_t last_status;//Stores the value of the last status
      //Output register
      IOOperation::OutputRegister *remote_enable;//Switch between remote and manual control
      //methods -- inherited virtual methods
      bool isFinished(){ return check_limit(idir)||test_lam_ff();};//It finished, if LAM ff has been set or limit switch reached
      bool getposHW(uint32_t &result) { bool retval=communicator->getData(POS,POS_F,addr,result,Camac::STATUS_X); if (retval) {last_status=(result>>20)&0xf; result&=0xfffff;} else {result=0; last_status=HRW_ST_COMM_ERR;}return retval;};//Implements the hardware dependent reading of the position. It returns with true and the result, if the communication was successful.
      bool setposHW_preparetomove(){return ((!check_limit(idir))&&remote_enable->Set(IOOperation::OutputRegister::HW_MOTORS_ON) && enable_disable_lam(true) && clear_lam_ff());};//Implements the preparation to move before starting the motor move. It returns with true, if the preparations are finished: enable LAM and clear LAM ff
      bool setposHW(uint32_t target){ return communicator->setData(POS,MOVE_F,addr,target,Camac::STATUS_X);};//Implements the hardware dependent moving of the motor. It returns with true, if the machine understand the command
      bool setposHW_postmove(){return (clear_lam_ff() && enable_disable_lam(false) && remote_enable->Reset(IOOperation::OutputRegister::HW_MOTORS_ON));};//Implements the task to do after successful motor move. It returns with true, if the it was successful: clears LAM ff and disables LAM
      bool emergencyStop(){ return (communicator->setData(POS,STOP_F,addr,0x0,Camac::STATUS_X) && setposHW_postmove());};//Stops the motor move
      bool check_limit(bool highlim){bool check=((last_status&( highlim ? HRW_ST_HIGH_LIMIT : HRW_ST_LOW_LIMIT))==( highlim ? HRW_ST_HIGH_LIMIT : HRW_ST_LOW_LIMIT)); if(check) std::cout<<"WARNING: "<<(highlim ? "HIGH" : "LOW")<<" LIMIT SWITCH touched!"<<std::endl; return check;}//Check forward(t) or backward(f) limit condition. Returns true, if limit switch touched.
      //elementary CAMAC methods -- LAM related
      bool test_lam_ff(){return communicator->Control(POS, TEST_F, addr,Camac::STATUS_X|Camac::STATUS_Q);}//Tests LAM flip-flop
      bool clear_lam_ff(){return communicator->Control(POS, LAM_F, addr,Camac::STATUS_X);}//Clears the LAM flip-flop
      bool enable_disable_lam(bool enable=true){return communicator->Control(enable? ENABLE_A : DISABLE_A, LAM_F, addr,Camac::STATUS_X);}//Enable(t)/Disable(f) LAM
  };
  
  
  class ParallelMotor: virtual public Motor, public IOOperation::InterruptReceiver
  //This class describes the functions related to the monochromator motors communicating trough parallel controller hardware
  //The motors M1 and M2 belong this class, and it will be the parent of the remaining motors
  //It contains the full implementation of "Camac Ablese modul" (as disp_slot/displayslot)
  {
    public:
      ParallelMotor(Camac &camcomm, std::string motor_name, uint8_t motor_ablese_number, uint8_t motor_output_number, uint32_t hilimp, uint32_t lowlimp, IOOperation::OutputRegister &power_cntrl, IOOperation::InterruptRegister &interr_register, Positioner::StepMotorController &step_controller, double ramp_sec, int low_step_sec, int hi_step_sec, uint8_t disp_slot=2, int acc_diff=0, int bits=13): Motor(camcomm, motor_name, motor_ablese_number, hilimp, lowlimp, bits, STRAIGHT, acc_diff), InterruptReceiver() {displayslot=disp_slot; power_control=&power_cntrl; controller=&step_controller; output_number=motor_output_number-1; ramp=ramp_sec; lospeed=low_step_sec; hispeed=hi_step_sec; interreg=&interr_register; interreg->Register(this,IOOperation::InterruptRegister::LIMIT_HI_OFF | IOOperation::InterruptRegister::LIMIT_LO_OFF); this->getPosition();};
      //Constructor: the camac communicator object; name of the motor; motor number (in ablese: starts from 1); motor number (output starts from 1); low and high limit of the motor in uniform pulses; motor power selection (output register object); incoming interrupt detection (interrupt register object); step motor controller object; ramp time from low to high speed in seconds; lowest speed of the motor in step/seconds; fastest speed of the motor in step/seconds; display slot number (Camac ablese modul No); the acceptable smallest difference in positioning; number of bits of the position decoder
      ~ParallelMotor(){while(!(this->isAllStopped())) usleep(1000u); power_control=0; controller=0; interreg->UnRegister(this); interreg=0;};//Destructor
      IOOperation::OutputRegister *getOutReg(){return power_control;};//Returns with the Output register, which allow this motor to run
      IOOperation::InterruptRegister *getIrReg(){return interreg;};//Returns with the Interrupt register, which is checked during the run (non-limit switch)
      Positioner::StepMotorController *getSMC(){return controller;};//Returns with the step-motor controller
      uint8_t getDisplayStation(){return displayslot;};//Returns with the CAMAC Ablese modul station number
      uint8_t getMotorOutputNo(){return output_number+1;}//Returns with the output number of the motor
      void setSpeed(int value, bool high){(high ? hispeed : lospeed)=value; return;};//Sets the fastest(t)/lowest(f) speed of the motor
      void setRamp(double ramp_sec){ramp=ramp_sec;}//Sets ramp time from low to high speed in seconds
      int getSpeed(bool high){return (high ? hispeed : lospeed);}//Returns with the fastest(t)/lowest(f) speed of the motor
      double getRamp(){return ramp;}//Returns with ramp time from low to high speed in seconds
      void Action(IOOperation::InterruptHandle *ih, uint32_t mask);//Performs the action related to the mask and interrupt register 
    protected:
      //Move related methods
      bool setposHW_preparetomove(){return powerOnOff(true)&&controller->EnableDisableLAM(MOT_CNTRL_NO, Positioner::StepMotorController::LAM_STOP|Positioner::StepMotorController::LAM_OUT|Positioner::StepMotorController::LAM_STACK, true);};//Switch on the motor, enable LAM service and clear LAMs
      bool setposHW_postmove();//Switch off the motors, disable and clear LAMS 
      bool emergencyStop(){return (controller->Stop(MOT_CNTRL_NO)&&setposHW_postmove());};//Stop motor and do postmove task
      IOOperation::OutputRegister *power_control;//Motor power slot number (Borer 1082A Output Register)
      IOOperation::InterruptRegister *interreg;//Interrupt register (Borer 1051A Interrupt Register)
      Positioner::StepMotorController *controller;//Motor move controller object
    private:
      //members
      uint8_t displayslot;//Display slot number
      uint8_t output_number;//Motor output number: this starts from 0
      double ramp;//Ramp time in seconds between high and low speed
      int lospeed;//Lowest motor speed
      int hispeed;//Fastest motor speed
      //static members - display related
      static const uint8_t DISP_OFF_A=0;//Disconnect display
      static const uint8_t DISP_VAL_F=0;//Ask display value function 
      static const uint8_t DISP_OFF_F=24;//Disconnect display function
      static const uint8_t DISP_SEL_F=26;//Motor selection for display function
      static ParallelMotor *ACTUAL_display;//Number of motor to be shown (initialized by 0 value)
      //static members - motor power related
      static bool HILIM_off;//High limit switch is off 
      static bool LOLIM_off;//Low limit switch is off 
      static const uint32_t mot_pow_arr[12];//Motor power array: 0 switch off, then the corresponding motor number
      //static members - motor move related
      static const uint8_t MOT_CNTRL_NO=1;//Steady state motor number at step motor controller
      //Move related methods
      bool setposHW(uint32_t target){ if (MOVING_motor==this) return (controller->SetRunningParameters(MOT_CNTRL_NO,ramp, lospeed, hispeed, acceptedzero(),false) && controller->MoveTo(MOT_CNTRL_NO, (uint32_t) target)); else return false;}//Start to move motor into the target position - NOT CHECKED
      bool powerOnOff(bool on_off);//Switch on (t) the power for the current motor or switch off the motor
      //Position asking related methods (Camac Ablese)
      bool getposHW(uint32_t &result);//Returns with the actual position of the motor. 
      bool isFinished(){return controller->TestAnyLAMMotor(MOT_CNTRL_NO) || (idir&&(!HILIM_off)) || (!idir&&(!LOLIM_off));};//It finished, if a LAM has been switched on or limit switch reached
      bool displayConnect(bool connect_disconnect=false) {return communicator->Control(connect_disconnect ? addr : DISP_OFF_A, connect_disconnect ? DISP_SEL_F : DISP_OFF_F, displayslot, Camac::STATUS_Q); };//Connect(t)/disconnect(f) display with this motor
      bool change_display();//this method try to change the display: true: success, false: motor is moving, changing to other is not allowed
  };
  
  class AngleParallelMotor : public ParallelMotor, public AngleMotor
  //This class describes a parallel motor, which has relationship with angles: M3 and M4 motors, but it is the basic class of M3
  {
    public:
      //public methods: constructor/destructor
      AngleParallelMotor(Camac &camcomm, std::string motor_name, uint8_t motor_ablese_number, uint8_t motor_output_number, uint32_t hilimp, uint32_t lowlimp, IOOperation::OutputRegister &power_cntrl, IOOperation::InterruptRegister &interr_register , Positioner::StepMotorController &step_controller, double ramp_sec, int low_step_sec, int hi_step_sec, double shift, double slope=-1.0/180.0, uint8_t disp_slot=2, int acc_diff=0, int bits=16): Motor(camcomm, motor_name, motor_ablese_number, hilimp, lowlimp, bits, STRAIGHT, acc_diff), ParallelMotor(camcomm, "", (uint8_t) 0x0, motor_output_number, (uint32_t) 0x0, (uint32_t) 0x0, power_cntrl, interr_register, step_controller, ramp_sec, low_step_sec, hi_step_sec, disp_slot, 0, 0), AngleMotor(camcomm, "", (uint8_t) 0x0, (uint32_t) 0x0, (uint32_t) 0x0, 0, (ORDER) 0, shift, slope, 0){};
      //Constructor: the camac communicator object; name of the motor; motor number starts from 1! (ablese and output, correspondingly); low and high limit of the motor in uniform pulses; motor power selection object (output register); incoming interrupt detection (interrupt register object); step motor controller object; ramp time from low to high speed in seconds; lowest speed of the motor in step/seconds; fastest speed of the motor in step/seconds; shift and slope according to degree=slope*(pos-shift); display slot number (Camac ablese modul No); the acceptable smallest difference in positioning; number of bits of the position decoder
      ~AngleParallelMotor(){while(!(this->isAllStopped())) usleep(1000u);};//Destructor
  };
  
  class CushionAPMotor: public AngleParallelMotor
  //This class describes the M4 motor, which need to operate cushion during the moving of the motor
  {
    public:
      //Public methods
      CushionAPMotor(Camac &camcomm, std::string motor_name, uint8_t motor_ablese_number, uint8_t motor_output_number, uint32_t hilimp, uint32_t lowlimp, IOOperation::OutputRegister &power_cntrl, IOOperation::InterruptRegister &interr_register,  Positioner::StepMotorController &step_controller, double ramp_sec, int low_step_sec, int hi_step_sec, double shift, double slope=-0.01, uint8_t disp_slot=2, int acc_diff=0, int bits=16): Motor(camcomm, motor_name, motor_ablese_number, hilimp, lowlimp, bits, STRAIGHT, acc_diff), AngleParallelMotor(camcomm, "", (uint8_t) 0, motor_output_number, (uint32_t) 0, (uint32_t) 0, power_cntrl, interr_register, step_controller, ramp_sec, low_step_sec, hi_step_sec, shift, slope, disp_slot, 0, 0){interreg->Register(this,IOOperation::InterruptRegister::COMPRESSEDAIR_ON);};
      //Constructor: the camac communicator object; name of the motor; low and high limit of the motor in uniform pulses; motor power selection object (output register); incoming interrupt detection (interrupt register object); step motor controller object; ramp time from low to high speed in seconds; lowest speed of the motor in step/seconds; fastest speed of the motor in step/seconds; shift according to degree=slope*(pos-shift); the acceptable smallest difference in positioning; display slot number (Camac ablese modul No)
      ~CushionAPMotor(){while(!(this->isAllStopped())) usleep(1000u); interreg->UnRegister(this,IOOperation::InterruptRegister::COMPRESSEDAIR_ON);};//Destructor
      bool EvacuateAir(){if (MOVING_motor!=0) return false; else return CushionOnOff(true,true);}
      void Action(IOOperation::InterruptHandle *ih, uint32_t mask);//Changes the status of the "comair" compressed air on indicator
    private:
      //members
      static bool COM_air;//Compressed air switch: true, if air is on
      //methods
      bool setposHW_preparetomove(){return (CushionOnOff(true,false) && ParallelMotor::setposHW_preparetomove());};//Switch on air cushions and the motor, as well - NOT CHECKED
      bool setposHW_postmove(){return (ParallelMotor::setposHW_postmove() && CushionOnOff(false,false));};//Switch off the motor and air cushions, as well
      bool emergencyStop(){return ((ParallelMotor::emergencyStop())&&CushionOnOff(false,false));};//Switch off the motor immediately and close compressed air
      bool CushionOnOff(bool on=false, bool evacuate=false);//Switch on/off air cushions or evacuate air from the system
  };
}
#endif
