//TESTED: 15.08.2013 by Laszlo Temleitner
#include "OHM.h"

const double OHM::TIMEOUT=.5;//1 s
const double OHM::WAITSUCCESS=0.2;//100 ms

OHM::OHM(const std::string nPort, const double timeout):Thermometer(TEMP_PARAMETERS)
//Constructor: device address, serial port name, Celsius(t)/Fahrenheit(f), baud rate, timeout
{
  //Setting the communication lime and address
  communicator=NULL;
  if(!ResetCommunication(nPort, timeout)) std::cout<<"OHM::OHM: error in opening the communication channel"<<std::endl;
}

bool OHM::ResetCommunication(const std::string nPort, const double timeout)
//Reset the communication parameters: serial port, transmission rate, timeout
{
  bool retval;//Returning value
  //Modify the settings or create a new channel
  if(communicator==NULL)
  {
    communicator= new Serial(nPort, 576000, 8, 'N', 1, true, true);
    retval=communicator->isOpened();
  }
  else retval=communicator->Open(nPort, 576000, 8, 'N', 1, true, true);
  
  //Set timeout
  if (retval) retval=communicator->setTimeout(timeout);
  
  //switch off printing
  printing=true;//Supposing printing
  if (retval)
  {
    retval=LockULKeyboard(true);//Lock
    if(retval) retval=StopPrint();
    if(retval) retval=LockULKeyboard(false);//Unlock
  }
  
  
  //Returning with the results
  return retval;
}

bool OHM::LockULKeyboard(bool lock)
//Locks (t) or unlock (f) the keyboard
{
  char command[2]={'P','1'};//Command: default unlock
  if(lock) command[1]='0';//Set to lock
  return setvalue(command,0, NULL);//Doing the task
}

bool OHM::SetModes(const bool sleep_mode_log_on, const bool auto_power_off, const bool relative_temperature)
//Set sleep mode logging, auto power off and relative temperature measurement on (t) or off(f)
{
  return (LockULKeyboard(true) && SleepModeLogOnOff(sleep_mode_log_on) && AutoPowerOff(auto_power_off) && RelativeTemp(relative_temperature) && LockULKeyboard(false));
}

bool OHM::SleepModeLogOnOff(bool on)
//Set on sleep mode log(t) or off
{
  //Initialize
  if (printing) return false;//If printing is on, do not allow to set it
  char command[3]={'W','C','0'};//Command string -default
  if(on) command[2]='1';//ON option
  
  //Doing the communication 
  return setvalue(command,1, (uint8_t*)(command+2));
}

bool OHM::AutoPowerOff(bool enable)
//Enable(t)/disable(f) auto poweroff function
{
  //Initialize
  if (printing) return false;//If printing is on, do not allow to set it
  char command[2]={'K','Q'};//Command string -default
  if(enable) command[1]='P';//Enable option
  
  //Doing the communication 
  return setvalue(command,0, NULL);
}

bool OHM::RelativeTemp(bool enable)
//Enable(t)/disable(f) relative temperature mode
{
  //Initialize
  if (printing) return false;//If printing is on, do not allow to set it
  char command[2]={'K','6'};//Command string -default
  if(enable) command[1]='7';//Enable option
  
  //Doing the communication 
  return setvalue(command,0, NULL);
}

bool OHM::DataLog(bool start)
//Start(t)/stop(f) data logging
{
  //Initialize
  if (printing) return false;//If printing is on, do not allow to set it
  char command[2]={'K','5'};//Command string -default
  if(start) command[1]='4';//Start option
  
  //Doing the communication 
  return setvalue(command,0, NULL);
}

bool OHM::GetTemperature(bool &chAOK, bool &chBOK, double &chA, double &chB)
//Returns with the successful communication(t). The first 2 variables indicates validity of the datasets, which are stored in the position of the last 2 variables
{
  std::string result=getabout('S','0',0);
  bool retval= (result!="Error");
  if (!retval) chAOK=chBOK=false;
  else
  {
    if((result.substr(0,12)).find("_BURN_") != std::string::npos) chAOK=false;
    else
    {
      chAOK=true;
      std::stringstream ss(result.substr(0,12));
      ss>>chA;
    }
    if((result.substr(12,12)).find("_BURN_") != std::string::npos) chBOK=false;
    else
    {
      chBOK=true;
      std::stringstream ss(result.substr(12,12));
      ss>>chB;
    }
  }
  return retval;
}

void OHM::getTempHW(double *result)
{
  double r[TEMP_PARAMETERS];
  bool b[TEMP_PARAMETERS];
  if(GetTemperature(b[0],b[1],r[0], r[1])) 
  {
    for(int i=0; i<TEMP_PARAMETERS; i++) 
    {
      if(b[i]) result[i]=r[i];//OK
      else result[i]=-5000.0;//Thermocouple is not connected
    }
  }
  else  for(int i=0; i<TEMP_PARAMETERS; i++) result[i]=-5000.0;//Communictaion error
  return;
}

bool OHM::ResetLog()
//Reset the logged data
{
  //Initialize
  if (printing) return false;//If printing is on, do not allow to set it
  char command[2]={'L','E'};//Command string -default
  
  //Doing the communication 
  return setvalue(command,0, NULL);
}

std::string OHM::GetTemperatureUnit(bool chnA)
//Returns with the unit of temperature of the selected (A:t, B:f) channel
{
  if (printing) return std::string("WARNING: stop print before asking information!");
  char command[3]={'R','U','A'};//Command string
  if(chnA) command[2]='B';
  uint8_t mess[50];//message string
  
  int ans=0;//answer
  ans=getvalue(command,3,50,mess);//Get the datasets
  //Transform to string
  if (ans<1) return std::string("Error");
  std::string str((char*) mess);
  replace(str.begin(),str.end(),(char) OHM::CR,'\n');
  return str.substr(2,ans-3);
}

double OHM::GetBatteryLevel()
//Returns with the battery level in volts
{
  std::string basic=getabout('R','P',0);//Getting the dataset
  
  //Error or warning
  if ((basic[0]=='W')||(basic[0]=='E'))
  {
    std::cout<<basic<<std::endl;
    return -1.0;
  }
  if(basic[0]!='&') return -2;//Not successful read
  
  //processing the dataset
  std::stringstream ss;
  ss<<basic.substr(1,4);
  int number;
  ss>>number;
  
  return ((double)number)*0.01;
}

int OHM::GetFreeLogPages()
//Returns with the number of free log pages
{
  std::string basic=getabout('L','N',0);//Getting the dataset
  
  //Error or warning
  if ((basic[0]=='W')||(basic[0]=='E'))
  {
    std::cout<<basic<<std::endl;
    return -1.0;
  }
  if(basic[0]!='&') return -2;//Not successful read
  
  //processing the dataset
  std::stringstream ss;
  ss<<basic.substr(1,4);
  int number;
  ss>>number;
  
  return number;
}

void OHM::About(std::string &model, std::string &model_desc, std::string &inst_sn, std::string &firm_ver, std::string &firm_date, std::string &cal_date, std::string &probe, std::string &probe_sn, std::string &probe_cal, double &batterylevel)
//This function returns with general information about instrument: model number, description, instrument serial number, firmware version and date, calibration date, probe set, probe serial number and calibration date
{
  if(LockULKeyboard(true))//Lock keyboard 
  {
    model=GetModel();//Returns the model number
    model_desc=GetModelDesc();//Returns with the model description
    inst_sn=GetInstSN();//Returns with the serial number of the instrument
    firm_ver=GetFirmVer();//Returns with firmware version
    firm_date=GetFirmDate();//Returns with firmware date
    cal_date=GetCalDate();//Returns with calibration date and time
    probe=GetProbe();//Returns with the type of the probe set
    probe_sn=GetProbeSN();//Returns with the serial number of the probe
    probe_cal=GetPCalDate();//Returns with the date of calibration of the probe
    batterylevel=GetBatteryLevel();//Battery level in volts
    LockULKeyboard(false);//Unlock keyboard
  }
  return;
}


bool OHM::SetUserCode(std::string user)
//Sets the user code, returns with success(t) or fail(f)
{
  //Initialize
  if (printing) return false;//If printing is on, do not allow to set it
  char command[2]={'T','2'};//Command string
  
  //Initialize message string
  return setvalue(command,(int) ((user.length()<16)? user.length() : 16), (uint8_t*) user.c_str());
}

bool OHM::SetTimeInterval(int interval)
//Set print/log interval in s
{
  //Initialize
  if (printing) return false;//If printing is on, do not allow to set it
  char command[3]={'W','A','0'};//Command string
  
  switch (interval)
  {
    case    1: command[2]='1'; break;//1s
    case    5: command[2]='2'; break;//5s
    case   10: command[2]='3'; break;//10s
    case   15: command[2]='4'; break;//15s
    case   30: command[2]='5'; break;//30s
    case   60: command[2]='6'; break;//1 min 
    case  120: command[2]='7'; break;//2 min 
    case  300: command[2]='8'; break;//5 min 
    case  600: command[2]='9'; break;//10 min 
    case  900: command[2]='A'; break;//15 min 
    case 1200: command[2]='B'; break;//20 min 
    case 1800: command[2]='C'; break;//30 min 
    case 3600: command[2]='D'; break;//1 h
    default: break;//set to immeadiate as default
  }
  
  //Doing the communication 
  return setvalue(command,1,(uint8_t*)(command+2));
}

int OHM::GetTimeInterval()
//Returns with the print/log interval in s
{
  std::string basic=getabout('R','A',16);//Getting the dataset
  
  //Error or warning
  if ((basic[0]=='W')||(basic[0]=='E'))
  {
    std::cout<<basic<<std::endl;
    return -1;
  }
  
  //processing the dataset
  std::stringstream ss;
  ss<<basic.substr(0,4);
  int number;
  ss>>number;
  
  return number; 
}

//private methods


//Medium level

std::string OHM::getabout(char clett1, char clett2, int cut)
//Skeleton to ask about instrument settings
{
  if (printing) return std::string("WARNING: stop print before asking information!");
  char command[2];//Command string
  command[0]=clett1;
  command[1]=clett2;
  uint8_t mess[2000];//message string
  
  int ans=0;//answer
  ans=getvalue(command,2,2000,mess);//Get the datasets
  //Transform to string
  if (ans<1) return std::string("Error");
  for(int i=cut; i<ans-1; i++) if (mess[i]==OHM::CR) mess[i]='\n';
  return (tostr(mess,ans)).substr(cut,ans-(cut+1));
}

std::string OHM::getbulky(char clett1, char clett2)
//Skeleton to ask bulky data
{
  //Initialize
  if (printing) std::string("WARNING: stop print before asking information!");//If printing is on, do not allow to set it
  
  //Initializing
  printing=true;
  std::string log;
  char command[2];//Command string
  command[0]=clett1;
  command[1]=clett2;
  uint8_t mess[2000];//message string
  
  //Read the datasets in
  bool flag=LockULKeyboard(true);//Lock keyboard
  int ans=0;//Answer
  if(flag) ans=getvalue(command,2,2000,mess);
  //Error check
  if ((ans<1)||(!flag))
  {
    printing=false;
    return std::string("Error");
  }
  //Copy to the log string
  std::string log2((char *)mess);
  log+=log2.substr(0,ans);
  
  //Continue to read the bulky datasets
  do
  {
    ans=getvalue(NULL,0,2000,mess);
    if(ans>0) 
    {
      std::string log3((char *)mess);
      log+=log3.substr(0,ans);
    }
  }while(ans>0);
  
  //Replace OHM::CR by /n
  std::replace( log.begin(), log.end(), (char) OHM::CR, '\n');
  
  //Finally, unlock keyboard
  flag=LockULKeyboard(false);//Unlock keyboard
  printing=false;
  
  return log;
}

//Medium-low level communication

bool OHM::setvalue(char *command, const int length, uint8_t *message)
//Send the command [2 letters] with the length long message
{
    uint8_t *sbuffer= new uint8_t [3+length];//Sendbuffer
    uint8_t *rbuffer= new uint8_t [2];//Replybuffer
    int replyl=2;
    //Set send buffer datasets
    sbuffer[0]=command[0];//Command letter1
    sbuffer[1]=command[1];//Command letter2
    for (int i=0; i<length; i++) sbuffer[2+i]=message[i];//Message
    
    //Sending the message
    bool retval=this->communicate(sbuffer,3+length,rbuffer,&replyl);//Returning value
    
    //Checking agreement: it should be "&+0xd"
    if(replyl<2) retval=false;
    if(retval) retval=((rbuffer[0]=='&')&&(rbuffer[1]==OHM::CR));

    //Delete buffers
    delete [] sbuffer;
    delete [] rbuffer;
  
    return retval;
}

int OHM::getvalue(char *command, const int cletters, const int length, uint8_t *message)
//Requesting a given value by command ["cletters" letters]. The available buffer for receiving messages is
//"message" which is "length" long. If command letters are 0, it will perform read only task!
{
    uint8_t *sbuffer=NULL;//Sendbuffer
    uint8_t *rbuffer= new uint8_t [length];//Replybuffer
    if(cletters!=0)
    {
      sbuffer= new uint8_t [cletters+1];//command + CR
      for(int i=0; i<cletters; i++) sbuffer[i]=command[i];
    }
    
    int l=length+1;
    //Send message, get the answer
    bool retval=this->communicate(sbuffer,((cletters==0) ? 0 :cletters+1), rbuffer,&l, (cletters==0));//Returning value

    //Writing the reply into message buffer
    if (retval) {for(int i=0; i<l; i++) message[i]=rbuffer[i];}
    else l=-1;//Error case
    
    //dispose memory
    if(cletters!=0) delete sbuffer;
    delete [] rbuffer;
    
    return l;
}

//Low level communication


bool OHM::communicate(uint8_t *sendbuffer, const int slen, uint8_t *replybuffer, int *rlen, const bool onlyreceive)
//Communicate with serial interface: it contains the data to send, its length, the reply message and its expected 
//length. If onlyreply true, then it read message only
{
    //Adding CR character
    if (slen!=0) sendbuffer[slen-1]=OHM::CR;
    
    //Writing it to the serial line
    int counterwhole=0;//Whole procedure counter
    int rep;//Local reply
    
    bool repeatfull=false;//Reapeat the full send and receive procedure
    do
    {
	repeatfull=false;
	int counter=0;//Local counter for send and reply
	bool repeatsend=false;//Repeat send procedure
	bool repeatreply=false;//Repeat reply procedure
	
	//Perform error checked writing
	if (!onlyreceive)
	{
	  do
	  {
  #ifdef DEBUG	    
	      std::cout<<"Sendbuffer:";
	      for(int i=0; i<slen; i++) std::cout<<i<<": (hex)"<<std::hex<<(int)sendbuffer[i]<<" dec:"<<std::dec<<sendbuffer[i]<<std::endl;
  #endif
	      //Send bytes
	      rep=communicator->Send(sendbuffer,slen);
	      
	      if(rep!=slen)
	      {
		  repeatsend=true;
		  //Communication error
		  std::cerr<<"Sending is unsuccessful(OHM::communicate) on the serial line!\n";
		  if(counter++<1) 
		  {
		      std::cerr<<"Trying again!"<<std::endl;
		      usleep((int)(100000.0*WAITSUCCESS));//Wait some time until the next trial
		  }
		  else return false;
	      }
	  }while(repeatsend);
	  usleep((int)(100000.0*WAITSUCCESS));//Wait some time until the answer
	}
	
	//Perform error checked reading
	counter=0;
	do
	{
	    //Receiving the reply
	    rep=communicator->Receive(replybuffer,*rlen,false);
#ifdef DEBUG	    
	    std::cout<<"Replybuffer:";
	    for(int i=0; i<rep; i++) std::cout<<i<<": (hex)"<<std::hex<<(int)replybuffer[i]<<" dec:"<<std::dec<<(int)replybuffer[i]<<" "<<(char)replybuffer[i]<<std::endl;
	    std::cout<<rep<<std::endl;
#endif
	    if ((rep<1)&&(!onlyreceive))
	    {	    
		//Possible reasons: 
		//timeout
		//SERIAL_ERROR --- try it igain
		repeatreply=true;
		usleep((int)(100000.0*WAITSUCCESS));//Wait some time until the next trial
		if(counter++==1)
		{
		    std::cerr<<"Communication error on the serial line (OHM::communicate) during the reading!\n";
		    if (counterwhole++==0) {std::cerr<<"Sending messages again\n"; repeatreply=false; repeatfull=true;} 
		    else return false;
		}
		
	    }
	    else 
	    {
	      //Leaving the cycle
	      *rlen=rep;
	      repeatreply=false;
	    }
	}while(repeatreply);
    }while(repeatfull);

    return true;
}

std::string OHM::tostr(uint8_t* buffer, const int length)
//Converting the digits containing the buffer into string
{
  std::stringstream ss;
  for(int i=0; i<length; i++) ss<<buffer[i];
  return ss.str();
}

