#include "Scaler.h"


namespace Scaler
{   
// -------------------------------------- PSDCounter class
  PSDCounter::PSDCounter(std::string name, double real_preset, double live_preset, int channels, int discriminator_low, int discriminator_hi, GATE_MODE gate, double deadtime_delay, int serial_number, double timeout_ms, uint16_t adc_offset) : Counter<uint32_t>(channels, {name}, {channels}), TrumpPCI(serial_number,timeout_ms)
  //Constructor: name, real and live presets (0.0: no preset); channel number(256/512/1024/2048), lower and upper levele discriminators; gate mode(GATE_OFF/GATE_COINCIDENCE/GATE_ANTICOINCIDENCE); deadtime delay for each incoming pulses in us, serial number of the card (-1: choose the first one among others), timeout of reading in ms; ADC offset in hardware unit (2048: no shift)
  {
    //Doing basic initialisation: clear all settings, set deadtime delay and ADC offset
    //ADC offset and deadtime delay are not changed frequently, thus they are allowed to configure only in constructor
    if(!(TrumpPCI::clearAll() && TrumpPCI::setDTimeDelay(deadtime_delay) && setOffset(adc_offset) && (getDTimeDelay()==deadtime_delay) && (getOffset()==adc_offset) )) std::cerr<<"ERROR: PSDCounter class is not initialized correctly!"<<std::endl;
    
    //Set the most common variables
    for(int i=0; i<2; i++) pres[i]=false;//Initialize preset shower
    if(!ChangePresets(real_preset, live_preset)) std::cerr<<"ERROR: PSDCounter::PDSCounter: trouble with setup of presets!"<<std::endl;
    if(!ChangeChannel(channels, discriminator_low, discriminator_hi)) std::cerr<<"ERROR: PSDCounter::PDSCounter: trouble with setup of channels!"<<std::endl;
    if(!ChangeGate(gate)) std::cerr<<"ERROR: PSDCounter::PDSCounter: trouble with setup of gates!"<<std::endl;
    for(int i=0; i<2; i++) reallivetime[i]=0.0;//Initialize real and live time belonging to the last read of counters
  }
  
  bool PSDCounter::ChangePresets(double real_preset, double live_preset)
  //Change presets: [0.0 means: not preset], should be integer multiple of 0.05
  {
    if(Counter::GetStatus()!=ST_OFF) return false;//Setup is available, if the counter has not been configured yet
    double presets[2];//Real and live preset corrected values
    enum PRESET_VALUE { PRESET_REAL, PRESET_LIVE};//Real and live time presets
    if(real_preset>=0.0) presets[PRESET_REAL]=std::round(real_preset*50.0)/50.0;//Adjust value to be acceptable for Trump
    else return false;
    if(live_preset>=0.0) presets[PRESET_LIVE]=std::round(live_preset*50.0)/50.0;//Adjust value to be acceptable for Trump
    else return false;
    bool retval=TrumpPCI::clearPresets();//Clears all of the presets
    for(int i=0; i<2; i++) if (retval && (presets[i]>0.0)) retval=setTime(presets[i],i%2 == 0,false);//Set real and live time presets
    if(retval)
    {
      double chk[2];
      retval=GetPresets(chk,chk+1);//Get the setup of presets
      if(retval) retval=((chk[0]==presets[0])&&(chk[1]==presets[1]));//Check consistency      
      if(retval) for(int i=0; i<2; i++) pres[i]=(presets[i]>0.0);//Update boolean list of presets
    }
    return retval;
  }
  
  bool PSDCounter::ChangeChannel(int channels, int discriminator_low, int discriminator_hi)
  //Change the number of applied channels: (256/512/1024/2048), low and high level of discriminators
  {
    if(Counter::GetStatus()!=ST_OFF) return false;//Setup is available, if the counter has not been configured yet
    if((channels!=256)&&(channels!=512)&&(channels!=1024)&&(channels!=2048)) return false;//Bad channel number
    if(!((discriminator_low>=0)&&(discriminator_hi>discriminator_low)&&(discriminator_hi<channels))) return false;//Bad setup of discriminators
    if(!Counter::changeCounterNo(channels,{channels})) return false;//Change the number of counter points
    //Update hardware
    bool retval=TrumpPCI::setGainConv(Counter::getCounterNo());//Set the conversion gain
    if (retval) retval=TrumpPCI::setDataRange(0,Counter::getCounterNo());//Set data range
    if (retval) retval=TrumpPCI::setDiscriminator(discriminator_low,true);//Set the low level discriminator
    if (retval) retval=TrumpPCI::setDiscriminator(discriminator_hi,false);//Set the high level discriminator
    //Check channel setting on hardware
    if(retval)
    {
      int values[3];//Returning values: channel no, low and high level of discriminator
      retval=GetChannel(values, values+1, values+2);//Get hardware values
      if(retval) retval=((values[0]==channels)&&(values[1]==discriminator_low)&&(values[2]==discriminator_hi));//Check consistency
    }
    return retval;
  }
  
  bool PSDCounter::ChangeGate(GATE_MODE gate)
  //Change gate mode: GATE_OFF/GATE_COINCIDENCE/GATE_ANTICOINCIDENCE are accepted
  {
    if(Counter::GetStatus()!=ST_OFF) return false;//Setup is available, if the counter has not been configured yet
    if(gate==GATE_ERROR) return false;//Not allow error input
    bool retval=TrumpPCI::setGate(gate);//Set gate mode
    if(retval) retval=(gate==GetGate());//Check gate setup
    return retval;
  }
  
  bool PSDCounter::GetPresets(double *real_preset, double *live_preset)
  //Returns with the current setup of real and live time presets in seconds. [0 in agument: not calculated], 0.0 at return: not set. Return: communication OK/error
  {
    bool retval=true;//Returning value
    if(real_preset!=0) retval=((*real_preset=getTime(true,true,false))>=0.0);
    if(live_preset!=0)  retval=((*live_preset=getTime(true,false,false))>=0.0);
    return retval;
  }
  
   bool PSDCounter::GetChannel(int *channel, int *discriminator_low, int *discriminator_high)
   //Returns with the current setup of channel related stuffs: number of channels, high and low level of discriminators. Returning value is true, if communication was OK
  {
    if((channel==0)||(discriminator_low==0)||(discriminator_high==0)) return false;//Expell non-proper request
    int start_chn=-1;//Start channel no.
    int chns=-1;
    //Check hardware setup
    bool retval=TrumpPCI::getGainConv(channel,0);
    if((retval)&&(*channel!=Counter::getCounterNo()))
    {
      retval=false;//Inconsistency has been found
      changeCounterNo(*channel,{*channel});//Try to update the actual counter number to be in consistency with reality (returning value is not checked)
    }
    if(retval) retval=TrumpPCI::getDataRange(&start_chn,&chns);
    if(retval) retval=((chns==*channel)&&(start_chn==0));
    if(retval) retval=((*discriminator_low=(int) TrumpPCI::getDiscriminator(true))!=(int)0xfffe);//Get higher level discriminator
    if(retval) retval=((*discriminator_high=(int) TrumpPCI::getDiscriminator(false))!=(int)0xfffe);//Get higher level discriminator
    return retval;
  }
  
  bool PSDCounter::TestWriteTimeData(PSD_CNT count, double real, double live)
  //Fill the counters with count counts and sets real and live times. It returns with communication success or fail. It works only, if counting has not been started, but the device is configured
  {
    if(Counter::GetStatus()!=ST_CONFIGURED) return false;//Expell non-configured case (configuration overwrites these values)
    bool retval=TrumpPCI::setData(count);//set data counts
    if(retval) retval=TrumpPCI::setTime(real,true,true);//set real time
    if(retval) retval=TrumpPCI::setTime(live,false,true);//set live time
    return retval;
  }
  
  bool PSDCounter::ReadCountsHW(PSD_CNT *counts, double *remaining)
  //Get number of counts, remaining time/counts and store the belonging live and real counts
  {
    std::string hw_stat;
    uint16_t active;
    bool retval=TrumpPCI::getDataSet(0,Counter::getCounterNo(),counts, 0);//Get counts
    double a=-29,b=-29;
    //if(retval) retval=TrumpPCI::getStatus(reallivetime+1,reallivetime,&active,hw_stat);//Get actual real and live times
    if(retval) retval=TrumpPCI::getStatus(&a,&b,&active,hw_stat);//Get actual real and live times
    reallivetime[0]=b;
    reallivetime[1]=a;
    if(retval&&remaining)
    {
      *remaining=-.5;//Set if it were no limit
      if((pres[0]==pres[1])&&(pres[0]==1)) *remaining=std::min(getTime(true,true, true),getTime(true,false, true));
      else for(int i=0; i<2; i++) if(pres[i]) *remaining=getTime(true,i%2 == 0, true);//Get remaining times: only the relevant one is stored
      retval= (*remaining >-1.0);
    }
    return retval;
  }

  //-------------------------MonitorCounter class--------------------------
  
  const uint8_t MonitorCounter::scode[6]={MonitorCounter::ALL_A, MonitorCounter::EVENTL_A, MonitorCounter::PRESETL_A, MonitorCounter::TIME_A, MonitorCounter::POOR_A, MonitorCounter::MODULE_A};//Contains the subcodes, belonging to MODES
  
  MonitorCounter::MonitorCounter(std::string name, Camac &cam, const uint8_t slotno, bool count_or_time_preset, const MON_CNT initial_preset, bool intensity_monitoring, const uint16_t low_intensity_warning_level, const uint8_t times_below_low_before_warning, bool warning_suspend, bool lam_in_use) : CamacCounter<uint64_t>(name, 2,cam,slotno,initial_preset,lam_in_use,true)
  // Constructor: name, camac communicator, slot number, count preset(t)/time preset(t), initial preset, do we need to monitor intensity level?, low intensity warning level in cps, how many times consecutively below the previous value to give intensity warning, suspend the acquisition in the case of warning?, wheather to use LAM service
  {
    //Give starting value for the fundamental variables
    first=true;//Before first run
    check_monitor=false;//The intensity will not be checked in the next round
    last_monitor=false;//The value of the last monitor is GOOD
    opmode=count_or_time_preset;//Operation mode: counter(t) or timer (f)
    SetIntensityParameters(intensity_monitoring,low_intensity_warning_level,times_below_low_before_warning,warning_suspend);//Set intensity related parameters
  }
  
  MonitorCounter::~MonitorCounter()
  //Destructor: we should disable LAM service of the device
  {
    if(GetLAMUse())
    {
      int i=0;
      for(i=0; ((i<=10)&& !enableDisable(opmode ? M_PRESET : M_TIME, false)); i++);//Disable preset LAM
      if(i>10) std::cerr<<"Error: MonitorCounter::~MonitorCounter: communication trouble!"<<std::endl;
      if(intensity_crg!=0x0) for(i=0; ((i<=10)&& !enableDisable(M_INTENSITY, false)); i++);//Disable monitor LAM
      if(i>10) std::cerr<<"Error: MonitorCounter::~MonitorCounter: communication trouble!"<<std::endl;
    }
  }
  
  bool MonitorCounter::ChangeOpMode(bool countertimer)
  //Changes the operation mode to counter(t) or timer(f), when status is OFF
  {
    if(GetStatus()!=ST_OFF) return false;
    if(countertimer==opmode) return true;//Already in the required state
    else
    {
      //If this is not the first run and LAM service in use, we have to disable one LAM and enable an other one
      if((!first)&&(GetLAMUse()&&(!enableDisable(opmode ? M_PRESET : M_TIME, false) || !enableDisable(countertimer ? M_PRESET : M_TIME, true)))) return false;
      opmode=countertimer;
      return true;
    }
  }
  
  bool MonitorCounter::SetIntensityParameters(const bool need, const uint16_t niv, const uint8_t consec, bool poor_suspend)
  //Sets the intensit warning levels in OFF status, if it is needed (true). If it were the case, warning will be issued, if the intensity is less than the lowest intensity niv(au) during "consec" number of consecutive sampling, at least. The poor_suspend regulates the behaviour of warning level during the measurement: if it is true, calculation is suspended, otherwise continues.
  {
    if(GetStatus()!=ST_OFF) return false;
    suspend_poor=poor_suspend;
    intensity_crg=(need ? (((uint32_t)niv)|(((uint32_t) consec&0xf)<<16)) : 0x0);//Set the configuration register
    return true;
  }
  
  bool MonitorCounter::SetPreset(MON_CNT preset)
  //Set preset for the current setup
  {
    //Expell counting and error on writing cases if the status is configured
    if((GetStatus()==ST_ON)||((GetStatus()==ST_CONFIGURED)&&!writeLL(opmode ? M_PRESET : M_TIME,preset))) return false;
    else return CamacCounter::SetPreset(preset);
  }
  
  bool MonitorCounter::SetLAMUse(bool use)
  //Set LAM service: only in OFF status -- CamacCounter will be overdefined
  {
    if(use==GetLAMUse()) return true;//We do not have to do anything
    else
    {
      //It should be change status, check the off status...
      if(GetStatus()!=ST_OFF) return false;//Actual status is other, than OFF...
      else
      {
	//If this is not the first run, we should enable(t)/disable(f) several LAMs..
	if(!first&&( ((intensity_crg!=0x0)&&suspend_poor&&!enableDisable(M_INTENSITY,use))|| !enableDisable(opmode ? M_PRESET : M_TIME, use) )) return false;
	return CamacCounter::SetLAMUse();//If everything is OK, update
      }
    }
  }
  
  bool MonitorCounter::SetupHW()
  //Set up the device according to present setting
  {
    //Set operation mode, preset, clear event counter and write intensity monitoring parameters
    if (!setOpMode(opmode)||((intensity_crg!=0x0)&& !writeLL(M_INTENSITY,(MON_CNT) intensity_crg)) || !writeLL(opmode ? M_PRESET : M_TIME,GetPreset())||!clearEvent()) return false;
    if(GetLAMUse() && !enableDisable(M_MODULE,true)) return false;//Enable LAM service of the module
    if(first)
    {
      if(GetLAMUse() && (!enableDisable(opmode ? M_PRESET : M_TIME, true) || ((intensity_crg!=0x0)&& suspend_poor&&!enableDisable(M_INTENSITY,true)))) return false;//If this is the first run: enable LAM for presets and intensity monitoring
      if(((intensity_crg!=0x0)&&!clearLAM(M_INTENSITY)) || !clearLAM(opmode ? M_PRESET : M_TIME)) return false;//Clear the LAM flags 
    }
    else first=false;
    return true;
  }
  
  bool MonitorCounter::StopHW()
  //Stops the data collection (ST_ON supposed)
  {
    //Stop module and clear LAM bistables
    if(!enableDisable(M_ALL,false)||((intensity_crg!=0x0)&&!clearLAM(M_INTENSITY))||!clearLAM(opmode ? M_PRESET : M_TIME)) return false;//Clear the LAM flags
    if(GetLAMUse()) return enableDisable(M_MODULE,false);//&&enableDisable(opmode ? M_PRESET : M_TIME, false)&&enableDisable(M_INTENSITY, false);//If LAM is operated: disable all module related LAMs
    else return true;
  }
  
  bool MonitorCounter::ReadCountsHW(MON_CNT *counts, double *remaining)
  //Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
  {
    //Get the counts
    counts[EVENT]=readLL(M_EVENT);
    MON_CNT countdown=0x0;//current value of the countdown
    if(remaining) countdown=readLL(opmode ? M_PRESET : M_TIME);//Get remaining
    //Check 
#ifdef DEBUG_SCALER
    if((counts[EVENT]==MON_BAD_OP)||(remaining&&(countdown==MON_BAD_OP)))
    {
      std::cerr<<"ERROR: MonitorCounter::ReadCountsHW: non-valid mode has been requested!"<<std::endl;
      return false;
    }
#endif
    if((counts[EVENT]==MON_TIMEOUT)||(remaining&&(countdown==MON_TIMEOUT))) return false;
    if(remaining) *remaining= ((countdown==(opmode ? COUNT_PRESET_MAX : TIME_PRESET_MAX)) ? 0.0 : (double) countdown);
    counts[PRESET]= (double) (((countdown==(opmode ? COUNT_PRESET_MAX : TIME_PRESET_MAX))||!remaining) ? GetPreset() : (GetPreset()-countdown));//Set elapsed counts
    if(check_monitor)
    {
      check_monitor=false;
      if((intensity_crg!=0x0) && !suspend_poor) last_monitor=testLFlag(M_INTENSITY);//If monitor should be checked, but not stop on POOR_INTENSITY and check monitor has been set to true, then the intensity flag is tested here
    }
    return true;
  }
  
  bool MonitorCounter::isFinishedHW()
  //Returns with true, if the measurement is finished (ST_ON supposed)
  {
    if (GetLAMUse()) return testLAM(M_ALL);
    else return (testLFlag(opmode ? M_PRESET : M_TIME) || ((intensity_crg!=0x0)&&suspend_poor&&testLFlag(M_INTENSITY)));
  }
  
  MON_CNT MonitorCounter::readLL(MODES mode)
  //Read the requested register. Mode can be M_EVENT, M_PRESET, M_TIME.
  {
    if((mode!=M_EVENT)&&(mode!=M_PRESET)&&(mode!=M_TIME)) return MON_BAD_OP;//Bad operation
    //Deal with the remaining ones 
    else if(mode==M_TIME)
    {
      //Read time preset from the register
      uint32_t result=0x0;
      int i=0;
      for(i=0; (i<10)&&((result=camacRead_LL(TIME_A))==CamacCounter::CERROR); i++);
      return ((i==10) ? MON_TIMEOUT  : result&0xffffff);
    }
    else
    {
      //In both cases, it will be a double read:
      // Initialize variables and subcodes
      uint8_t subcodes[2];//Defines the subcodes
      uint64_t answer=0x0;//Answer
      const uint32_t mask[2]={0xffffff,0xfff};//Masks for low(24bit) and high(12bit) of the registered value
      subcodes[0]=((mode==M_EVENT) ? EVENTL_A : PRESETL_A);
      subcodes[1]=((mode==M_EVENT) ? EVENTM_A : PRESETM_A);
      //Read the fragments of the registers
      for(int i=0; i<2; i++)
      {
	int j=0;
	 uint32_t result=0x0;
	for(j=0; (j<10)&&((result=camacRead_LL(subcodes[i]))==CamacCounter::CERROR); j++);
	if(j==10) return MON_TIMEOUT;//Immediately return, if communication trouble happened
	else answer+= (result&mask[i])<<(24*i);//Add it to the answer
      }
      return answer;
    }
  }
  
  bool MonitorCounter::writeLL(MODES mode, MON_CNT value)
  //Writes the value into the required mode register. Mode can be M_PRESET, M_TIME, M_INTENSITY.
  {
    if((mode!=M_INTENSITY)&&(mode!=M_PRESET)&&(mode!=M_TIME)) return false;//Bad module
    //Deal with the remaining ones 
    else if(mode==M_TIME) return camacWrite_LL(TIME_A, value&0xffffff);//Write time register
    else if(mode==M_INTENSITY) return camacWrite_LL(POOR_A, value&0xffffff,false);//Write intensity registers
    else return (camacWrite_LL(PRESETM_A, (value>>24)&0xfff) && camacWrite_LL(PRESETL_A, value&0xffffff));//Write preset
  }
  
  bool MonitorCounter::testLAM(MODES mode)
  //Test LAM enabled/disabled status, returns with true, if it is on. Mode can be M_ALL, M_PRESET, M_TIME, M_INTENSITY.
  {
    if((mode!=M_PRESET)&&(mode!=M_TIME)&&(mode!=M_INTENSITY)&&(mode!=M_ALL)) return false;
    else return camacTestLAM_LL(scode[mode]);
  }
  
  bool MonitorCounter::clearLAM(MODES mode)
  //Clears LAM bistable, returns with communication success or fail. Mode can be M_PRESET, M_TIME, M_INTENSITY.
  {
    if((mode!=M_PRESET)&&(mode!=M_TIME)&&(mode!=M_INTENSITY)) return false;
    else return camacClearLAM_LL(scode[mode]);
  }
  
  bool MonitorCounter::enableDisable(MODES mode, bool enable)
  //Enables(t) or disables (f) operation. Mode can be M_ALL, M_PRESET, M_TIME, M_INTENSITY, M_MODULE
  {
    if((mode!=M_ALL)&&(mode!=M_PRESET)&&(mode!=M_TIME)&&(mode!=M_INTENSITY)&&(mode!=M_MODULE)) return false;
    else return camacEnableDisable_LL(scode[mode], enable);
  }
  
  bool MonitorCounter::testLFlag(MODES mode)
  //Tests the value of L-bistable(DONE) flag. Mode can be M_PRESET, M_TIME, M_INTENSITY.
  {
    if((mode!=M_PRESET)&&(mode!=M_TIME)&&(mode!=M_INTENSITY)) return false;
    else return camacTestStatus_LL(scode[mode]);
  }
  
  //---------------------HarwellCounter class
  bool HarwellCounter::SetControlRegister(const uint16_t cr)
  //This method sets the control register
  {
    if(GetStatus()!=ST_OFF) return false;
    control_word=cr&0x1ff;
    return true;
  }

  bool HarwellCounter::SetupHW()
  //Set up the device according to present setting
  {
    //Before any operations, set some of the variables
    if((control_word&LAMENABLE_C)==LAMENABLE_C) SetLAMUse(true);//Control register contain LAM enable
    no_incoming=true;//Adjust no incoming pulse arrived indicator
    if((control_word&A1NXTCRRY_C) == A1NXTCRRY_C) defaultlevel=true;
    else if((control_word&A0NXTCRRY_C) == A0NXTCRRY_C) defaultlevel=false;
    else defaultlevel=GetALevel();//Get the actual level of A: this will be the default. But more exact, if we set it by carry setup of the control register
    //setup of the module: enable LAM (if used) and write control register with inhibit feature
    bool retval=true;
    if(GetLAMUse()) retval=camacEnableDisable_LL(START_A,true);//Enable LAM if it has been requested
    if(retval&&(GetPreset()!=0x0)) retval=camcomm->setData(INHIBIT_A, WRCNTRL_F, getSlot(), control_word&0x1ff, 0x0);//Configure counter, if preset will be given;
    return retval;
  }
  
  bool HarwellCounter::StartHW()
  //Starts counting (should be implemented by the actual hardware)
  {
    if(GetPreset()!=0x0) return camacWrite_LL(START_A,(uint32_t) GetPreset());//Write preset count register and start
    else return camcomm->setData(START_A, WRCNTRL_F, getSlot(), control_word&0x1ff, 0x0);
  }
  
  bool HarwellCounter::StopHW()
  //Stops the data collection (ST_ON supposed)
  {
    if(!camacClearLAM_LL(INHIBIT_A)) return false;//Clear LAM(DONE) bistable, but prevent to continue counting by INHIBIT_A
    if(GetLAMUse() && !camacEnableDisable_LL(START_A,false)) return false;//Disable LAM
    return true; 
  }
  
  bool HarwellCounter::ReadCountsHW(HW_CNT *counts, double *remaining)
  //Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
  {
    uint32_t signal=camacRead_LL(START_A);//Read signal
    if(signal == CamacCounter::CERROR) return false;//Communication error
    *counts=((GetPreset()>=(signal&PRESET_MAX)) ? GetPreset()-(signal&PRESET_MAX) : GetPreset()+(PRESET_MAX-(signal&PRESET_MAX)+1) );//Get real number of counts
    if (*counts>0) no_incoming=false;//Set number of incoming pulses
    if(remaining) *remaining=((GetPreset()>=(signal&PRESET_MAX)) ? signal&PRESET_MAX : 0.0);
    return true;
  }

  bool HarwellCounter::isFinishedHW()
  //Returns with true, if the measurement is finished (ST_ON supposed)
  {
    if(GetLAMUse()) return camacTestLAM_LL(START_A);//If LAM service is available
    else if(no_incoming)
    {
      //There is a chance, that counting has been started, but incoming pulse has not been detected yet: 
      //this situation, the A level can be the same as in carry. We can avoid it, if we check number of counts
      HW_CNT cnt=0x0;
      double rem=0.0;
      if(ReadCountsHW(&cnt,&rem)&&!no_incoming) return isFinishedHW();//Read counts, try to update no_incoming: if intensity has been arrived, call the routine again
      else return false;//Any case
    }
    else return (defaultlevel==GetALevel());
  }
  
  uint16_t HarwellCounter::str2code(std::string str)
  //Converting string into control code
  {
    uint16_t controlregister=0;
    if(str.find("A0N")!=std::string::npos) controlregister|=Scaler::HarwellCounter::A0NOW_C;
    if(str.find("A1N")!=std::string::npos) controlregister|=Scaler::HarwellCounter::A1NOW_C;
    if(str.find("A0X")!=std::string::npos) controlregister|=Scaler::HarwellCounter::A0NXTINP_C;
    if(str.find("A1X")!=std::string::npos) controlregister|=Scaler::HarwellCounter::A1NXTINP_C;
    if(str.find("A0C")!=std::string::npos) controlregister|=Scaler::HarwellCounter::A0NXTCRRY_C;
    if(str.find("A1C")!=std::string::npos) controlregister|=Scaler::HarwellCounter::A1NXTCRRY_C;
    if(str.find("BPU")!=std::string::npos) controlregister|=Scaler::HarwellCounter::BPULSE_C;
    if(str.find("LAX")!=std::string::npos) controlregister|=Scaler::HarwellCounter::LAMNEXT_C;
    if(str.find("LAE")!=std::string::npos) controlregister|=Scaler::HarwellCounter::LAMENABLE_C;
    return controlregister;
  }
  
  std::string HarwellCounter::code2str(uint16_t num)
  //Converting control code into string
  {
    //Convert codeword
    std::string cdw;//codeword
    if((num&Scaler::HarwellCounter::A0NOW_C)!=0x0) cdw+="|A0N";
    if((num&Scaler::HarwellCounter::A1NOW_C)!=0x0) cdw+="|A1N";
    if((num&Scaler::HarwellCounter::BPULSE_C)!=0x0) cdw+="|BPU";
    if((num&Scaler::HarwellCounter::A0NXTINP_C)!=0x0) cdw+="|A0X";
    if((num&Scaler::HarwellCounter::A1NXTINP_C)!=0x0) cdw+="|A1X";
    if((num&Scaler::HarwellCounter::LAMNEXT_C)!=0x0) cdw+="|LAX";
    if((num&Scaler::HarwellCounter::A0NXTCRRY_C)!=0x0) cdw+="|A0C";
    if((num&Scaler::HarwellCounter::A1NXTCRRY_C)!=0x0) cdw+="|A1C";
    if((num&Scaler::HarwellCounter::LAMENABLE_C)!=0x0) cdw+="|LAE";
    cdw=cdw.substr(1);//Cut the first | of the string
    return cdw;
  }
  
  //---------------------------------------------------------------------------RTCCounter class
  bool RTCCounter::SetupHW()
  //Set up the device according to present setting
  {
    //setup of the module: enable LAM (if used) and write frequency divisor
    bool retval=true;
    preset_reached=false;
    if (GetLAMUse()) retval=camacEnableDisable_LL(GEN_A,true);
    if (retval) retval=camacWrite_LL(FDIV_A, divisor_word);
    return retval;
  }
  
  bool RTCCounter::StartHW()
  //Starts counting (should be implemented by the actual hardware)
  {
    if(GetPreset()!=0) return camacWrite_LL(GEN_A,GetPreset());//Write preset count to register and start
    else return camcomm->Control(GEN_A, STMAX_F, getSlot(), (uint8_t)Camac::STATUS_Q);
  }
  
  bool RTCCounter::StopHW()
  //Stops the data collection (ST_ON supposed)
  {
    if(!camacWrite_LL(FDIV_A,DIV_STOP)) return false;//Stops to accept incoming pulses
    if(GetLAMUse() && (!camacClearLAM_LL(GEN_A) || !camacEnableDisable_LL(GEN_A,false))) return false;//Clear and Disable LAM
    return true; 
  }
  
  bool RTCCounter::ReadCountsHW(RTC_CNT *counts, double *remaining)
  //Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
  {
    uint32_t signal=camacRead_LL(GEN_A);//Read signal
    if(signal == CamacCounter::CERROR) return false;//Communication error
    if(GetPreset()==0)
    {
      *counts=signal&MAX_PRESET;
      if((*counts == 0x0)&&preset_reached) *counts=MAX_PRESET;
    }
    else *counts=(signal&MAX_PRESET)-(MAX_PRESET-GetPreset());//Get real number of counts
    if(remaining) *remaining=MAX_PRESET-(signal&MAX_PRESET);
    return true;
  }

  bool RTCCounter::isFinishedHW()
  //Returns with true, if the measurement is finished (ST_ON supposed)
  {
    if(GetLAMUse()) return (preset_reached=camacTestLAM_LL(GEN_A));//If LAM service is available
    else return (preset_reached= !GetBusyStatus());
  }
  
  RTCCounter::DIVISOR RTCCounter::GetFrequencyDivisor()
  //Returns with the value of the actual frequency divisor
  {
    int result=-1;//Set default value
    if(divisor_word!=0x0)
    {
      uint16_t dw=0x1;//Get a copy
      //Find out the power
      result++;
      for(int i=0; dw==divisor_word; i++)
      {
	result++;
	dw=(dw<<1);
      }
    }
    //Converting then to DIVISOR units
    RTCCounter::DIVISOR res=DIV_STOP;
    if(divisor_word==0x1) res=DIV_1;
    else if(divisor_word==0x2) res=DIV_8;
    else if(divisor_word==0x4) res=DIV_64;
    else if(divisor_word==0x8) res=DIV_512;
    else if(divisor_word==0x10) res=DIV_4096;
    else if(divisor_word==0x20) res=DIV_32768;
    else if(divisor_word==0x40) res=DIV_262144;
    return res;
  }
  
  RTCCounter::DIVISOR RTCCounter::num2div(const int num)
  //This method converts integer divisor into DIVISOR type
  {
    DIVISOR divisor_value=DIV_STOP;
    if(num>0)
    {
      divisor_value=DIV_1;
      if(num>=8) 
      {
	divisor_value=DIV_8;
	if(num>=64) 
	{
	  divisor_value=DIV_64;
	  if(num>=512) 
	  {
	    divisor_value=DIV_512;
	    if(num>=4096) 
	    {
	      divisor_value=DIV_4096;
	      if(num>=32768) 
	      {
		divisor_value=DIV_32768;
		if(num>=262144) divisor_value=DIV_262144;
	      }
	    }
	  }
	}
      }
    }
    return divisor_value;
  }
}


