#include <algorithm>
#include <cmath>
#include <cstddef>
#include <cstdint>

#include <iostream>
#include <string>
#include <unistd.h>


#include "Camac.h"
#include "TrumpPCI.h"
#include "Mesy.h"


#ifndef SCALER_H
#define SCALER_H

namespace Scaler
{
  enum STATUS {ST_OFF, ST_CONFIGURED, ST_ON};//Counter basic status
  
  template<class T> class Counter
  //General counter object
  {
    public:
    	//Methods
    	Counter(int counter_no, std::vector<std::string> nm, std::vector<int> dims): dimensions(dims){no_counter=counter_no; cnt = new T [no_counter]; stat=ST_OFF; already_read=false; setNames(nm);};//Initialize counter object. Argument: number of counters belonging to this class, the names of the counter and dimensions of the counter array
    	~Counter(){delete [] cnt;}//Delete counts
    	bool Configure(){if(stat==ST_ON) return false; if(!SetupHW()) return false; stat=ST_CONFIGURED; already_read=false; return true;};//Configure the counter to be ready for start, retuns with success or false, if it is still counting, or setup was unsuccessful
    	bool Start(){if(stat!=ST_CONFIGURED) return false; if(!StartHW()) return false; stat=ST_ON; return true;};//Start counter, returns with success, if everything is OK. Returns with false: if counter is not configured or starting by hardware has not been succeeded
    	bool Read(T* counts, double &remaining, bool &isfinished);//Read the actual counts in: they apper in counts, the remeining time/counts in remaining and finished status
    	bool Stop(T* counts);//Stops counting: returns with true. The counts is overwritten by the number of detected counts
    	Scaler::STATUS GetStatus(){return stat;};//Returns with the actual status
    	int getCounterNo(){return no_counter;};//Returns with the actual number of counters
    	std::vector<int> getCounterDimensions(){return dimensions;};//Returns with the dimensions of hte counter in [i][j][k].. order
    	void setNames(std::vector<std::string> nm){name=nm;};//Sets the name of the counter
    	std::vector<std::string> getNames(){return name;};//Returns with the name of the counter
    protected:
        bool changeCounterNo(int counter_no, std::vector<int> dims);//Changes the dimension of the cnt array (the content of the previous dataset will be cleared). This operation is allowed to be performed, when Counter is OFF
    	virtual bool SetupHW()=0;//Set up the device according to present setting
    	virtual bool StartHW()=0;//Starts counting (should be implemented by the actual hardware)
    	virtual bool StopHW()=0;//Stops the data collection (ST_ON supposed)
    	virtual bool ReadCountsHW(T* counts, double *remaining)=0;//Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
    	virtual bool isFinishedHW()=0;//Returns with true, if the measurement is finished (ST_ON supposed)
    private:
    	//Private members
    	Scaler::STATUS stat;//Stores the current status of the counter
    	int no_counter;//Number of counters
        T* cnt;//Stores the actual value of counts
        bool already_read;//This flag set to true, if dataset has been read in during the ST_ON status
        std::vector<std::string> name;//Names of the counters belonging to 0,0; 01...etc. indices
        std::vector<int> dimensions;//Dimension of the cnt array in [i][j][k] order
  };
  
  typedef uint32_t PSD_CNT;//Resolution of the PSD counter
  
  class PSDCounter: public Counter<PSD_CNT>, private TrumpPCI
  //This class reduces the ability of TrumpPCI card to the relevant ones used as position sensitive detector counter and extends Counter capabilities
  {
    public:
      //public methods
      PSDCounter(std::string name, double real_preset=0.0, double live_preset=0.0, int channels=1024, int discriminator_low=10, int discriminator_hi=1014, GATE_MODE gate=GATE_OFF, double deadtime_delay=1.6, int serial_number=-1, double timeout_ms=0.1, uint16_t adc_offset=2048);//Constructor: name, real and live presets (0.0: no preset); channel number(256/512/1024/2048), lower and upper levele discriminators; gate mode(GATE_OFF/GATE_COINCIDENCE/GATE_ANTICOINCIDENCE); deadtime delay for each incoming pulses in us, serial number of the card (-1: choose the first one among others), timeout of reading in ms; ADC offset in hardware unit (2048: no shift)
      ~PSDCounter(){};//Destructor
      //Counting extended methods: get the time informations
      void GetTimes(double *real, double *live){*real=reallivetime[0]; *live=reallivetime[1];};//Returns with the last stored real and live times
      bool ReadPSD(PSD_CNT *counts, double *actual_real, double *actual_live, double &remaining, bool &isfinished){bool retval=Read(counts, remaining, isfinished); GetTimes(actual_real, actual_live); return retval;};//Read the actual counts in: they apper in counts, the actual real and live times, the remeining time and finished status
      bool StopPSD(PSD_CNT *counts, double *real, double *live){bool retval=Stop( counts); GetTimes(real,live); return retval;};//Stops counting: returns with true. The counts is overwritten by the number of detected counts, real and live times
      //Setup related methods
      bool ChangePresets(double real_preset=0.0, double live_preset=0.0);//Change presets: [0.0 means: not preset], should be integer multiple of 0.05
      bool ChangeChannel(int channels=1024, int discriminator_low=10, int discriminator_hi=1014);//Change the number of applied channels: (256/512/1024/2048), low and high level of discriminators (discriminator numbers from 0..(channel-1))
      bool ChangeGate(GATE_MODE gate=GATE_OFF);//Change gate mode: GATE_OFF/GATE_COINCIDENCE/GATE_ANTICOINCIDENCE are accepted
      //Setup checking methods
      bool GetPresets(double *real_preset, double *live_preset);//Returns with the current setup of real and live time presets in seconds. [0 in agument: not calculated], 0.0 at return: not set. Return: communication OK/error
      bool GetChannel(int *channel, int *discriminator_low, int *discriminator_high);//Returns with the current setup of channel related stuffs: number of channels, high and low level of discriminators. Returning value is true, if communication was OK (discriminator numbers from 0..(channel-1))
      GATE_MODE GetGate(){return TrumpPCI::getGate();};//Returns with the current setup of gate
      //Methods for test purpose and About
      bool About(int *serial_number, std::string *firmware){bool feat; std::string m,r; return TrumpPCI::About(&feat,&m,&r,serial_number, firmware);};//Provides information about the card: serial number and firmware model&version. It returns with true, if communication was successful.
      bool TestWriteTimeData(PSD_CNT count, double real, double live);//Fill the counters with count counts and sets real and live times. It returns with communication success or fail. It works only, if counting has not been started, but the device is configured
      double getDTimeDelay(){return TrumpPCI::getDTimeDelay();};//Returns with the required time to restore initial level from the end of digitalization in microsec units (0 means infinite delay). Communication error is also reported.
      uint16_t getOffset(){return TrumpPCI::getOffset();};//Returns with the ADC offset in hardware units or with error
      double getTimeout(){return TrumpPCI::getTimeout();};//Returns with the value of timeout in ms
    private:
      //private methods
      bool SetupHW(){for(int i=0; i<2; i++) reallivetime[i]=0.0; return clearTimeData();};//Set up the device: clear all registered data and timers
      bool StartHW(){return TrumpPCI::StopStart(false);};//Starts counting (should be implemented by the actual hardware)
      bool StopHW(){return TrumpPCI::StopStart(true);};//Stops the data collection
      bool ReadCountsHW(PSD_CNT *counts, double *remaining);//Get number of counts and store the belonging live and real counts and remaining time/counts
      bool isFinishedHW(){return !TrumpPCI::getStatus();};//Returns with true, if the measurement is finished
      //private members
      double reallivetime[2];//Stores the real and live time values belonging to the previous reques of ReadCountsHW
      bool pres[2];//Stores, which preset has been set up
  };
  
  class MesyCounter: public Counter<Mesy::MESY_CNT>, private Mesy::MesyCommand
  //This class describes a Mesy PSD array embedded into the framework of Counter
  {
    public:
        MesyCounter(std::vector<std::string> tube_names, UDPClient &command, UDPServerTimeOut &data, Mesy::MesyDAQ &msdaq, std::vector<uint8_t> &thresholds, std::vector<uint8_t> &gains, const uint8_t master=0): Counter<Mesy::MESY_CNT>(msdaq.GetNoDatasets(),tube_names, msdaq.GetDatasetDimensions()),  Mesy::MesyCommand(command, data, msdaq, thresholds, gains, master){};//Constructor: names of the tubes, sets the UDP communication objects for command (might be the Mesytec device or data collector computer, which act as relay) and for data channels (to the data collector computer), the data acquisition object (doing the high-level acquisition), thresholds (discriminator) for each MPSD modules,  gains for each active tubes and the master channel number.
        void WriteRaws(std::string fname=""){Mesy::MesyCommand::WriteRaws(fname); return;};//It writes raw datasets into fname_rawX.dat file, if filename is not provided, then the already set up one will be saved
    private:
        //members
        //methods
        bool SetupHW(){return Mesy::MesyCommand::UpdateSetup()&&Mesy::MesyCommand::MCPDDAQReset();};//Set up the device according to present setting and reset the counter
        bool StartHW(){return Mesy::MesyCommand::MCPDDAQStart();};//Starts counting (should be implemented by the actual hardware)
        bool StopHW(){return Mesy::MesyCommand::MCPDDAQStop();};//Stops the data collection (ST_ON supposed)
        bool ReadCountsHW(Mesy::MESY_CNT* counts, double *remaining){if(remaining!=0) *remaining=-.5; return Mesy::MesyCommand::GetRecordedData(counts);};//Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
        bool isFinishedHW(){return !Mesy::MesyCommand::IsMeasurementON();};//Returns with true, if the measurement is finished (ST_ON supposed)
  };
  
  
  template<class T> class CamacCounter: public Counter<T>
  //This class describes a Counter class operated trough Camac bus and provides low-level methods for further use by children classes
  {
    public:
      //public methods
      CamacCounter(std::string name, int counter_no, Camac &cam, const uint8_t slotno, const T initial_preset=0L, const bool lam_in_use=true, bool xanswer=true):Counter<T>(counter_no, {name}, {counter_no}){camcomm=&cam; slot=slotno; xans=xanswer; SetPreset(initial_preset); SetLAMUse(lam_in_use);};// Constructor: name, camac communicator, slot number, initial preset, wheather to use LAM service, does module should give X answer?
      ~CamacCounter(){camcomm=0;};//Destructor
      bool SetLAMUse(bool use=true){if(Counter<T>::GetStatus()!=ST_OFF) return false; else{ulam=use; return true;}};//Set LAM service: only in OFF status
      bool GetLAMUse(){return ulam;};//Returns with LAM service indicator
      bool SetPreset(T preset){if(Counter<T>::GetStatus()==ST_ON) return false; else{start_preset=preset; return true;}};//Stores the preset value, if counting has not been started
      T GetPreset(){return start_preset;};//Returns with the preset start value
      std::string getCAMName(){return camcomm->getName();}//Returns with the ID of the CAMAC controller
      inline uint8_t getSlot(){return slot;};//Returns with the slot number
    protected:
      //members
      Camac *camcomm;//Camac communicator object
      static const uint32_t CERROR=0xffffffff;//Communication error
      //virtual methods
      virtual bool SetupHW()=0;//Set up the device according to present setting
      virtual bool StartHW()=0;//Starts counting (should be implemented by the actual hardware)
      virtual bool StopHW()=0;//Stops the data collection (ST_ON supposed)
      virtual bool ReadCountsHW(T *counts, double *remaining)=0;//Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
      virtual bool isFinishedHW()=0;//Returns with true, if the measurement is finished (ST_ON supposed)
      //Low-level Camac methods
      inline uint32_t camacRead_LL(uint8_t subcode){uint32_t aux; bool retval=camcomm->getData(subcode, READ_F, slot, aux, (uint8_t) ((xans? Camac::STATUS_X : 0x0) | Camac::STATUS_Q)); return (retval ? aux&0xffffff : CERROR);};//Read register value with subcode, returns with that or error code
      inline bool camacTestLAM_LL(uint8_t subcode){return camcomm->Control(subcode, TLAM_F, slot, (uint8_t) ((xans? Camac::STATUS_X : 0x0) | Camac::STATUS_Q));};//Tests LAM bistable status, returns with true, if it is on
      inline bool camacClearLAM_LL(uint8_t subcode){return camcomm->Control(subcode, CLAM_F, slot, (uint8_t)  (xans? Camac::STATUS_X : 0x0));};//Clears LAM bistable, returns with communication success or fail
      inline bool camacWrite_LL(uint8_t subcode, uint32_t value, bool qanswer=true){return camcomm->setData(subcode, WRITE_F, slot, value&0xffffff, (uint8_t) ((xans ? Camac::STATUS_X : 0x0) | (qanswer ? Camac::STATUS_Q : 0x0)));};//Writes the value into register, if qanswer is true, Q answer is required over X
      inline bool camacEnableDisable_LL(uint8_t subcode, bool enabledisable){return camcomm->Control(subcode, ( enabledisable ? ENABLE_F : DISABLE_F), slot, (uint8_t)  (xans? Camac::STATUS_X : 0x0));};//Enables(t) or disables(f) operation, returns with communication success or fail
      inline bool camacTestStatus_LL(uint8_t subcode){return camcomm->Control(subcode, TSTAT_F, slot, (uint8_t) ((xans? Camac::STATUS_X : 0x0) | Camac::STATUS_Q));};//Test some status flag
    private:
      //private members
      uint8_t slot;//Slot number
      T start_preset;//Stores the start preset: necessary to calculate elapsed counts
      bool ulam;//LAM service indicator
      bool xans;//X answer on CAMAC operation (t: present, f: absent)
      // CAMAC function codes
      static const uint8_t READ_F=0;//Read function code
      static const uint8_t TLAM_F=8;//Test  LAM function code
      static const uint8_t CLAM_F=10;//Clear  LAM function code
      static const uint8_t WRITE_F=16;//Write function code
      static const uint8_t DISABLE_F=24;//Disable function code
      static const uint8_t ENABLE_F=26;//Enable function code
      static const uint8_t TSTAT_F=27;//Test status function code
  };
  
  typedef uint64_t MON_CNT;//Define the resolution of the monitor counter
  
  class MonitorCounter: public CamacCounter<MON_CNT>
  //This class implements the Scaler Module monitor counter
  {
    public:
      //public methods
      MonitorCounter(std::string name, Camac &cam, const uint8_t slotno, bool count_or_time_preset=false, const MON_CNT initial_preset=0L, bool intensity_monitoring=true, const uint16_t low_intensity_warning_level=DEFAULT_INTENSITY_NIVEAU, const uint8_t times_below_low_before_warning=DEFAULT_CONSEC, bool warning_suspend=true, bool lam_in_use=true);// Constructor: name, camac communicator, slot number, count preset(t)/time preset(t), initial preset, do we need to monitor intensity level?, low intensity warning level in cps, how many times consecutively below the previous value to give intensity warning, suspend the acquisition in the case of warning?, wheather to use LAM service
      ~MonitorCounter();//Destructor: we should disable LAM service of the device
      bool ChangeOpMode(bool countertimer);//Changes the operation mode to counter(t) or timer(f), when status is OFF
      bool SetIntensityParameters(const bool need=true, const uint16_t niv=DEFAULT_INTENSITY_NIVEAU, const uint8_t consec=DEFAULT_CONSEC, bool poor_suspend=true);//Sets the intensit warning levels in OFF status, if it is needed (true). If it were the case, warning will be issued, if the intensity is less than the lowest intensity niv(au) during "consec" number of consecutive sampling, at least. The poor_suspend regulates the behaviour of warning level during the measurement: if it is true, calculation is suspended, otherwise continues.
      void GetIntensityParameters(uint16_t *niv, uint8_t *consec){ *niv=intensity_crg&0xffff; *consec=(intensity_crg>>16)&0xf; return;};//This function returns with the intensity niveau and consecutive appearence, used to qualify signal as poor
      inline void CheckMonitorNextRead(){if((!suspend_poor)&&(intensity_crg!=0x0)) check_monitor=true;return;};//Set check monitor POOR_INTENSITY flag at next read (only has an effect, if intensity monitoring requested, but not suspend on warning)
      inline bool PoorIntensity(){return (((!suspend_poor)&&(intensity_crg!=0x0)) ? last_monitor : false);};//Returns with the value at the last monitor counter check. T: POOR_INTENSITY (valid only, if intensity monitoring)
      inline bool GetOpMode(){return opmode;}//T: counter, F: timer
      bool SetPreset(MON_CNT preset);//Set preset for the current setup
      bool SetLAMUse(bool use);//Set LAM service: only in OFF status -- CamacCounter will be overdefined
      //public constants
      static const MON_CNT TIME_PRESET_MAX=0xffffff;//Maximal available time preset (about 194 days)
      static const MON_CNT COUNT_PRESET_MAX=0xffffffffffff;//Maximal available count preset
      static const uint16_t DEFAULT_INTENSITY_NIVEAU=5;
      static const uint8_t DEFAULT_CONSEC=7;
    private:
      //private members
      bool first;//Does it the first configuration? (if not, it saves some CAMAC communication)
      bool opmode;//Operation mode: counter(t) or timer (f)
      bool suspend_poor;//Suspend counting for low rate, or continue
      bool check_monitor;//If this flag is true, then the monitor intensity bistable will be checked
      bool last_monitor;//Returns with the value of the last monitor (t: POOR_INTENSITY)
      uint32_t intensity_crg;//Stores the configuration register of intensity. If it is 0x0, means, that this kind of monitoring will not be required
      //Constants
      static const MON_CNT MON_BAD_OP=0xffffffffffffffff;//Not suitable operation mode
      static const MON_CNT MON_TIMEOUT=0xfffffffffffffffe;//Bad communication error or not possible to read
      enum MODES{M_ALL, M_EVENT, M_PRESET, M_TIME, M_INTENSITY, M_MODULE};//Defines different modes
      enum REGISTERED_COUNTS{EVENT,PRESET};//Labeling the counts
      static const uint8_t scode[6];//Contains the subcodes, belonging to MODES
      //CAMAC subcodes
      static const uint8_t ALL_A=0;//Realted to overall
      static const uint8_t EVENTL_A=1;//Least significant bits of event counter
      static const uint8_t EVENTM_A=2;//Most significant bits of event counter
      static const uint8_t PRESETL_A=3;//Least significant bits of preset counter
      static const uint8_t PRESETM_A=4;//Most significant bits of preset counter
      static const uint8_t TIME_A=5;//Time preset
      static const uint8_t POOR_A=6;//Intensity monitoring
      static const uint8_t MODULE_A=11;//Module related
      //CAMAC function codes
      static const uint8_t CLEAR_F=9;//Clear (event counter) function code
      static const uint8_t MODESET_F=17;//Mode set function code
      //CAMAC writing codes
      static const uint32_t COUNT_W=1;//Count preset mode
      static const uint32_t TIME_W=2;//Time preset mode
      //Private methods
      //Realisation of the pure virtual methods of the parent
      bool SetupHW();//Set up the device according to present setting
      inline bool StartHW(){return enableDisable(M_ALL,true);};//Starts counting: enable module (DW or ext INHIBIT should be cleared)
      bool StopHW();//Stops the data collection (ST_ON supposed)
      bool ReadCountsHW(MON_CNT *counts, double *remaining);//Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
      bool isFinishedHW();//Returns with true, if the measurement is finished (ST_ON supposed)
      //interface methods for the members
      //Getting the original CAMAC into friendly format
      MON_CNT readLL(MODES mode);//Read the requested register. Mode can be M_EVENT, M_PRESET, M_TIME.
      bool testLAM(MODES mode);//Test LAM enabled/disabled status, returns with true, if it is on. Mode can be M_ALL, M_PRESET, M_TIME, M_INTENSITY.
      bool clearEvent(){return camcomm->Control(EVENTL_A, CLEAR_F, getSlot(),Camac::STATUS_X);};//Clears the event counter. 
      bool clearLAM(MODES mode);//Clears LAM bistable, returns with communication success or fail. Mode can be M_PRESET, M_TIME, M_INTENSITY.
      bool writeLL(MODES mode, MON_CNT value);//Writes the value into the required mode register. Mode can be M_PRESET, M_TIME, M_INTENSITY.
      bool setOpMode(bool counter){return camcomm->setData(MODULE_A, MODESET_F, getSlot(), (counter ? COUNT_W : TIME_W), Camac::STATUS_X);};//Sets the required mode of the module: count preset (t) or time preset (f)
      bool enableDisable(MODES mode, bool enable);//Enables(t) or disables (f) operation. Mode can be M_ALL, M_PRESET, M_TIME, M_INTENSITY, M_MODULE
      bool testLFlag(MODES mode);//Tests the value of L-bistable(DONE) flag. Mode can be M_PRESET, M_TIME, M_INTENSITY.
  };
  
  typedef uint32_t HW_CNT;//Resolution of Harwell counter: get consistent with PRESET_MAX
  
  class HarwellCounter: public CamacCounter<HW_CNT>
  //This class implements the 95/7039-1/6 Preset Counting Register counter made in Harwell
  {
    public:
      //Public methods
      HarwellCounter(std::string name, Camac &cam, const uint8_t slotno, const HW_CNT initial_preset=0L, const uint16_t control_register=0, bool lam_in_use=true):CamacCounter<HW_CNT>(name, 1, cam, slotno, initial_preset, lam_in_use,false){SetControlRegister(control_register);}; //Constructor: name, camac communicator, slot number, initial preset [zero means: does not need preset], control register, wheather to use LAM service
      ~HarwellCounter(){};//Destructor
      inline bool GetALevel(){return camacTestStatus_LL(START_A);};//Returns with the actual level of "A" output
      inline bool StepDownPreset(){return camcomm->Control(START_A, XECDEC_F,getSlot(),0x0);};//Decrease the preset count register by one
      bool SetControlRegister(const uint16_t cr);//This method sets the control register
      uint16_t GetControlRegister(){return control_word;};//Returns with the actual value of the control register
      //public members
      static const HW_CNT PRESET_MAX=0xffffff;//Preset maximum(24 bits)
      //Control register constants
      static uint16_t str2code(std::string str);//Converting string into control code
      static std::string code2str(uint16_t num);//Converting control code into string
      static const uint16_t A0NOW_C=0x1;//Clear A (high voltage)
      static const uint16_t A1NOW_C=0x2;//Set A (low voltage)
      static const uint16_t BPULSE_C=0x4;//Give nove a pulse on B
      static const uint16_t A0NXTINP_C=0x8;//Clear A at next input
      static const uint16_t A1NXTINP_C=0x10;//Set A at next input
      static const uint16_t LAMNEXT_C=0x20;//Clears LAM bistable and give LAM at next input
      static const uint16_t A0NXTCRRY_C=0x40;//Clear A at next carry
      static const uint16_t A1NXTCRRY_C=0x80;//Set A at next carry
      static const uint16_t LAMENABLE_C=0x100;//Enables LAM
    private:
      //Private members
      uint16_t control_word;//Contains the word should be written to control register
      bool defaultlevel;//Default level of A (this level indicates, that counting is finished by not using LAM service)
      bool no_incoming;//Indicates, that any impulse has not been detected yet.
      //CAMAC codes
      static const uint8_t WRCNTRL_F=17;//Write control register function code
      static const uint8_t XECDEC_F=25;//Execute function code: decrease preset register by 1
      static const uint8_t START_A=0;//General subcode: for clearLAM, write_LL and F(17), it starts to count
      static const uint8_t IQ_A=1;//Does not give Q-answer: for testing purposes
      static const uint8_t INHIBIT_A=2;//Inhibit counting for clearLAM, write_LL and F(17)
      //private methods
      //implementation of the parent's pure virtuals
      bool SetupHW();//Set up the device according to present setting
      bool StartHW();//Starts counting (should be implemented by the actual hardware)
      bool StopHW();//Stops the data collection (ST_ON supposed)
      bool ReadCountsHW(HW_CNT *counts, double *remaining);//Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
      bool isFinishedHW();//Returns with true, if the measurement is finished (ST_ON supposed)
  };
  
  typedef uint16_t RTC_CNT;//Define Real Time Clock resolution: get consistent with MAX_PRESET
  
  class RTCCounter: public CamacCounter<RTC_CNT>
  //This class implements the RTC 2014 Real Time Clock counter
  {
    public:
      //public members
      static const uint32_t MAX_PRESET=0xffff;//Maximal preset start from 0
      enum DIVISOR{DIV_STOP, DIV_1, DIV_8, DIV_64, DIV_512, DIV_4096, DIV_32768, DIV_262144};//Divisor of frequency/counts by the value after "_"
      //Public methods
      RTCCounter(std::string name, Camac &cam, const uint8_t slotno, const RTC_CNT initial_preset=0L, DIVISOR freq_div=DIV_1, bool lam_in_use=true):CamacCounter<RTC_CNT>(name, 1, cam, slotno, initial_preset, lam_in_use,false){SetFrequencyDivisor(freq_div);}; //Constructor: name, camac communicator, slot number, initial preset [zero means:with maximal preset], divisor for frequancy [0: stop clock, 1-7: div:2^{3*(n-1)}] , wheather to use LAM service
      ~RTCCounter(){};//Destructor
      inline bool GetBusyStatus(){return camacTestStatus_LL(GEN_A);};//Returns with the actual status of BUSY flag
      void SetFrequencyDivisor(DIVISOR fdiv){divisor_word=((fdiv==DIV_STOP) ? 0x0 : 0x1<<(fdiv-1));};//This method sets the frequency divisor word [0: stop clock, 1-7: div:2^{3*(n-1)}]
      DIVISOR GetFrequencyDivisor();//Returns with the value of the actual frequency divisor
      static DIVISOR num2div(const int num);//This method converts integer divisor into DIVISOR type
      inline static int div2num(const DIVISOR div){return (int)((div==DIV_STOP) ? 0x0 : 0x1<<3*(div-1));};//This method converts DIVISOR into integer type
    private:
      //Private members
      uint16_t divisor_word;//Contains the word should be written to control register
      bool preset_reached;//If the counting process finished by reaching the preset, it is set to true
      //CAMAC codes
      static const uint8_t STMAX_F=28;//Start maximal run function code
      static const uint8_t GEN_A=0;//General subcode
      static const uint8_t FDIV_A=1;//Frequency divisor subcode
      //private methods
      //implementation of the parent's pure virtuals
      bool SetupHW();//Set up the device according to present setting
      bool StartHW();//Starts counting (should be implemented by the actual hardware)
      bool StopHW();//Stops the data collection (ST_ON supposed)
      bool ReadCountsHW(RTC_CNT *counts, double *remaining);//Get number of counts and remaining time/counts (ST_ON (it is not sure, still in it by hardware -- if communication with HW is the bottleneck, please do not check finished status) /ST_OFF(surely finished) are supposed)
      bool isFinishedHW();//Returns with true, if the measurement is finished (ST_ON supposed)
  
  };
 
 //Template functions
 template <class T> bool Counter<T>::Read(T* counts, double &remaining, bool &isfinished)
  //Read the actual counts in: they apper in counts, the remeining time/counts in remaining and finished status
  {
    remaining=0.0;//Initialize remaning time
    if(stat==ST_ON)
    {
      already_read=true;//Set already read flag
      isfinished=isFinishedHW();//Figure out: still counting?
      if(isfinished) return Stop(counts);//If finished, call the stop routine
      else return ReadCountsHW(counts,&remaining);//Call hardware related routine, return false in the case of bad answer
    }
    else if(!already_read)
    {
      //We have not been read in any values, but it would provide some datasets
      isfinished=true;
      return ReadCountsHW(counts,&remaining);//Read counts
    }
    else
    {
      //The counting already finished, read it from local storage
      if(counts!=0) for (int i=0; i<no_counter; i++) counts[i]=cnt[i];
      isfinished=true;
      return true;
    }
  }
  
  template <class T> bool Counter<T>::Stop(T* counts)
  //Stops counting: returns with true. The counts is overwritten by the number of detected counts
  {
    if(stat!=ST_ON) return false;//Counting already finished
    else
    {
    	//We should stop it now
    	bool retval=StopHW();//Stop the counter
    	if(!retval) return false;
    	stat=ST_OFF;//Set status
    	if(!ReadCountsHW(cnt,0)) return false;//Read detected counts by temporary storage
    	if(counts!=0) for (int i=0; i<no_counter; i++) counts[i]=cnt[i];//Set the counts
    	return true;
    }
  }
  
  template <class T> bool Counter<T>::changeCounterNo(int counter_no, std::vector<int> dims)//Changes the dimension of the cnt array (the content of the previous dataset will be cleared). This operation is allowed to be performed, when Counter is OFF
  {
    if(stat!=ST_OFF) return false;//It is allowed to do, when the counter is off
    if(no_counter!=counter_no)
    {
    	//We have to do anything, if the present and the requested number is not the same
    	delete [] cnt;
    	no_counter=counter_no;
        dimensions=dims;
    	cnt = new T [no_counter];
    }
    return true;
  }
}

#endif
  
