#include "Serial.h"

bool Serial::Open(const std::string nPort, const int nBaud, const int nDatabits, const char nParity, const int nStopbits, const bool rawcanonical, const bool softhand, const bool hardhand)
//This function opens a serial communication line: port name (device), baud rate, number of databits, parity: 'N'-none,'E'-even,'O'-odd, number of stopbits, raw (t) or canonical input(f), software handshake (Xon/Xoff), hardwarehandshake --- if everything is OK, retuns with true
{
	//Setting the raw/canonical reading value
	rc=rawcanonical;
  
	//Checking opened status:
	if (fd>0) Close();//If is it already open: close it

    //allocating the device
#ifdef DEBUG_SERIAL
    std::cout<<"(Serial::Open)The requested device name: "<<nPort<<std::endl;
#endif

    int flags= O_RDWR;
    if (rc) flags |= (O_NOCTTY | O_EXCL | O_NDELAY);
    else flags &= ~O_NDELAY;
    std::string filename = "/dev/" + nPort;
    fd=open(filename.c_str(), flags);//RW
    //Checking status
    if(fd<0)
    {
        std::cerr<<"ERROR(Serial::Open): Error occured during the opening of the \""<<nPort<<"\" device."<<std::endl;
        return false;
    }

    //Getting the old structure and store it until the end
    tcgetattr(fd,&oldstruct);

	//Set the new structure
	bzero(&newstruct,sizeof(newstruct));//Set all bits to zero
	//Control flags --set the baudrate
	switch (nBaud)
	{
		case    50:  newstruct.c_cflag = B50; break;
		case    75:  newstruct.c_cflag = B75; break;
		case   110:  newstruct.c_cflag = B110; break;
		case   134:  newstruct.c_cflag = B134; break;
		case   150:  newstruct.c_cflag = B150; break;
		case   200:  newstruct.c_cflag = B200; break;
		case   300:  newstruct.c_cflag = B300; break;
		case   600:  newstruct.c_cflag = B600; break;
		case  1200:  newstruct.c_cflag = B1200; break;
		case  1800:  newstruct.c_cflag = B1800; break;
		case  2400:  newstruct.c_cflag = B2400; break;
		case  4800:  newstruct.c_cflag = B4800; break;
		case  9600:  newstruct.c_cflag = B9600; break;
		case 19200:  newstruct.c_cflag = B19200; break;
		case 38400:  newstruct.c_cflag = B38400; break;
		case 57600:  newstruct.c_cflag = B57600; break;
		case 115200:  newstruct.c_cflag = B115200; break;
		case 230400:  newstruct.c_cflag = B230400; break;
		case 460800:  newstruct.c_cflag = B460800; break;
		case 576000:  newstruct.c_cflag = B576000; break;
		default: std::cerr<<"ERROR(Serial::Open): Invalid baudrate! Only 50, 75, 110, 134, 150, 200,\n300, 600, 1200, 1800, 2400, 4800, 9600\n19200, 38400, 57600, 115200, 230400, 460800 are allowed!\n"; return false;
	}

	//setting the databits
	newstruct.c_cflag &=~CSIZE;//Bit mask for databits
	switch (nDatabits)
	{
		case 5: newstruct.c_cflag |= CS5; break;
		case 6: newstruct.c_cflag |= CS6; break;
		case 7: newstruct.c_cflag |= CS7; break;
		case 8: newstruct.c_cflag |= CS8; break;
		default: std::cerr<<"ERROR(Serial::Open): Invalid databit! Only 5, 6, 7, 8 are allowed!\n"; return false;
	}
	
	//Stopbits: 1 or 2
	if(nStopbits==1) newstruct.c_cflag &= ~CSTOPB;
	else newstruct.c_cflag |= CSTOPB;
	
	//Parity bits and parity check
	if(nParity=='N')//NONE
	{
	   newstruct.c_cflag &= ~(PARENB | PARODD);
	   newstruct.c_iflag &= ~INPCK;
	}
	else if (nParity=='E')//Even
	{
	  newstruct.c_cflag |= PARENB;
	  newstruct.c_cflag &=~ PARODD;
	  newstruct.c_iflag |= INPCK;
	}
	else if (nParity=='O')//ODD
	{
	  newstruct.c_cflag |= PARENB;
	  newstruct.c_cflag |= PARODD;
	  newstruct.c_iflag |= INPCK;
	}
	else {std::cerr<<"ERROR(Serial::Open): Parity! Only 'N':none, 'E':even, 'O':odd are allowed!\n"; return false;}

	newstruct.c_cflag|=(CLOCAL | CREAD);//Enable receiver and set local mode
	if (softhand) newstruct.c_iflag|= IXON | IXOFF;
	else newstruct.c_iflag &= ~(IXON | IXOFF | IXANY);//Software flow control disabled
	if(rc)
	{
	  newstruct.c_oflag &= ~OPOST;//Raw output
	  newstruct.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);//non-canonic mode, no inp char echo, not allowed signals => RAW_INPUT
	  newstruct.c_cc[VTIME]=0;//timer is not used
	  newstruct.c_cc[VMIN]=0;//number of characters to be read 
	}
	else
	{
	  newstruct.c_cflag&=~CRTSCTS;//No hardware handshake
	  newstruct.c_iflag |=IGNBRK;//Ignore break
	  newstruct.c_oflag=0;//Output standard
	  newstruct.c_lflag=0;//Input standard
	  newstruct.c_cc[VTIME]=1;//timer is used: 1/10 seconds
	  newstruct.c_cc[VMIN]=0;//number of characters to be read 
	}
	
	//Clear puffers and activate the new setup...
	tcflush(fd, TCIFLUSH);
	tcsetattr(fd,TCSANOW,&newstruct);
	tcflush(fd, TCIOFLUSH);
	
	//Set high level status
	if(hardhand||softhand)
	{
	  int data=0;
	  ioctl(fd, TIOCMGET, &data);
	  data |= TIOCM_RTS;
	  ioctl(fd, TIOCMSET, &data);
	  //Set hardware handshake
	  tcgetattr(fd, &newstruct);
	  if (hardhand) newstruct.c_cflag |= CRTSCTS;
	  else newstruct.c_cflag &= ~CRTSCTS;
	  tcsetattr(fd, TCSANOW, &newstruct);
	  tcflush(fd, TCIOFLUSH);
	}
	
	return true;
}

bool Serial::Close()
{
  if (fd>0)
  {
    tcsetattr(fd,TCSANOW,&oldstruct);
    
    //Free descriptor
    if(close(fd)!=0)
    {
      std::cerr<<"ERROR(Serial::Close): Closing the serial line is not successful!"<<std::endl;
      fd=0;
      return false;
    }
    fd = 0;
  }
  
  return true;
}

int Serial::Send(const uint8_t *buffer, const int length)
//This method send the buffer content to serial line. Buffer is consits at least length number of bytes
{

#ifdef DEBUG_SERIAL
        std::cout<<"(Serial::Send) Content of the object's buffer: "<<std::endl;
        for(int i=0; i<length; i++) std::cout<<(int)buffer[i]<<std::endl;
#endif
  //Checking port open status
  if(fd<=0)
  {
    std::cerr<<"ERROR(Serial::Send): The device is not open! Writing has not been done!\n";
    return -1;
  }
  return (int) write(fd,buffer,(size_t)length);
}

int Serial::Receive(uint8_t *buffer, const int length, bool flush)
//This method read length bytes and stores them into the buffer. Returns with the length of the data
{
  //Checking port open status
  if(fd<=0)
  {
    std::cerr<<"ERROR(Serial::Receive): The device is not open! Writing has not been done!\n";
    return -1;
  }
  
  uint8_t lbuf[Serial::LOCAL_BUFFER_SIZE];//Local buffer
  int actual=0;//Counter of the returning array
  
  
  if (flush)
  {
    //Clear the content of the buffer -- then go to the reading from port
    while(!temp.empty()) temp.pop();
  }
  else
  {
    //read values from buffer, if are there any
    while((!temp.empty())&&(actual<length))
    {
	buffer[actual++]=temp.front(); 
#ifdef DEBUG_SERIAL
        std::cout<<"(Serial::Receive) Content of the object's buffer: "<<(int)temp.front()<<std::endl;
#endif
	temp.pop();
    }
  }
  
  int returnval=actual;//Return value
  
  //Reading from serial line
  if(rc)
  {
    if(actual<length)
    {
      //Setting task list
      fd_set watchset, inset;
      
      FD_ZERO(&watchset);
      FD_SET(this->fd,&watchset);
      struct timeval ltimeout=*timeout;
    
      while(FD_ISSET(this->fd,&watchset))
      {
	  inset=watchset;//Copy the content to avoid overwriting
	  if(select(fd+1,&inset,NULL,NULL,&ltimeout)<0)//Suspend the execution of the program
	  {
	      std::cerr<<"ERROR(Serial::Receive): occured during the calling of the select()!"<<std::endl;
	      return -2;
	  }
      
	  //Receiving the data, if does it contain any
      
	  int res=(int)read(fd,lbuf,(size_t)(sizeof(lbuf)-1));
	  if (res>0) //Data has been read
	  {
	      //Copy received dataset into buffer
#ifdef DEBUG_SERIAL
	      std::cout<<"(Serial::Receive) Received data ("<<(int)res<<") by direct reading of the port"<<std::endl;
	      for(int i=0; i<res; i++) std::cout<<"(Serial::Receive) "<<i+1<<" byte: "<<std::dec<<(int)lbuf[i]<<" (int), "<<std::hex<<(int)lbuf[i]<<" (hexa)"<<std::dec<<std::endl;
#endif
	      int counter=0;//Counter of the local array
	      while((length>actual)&&(counter<res))
	      {
		  buffer[actual++]=lbuf[counter++];
	      }
	      if(actual==length)
	      { 
		  while(counter<res) temp.push(lbuf[counter++]);//Set the remaining part into the temporary buffer
		  returnval=length+temp.size();//Returns with the all availabe dataset (returned+buffer)
		  FD_CLR(fd,&watchset);//Go out from cycle
	      }
	  }
	  else if (res==0)
	  {
	      //Timeout
#ifdef DEBUG_SERIAL
	      std::cout<<"(Serial::Receive) Timeout\n"<<std::endl;
#endif
	      returnval=actual;//Returns with the number of available bytes
	      FD_CLR(fd,&watchset);//Go out from cycle
	  }
	  else
	  {
	      std::cerr<<"ERROR(Serial::Receive): occured during the receiving!"<<std::endl;
	      return -3;
	  }
      }    
    }
  }
  else
  {
    //Canonical reading-- receiving the data
    int res=(int) read(fd,lbuf,(size_t)(sizeof(lbuf)-1));
    if (res>0)
    {
      //Data has been read in
#ifdef DEBUG_SERIAL
	std::cout<<"(Serial::Receive) Received data ("<<(int)res<<") by canonical reading of the port"<<std::endl;
	for(int i=0; i<res; i++) std::cout<<"(Serial::Receive) "<<i+1<<" byte: "<<std::dec<<(int)lbuf[i]<<" (int), "<<std::hex<<(int)lbuf[i]<<" (hexa)"<<std::dec<<(char) lbuf[i]<<std::endl;
#endif
      int counter=0;//Counter of the local array
      while((length>actual)&&(counter<res))
      {
	  buffer[actual++]=lbuf[counter++];
      }
      if(actual==length)
      { 
	  while(counter<res) temp.push(lbuf[counter++]);//Set the remaining part into the temporary buffer
	  returnval=length+temp.size();//Returns with the all availabe dataset (returned+buffer)
      }
      else returnval=actual;
    }
    else if (res==0)
    {
	//Timeout
#ifdef DEBUG_SERIAL
	std::cout<<"(Serial::Receive) Timeout\n"<<std::endl;
#endif
	returnval=actual;//Returns with the number of available bytes
    }
    else
    {
	std::cerr<<"ERROR(Serial::Receive): occured during the receiving!"<<std::endl;
	return -3;
    }
  }
  return returnval;
}

bool Serial::setTimeout(double second)
//Sets the timeout into the requested time
{
  if ((second==0.0)&&(timeout!=NULL))//It should set to zero
  {
    delete timeout;
    timeout=NULL;
  }
  else if (rc) //Raw format
  {
    if(timeout==NULL) timeout = new struct timeval;//Not allocated case
    timeout->tv_sec=(int) second;
    timeout->tv_usec=((int)(second*1000000))%1000000;
  }
  else //Canonical format
  {
    termios prevstruct;//Previous structure
    tcgetattr(fd, &prevstruct);//Reading the previous setting
    prevstruct.c_cc[VTIME]=((((int) (second*10)) >0) ? (int) (second*10) : 1);//timer is used: last part: prevent to be zero, so it will be at least 1/10 sec
    tcsetattr(fd, TCSANOW, &prevstruct);//Set the new changes
    tcflush(fd, TCIOFLUSH);//Clear datasets
  }
  
  return true;
}

std::string Serial::getPortName()
//Returns with the port number
{
  if(fd<=0)
  {
      std::cerr<<"ERROR(Serial::getPortName): The device is not open!"<<std::endl;
      return "ERROR(Serial::getPortName): The device is not open!";
  }
  else return device;
}

double Serial::getTimeout()
//Returns with the current value of timeout
{
  if(timeout==NULL) return 0.0;//This is not set
  else return (double)(timeout->tv_sec)+0.000001*((double)(timeout->tv_usec));
}

void Serial::getCommParameters(int *baud, int *databits, char *parity, int *stopbits)
//This function returns with the currently set communication parameters as baud rate, data bits, parity and number of stopbits
{
  //Get the baud rate
  if ((newstruct.c_cflag&B576000)==B576000) *baud = 576000;
  else if ((newstruct.c_cflag&B460800)==B460800) *baud = 460800;
  else if ((newstruct.c_cflag&B230400)==B230400) *baud = 230400;
  else if ((newstruct.c_cflag&B115200)==B115200) *baud = 115200;
  else if ((newstruct.c_cflag&B57600)==B57600) *baud = 57600;
  else if ((newstruct.c_cflag&B38400)==B38400) *baud = 38400;
  else if ((newstruct.c_cflag&B19200)==B19200) *baud = 19200;
  else if ((newstruct.c_cflag&B9600)==B9600) *baud = 9600;
  else if ((newstruct.c_cflag&B4800)==B4800) *baud = 4800;
  else if ((newstruct.c_cflag&B2400)==B2400) *baud = 2400;
  else if ((newstruct.c_cflag&B1800)==B1800) *baud = 1800;
  else if ((newstruct.c_cflag&B1200)==B1200) *baud = 1200;
  else if ((newstruct.c_cflag&B600)==B600) *baud = 600;
  else if ((newstruct.c_cflag&B300)==B300) *baud = 300;
  else if ((newstruct.c_cflag&B200)==B200) *baud = 200;
  else if ((newstruct.c_cflag&B150)==B150) *baud = 150;
  else if ((newstruct.c_cflag&B134)==B134) *baud = 134;
  else if ((newstruct.c_cflag&B110)==B110) *baud = 110;
  else if ((newstruct.c_cflag&B75)==B75) *baud = 75;
  else if ((newstruct.c_cflag&B50)==B50) *baud = 50;
  else *baud=0;
  //Get the number of databits
  if((newstruct.c_cflag&CS8) == CS8) *databits=8;
  else if((newstruct.c_cflag&CS7) == CS7) *databits=7;
  else if((newstruct.c_cflag&CS6) == CS6) *databits=6;
  else if((newstruct.c_cflag&CS5) == CS5) *databits=5;
  //Decode parity
  if((newstruct.c_cflag&PARENB)==PARENB)
  {
    if((newstruct.c_cflag&PARODD)==PARODD) *parity='O';//odd parity
    else *parity='E';//even parity
  }
  else *parity='N';//none
  //Determine the stopbits
  if ((newstruct.c_cflag&CSTOPB) ==CSTOPB) *stopbits=2;
  else *stopbits=1;
  return;
}
