#include "SerialCamac.h"

//Static arrays
const uint8_t SerialCamac::SPACE_STRING[16]={ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB,ENB};//Reset string
uint8_t SerialCamac::READ_STRING[16]={0,0,0,0,0,SPC,SPC,SPC,SPC,SPC,SPC,SPC,SPC,ENB,ENB,ENB};//Contains the initialisation of the read string
uint8_t SerialCamac::WRITE_STRING[16]={0,0,0,0,0,0,0,0,0,SPC,SPC,SPC,SPC,ENB,ENB,ENB};//Contains the initialisation of the write string

SerialCamac::SerialCamac(std::string portName ,uint8_t cNo, const std::string nm, double timeout, int nBaud, int nDataBits, char nParity, int nStopbits):Camac(cNo,nm)
//Constructor
{
  pS=NULL;
  max_trials=1500000;
  //Opening the communication channel and reset the Harwell Camac unit
  if(!ResetComm(portName, timeout, nBaud, nDataBits, nParity, nStopbits))
  {
    std::cerr<<"ERROR (SerialCamac::SerialCamac): at reset communication line and Harwell unit!\n";
  }
}

bool SerialCamac::Reset()
//Resets the Harwell Camac unit
{
  //Checking open status of the serial communication line
  if (!(pS->isOpened())) return false;//The serial line is not opened yet
  //Send space, reset strings, then clear status register
  bool retval=false;
  for (int i=0; (i<5)&&(!retval); i++) retval=clear_status_register();
#ifdef DEBUG_SERIAL_CAMAC
  if (!retval) std::cerr<<"SerialCamac::Reset: unsuccessful attemption to clear the status register!"<<std::endl;
#endif
  return send_space_string() && (send_initfinal_string(true) && retval);
}

bool SerialCamac::GetInhibit(bool test)
//Get the dataway inhibit value (t: set, f: unset). If test is true, it returns the setting of the controller (it might be other sources of dataway inhibit: test is tests only the controller, but general the dataway directly).
{
  uint32_t status=0x0;
  if(getData(STREG_A, READ_F, STATION, status, STATUS_X|STATUS_Q))
    return (status&(test ? SerialCamac::STR_INHIBIT_FLAG : SerialCamac::STR_INHIBIT_DW))==(test ? SerialCamac::STR_INHIBIT_FLAG : SerialCamac::STR_INHIBIT_DW);//Checking the corresponding flags
  else return false;//Wrong communication
}


bool SerialCamac::ResetComm(const std::string portName, const double timeout, const int nBaud, const int nDataBits, const char nParity, const int nStopbits)
//Resets the communication channel and Harwell Camac unit
{
  bool retval;//Returning value
  //Modify the settings or create a new channel
  if(pS==NULL)
  {
    pS= new Serial(portName, nBaud, nDataBits, nParity, nStopbits);
    retval=pS->isOpened();
  }
  else retval=pS->Open(portName, nBaud, nDataBits, nParity, nStopbits) && (send_initfinal_string(false) && clear_status_register());
  
  //Set timeout
  if (retval)
  {
    retval=pS->setTimeout(0.001);
    max_trials=timeout*1000;
  }
  
  //Reset the device
  if(retval) retval=this->Reset();
  
  //Returning with the results
  return retval;
}

bool SerialCamac::SetResetDeviceTest(TEST_mode tm, bool setreset)
//Sets the device into the required test modes: only LOOP_COLLAPSE, BY_PASS, OFF_LINE are allowed
{
  uint32_t status=0x0;//Status word 
  switch(tm)
  {
    case LOOP_COLLAPSE	: status=SerialCamac::STR_LOOP_COLL; break;//Loop collapse
    case BY_PASS	: status=SerialCamac::STR_BYPASS; break;//By-pass
    case OFF_LINE	: status=SerialCamac::STR_OFFLINE; break;//Software offline
    default: return false;//Non-conform request
  }
  bool retval=false;
  for(int i=0; (i<5)&&(!retval); i++) retval=setData(STREG_A, (setreset ? SSET_F : SCLR_F), STATION, status, STATUS_X|STATUS_Q);//Set/reset the requested status bits
  return retval;
}

bool SerialCamac::CheckDeviceTest(TEST_mode tm)
//Returns with the actual situation of test device as requested as LOOP_COLLAPSE, OFF_LINE, HW_OFFLINE
{
  if(tm==BY_PASS) return false;//Non-valid request
  uint32_t status=0x0;//Status register
  if(getData(STREG_A, READ_F, STATION, status, STATUS_X|STATUS_Q))
  {
    switch(tm)
    {
      case LOOP_COLLAPSE: return ((status&STR_LOOP_COLL)==STR_LOOP_COLL); break;//Loop collapse
      case OFF_LINE	: return ((status&STR_OFFLINE)==STR_OFFLINE); break;//Software offline
      case HW_OFFLINE	: return ((status&STR_HW_OFF)==STR_HW_OFF); break;//Hardware offline
      default: return false;//Non-conform request
    }
  }
  else return false;//Not successful operation
}

uint8_t SerialCamac::CheckPDataFlags()
//Returns with the flags belonging to the previous data
{
  uint32_t status=0x0;//Status register
  uint8_t answer=0x0;//Answer coded uniformly
  if(getData(STREG_A, READ_F, STATION, status, STATUS_X|STATUS_Q))
  {
    if((status&STR_DERR)==STR_DERR) answer+=STATUS_S_ERR;//Error
    if((status&STR_DSX)==STR_DSX) answer+=STATUS_X;//X answer
    if((status&STR_DSQ)==STR_DSQ) answer+=STATUS_Q;//Q answer
  }
  else answer+=STATUS_S_DERR;//Delayed error as communication error indicator
  return answer;
}

bool SerialCamac::GetPData(uint32_t& previous_data, bool q_status)
//Returns with the previously read data and communication errors. If q_status is true, then Q answer will be checked
{
  //This is not implemented yet correctly
  return getData(STORED_A, STORED_F, STATION, previous_data, (q_status ? (STATUS_X|STATUS_Q) : STATUS_X));
}

uint32_t SerialCamac::getdata(const uint8_t subaddress, const uint8_t functioncode, const uint8_t station, uint8_t *response, const bool only_status)
//Get dataset from the station using function code and subaddress.
//If the only_status bit is true, then it returns with the dataset and the response code
//If it were false, it returns only with the status byte information, regardless the returning value
{
  //Creating array to be send
  uint8_t comm[16];
  for(int i=0; i<16; i++) comm[i]=READ_STRING[i];//Set the remainings
  comm[0]=getCrate();//Crate number
  comm[1]=subaddress&0x0f;//Subaddress: only 0~15 + M1(0), M2(O) command
  comm[2]=(functioncode&0x1f)|0x20;//function code + reserved bit (1)
  comm[3]=(station&0x1f)|0x20;//Station number + reserved bit
  comm[4]=0;
  for(int i=0;i<4; i++) comm[4]^=((comm[i]) & 0x3f);//Check parity
  
  //Array is ready -- communiate with the controller
  uint8_t reply[20];
  int nofchar=communicate(comm,reply,(only_status ? 3 : 7),20,false);//Get number of characters read
  if(nofchar<0)
  {
    //Communication trouble
    std::cerr<<"ERROR(SerialCamac::getdata): communication trouble!\n";
    return Camac::CAMAC_ERR;
  }
  
#ifdef DEBUG_SERIAL_CAMAC
  std::cout<<"SerialCamac:getdata: number of characters received: "<<nofchar<<std::endl;
  for(int i=0; i<nofchar; i++) std::cout<<i+1<<" "<<(int) reply[i]<<std::endl;
#endif
  
  //Returning with the result
  if(only_status) 
  {
    if(response) *response=transforms2u(reply[1]&0xf);
    return transforms2u(reply[1]&0xf);
  }
  else
  {
    if(response) *response=transforms2u(reply[1]&0xf);
    return ((reply[2]&0x3f)<<18 | (reply[3]&0x3f)<<12 | (reply[4]&0x3f)<<6 | (reply[5]&0x3f));
  }
}
uint8_t SerialCamac::setdata(const uint8_t subaddress, const uint8_t functioncode, const uint8_t station, const uint32_t data)
//Set dataset of the station's using function code and subaddress. It returns with proper response. Otherwise: with error code 0x0
{
  //Creating array to be send
  uint8_t comm[16];
  for(int i=0; i<16; i++) comm[i]=WRITE_STRING[i];//Set the remainings
  comm[0]=getCrate();//Crate number
  comm[1]=subaddress&0x0f;//Subaddress: only 0~15+command(0,0)
  comm[2]=(functioncode&0x1f)|0x20;//function code+reserved bit (1)
  comm[3]=(station&0x1f)|0x20;//Station number+reserved bit (1)
  //Cutting the 24 bit (3 byte) dataset into 6 bit fragments
  comm[4]=(data>>18)&0x3f;//Upper
  comm[5]=(data>>12)&0x3f;//UMiddle
  comm[6]=(data>> 6)&0x3f;//LMiddle
  comm[7]=data&0x3f;//Lower
  comm[8]=0;//Sum byte
  for(int i=0;i<8; i++) comm[8]^=((comm[i]) & 0x3f);//Sum byte
  
  //Array is ready -- communicate with the controller
  uint8_t reply[20];
  int nofchar=communicate(comm,reply,3,20,true);
  if(nofchar<0)
  {
    //Communication trouble
    std::cerr<<"ERROR(SerialCamac::setdata): communication trouble!\n";
    return 0x0;
  }

#ifdef DEBUG_SERIAL_CAMAC
  std::cout<<"SerialCamac:setdata: number of characters received: "<<nofchar<<std::endl;
  for(int i=0; i<nofchar; i++) std::cout<<i+1<<" "<<(int) reply[i]<<std::endl;
#endif
  
  //Returning with the result
   return transforms2u(reply[1]&0xf);
}

int SerialCamac::communicate(uint8_t *sendbytes, uint8_t *replybytes, const int min_response, const int max_response, const bool writemode)
//Communicate on serial interface by sending 16 sendbytes, receving the feedback and put the reply into replybytes array.
//The maximum array's maximal index is max_response and filtering should be done to get at least min_response. It returns 
//with the number of response bytes or less than 0 if error occured
{
  //Checking the opened status of the communication line
  if(!(pS->isOpened()))
  {
    std::cerr<<"ERROR(SerialCamac::communicate): Communication line is not available!\n";
    return -1;
  }
  
  //Creating array to receive datasets
  const int MAX_RCV=100;//Size of the receiving buffer
  uint8_t receive[MAX_RCV];
  
#ifdef DEBUG_SERIAL_CAMAC
  std::cout<<"SerialCamac::communicate: sendbytes\n"; for (int i=0; i<16; i++) std::cout<<i+1<<" "<<(int)sendbytes[i]<<std::endl;
  std::cout<<"Mode: "<<(writemode? "W\n" : "R\n");
#endif
  
  //Sending the datasets
  if (pS->Send(sendbytes, 16) <16) return -2;//Serial communication error
    
  //Getting the reply
  int rcv_start_index=0;//Start index of the receiving buffer
  bool crate_found=false;//Crate number found: this indicates the start of the answer
  int headerstart=0;//Header start index
  int replystop=0, replystart=0;//They are marking the expected end and start indices of the reply
  int rel;//Length of the response
  bool solution_found=false;
  int trial_no=max_trials;
  do
  {
    int rnum=pS->Receive(receive+rcv_start_index, MAX_RCV-rcv_start_index);//Get the actual value
    if(rnum>0)
    {
#ifdef DEBUG_SERIAL_CAMAC
      for(int i=rcv_start_index; i<rcv_start_index+rnum; i++ ) std::cout<<i<<" "<<(int) receive[i]<<std::endl;
#endif
      if(!crate_found)
      {
	while((headerstart<rnum+rcv_start_index)&&(receive[headerstart++]!=getCrate()));//Get candidate for header
	crate_found=(receive[headerstart-1]==getCrate());
	if(crate_found) replystop=replystart=headerstart+ (writemode ? 9 : 5);//Set the expected reply start
      }
      if(crate_found && (replystop<rnum+rcv_start_index))
      {
	while ((replystop< rnum+rcv_start_index) && (receive[replystop++]<REND));//Looking for the end of the reply block
#ifdef DEBUG_SERIAL_CAMAC
	std::cout<<"Replystart: "<<replystart<<", replystop: "<<replystop<<", receive: "<<receive[replystop-1]<<std::endl;
#endif
	if(receive[replystop-1]>=REND)
	{
	  //A candidate has been found for the end
	  rel=replystop-replystart+1;//Determining the length
	  if(rel<min_response) crate_found=false;//Probably wrong block: try it again
	  else if(STE==(receive[replystart+1]&0x1))
	  {
	    std::cerr<<"ERROR: SerialCamac::communicate: Status error"<<std::endl;
	    return -2;//Status error happened
	  }
	  else
	  {
	    //Check parity
	    uint8_t sumr=REND;//starting value of parity: this will be the end byte of the response
            for(int i=replystart; i<replystart+min_response; i++) sumr^=receive[i];
	    if(sumr==0) solution_found=true;//OK!!!!!!!!!!
	    else crate_found=false;//Wrong parity... probably other
	  }
	}
	else if(replystop-replystart>max_response) crate_found=false;//Probably wrong block: try it again
      }
      rcv_start_index+=rnum;//Update receiving start index
    }
  }while(!solution_found && --trial_no>0);
  
  
  
  if(trial_no==0)
  {
    std::cerr<<"WARNING: SerialCamac::communicate: trial number limit exceeded."<<std::endl;
    return -1;
  }
  //Set the response bytes
  for(int i=0; i<min_response; i++) replybytes[i]=receive[replystart+i];
  
  return min_response;
}    

uint8_t SerialCamac::transforms2u(const uint8_t inp)
//Transforms serial error codes into the uniform error codes
{
    uint8_t answer=0x0;
    if (inp&S_X) answer+=Camac::STATUS_X;//X answer
    if (inp&S_Q) answer+=Camac::STATUS_Q;//Q answer
    if (inp&S_ERR) answer+=this->STATUS_S_ERR;//Parity error answer
    if (inp&S_DERR) answer+=this->STATUS_S_DERR;//Delayed error answer
#ifdef DEBUG_SERIAL_CAMAC
    std::cout<<"SerialCamac::transform2u original: "<<(int) inp<<", transformed: "<<(int)answer<<std::endl;
#endif
    return answer;
}
