#include "Serial.h"
#include "Camac.h"

#ifndef SERIAL_CAMAC_H
#define SERIAL_CAMAC_H

enum TEST_mode {LOOP_COLLAPSE, BY_PASS, OFF_LINE, HW_OFFLINE};//Test modes flags: loop collapsed (both end will be out of operation), by pass (only one end goes out from operation), software offline, hardware offline 

class SerialCamac : public Camac
{
  public:
    //methods
    SerialCamac(std::string portName ,uint8_t cNo, const std::string nm, double timeout=0.5, int nBaud=2400, int nDataBits=7, char nParity='O', int nStopbits=2);//Constructor
    ~SerialCamac(){if(pS) {send_initfinal_string(false); clear_status_register();}  delete pS;};//Destructor -- delete the serial channel object
    std::string getPortName(){return pS->getPortName();};//Returns with the actual port name
    void getCommParameters(int *baud, int *databits, char *parity, int *stopbits){pS->getCommParameters(baud,databits,parity,stopbits);};//This function returns with the currently set communication parameters as baud rate, data bits, parity and number of stopbits
    double getTimeout(){return pS->getTimeout()*max_trials;};//Returns with the current value of timeout
    //Inherited virtual methods
    bool Reset();//Resets the Harwell Camac unit
    bool SetResetInhibit(bool reset){return setData(STREG_A, reset ? SCLR_F : SSET_F, STATION, STR_INHIBIT_FLAG, STATUS_X|STATUS_Q);};//Set(f)/clear(t) dataway inhibit. Returns with communication success or error
    bool GetInhibit(bool test);//Get the dataway inhibit value (t: set, f: unset). If test is true, it returns the setting of the controller (it might be other sources of dataway inhibit: test is tests only the controller, but general the dataway directly).
    //control
    bool ResetComm(const std::string portName="camac_serial", const double timeout=0.5, const int nBaud=2400, const int nDataBits=7, const char nParity='O', const int nStopbits=2);//Resets the communication channel and the Harwell Camac unit
    bool SetResetDeviceTest(TEST_mode tm, bool setreset=false);//Sets the device into the required test modes: only LOOP_COLLAPSE, BY_PASS, OFF_LINE are allowed
    bool CheckDeviceTest(TEST_mode tm);//Returns with the actual situation of test device as requested as LOOP_COLLAPSE, OFF_LINE, HW_OFFLINE
    uint8_t CheckPDataFlags();//Returns with the flags belonging to the previous data
    bool GetPData(uint32_t& previous_data, bool q_status=true);//Returns with the previously read data and communication errors. If q_status is true, then Q answer will be checked
    //static members
    static const uint8_t STATUS_S_ERR=0x4;//Serial error: related to wrong command format or bad checksum
    static const uint8_t STATUS_S_DERR=0x8;//Serial delayed error message
  private:
    //static members
    static const uint8_t RESET_STRING[16];//Contains the Camac SCC reset string
    static const uint8_t SPACE_STRING[16];//Contains space bytes
    static uint8_t READ_STRING[16];//Contains the initialisation of the read string
    static uint8_t WRITE_STRING[16];//Contains the initialisation of the write string
    static const uint8_t SPC=0x3f;//Spacebyte
    static const uint8_t REND=0x40;//End bit of the response
    static const uint8_t ENB=0x60;//End byte
    static const uint8_t STE=0x1;//Status error
    //static members : status byte information 
    static const uint8_t S_ERR=0x1;//Serial error: related to wrong command format or bad checksum
    static const uint8_t S_X=0x2;//Serial X answer
    static const uint8_t S_Q=0x4;//Serial Q answer
    static const uint8_t S_DERR=0x8;//Serial delayed error message
    //statis members - CAMAC commands
    static const uint8_t STATION=30;//Station code
    static const uint8_t STREG_A=0;//Status register sub code
    static const uint8_t STORED_A=1;//Stored data register sub code
    static const uint8_t STORED_F=0;//Read stored data register function code
    static const uint8_t READ_F=1;//Read function code
    static const uint8_t WRITE_F=17;//Write function code
    static const uint8_t SSET_F=19;//Selective set function code
    static const uint8_t SCLR_F=23;//Selective clear function code
    //static members - Status register
    static const uint16_t STR_CLEAR=0x0;//Clear the status register
    static const uint16_t STR_GENERATE_C=0x1;//Generate Clear signal WO
    static const uint16_t STR_GENERATE_Z=0x2;//Generate initialiZe signal WO
    static const uint16_t STR_INHIBIT_FLAG=0x4;//Inhibit flag R/W
    static const uint16_t STR_DERR=0x8;//Previous data delayed error flag
    static const uint16_t STR_DSX=0x10;//Previous data X answer flag
    static const uint16_t STR_DSQ=0x20;//Previous data Q answer flag
    static const uint16_t STR_INHIBIT_DW=0x40;//Dataway inhibit RO
    static const uint16_t STR_LOOP_COLL=0x400;//Loop collapse flag: when both end of wire is intended to switch off R/W
    static const uint16_t STR_BYPASS=0x800;//Bypass flag: when one device is get out of the chain by a shortcut (only one side is get switched off) WO
    static const uint16_t STR_OFFLINE=0x1000;//Offline flag: set it by software R/W
    static const uint16_t STR_HW_OFF=0x2000;//Offline flag: set it on hardware RO
    //members
    Serial *pS;//Serial channel
    int max_trials;//Number of maximal allowed trials before real TimeOut msg
    //inherited virtual methods
    uint32_t getdata(const uint8_t subaddress, const uint8_t functioncode, const uint8_t station, uint8_t *response, const bool only_status=false);
    //Get dataset from the station using function code and subaddress.
    //If the only_status bit is false, then it returns with the dataset and the response code
    //If it were true, it returns only with the status byte information, regardless the returning value
    uint8_t setdata(const uint8_t subaddress, const uint8_t functioncode, const uint8_t station, const uint32_t data);
    //Set dataset of the station's using function code and subaddress. It returns with proper response. Otherwise: with error codes 00
    //local methods
    bool send_initfinal_string(bool init=true) {return (setdata(STREG_A,WRITE_F,STATION, init ? STR_GENERATE_Z : STR_GENERATE_C )==(STATUS_X | STATUS_Q));};//Send init/clear string
    bool send_space_string() {return (pS->Send(SPACE_STRING,16)==16);};//Send space string
    bool clear_status_register(){ return (setdata(STREG_A,WRITE_F,STATION,STR_CLEAR)==(STATUS_X | STATUS_Q));};//Clear the status register
    int communicate(uint8_t *sendbytes, uint8_t *replybytes, const int min_response, const int max_response, const bool writemode);
    //Communicate on serial interface by sending 16 sendbytes, receving the feedback and put the reply into replybytes array.
    //The maximum array's maximal index is max_response and filtering should be done to get at least min_response. It returns 
    //with the number of response bytes or less than 0 if error occured 
    uint8_t transforms2u(const uint8_t inp);//Transforms serial error codes into the uniform error codes
};
#endif
