#include "TrumpPCI.h"

TrumpPCI::TrumpPCI(int ser_number, double timeout_ms)
//Constructor:  serial number of the card (-1: first valid card), timeout of reading in milliseconds
{
	//Allocate memory for IO buffers, initialize timeout
	timeOut=(long)(1000.0/FIFO_WAIT_SHORT*timeout_ms);//sets the timeout in ticks units
	recvBuffer = new uint32_t [RCV_MAX_SIZE/4];//Allocating memory for low level data receiving transmission
	sendBuffer = new uint32_t [COMMAND_MAX_SIZE/4];//Allocating memory for low level data sending transmission
	//Looking for BAR of Trump PCI card
	std::vector<uint32_t> barr;//Stores the base address registers
	looking_for_cards(&barr);//Looking
	bool found=false;
	if (barr.size()==0) std::cerr<<"There is not any TrumpPCI card in the computer!"<<std::endl;
	else
	{    
	    for(int i=0; i<(barr.size()&&(!found)); i++)
	    //Check the valid list of items
	    {
		portNumber=barr[i];;//Set the base address register
		//Open the card for communication and set registers of the card
	        commOK=card_openclose(true);//Setting the communication OK flag
	        if(commOK)
	        {
	    	    card_set_operation();//Configure card
	    	    
	    	    if (ser_number==-1) found=true;//First valid card found, stop here...
	    	    else
	    	    {
	    		//Check serial number
	    		bool features;
	    		std::string mode, radix, firmware;
	    		int snum;
	    		bool ans=About(&features, &mode, &radix, &snum, &firmware);
	    		if(ans&&(snum==ser_number)) 
	    		{
	    		    //Card has been found
	    		    found=true;
	    		    std::cout<<"Card with "<<snum<<" serial number has been found"<<std::endl;
	    		}
	    		else
	    		{
	    		    //Not found, reconfigure card and report on serial number (if possible)
	    		    if(ans) std::cout<<"Non-matching serial number: "<<snum<<std::endl;
	    		    //Close this connection
	    		    card_reset_operation();//Reset status registers of the card
			    commOK= (!card_openclose(false));//Closing the connection
			    if(commOK) std::cerr<<"Error on closing the connection with the card no."<<i+1<<std::endl;
	    		}
	    	    }
	    	}
	    }
	}
	
	if(!found)
	{
	    std::cerr<<"There is not found any unallocated card with serial number "<<ser_number<<std::endl;
	    commOK=false;//Set communication unsuccessful
	}
	else
	{
	    //We need to initialize some important variables
	    commOK=this->getDataMask(&d_mask,&r_mask);//Get the data and region-of-interest masks
	}
}

TrumpPCI::~TrumpPCI()
//Destructor
{
	//Deleting the allocated buffers for low level data transmission
	delete [] recvBuffer;
	delete [] sendBuffer;
	if(commOK) 
	{
	  card_reset_operation();//Reset status registers of the card
	  commOK= (!card_openclose(false));//Closing the connection
	  if(commOK) std::cerr<<"Error on closing the connection with the card"<<std::endl;
	}
}


bool TrumpPCI::About(bool* features_check, std::string *mode, std::string *radix, int *serial_number, std::string *firmware)
//Provides hard wired information about the card: its features (does it identical to manual), mode (pule height analysis), radix type, serial number and firmware model&version. It returns with true, if communication was successful.
{
	std::stringstream res;//Result stringstream
	E_CODE ans=SUCCESS;//Returning error code
	if (features_check)
	{
	    ans=processing_communication(std::string("SHOW_FEATURES"), &res, NULL);//Get features
	    if(ans!=SUCCESS) return false;
	    //Processing features
	    uint32_t first, second, third, fourth;//Segments
	    res>>first>>second>>third>>fourth;//Get the values
	    if((first==2149581857ul)&& //first word: 1 (Software-selectable conversion gain) + 32 (PHA-mode functions available) + 1024 (Software-selectable analog offset) + 2097152 (Software retrievable serial number (SHOW_SNUM)) + 2147483648 (Software-selectable ADC gate mode (SET_GATE...))= 2149581857;
	     (second==1075318797ul)&& //second word: 1 (Software downloadable firmware) + 4 (Software-selectable LLD) + 8 (Software-selectable ULD) + 4096 (Statistical preset (SET_UNCERT_PRES)) + 524288 (SHOW_STATUS command supported with $M response) + 1048576 (Overflow preset (ENA/DIS/SHO_OVER_PRES)) + 1073741824 (Explorable gains (LIST/VERI_GAIN_FINE, ...COAR,...CONV))=1075318797;
	     (third==16ul) &&(fourth==0ul)) *features_check=true;// third word: 16 (MDA Preset)
	    else *features_check=false;
	    res.str(""); res.clear();//Flush buffer
	}
	if(mode)
	{
	    ans=processing_communication(std::string("SHOW_MODE"), &res, NULL);//Get mode (pulse height analysis)
	    if(ans!=SUCCESS) return false;
	    *mode=res.str();//Set the mode variable
	    res.str(""); res.clear();//Flush buffer
	}
	if(radix)
	{
	    ans=processing_communication(std::string("SHOW_RADIX"), &res, NULL);//Get radix (binary)
	    if(ans!=SUCCESS) return false;
	    *radix=res.str();//Set the radix variable
	    res.str(""); res.clear();//Flush buffer
	}
	if(serial_number)
	{
	    ans=processing_communication(std::string("SHOW_SNUM"), &res, NULL);//Get the serial number
	    if(ans!=SUCCESS) return false;
	    res>>*serial_number;//Set the SN variable
	    res.str(""); res.clear();//Flush buffer
	}
	if(firmware)
	{    
	    ans=processing_communication(std::string("SHOW_VERSION"), &res, NULL);//Get firmware version number
	    if(ans!=SUCCESS) return false;
	    *firmware=res.str();//Set the firmware: mmmm-vvv (model-version)
	}
	return true;//Successful end	
}

bool TrumpPCI::SetDevice()
//Sets the device --- useless routine
{
	return (processing_communication(std::string("SET_DEVICE 1"), NULL, NULL)==SUCCESS);//Communicate
}

//Dataset related methods


bool TrumpPCI::setDataRange(const int start, const int length)
//Sets the channel range exact value. Returns with its success
{
	std::stringstream ss;
	ss<<"SET_WINDOW "<<start<<","<<length;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::getDataRange(int *start, int *length)
//Asks the actual data range: number of channels and start channel. Returns with the communication success/fail
{
	std::stringstream res;//Result
	if((!start)||(!length)) return false;//Checking non-null pointers
	E_CODE ans=processing_communication(std::string("SHOW_WINDOW"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		res>>*start>>*length;//Set the datasets
		return true;
	}
	else return false;
}

bool TrumpPCI::setData(uint32_t count)
//Sets the count for data channels, which are in the range. It returns with communication success or fail
{
	std::stringstream ss;
	if (count==0) ss<<"CLEAR_DATA";
	else ss<<"SET_DATA "<<(unsigned long) count;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

uint16_t TrumpPCI::lastDataChannel()
//Returns with the last digitized channel number by ADC. Otherwise with error
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(std::string("SHOW_ADC_CONVERSION"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		uint16_t data;
		res>>data;//Set the datasets
		return data;
	}
	else return 0xfffe;//Error indicator
}



bool TrumpPCI::getDataSet(int start, int length, uint32_t *counts, bool *roi_flag)
//This method ask the recorded counts and region of interest flags of channels from start and length long. Returns with communication errors 
{
	std::stringstream ss;
	ss<<"SHOW_SPECTRUM "<<start<<","<<length;
	if(!counts) return false;//It is necessary to be given
	E_CODE ans=processing_communication(std::string(ss.str()), NULL, counts);//Communicate
	if(ans==SUCCESS)
	{
		//Masking the dataset
		if (roi_flag !=NULL) for(int i=0; i<length; i++) roi_flag[i]=((counts[i]&r_mask)==r_mask);
		for(int i=0; i<length; i++) counts[i]&=d_mask;
		return true;
	}
	else return false;
}


//Gate input configuration related methods

bool TrumpPCI::setGate(GATE_MODE gm)
//Sets the gate mode. Returns with the result of communication
{
	std::string command;
	switch(gm)
	{
		case         GATE_OFF    : command="SET_GATE_OFF"; break;
		case GATE_COINCIDENCE    : command="SET_GATE_COINCIDENT"; break;
		case GATE_ANTICOINCIDENCE: command="SET_GATE_ANTICOINCIDENT"; break;
		default: return false;//Non-valid setting
	}
	return (processing_communication(command, NULL, NULL)==SUCCESS);//Communicate
}

GATE_MODE TrumpPCI::getGate()
//Returns with the status of the gate mode or with error message
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(std::string("SHOW_GATE"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		if(res.str()=="OFF") return GATE_OFF;
		else if(res.str()=="COI")	return 	GATE_COINCIDENCE;
		else if(res.str()=="ANT")	return 	GATE_ANTICOINCIDENCE;
		else return GATE_ERROR;
	}
	else return GATE_ERROR;
}


//Discriminator related methods

bool TrumpPCI::setDiscriminator(int channel, bool lowhi)
//Sets the low(true) or high(false) discriminator level to channel value.  Returns with the result of communication
{
	std::stringstream ss;
	ss<<"SET_"<<(lowhi ? "L" : "U")<<"LD "<<channel;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

uint16_t TrumpPCI::getDiscriminator(bool lowhi)
//Returns with the low(true) or high(false) discriminator level or with error message
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(lowhi ? std::string("SHOW_LLD") : std::string("SHOW_ULD"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		uint16_t value;
		res>>value;//Get the data
		return value;
	}
	else return 0xfffe;
}

//Deadtime related methods

bool TrumpPCI::setDTimeDelay(double delay)
//Set the required time to restore initial level from the end of digitalization in microsec (0 means infinite delay).  Returns with the result of communication
{
	std::stringstream ss;
	ss<<"SET_RESET_DELAY "<<(int)(delay/0.08);
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

double TrumpPCI::getDTimeDelay()
//Returns with the required time to restore initial level from the end of digitalization in microsec units (0 means infinite delay). Communication error is also reported.
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(std::string("SHOW_RESET_DELAY"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		long ticks;
		res>>ticks;//Get the data
		return ((double) (ticks*0.08));//Converting to microseconds
	}
	else return -1.0;
}

//Conversion gain related methods

bool TrumpPCI::setGainConv(int channel)
//Sets the gain conversion. Returns with communication error or non-valid input errors
{
	if ((channel!=0)&&(channel!=256)&&(channel!=512)&&(channel!=1024)&&(channel!=2048)) return false;//Invalid number error
	std::stringstream ss;
	ss<<"SET_GAIN_CONVERSION "<<channel;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::getGainConv(int *current, int *maximal)
//Provides information about the current and maximal (if it has been set) gain conversion. Returns with the result of communication
{
	std::stringstream res;//Result
	if(!current) return false;//Error due to NULL pointer
	E_CODE ans=processing_communication((maximal==0) ? std::string("SHOW_GAIN_CONVERSION") : std::string("SHOW_CONFIGURATION"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		if(maximal==0)
		{
			res>>*current;
		}
		else
		{
			int segment_no;//Segment number (always 1)
			res>>*maximal>>segment_no>>*current;//Set the datasets
		}
		return true;
	}
	else return false;
}

//Input signal offset setup related methods

bool TrumpPCI::setOffset(uint16_t offset)
//Sets the ADC offset in hardware units (2048: no shift). Returns with communication error
{
	std::stringstream ss;
	ss<<"SET_OFFSET_FINE "<<(unsigned long)offset;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::getOffset(int *min_ch, int *max_ch, int *min_hw, int *max_hw)
//Get the limits of offset setup: min&max in channel and in the corresponding hardware units. Returns with communication error
{
	std::stringstream res;//Result
	if((!min_ch)||(!max_ch)||(!min_hw)||(!max_hw)) return false;//Error due to NULL pointers
	E_CODE ans=processing_communication(std::string("LIST_OFFSET_FINE"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		res>>*min_ch>>*max_ch>>*min_hw>>*max_hw;//Set the datasets
		return true;
	}
	else return false;
}

uint16_t TrumpPCI::getOffset()
//Returns with the offset in hardware units or with error
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(std::string("SHOW_OFFSET_FINE"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		uint16_t setting;
		res>>setting;//Get the datasets
		return setting;//Returns with the new setting
	}
	else return 0xfffe;//Error code
}

uint16_t TrumpPCI::incrOffset()
//Increases the offset by 1 hardware unit, returns with the new value in hardware units.
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(std::string("STEP_OFFSET_FINE"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		uint16_t setting;
		res>>setting;//Get the datasets
		return setting;//Returns with the new setting
	}
	else return 0xfffe;//Error code
}

//Start/Stop and diagnostics related methods

bool TrumpPCI::getStatus()
//Returns with on measurement(t) or idle (f) status
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(std::string("SHOW_ACTIVE"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		int status;
		res>>status;//Get the datasets
		return (status==1);
	}
	else return false;
}

bool TrumpPCI::getStatus(double *live_time, double *real_time, uint16_t *active_device, std::string &hw_status)
//Gets status information about live and true time in seconds, active device and on hardware status. Returns with communication error
{
	std::stringstream res;//Result
	if((!live_time)||(!real_time)||(!active_device)) return false;//Error due to NULL pointers
	E_CODE ans=processing_communication(std::string("SHOW_STATUS"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		res>>*live_time>>*real_time>>*active_device>>hw_status;//Set the datasets
		*live_time/=50.0;//Convert time to seconds
		*real_time/=50.0;
		return true;
	}
	else return false;
}

bool TrumpPCI::StopStart(bool stop)
//Starts(f) / stops (t) acquisition. Returns with communication error (true in the case of SUCCESS, ST_ALREADY_ST)
{
	E_CODE ans=processing_communication(stop ? std::string("STOP") : std::string("START") , NULL, NULL);//Communicate
	return ((ans==SUCCESS)||(ans==ST_ALREADY_ST));
}


//Extended clear methods

bool TrumpPCI::clearTimeData()
//Clears real, live time counters and datasets in window of interest. Returns with communication errors
{
	return (processing_communication(std::string("CLEAR"), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::clearPresets()
//Clears real, live time, roi integral, roi maximum and overflow presets. Returns with communication errors
{
	return (processing_communication(std::string("CLEAR_PRESETS"), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::clearAll()
//Clears the (live+preset) counters, the dataset counters in window of interest. Additionally, real, live time, roi integral, roi maximum and overflow presets will be set to zero and disabled. Moreover, roi regions will be cleared. Returns with communication errors
{
	return (processing_communication(std::string("CLEAR_ALL"), NULL, NULL)==SUCCESS);//Communicate
}


//Real and live time counts/preset related methods

bool TrumpPCI::setTime(double time, bool truelive, bool countpreset)
//Sets the real(truelive=true)/live(truelive=false) counter(countpreset=true)/preset(countpreset=false) in seconds. Returns with communication errors - 0 disable
{
	std::stringstream ss;
	ss<<"SET_"<<(truelive ? "TRUE" : "LIVE")<<(countpreset ? " " : "_PRESET ")<<(unsigned long)(time*50.0);
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

double TrumpPCI::getTime(bool remainingelapsed, bool truelive, bool countpreset)
//Returns with the real(truelive=true)/live(truelive=false) counter(countpreset=true)/preset(countpreset=false) value in seconds. If counter is asked, there is possible to ask for the elapsed (remainingelapsed=false) or remaining time (remainingelapsed=true). If preset is requested, remainingelapsed will be omitted.
{
	std::stringstream res;//Result
    std::string command("SHOW_");
	command+=( truelive ? "TRUE" : "LIVE");
	if (!countpreset) command+="_PRESET";
	else if (remainingelapsed) command+="_REMAINING";
	E_CODE ans=processing_communication(command, &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		long ticks;
		res>>ticks;//Get the data
		return ((double) (ticks/50.0));//Converting to seconds
	}
	else return -1.0;
}

bool TrumpPCI::clearTime()
//Clears real and live time counter. Returns with communication error
{
	return (processing_communication(std::string("CLEAR_COUNTER"), NULL, NULL)==SUCCESS);//Communicate
}


//Overflow preset related methods

bool TrumpPCI::setOFPreset(bool enabledisable)
//Enable (t) (at channel with 2147483647 counts stops)/ disable(f) (continues to count) overflow preset. Returns with communication error
{
	std::stringstream res;//Result
    std::string command;
	if(enabledisable) command="ENABLE";
	else command="DISABLE";
	command+="_OVERFLOW_PRESET";
	return (processing_communication(command, NULL, NULL)==SUCCESS);//Communicate
}
		
bool TrumpPCI::getOFPreset()
//Returns with the Enable (t)/ disable(f) status of overflow preset.
{
	std::stringstream res;//Result
	E_CODE ans=processing_communication(std::string("SHOW_OVERFLOW_PRESET"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		int result;
		res>>result;//Get the data
		return (result==1);
	}
	else return false;//Error
}


//Region of interest (ROI) setup related methods

bool TrumpPCI::setROI(int start, int number_ch)
//Set the roi for channels 1..number_ch from start. Returns with communication error
{
	std::stringstream ss;
	ss<<"SET_ROI "<<start<<","<<number_ch;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::getROI(int *start, int *number_ch, bool first)
//Gets start channel and number of channel of continous group of roi. It shows for the first (true) or the followings(false). If there is no more, it returns with 0,0. Returns with communication error
{
	std::stringstream res;//Result
	if((!start)||(!number_ch)) return false;//Error due to NULL pointers
	E_CODE ans=processing_communication(first ? std::string("SHOW_ROI") : std::string("SHOW_NEXT"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		res>>*start>>*number_ch;//Set the datasets
		return true;
	}
	else return false;
}

bool TrumpPCI::clearROI()
//Clears all ROI in the window of interest. Returns with communication error
{
	return (processing_communication(std::string("CLEAR_ROI"), NULL, NULL)==SUCCESS);//Communicate
}


//ROI-Integral and maximum preset common method

bool TrumpPCI::setIntegralMax(uint32_t preset, bool max)
//Sets integral(f) or maximal value(t) preset counts for the whole roi, 0: disabled. 4294967295: integral maximal, 2147483647: maximum maximal. Returns with communication error
{
	std::stringstream ss;
	ss<<"SET_"<<(max ? "PEAK" : "INTEGRAL")<<"_PRESET "<<(unsigned long)preset;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

//ROI-Integral preset related method

uint32_t TrumpPCI::getIntegral(bool preset, int start, int num_ch)
//Returns with integral preset (true)/actual integral value (false) for start, num_ch(non-zero) or for the whole roi (num_ch=0)
{
	std::stringstream res;//Result
    std::string command("SHOW_INTEGRAL");
	if(preset) command+="_PRESET";
	else if(num_ch!=0)
	{
		res<<command<<" "<<start<<","<<num_ch;
		command=res.str();
		res.str(""); res.clear();//Clear stringstream
	} 
	else command+=" ";
	E_CODE ans=processing_communication(command, &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		uint32_t result;
		res>>result;//Get the data
		return result;
	}
	else return 0xfffffffe;//Error code
}

//ROI-maximum and maximum preset related method

uint32_t TrumpPCI::getMax(bool preset, bool channel)
//Returns the maximal value among ROI: preset (preset=t)/ actual value (f,f)/ the belonging channel number (f,t), answer: 1..2147483647: counts, 0: 0 counts/channel (f,*) or no roi (f,*) or disabled(t,*)
{
	std::stringstream res;//Result
    std::string command("SHOW_PEAK");
	if(preset) command+="_PRESET";
	else if(channel) command+="_CHANNEL";
	E_CODE ans=processing_communication(command, &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		uint32_t result;
		res>>result;//Get the data
		return result;
	}
	else return 0xfffffffe;//Error code
}

//Uncertainty-MDA common methods

bool TrumpPCI::setUncMDA(double preset, bool uncertainty)
//Sets the uncertainty(t) (in % units)/MDA(f) (in gamma/s units) preset (0.0 means disabled). Returns with communication error
{
	std::stringstream ss;
	ss<<"SET_"<<(uncertainty ? "UNCERTAINTY" : "MDA")<<"_PRESET "<<preset;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::setUncMDARange(int start, int num_chn, bool uncertainty)
//Sets the uncertainty preset(t)/MDA(f) range from start with num_chn. Returns with communication error
{
	std::stringstream ss;
	ss<<"SET_ROI_"<<(uncertainty ? "UNCERTAINTY " : "MDA ")<<start<<","<<num_chn;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

double TrumpPCI::getUncMDA(bool preset, bool uncertainty)
//Returns with the uncertainty (*,t) /MDA (*,f) actual (f,*) or preset (t,*) value
{
	std::stringstream res;//Result
    std::string command("SHOW_");
	if(uncertainty) command+="UNCERTAINTY";
	else command+="MDA";
	if(preset) command+="_PRESET";
	E_CODE ans=processing_communication(command, &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		double result;
		res>>result;//Get the data
		return result;
	}
	else return -1.0;
}

bool TrumpPCI::getUncMDARange(int *start, int *num_chn, bool uncertainty)
//Gets the Uncertainty preset/MDA range at roi from start with num_chn. Returns with communication error
{
	std::stringstream res;//Result
	if((!start)||(!num_chn)) return false;//Error due to NULL pointers
	E_CODE ans=processing_communication(uncertainty ? std::string("SHOW_ROI_UNCERTAINTY") : std::string("SHOW_ROI_MDA"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		res>>*start>>*num_chn;//Set the datasets
		return true;
	}
	else return false;
}

		
//MDA (minimum detectable activity) preset related methods


bool TrumpPCI::setMDACoeffs(double a, double b, double c)
//Sets the MDA coefficients used to calculate the MDA=(a+sqrt(b+c*roi_counts))/live_time in gamma/s units. Returns with communication error
{
	std::stringstream ss;
	ss<<"SET_MDA_COEF "<<a<<","<<b<<","<<c;
	return (processing_communication(std::string(ss.str()), NULL, NULL)==SUCCESS);//Communicate
}

bool TrumpPCI::getMDACoeffs(double *a, double *b, double *c)
//Gets the MDA coefficients used to calculate the MDA=(a+sqrt(b+c*roi_counts))/live_time in gamma/s units. Returns with communication error
{
	std::stringstream res;//Result
	if((!a)||(!b)||(!c)) return false;//Error due to NULL pointers
	E_CODE ans=processing_communication(std::string("SHOW_MDA_COEFFICIENT"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		res>>*a>>*b>>*c;//Set the datasets
		return true;
	}
	else return false;
}


//---------------------private methods -----------------------------

bool TrumpPCI::getDataMask(uint32_t *data_mask, uint32_t *roi_mask)
//Get the masks, which should be "&" to get data and roi bits. It returns with communication success or fail
{
	std::stringstream res;//Result
	if((!data_mask)||(!roi_mask)) return false;//Checking that both pointers are provided
	E_CODE ans=processing_communication(std::string("SHOW_CONFIGURATION_MASK"), &res, NULL);//Communicate
	if(ans==SUCCESS)
	{
		res>>*data_mask>>*roi_mask;//Set the datasets
		return true;
	}
	else return false;
}

E_CODE TrumpPCI::processing_communication(std::string command, std::stringstream *result, uint32_t *dataset)
//This function process the communication: gives a checksum for the command, gives this expression to communicate with hardware and process the answer into more relevant format: in the case of simple command, the result processed into result. In the case of raw channel, the result processed into the dataset array. It returns with an appropriate error code
{
	if(!commOK) return TIMEOUT;//If not successful open of the channel
	char lf=0xa;//New line character
	command+=lf;//Adding <LF>
#ifdef DEBUG_TRUMP
	std::cout<<"Command to be send: "<<command<<std::endl;
#endif
	//Communicate with the hardware
	int length=command.length();//Length of the answer in double word units
	uint32_t *answer_array;//Pointer pointing to the answer array
	bool correct=false;//It is true, if enough number of dataset has been read in
	for(int j=0; (j<30)&&(!correct); j++)
	{	
	  answer_array=card_communicate(((j==0) ? (uint8_t *)command.c_str() : 0),length);//Communicate with card, receive the answer
#ifdef DEBUG_TRUMP
	std::cout<<"Trial number"<<j+1<<", length is "<<length<<std::endl;
#endif
	  //Check, whether we have got the full answer?
	  if(dataset) for(int i=256; (i<=8196)&&(!correct); i*=2) correct=(length==2*i+COMMAND_MAX_SIZE);
	  else correct=(length>=COMMAND_MAX_SIZE);

	}
#ifdef DEBUG_TRUMP
	if(!correct) std::cout<<"ERROR: Not successful read to process "<<length<<" number of datasets"<<std::endl;
#endif	
	if((answer_array!=0)&&correct) return process_answer(answer_array, result, dataset);////Communicating with the device and processing the answer
	else return TIMEOUT;
}

E_CODE TrumpPCI::process_answer(const uint32_t *answer, std::stringstream *result, uint32_t *dataset)
//This method process the "answer". It checks the checsum (both response and command, as well), cut pure answer and splits them into relevant units, putting them into the "result". If raw channel datasets are asked, they will be stored in dataset array instead of "result". It returns with an error code in each cases.
{
  //Creating char array for the command part
  std::string str("");//String to store the command part
  bool end_command=true;//End of command indicator in false position
  bool next=false;//Start of % answer
  for(int i=0; (i<RCV_MAX_SIZE/4)&&end_command; i++)
  {
    for(int j=0; j<2; j++)
    {
      char value=(char)(((answer[i])>>(8*j))&0xff);
      if(value==0x25) next=true;
      else if((value==0xa)&&next) end_command=false;
#ifdef DEBUG_TRUMP
      if((i==0)&&(j==0))std::cout<<"Answer: "<<value;
      else if (value==0xa) std::cout<<"<LF>";
      else std::cout<<value;
      if(!end_command) std::cout<<std::endl;
#endif
      str+=value;//Add it to the string
    }
  }
  //Dealing with error codes: checksum and error codes
  std::size_t percent=str.find('%');//Looking for the first occurence of % character, which signifies the start of the end of block
  if(percent==std::string::npos) return EXEC_OTHER;//Missing % response
  std::stringstream conv(str.substr(percent+7,3));//Put checksum characters
  int chk;//Provided checksum
  conv>>chk;
  if(((uint8_t)chk)!=checksum(str.substr(percent,7))) return COMM_CHKS_RCV;//Returns with checksum error
  E_CODE retval=raw2ecode(str.substr(percent+1,6));//Get the error code
  if((percent==0)||(retval!=SUCCESS)) return retval;//Return, if % answer expected or unsuccessful operation done
  else
  {
    //It seems, that the answer contains a message, before the "<LF>%" part
    if(str[0]=='$')
    {
      if((str[1]!='F')&&(str[1]!='I'))
      {
	//$ response record: checking the checksum, if it has
	conv.str(""); conv.clear();//Clear buffer
	conv<<str.substr(percent-4,3);
	conv>>chk;
	if(((uint8_t)chk)!=checksum(str.substr(0,percent-4))) return COMM_CHKS_RCV;//Returns with checksum error
      }
      //Checksum OK, transforming the characters to the stringstream
      switch(str[1])
      {
	case 'A':
	case 'C':
	case 'E': //Not checked
	case 'G': *result<<str.substr(2,percent-6); break;
	case 'D': *result<<str.substr(2,5)<<" "<<str.substr(7,5); break;
	case 'F': *result<<str.substr(2,percent-3); break;
	case 'I': *result<<((str[2]=='T') ? "1" : "0"); break;
	case 'J': for(int i=0;i<18;i++) *result<<str.substr(2+5*i,5)<<" "; break;
	case 'M': *result<<str.substr(2,10)<<" "<<str.substr(12,10)<<" "<<str.substr(22,5)<<" "<<str.substr(27,5); break;
	case 'N': for(int i=0;i<3;i++) *result<<str.substr(2+3*i,3)<<" "; break;//Not checked
	default: retval=EXEC_OTHER;//Non-valid command string - not checked
      }
    }
    else if (std::string::npos!=str.find("SPECTRUM"))
    {
      //Getting the raw channel values
      std::stringstream localss(str.substr(str.find(" ")+1,percent-str.find(" ")-2));//Set number of bytes to be read in stringstream
      int bytes;
      localss>>bytes;
      for(int i=COMMAND_DWORD; i<COMMAND_DWORD+bytes/4; i+=2) dataset[(i-COMMAND_DWORD)/2]=answer[i]+(answer[i+1]<<16);//Putting the counts into right numbers
    }
    else if (str.find(" ")>percent) retval=EXEC_OTHER;//Non-valid command string
    else *result<<str.substr(str.find(" ")+1,percent-str.find(" ")-2);//Writing everyting directly to stringsream between KEYWORD and <LF>%
  }

  return retval;//Returning with the result
}


E_CODE TrumpPCI::raw2ecode(const std::string &answer)
//Converts the raw answer format into error codes. It is supposed, that checksum already checked
{
	//Converting error codes into maior and minor codes
	std::stringstream ss(answer.substr(0,3)+" "+answer.substr(3,3));
	int ma,mi;//Maior and minor error codes
	ss>>ma>>mi;//Get the values
	if (ma==0)//START/STOP Warnings/ Success
	{
		switch(mi)
		{
			case 0: return SUCCESS; break;//No errors detected
			case 5: return ST_ALREADY_ST; break;//MCB already started or stopped
			case 6: return ST_PRESET_EXC; break;//Preset already exceeded
			default: return NOT_KNOWN;
		}
	}
	else if (ma==129)  //Command Syntax Errors
	{
		switch(mi)
		{
			case 1: return CMD_INV_VERB; break;//Invalid verb in command
			case 2: return CMD_INV_NOUN; break;//Invalid noun in command
			case 3: return CMD_INV_VERBNOUN; break;//Invalid verb and noun in command
			case 4: return CMD_INV_MOD; break;//Invalid modifier in command
			case 5: return CMD_INV_VERBMOD; break;//Invalid verb and modifier in command
			case 6: return CMD_INV_NOUNMOD; break;//Invalid noun and modifier in command
			case 7: return CMD_INV_VERBNOUNMOD; break;//Invalid verb, noun, and modifier in command
			case 132: return CMD_INV; break;//Invalid command (verb, noun, and modifier valid, but not together)
			default: return NOT_KNOWN;
		}
	}
	else if (ma==130)//Communication Errors
	{
		switch(mi)
		{
			case 128: return COMM_CHKS; break;//Incorrect checksum (only if checksum provided)
			case 129: return COMM_LONG; break;//Command record too long
			default: return NOT_KNOWN;
		}
	}
	else if (ma==131)  //Execution Errors
	{
		switch(mi)
		{
			case 128: return EXEC_INV_1ST; break;//Invalid 1st parameter
			case 129: return EXEC_INV_2ND; break;//Invalid 2nd paramete
			case 130: return EXEC_INV_3RD; break;//Invalid 3rd parameter
			case 132: return EXEC_INV_NPAR; break;//Invalid number of parameters
			case 135: return EXEC_ILL_COMM_ACQ; break;//Illegal command while acquisition is in progress
			case 136: return EXEC_ILL_COMM_MODE; break;//Illegal command in current mode of operation
			default: return NOT_KNOWN;
		}
	}

	return NOT_KNOWN;//Unknown maior number
}

uint8_t TrumpPCI::checksum(const std::string &array)
//This method returns with the checksum of the array
{
	uint32_t sum=0l;//initialisation
	for(int i=0; i<(int)array.length(); i++) sum+=(uint8_t)(array[i]);//Summing up the ascii representation of the characters
	return (uint8_t)(sum&0xff);//Returns with the low byte
}

//---------------------------------------------------------Card related methods

bool TrumpPCI::card_openclose(bool bopen)
//Open(t)/close(f) the present communication port. Returns with success (t) or fail
{
  //We are intend to use port_number between 0 to max_offset
  seteuid(0);//Get root access temporary
  int ecode=ioperm((unsigned long)portNumber, (unsigned long)MAX_OFFSET, (bopen ? 1 : 0));
  seteuid(getuid());//Set it back to normal user
  if (ecode!=0)
  {
    std::cerr<<"ERROR: ("<<(bopen ? "open)" : "close)")<<std::endl;
    switch(errno)
    {
      case EINVAL: std::cerr<<"Invalid values of port number ("<<std::hex<<portNumber<<std::dec<<") or maximal offset ("<<MAX_OFFSET<<")"<<std::endl; break;
      case EIO: std::cerr<<"This call is not supported!"<<std::endl; break;
      case ENOMEM: std::cerr<<"Out of memory!"<<std::endl; break;
      case EPERM: std::cerr<<"Your privilege is insufficient!"<<std::endl; break;
      default: std::cerr<<"Other Error"<<std::endl;
    }
    return false;
  }
  return true;
}

void TrumpPCI::card_set_operation()
//This function initializes the registers of the card to be ready for communication
{
  card_write(INTCSR,RESET_INTCSR);//Reset the interrupt registers of the card
  //DMA operation: Here should be enable interrupt if the write transfer (DMA reading from card) has been completed and reset all kind of interrupts, then start interrupt service
  return;
}

void TrumpPCI::card_reset_operation()
//This function resets the registers of the card to return to the default position
{
    //DMA operation: here should be finish operating the interrupt wired to WTC_INTCSR
    card_write(INTCSR, RESET_INTCSR);//reset interrupt registers
    card_write(MCSR, RESET_MCSR);//reset status registers
    return;
}

void TrumpPCI::card_reset_incoming_fifo()
//This function resets incoming (Add-on->PCI) fifo 
{
  card_write(MCSR,card_read(MCSR)|A2PFSR_MCSR);//Reset the add on 2 pci fifo
  return;
}

void TrumpPCI::card_write_fifo(uint8_t *msg, int length)
//This function writes the length long msg into the fifo
{
    //First of all, coding the incoming message into dword-s in 3210; 7654 order.....
    int buffer_length=length/4+ (((length%4)!=0)? 1 : 0);//buffer length: if it is not integer dword long, an additional one will need
    for (int i=0; i<length/4; i++) sendBuffer[i]=(((uint32_t)msg[i*4+3])<<24)+(((uint32_t)msg[i*4+2])<<16)+(((uint32_t)msg[i*4+1])<<8)+((uint32_t)msg[i*4]);//Writing the whole dword part of the command
    if(length%4!=0)
    {
		//Writing the remaining part, if any...
		sendBuffer[buffer_length-1]=0x0;
		for (int i=4*(length/4); i<length; i++) sendBuffer[buffer_length-1]+= (((uint32_t) msg[i])<<(8*(i%4)));
    }    

    //message has been coded into the buffer, sending them to the outgoing fifo
    for (int i=0; i<buffer_length; i++) 
    {
		while (card_read(MCSR)&P2A_FIFO_FULL_MCSR);//waiting for free space in PCI->Add-On fifo
#ifdef DEBUG_TRUMP
		std::cout<<"Raw dataset to send ("<<i+1<<"): "<<std::hex<<(long)sendBuffer[i]<<std::dec<<std::endl;
#endif
		card_write(FIFO,sendBuffer[i]);//Write buffer to outgoing fifo
    }
    return;//message has been sent
}

int TrumpPCI::card_read_fifo(long start_indx)
//This function read the content of the incoming fifo, stores its content into standard receive buffer (recvBuffer) strating from start_indx. It returns with the length of message or with timeout (-1)
{
        long counter=0;
	while(((card_read(MCSR)&A2P_FIFO_EMPTY_MCSR))&&(counter++<timeOut)) usleep(FIFO_WAIT_SHORT);//Waiting for message in the incoming fifo
	if(counter>=timeOut) return 0;//Timeout
	else
	{
		//Regular situation, read buffer until it becomes empty
		counter=start_indx;//Set the counter to receive buffer
    		while (!(card_read(MCSR)&A2P_FIFO_EMPTY_MCSR)) 
    		{
    		    recvBuffer[counter++]=card_read(FIFO);
    		    usleep(1);
#ifdef DEBUG_TRUMP
    		    std::cout<<"Received raw message ("<<counter<<"): "<<std::hex<<recvBuffer[counter-1]<<std::dec<<std::endl;
#endif
    		}
		return counter;//Return with number of dwords has been read
	}
}

uint32_t*  TrumpPCI::card_communicate(uint8_t *command, int &length)
//Does the task of communication: send the length (in) long command to the mcb and receive a long (out) answer. If the command is NULL, then function assumes, that this is a continuation. Thus, comand is not transmitted, but reads. This case, the length(in) would be the length of the previous read. It returns with the answer pointer (success) or with null pointer (timeout)
{
        int savedlength=length;
	if(command!=0)
        {
          card_reset_incoming_fifo();//Reset incoming fifo
	  card_write_fifo(command,length);//Transfer command
          savedlength=0;
        }
	length=((command!=0) ? card_read_fifo() : card_read_fifo(length));//Read the answer
	if(length!=0) card_reset_incoming_fifo();//Reset incoming fifo
	//Collecting the answers
	if(length>0) return recvBuffer;//regular answer
	else
        {
         length=savedlength;
         return 0;//timeout
        }
}

void TrumpPCI::looking_for_cards(std::vector<uint32_t> *addr)
//This function is looking for the TrumpPCI card identifying them by vendor and device id. Then, put BAR into addr
{
    //Get the list of all pci devices
    struct pci_access *pciacc= pci_alloc();//Get the pci_access structure
    pci_init(pciacc);//Initialize PCI library
    pci_scan_bus(pciacc);//Scan the bus
    
    //Going through on each devices
    for(struct pci_dev *pcidev=pciacc->devices; pcidev; pcidev=pcidev->next)
    {
	if(( pcidev->vendor_id == (u16) TRUMP_VENDOR_ID) && (pcidev->device_id == (u16) TRUMP_DEV_ID))
	{
	    //Trump card has been found... add datasets for the array
	    addr->push_back((uint32_t) (pcidev->base_addr[0] &0xfffffffc));//BAR register(pciaddr_t) - masking its end due to io port number
	}
    }
    pci_cleanup(pciacc);//Close everything
    
    return;
    
}


