#include <iomanip>
#include <iostream>
#include <string>
#include <sstream>
#include <vector>

//Card io related declarations
#include <sys/io.h>
#include <unistd.h>

extern "C"
{ 
#include <pci/pci.h>
}             //libpci-dev is required


#ifndef TRUMPPCI_CLASS_H
#define TRUMPPCI_CLASS_H

enum GATE_MODE { GATE_OFF, GATE_COINCIDENCE, GATE_ANTICOINCIDENCE, GATE_ERROR};//Gate modes and an additional error field
enum E_CODE { SUCCESS, ST_ALREADY_ST, ST_PRESET_EXC, CMD_INV_VERB, CMD_INV_NOUN, CMD_INV_VERBNOUN, CMD_INV_MOD, CMD_INV_VERBMOD, CMD_INV_NOUNMOD, CMD_INV_VERBNOUNMOD, CMD_INV, COMM_CHKS, COMM_LONG, COMM_CHKS_RCV, EXEC_INV_1ST, EXEC_INV_2ND, EXEC_INV_3RD, EXEC_INV_NPAR, EXEC_ILL_COMM_ACQ, EXEC_ILL_COMM_MODE, EXEC_OTHER, NOT_KNOWN, TIMEOUT};
/*
%000000069 SUCCESS: No errors detected
START/STOP Warnings:
%000005074 ST_ALREADY_ST: MCB already started or stopped
%000006075 ST_PRESET_EXC: Preset already exceeded
Command Syntax Errors:
%129001082 CMD_INV_VERB: Invalid verb in command
%129002083 CMD_INV_NOUN: Invalid noun in command
%129003084 CMD_INV_VERBNOUN:Invalid verb and noun in command
%129004085 CMD_INV_MOD: Invalid modifier in command
%129005086 CMD_INV_VERBMOD: Invalid verb and modifier in command
%129006087 CMD_INV_NOUNMOD: Invalid noun and modifier in command
%129007088 CMD_INV_VERBNOUNMOD: Invalid verb, noun, and modifier in command
%129132087 CMD_INV: Invalid command (verb, noun, and modifier valid, but not together)
Communication Errors:
%130128084 COMM_CHKS: Incorrect checksum (only if checksum provided)
%130129085 COMM_LONG: Command record too long
COMM_CHKS_RCV: Incorrect checksum received
Execution Errors:
%131128085 EXEC_INV_1ST: Invalid 1st parameter
%131129086 EXEC_INV_2ND: Invalid 2nd parameter
%131130078 EXEC_INV_3RD: Invalid 3rd parameter
%131132080 EXEC_INV_NPAR: Invalid number of parameters
%131135083 EXEC_ILL_COMM_ACQ: Illegal command while acquisition is in progress
%131136084 EXEC_ILL_COMM_MODE: Illegal command in current mode of operation
EXEC_OTHER: other, nor % starting, nor specific response provided
NOT_KNOWN: does not known minor or maior code error
TIMEOUT: timeout error during the waiting for answer
*/

class TrumpPCI
//This class describes all of the functions required to deal with this card
{
  public:
    //Public methods
    TrumpPCI(int ser_number=-1, double timeout_ms=0.1);//Constructor:  serial number of the card (-1: first valid card), timeout of reading in milliseconds
    ~TrumpPCI();//Destructor
    
    double getTimeout(){return ((double)(timeOut*FIFO_WAIT_SHORT))/1000.0;};//Returns with the value of timeout in ms

    bool About(bool* features_check, std::string *mode, std::string *radix, int *serial_number, std::string *firmware);//Provides hard wired information about the card: its features (does it identical to manual), mode (pule height analysis), radix type, serial number and firmware model&version. It returns with true, if communication was successful.
    bool SetDevice();//Sets the device and remap shared memory - useless routine

    //Dataset related methods
    bool setDataRange(const int start=0, const int length=2048);//Sets the channel range exact value. Returns with its success
    bool getDataRange(int *start, int *length);//Asks the actual data range: number of channels and start channel. Returns with the communication success/fail
    bool setData(uint32_t count=0);//Sets the count for data channels, which are in the range. It returns with communication success or fail
    uint16_t lastDataChannel();//Returns with the last digitized channel number by ADC. Otherwise with error 
    bool getDataSet(int start, int length, uint32_t *counts, bool *roi_flag=NULL);//This method ask the recorded counts (necessary to provide!) and region of interest flags (optional) of channels from start and length long. Returns with communication errors 

    //Gate input configuration related methods
    bool setGate(GATE_MODE gm=GATE_OFF);//Sets the gate mode. Returns with the result of communication
    GATE_MODE getGate();//Returns with the status of the gate mode or with error message

    //Discriminator related methods
    bool setDiscriminator(int channel, bool lowhi=true);//Sets the low(true) or high(false) discriminator level to channel value.  Returns with the result of communication
    uint16_t getDiscriminator(bool lowhi=true);//Returns with the low(true) or high(false) discriminator level or with error message
    
    //Deadtime related methods
    bool setDTimeDelay(double delay=1.6);//Set the required time to restore initial level from the end of digitalization in microsec (0 means infinite delay).  Returns with the result of communication
    double getDTimeDelay();//Returns with the required time to restore initial level from the end of digitalization in microsec units (0 means infinite delay). Communication error is also reported.

    //Conversion gain related methods
    bool setGainConv(int channel=0);//Sets the gain conversion. Returns with communication error or non-valid input errors
    bool getGainConv(int *current, int *maximal=0);//Provides information about the current and maximal (if it has been set) gain conversion. Returns with the result of communication

    //Input signal offset setup related methods
    bool setOffset(uint16_t offset=2048);//Sets the ADC offset in hardware units (2048: no shift). Returns with communication error
    bool getOffset(int *min_ch, int *max_ch, int *min_hw, int *max_hw);//Get the limits of offset setup: min&max in channel and in the corresponding hardware units. Returns with communication error
    uint16_t getOffset();//Returns with the offset in hardware units or with error.
    uint16_t incrOffset();//Increases the offset by 1 hardware unit, returns with the new value in hardware units.
    
    //Start/Stop and diagnostics related methods   
    bool getStatus();//Returns with on measurement(t) or idle (f) status
    bool getStatus(double *live_time, double *real_time, uint16_t *active_device, std::string &hw_status);//Gets status information about live and true time in seconds, active device and on hardware status. Returns with communication error
/*		Bit 0 (LSB) Bias supply polarity (0 = positive, 1 = negative).
    Bit 1 Bias supply overload (0 = overload, 1 = normal).
    Bit 11 Waiting for start delay (0 = ready, 1 = waiting)
    Bit 12 Main Battery 2 In Use (0 = battery 1 in use, 1 = battery 2 in use).
    Bit 13 Main Battery 1 Low Warning (0 = normal, 1 = battery voltage low).
    Bit 14 Main Battery 2 Low Warning (0 = normal, 1 = battery voltage low).
*/
    bool StopStart(bool stop=true);//Starts(f) / stops (t) acquisition. Returns with communication error (true in the case of SUCCESS, ST_ALREADY_ST)

    //Extended clear methods
    bool clearTimeData();//Clears real, live time counters and datasets in window of interest. Returns with communication errors
    bool clearPresets();//Clears real, live time, roi integral, roi maximum and overflow presets. Returns with communication errors
    bool clearAll();//Clears the (live+preset) counters, the dataset counters in window of interest. Additionally, real, live time, roi integral, roi maximum and overflow presets will be set to zero and disabled. Moreover, roi regions will be cleared. Returns with communication errors	

    //Real and live time counts/preset
    bool setTime(double time=0.0, bool truelive=true, bool countpreset=false);//Sets the real(truelive=true)/live(truelive=false) counter(countpreset=true)/preset(countpreset=false) in seconds. Returns with communication errors - 0 disable
    double getTime(bool remainingelapsed=true, bool truelive=true, bool countpreset=true);//Returns with the real(truelive=true)/live(truelive=false) counter(countpreset=true)/preset(countpreset=false) value in seconds. If counter is asked, there is possible to ask for the elapsed (remainingelapsed=false) or remaining time (remainingelapsed=true). If preset is requested, remainingelapsed will be omitted.
    bool clearTime();//Clears real and live time counter. Returns with communication error

    //Overflow preset related
    bool setOFPreset(bool enabledisable=false);//Enable (t) (at channel with 2147483647 counts stops)/ disable(f) (continues to count) overflow preset. Returns with communication error
    bool getOFPreset();//Returns with the Enable (t)/ disable(f) status of overflow preset.

    //Region of interest (ROI) setup related methods
    bool setROI(int start, int number_ch);//Set the roi for channels 1..number_ch from start. Returns with communication error
    bool getROI(int *start, int *number_ch, bool first=true);//Gets start channel and number of channel of continous group of roi. It shows for the first (true) or the followings(false). If there is no more, it returns with 0,0. Returns with communication error
    bool clearROI();//Clears all ROI in the window of interest. Returns with communication error

    //ROI-Integral and maximum preset common method
    bool setIntegralMax(uint32_t preset=0, bool max=true);//Sets integral(f) or maximal value(t) preset for the whole roi, 0: disabled. 4294967295: integral maximal, 2147483647: maximum maximal. Returns with communication error
    
    //ROI-Integral preset related method
    uint32_t getIntegral(bool preset=false, int start=0, int num_ch=0);//Returns with integral preset (true)/actual integral value (false) for start, num_ch(non-zero) or for the whole roi (num_ch=0) (0..4294967295, this last means <=real for non-preset)

    //ROI-maximum and maximum preset related method
    uint32_t getMax(bool preset=false, bool channel=false);//Returns the maximal value among ROI: preset (preset=t)/ actual value (f,f)/ the belonging channel number (f,t), answer: 1..2147483647: counts, 0: 0 counts/channel (f,*) or no roi (f,*) or disabled(t,*)

    //Uncertainty-MDA common methods
    bool setUncMDA(double preset=0.0, bool uncertainty=true);//Sets the uncertainty(t) (in % units)/MDA(f) (in gamma/s units) preset (0 means disabled). Returns with communication error
    bool setUncMDARange(int start, int num_chn, bool uncertainty=true);//Sets the uncertainty preset(t)/MDA(f) range from start with num_chn. Returns with communication error
    double getUncMDA(bool preset=false, bool uncertainty=true);//Returns with the uncertainty (*,t)(% units) /MDA (*,f) (gamma/s units) actual (f,*) or preset (t,*) value. 0 means preset disabled
    bool getUncMDARange(int *start, int *num_chn, bool uncertainty=true);//Gets the Uncertainty preset/MDA range at roi from start with num_chn. Returns with communication error
    
    //MDA (minimum detectable activity) preset related methods
    bool setMDACoeffs(double a, double b, double c);//Sets the MDA coefficients used to calculate the MDA=(a+sqrt(b+c*roi_counts))/live_time in gamma/s units. Returns with communication error
    bool getMDACoeffs(double *a, double *b, double *c);//Gets the MDA coefficients used to calculate the MDA=(a+sqrt(b+c*roi_counts))/live_time in gamma/s units. Returns with communication error

  private:
    //private methods
    //Non-public commands due to simplification
    bool getDataMask(uint32_t *data_mask, uint32_t *roi_mask);//Get the masks, which should be "&" to get data and roi bits. It returns with communication success or fail

    //Communication and answer processing methods
    E_CODE processing_communication(std::string command, std::stringstream *result, uint32_t *dataset);//This function process the communication: gives a checksum for the command, gives this expression to communicate with hardware and process the answer into more relevant format: in the case of simple command, the result processed into result. In the case of raw channel, the result processed into the dataset array. It returns with an appropriate error code
    E_CODE process_answer(const uint32_t *answer, std::stringstream *result, uint32_t *dataset);//This method process the "answer". It checks the checsum (both response and command, as well), cut pure answer and splits them into relevant units, putting them into the "result". If raw channel datasets are asked, they will be stored in dataset array instead of "result". It returns with an error code in each cases.
    E_CODE raw2ecode(const std::string &answer);//Converts the raw answer format into error codes. It is supposed, that checksum already checked
    uint8_t checksum(const std::string &array);//This method returns with the checksum of the array
    
    //Card basic io operations
    bool card_openclose(bool bopen);//Opens(t)/closes(f) the present communication port. Returns with success (t) or fail
    uint32_t card_read(uint8_t offset) {return (uint32_t) inl((unsigned)(portNumber+offset));}; //This function performs reading a double word from a register, whose address is shifted by offset bytes from the starting port.
    void card_write(uint8_t offset, uint32_t data) {outl((unsigned long) data,(unsigned)(portNumber+offset)); return;};//This function performs writing of double word data to register, whose address is shifted by offset bytes from the starting port.
    
    //Card medium level operations
    void card_set_operation();//This function initializes the registers of the card to be ready for communication
    void card_reset_operation();//This function resets the registers of the card to return to the default position
    void card_reset_incoming_fifo();//This function resets the incoming (Add-on->PCI) fifo
    void card_write_fifo(uint8_t *msg, int length);//This function writes the length long msg into the fifo
    int card_read_fifo(long start_indx=0);//This function read the content of the incoming fifo, stores its content into standard receive buffer (recvBuffer) strating from start_indx. It returns with the length of message or with timeout (-1)
    uint32_t *card_communicate(uint8_t *command, int &length);//Does the task of communication: send the length (in) long command to the mcb and receive a long (out) answer. If the command is NULL, then function assumes, that this is a continuation. Thus, comand is not transmitted, but reads. This case, the length(in) would be the length of the previous read. It returns with the answer pointer (success) or with null pointer (timeout)
    
    //Get card info (pci)
    void looking_for_cards(std::vector<uint32_t> *addr);//This function is looking for the TrumpPCI card identifying them by vendor and device id. Then, put BAR into addr
    
    
    //Constants

    //Relevant offsets
    static const uint8_t MAX_OFFSET=64;//Maximal allowed offset of the data
    static const uint8_t FIFO=0x20;//FIFO
    static const uint8_t MWAR=0x24;//Master write address register
    static const uint8_t MWTC=0x28;//Master write transfer count register
    static const uint8_t MRAR=0x2c;//Master read address register
    static const uint8_t MRTC=0x30;//Master read transfer count register
    static const uint8_t INTCSR=0x38;//Interrupt control/status register
    static const uint8_t MCSR=0x3c;//Bus Master control/status register

    //INTCSR bits
    static const uint32_t ANYTC_INTCSR=0x800000;//Any transfer completed
    static const uint32_t RTC_INTCSR=0x80000;//DMA read transfer completed (write to card by DMA finished) 
    static const uint32_t WTC_INTCSR=0x40000; //DMA write transfer completed  (read from card by DMA finished) 
    static const uint32_t EIRTC_INTCSR=0x8000; //enable interrupt for DMA read transfer completed  (write to card by DMA finished)
    static const uint32_t EIWTC_INTCSR=0x4000; //enable interrupt for DMA write transfer completed  (read from card by DMA finished) 
    static const uint32_t RESET_INTCSR=0x3f0000; //Resets all of bits 

    //MCSR bits
    static const uint32_t A2PFSR_MCSR=0x4000000;//Add-on to PCI (read card) FIFO status reset-rwc
    static const uint32_t P2AFSR_MCSR=0x2000000;//PCI to Add-on (write card) FIFO status reset-rwc
    static const uint32_t RTE_MCSR=0x4000;//DMA read transfer enable (write to card by DMA) 
    static const uint32_t WTE_MCSR=0x400; //DMA write transfer enable (read from card by DMA)
    static const uint32_t A2P_FIFO_EMPTY_MCSR=0x20;//Add-on to PCI (read card) FIFO is empty
    static const uint32_t P2A_FIFO_FULL_MCSR=0x1;//PCI to Add-on (write card) FIFO is full
    static const uint32_t RESET_MCSR=A2PFSR_MCSR|P2AFSR_MCSR; //Reset transfers and fifo

    //Generals
    static const uint16_t TRUMP_VENDOR_ID=0x10e8;//Vendor ID
    static const uint16_t TRUMP_DEV_ID=0x82b7;//Vendor ID
    static const int COMMAND_DWORD=128;//Size of the command part of the receiving buffer in dword
    static const int RCV_MAX_SIZE=16896;//Maximal size of the receiving buffer in bytes (2(internal bus is only 16bit instead of 32)x4(size of one channel)x2048+512(256 character)x2 (16bit to 32bite storage))
    static const int COMMAND_MAX_SIZE=128;//Maximal size of a command including checksum in bytes
    static const int FIFO_WAIT_SHORT=10;//Time in microsec between two attemption to read the incoming fifo full register

    //Private members
    uint32_t *sendBuffer;//Send buffer for the low level communication
    uint32_t *recvBuffer;//Receiving buffer for the low level communication
    uint32_t portNumber;//Base address register
    uint32_t d_mask;//Data mask on direct access of data
    uint32_t r_mask;//Region of interest mask
    long timeOut;//Timeout in ticks of FIFO_WAIT_SHORT
    bool commOK;//Channel is open to communicate with the card

};
#endif

