#include "UDP.h"

bool operator!= (const struct sockaddr_in &a, const struct sockaddr_in &b) { return (a.sin_addr.s_addr!=b.sin_addr.s_addr)||(a.sin_port!=b.sin_port);}
bool operator< (const struct sockaddr_in &a, const struct sockaddr_in &b) { return (a.sin_addr.s_addr<b.sin_addr.s_addr)||((a.sin_addr.s_addr==b.sin_addr.s_addr)&&(a.sin_port<b.sin_port));}

UDP::UDP(const int local_port_no, const std::string local_ip, std::vector<std::string> &far_ips, std::vector<int> &far_ports, std::vector<int> &far_ids)
//Constructor by setting the client/server to the local port number acting as server (local ip should be="") or client(local_ip should be specified) and set the IP/port/module_id structures up
{
    if((far_ips.size()!=far_ids.size())||(far_ids.size()!=far_ports.size()))
    {
        std::cerr<<"ERROR (UDP::UDP) The sizes of far_ips, far_ports and far_ids do not conform!"<<std::endl;
        return;
    }
    else
    {
        Socket_fd=-1;
        for(int i=0; i<static_cast<int>(far_ips.size());i++) SetFar(far_ips[i],far_ports[i],far_ids[i]);
        Partnersize=sizeof(Partner);
        SetLocal(local_port_no,local_ip);
    }
}


UDP::~UDP()
{
  if((Socket_fd>=0)&&(close(Socket_fd)!=0)) std::cerr<<"ERROR (UDP::~UDP) Error at closing of the socket."<<std::endl;
  Socket_fd=-1;
}

//ID-ip-port related methods

std::string UDP::GetFarIP(int id) const
//Returns with the far IP address belonging to id
{
    char addr[INET_ADDRSTRLEN];
    std::map<int,struct sockaddr_in>::const_iterator it=Far_id.find(id);
    if(it==Far_id.cend())
    {
        std::cerr<<"ERROR (UDP::GetFarIP): the requested module id ("<<id<<") is missing from the list!"<<std::endl;
        return "-1";
    }
    return std::string(inet_ntop(AF_INET,&((it->second).sin_addr),addr,INET_ADDRSTRLEN));
}

int UDP::GetFarPort(int id) const
//Returns with the far port number belonging to id
{
   std::map<int,struct sockaddr_in>::const_iterator it=Far_id.find(id);
   if(it==Far_id.cend())
    {
        std::cerr<<"ERROR (UDP::GetFarPort): the requested module id ("<<id<<") is missing from the list!"<<std::endl;
        return -1;
    }
   return ntohs((it->second).sin_port);
}

bool UDP::SetFar(std::string ip, int port_no, int module_id)
//Add an item to the IP/port/module_id structure
{
    struct sockaddr_in addr;
    memset(reinterpret_cast<struct sockaddr_in*>(&addr),0,sizeof(addr));//Set all values to zero
    addr.sin_family=AF_INET;
    addr.sin_port=htons(port_no);
    if(inet_pton(AF_INET,ip.c_str(),&(addr.sin_addr.s_addr))!=1)
    {
        std::cerr<<"ERROR (UDP::SetFar): the address is not converted"<<std::endl;
        return false;
    } 
    else
    {
        bool retval=(Far_id.insert(std::pair<int,struct sockaddr_in>(module_id,addr))).second;
        if(retval) retval=(Far_sock.insert(std::pair<struct sockaddr_in,int>(addr,module_id))).second;
        return retval;
    }
}

void UDP::DeleteFar(int module_id)
//Delete the module_id from the list
{
    std::map<int,struct sockaddr_in>::iterator it=Far_id.find(module_id);
    if(it==Far_id.end()) std::cerr<<"ERROR (UDP::DeleteFar): the requested module id ("<<module_id<<") is missing from the list!"<<std::endl;
    else
    {
        Far_sock.erase(it->second);
        Far_id.erase(module_id);
    }
    return;
}

bool UDP::ModifyFar(int prev_id,int new_id)
//Modify the prev_id to new_id
{
    std::map<int,struct sockaddr_in>::iterator it=Far_id.find(prev_id);
    if(it==Far_id.end())
    {
        std::cerr<<"ERROR (UDP::ModifyFar): the previous module id ("<<prev_id<<") is missing from the list!"<<std::endl;
        return false;
    }
    struct sockaddr_in addr=it->second;
    DeleteFar(prev_id);
    bool retval=(Far_id.insert(std::pair<int,struct sockaddr_in>(new_id,addr))).second;
    if(retval) retval=(Far_sock.insert(std::pair<struct sockaddr_in,int>(addr,new_id))).second;
    return retval;
}

bool UDP::ModifyFar(int module_id, std::string new_ip, int new_port_no)
//Modifies the current setup of module_id by setting the new_ip and new_port_no
{
    DeleteFar(module_id);
    return SetFar(new_ip,new_port_no,module_id);
}

 std::vector<int> UDP::GetFarIDs()
 //Returns with the far id's 
 {
     std::vector<int> ids;
     for(std::map<int,struct sockaddr_in>::iterator it=Far_id.begin(); it!=Far_id.end(); it++) ids.push_back(it->first);
     return ids;
 }

//Local port settings and initialization

std::string UDP::GetLocalIP() const
//Returns with the local IP address
{
    char addr[INET_ADDRSTRLEN];
    return std::string(inet_ntop(AF_INET,&(Local.sin_addr),addr,INET_ADDRSTRLEN));
}

int UDP::GetLocalPort() const
//Returns with the local port number
{
   return ntohs(Local.sin_port);
}

bool UDP::SetLocal(const int local_port_no, const std::string &local_ip)
//Configures the server(local_ip="")/client(specific local_ip) to use the local_port_no. If a socket already in operation, it reset to the new values
{
  struct addrinfo myaddress;
  struct addrinfo *result;
  std::string port=std::to_string(local_port_no);
  
  memset(&myaddress, 0, sizeof(myaddress));//Initialize
  if(local_ip=="") myaddress.ai_flags = AI_PASSIVE;//Server
  myaddress.ai_family = AF_INET;//IPv4
  myaddress.ai_socktype = SOCK_DGRAM;//UDP
  
  if((Socket_fd>=0)&&(!close(Socket_fd)))
  {
      std::cerr<<"ERROR (UDP::SetLocal) Error at closing of the socket."<<std::endl;
      return false;
  }
  
  int err=getaddrinfo( (local_ip=="") ? NULL : local_ip.c_str(),port.c_str(),&myaddress,&result);
  if(err!=0)
  {
      std::cerr<<"ERROR (UDP::SetLocal) The address has not been converted. Code: "<<std::string(gai_strerror(err))<<std::endl;
      return false;
  }
  Local=*(reinterpret_cast<struct sockaddr_in*>((result->ai_addr)));//Set the local address structure
  
  if(result==NULL)
  {
      std::cerr<<"ERROR (UDP::SetLocal) The address has not been converted."<<std::endl;
      return false;
  }
  
  Socket_fd=socket(result->ai_family,result->ai_socktype,result->ai_protocol);
  if(Socket_fd<0)
  {
      std::cerr<<"ERROR (UDP::SetLocal) The socket has not been created."<<std::endl;
      return false;
  }

  if(bind(Socket_fd,result->ai_addr,result->ai_addrlen)!=0)
  {
    std::cerr<<"ERROR (UDPServer::SetLocal) The socket was a trouble at bounding to an address."<<std::endl;
    return false;
  }
  
  freeaddrinfo(result);
  return true;
}

//Communication routines
bool UDP::SendMessage(const char *msgtosend, const int &send_length, const int &destination_id)
//This method send the send_length long msgtosend to destination id. Returns with successfull delivery report
{
    std::map<int,struct sockaddr_in>::iterator it=Far_id.find(destination_id);
    if(it==Far_id.end())
    {
        std::cerr<<"ERROR (UDP::SendMessage): the destination module id ("<<destination_id<<") is missing from the list!"<<std::endl;
        return false;
    }
    struct sockaddr_in addr=it->second;
#ifdef DEBUG_UDP
    std::cout<<"UDP::SendMessage: "<<std::endl;
    for(int i=0; i<send_length; i++) std::cout<<std::hex<<"0x"<<(short)msgtosend[i]<<std::dec<<std::endl;
    std::cout<<"Send length: "<<send_length<<std::endl;
    std::cout<<"Socket number: "<<Socket_fd<<std::endl;
    std::cout<<"Target ID: "<<destination_id<<std::endl;
    std::cout<<"Target IP   : "<<GetFarIP(destination_id)<<"/"<<GetFarPort(destination_id)<<std::endl;
    std::cout<<"By other way: "<<inet_ntoa(addr.sin_addr)<<"/"<<ntohs(addr.sin_port)<<std::endl;
#endif
    if(sendto(Socket_fd,msgtosend,(size_t)send_length,0,(struct sockaddr*) &addr,(socklen_t)sizeof(addr))!=send_length)
    {
        int err=errno;
        perror("Sendmessage");
        std::cerr<<"ERROR (UDP::SendMessage): the sent message size is less than the requested: "<<err<<std::endl;
        return false;
    }
    return true;
}

int UDP::ReceiveMessage(char *buffer, const int &buffer_size, bool &tolerate_other_sender, const int &expected_sender_id)
//This method writes the received message to the buffer_size long buffer and returns with the length of the message. If error happens, it returns with negative value. If tolerate_other_sender is set to false the id of the sender should be the same as expected_sender_id 
{
  ssize_t alength=recvfrom(Socket_fd,buffer,buffer_size,0,reinterpret_cast<struct sockaddr*>(&Partner),reinterpret_cast<socklen_t *>(&Partnersize));
  if(alength<0)
  {
    int errsv=errno;
    if(errsv!=EAGAIN) std::cerr<<"ERROR (UDP::ReceiveMessage): error appeared at the receiving a message: "<<errsv<<std::endl;
    return -1;
  }
  
  if(!tolerate_other_sender)
  {
    std::map<int,struct sockaddr_in>::iterator it=Far_id.find(expected_sender_id);
    if(it==Far_id.end())
    {
        std::cerr<<"ERROR (UDP::ReceiveMessage): the module id of the expected sender ("<<expected_sender_id<<") is missing from the list!"<<std::endl;
        return -1;
    }
    if(Partner!=it->second)
    {
        char addr[INET_ADDRSTRLEN];
        std::cerr<<"ERROR (UDP::ReceiveMessage): the answer arrived from not the requested address:\n\t"<<inet_ntop(AF_INET,&(Partner.sin_addr),addr,INET_ADDRSTRLEN)<<"/"<<ntohs(Partner.sin_port)<<" instead from "<<GetFarIP(expected_sender_id)<<"/"<<GetFarPort(expected_sender_id)<<std::endl;
        return -1;
    }
  }
  return static_cast<int>(alength);
}

int UDPClient::SendAndReceive(const char *msgtosend, const int send_length, char *answer, const int &answer_maxsize, const int &destination_id, bool only_receive, bool tolerate_other_sender)
//If onlyrecive is false it sends a send_length long msgtosend to the destination_id address and receive the answer to an answer_maxsize long answer buffer. If only_receive is true, it waits for the reply only. If tolerate_other_sender is false, it returns with error if other than the computer with destination_id is answering. Returns with the valid length of the answer or <0 in the case of error, and puts the answer to answer_length array
{
  if((!only_receive)&&(!SendMessage(msgtosend,send_length,destination_id))) return -1;
  return ReceiveMessage(answer,answer_maxsize,tolerate_other_sender, destination_id);
}

UDPServerTimeOut::UDPServerTimeOut(const int local_port_no, std::vector<std::string> &far_ips, std::vector<int> &far_ports, std::vector<int> &far_ids,int timeout_milliseconds):UDP(local_port_no,"",far_ips,far_ports,far_ids)
//Constructor by setting to the local port number acting as server and set the IP/port/module_id structures up, also timeout is set in milliseconds
{
    Tv.tv_sec=timeout_milliseconds/1000;
    Tv.tv_usec=1000*(timeout_milliseconds%1000);
    if(setsockopt(Socket_fd,SOL_SOCKET,SO_RCVTIMEO,&Tv,sizeof(Tv))<0) std::cerr<<"ERROR (UDP::UDPServerTimeOut): setting the timeout has been failed!"<<std::endl;
    return;
}
