//TESTED: 04.10.2013 by Laszlo Temleitner
#include "UPP.h"

const double UPP::TIMEOUT=1.0;//1 s
const double UPP::WAITSUCCESS=0.005;//5 ms

UPP::UPP(const int address, const std::string nPort, const bool cf, const int nBaud, const double timeout):Thermometer(TEMP_PARAMETERS)
//Constructor: device address, serial port name, Celsius(t)/Fahrenheit(f), baud rate, timeout
{
  //Set the unit
  cfunit=cf;
  //Setting the communication lime and address
  communicator=NULL;
  SetAddress(address);
  if(!ResetCommunication(nPort, nBaud, timeout)) std::cout<<"UPP::UPP: error in opening the communication channel"<<std::endl;
}

bool UPP::ResetCommunication(const std::string nPort, const int nBaud, const double timeout)
//Reset the communication parameters: serial port name, transmission rate, timeout
{
  bool retval;//Returning value
  //Modify the settings or create a new channel
  if(communicator==NULL)
  {
    communicator= new Serial(nPort, nBaud, 8, 'E', 1);
    retval=communicator->isOpened();
  }
  else retval=communicator->Open(nPort, nBaud, 8, 'E', 1);
  
  //Set timeout
  if (retval) retval=communicator->setTimeout(timeout);
  
  //Set the unit of display
  if(retval) retval=SetTemperatureUnit(cfunit);
  
  //Returning with the results
  return retval;
}

bool UPP::GetTemperature(double &onechannel, double &ratio)
//Get the one channel temperature in the measurement unit and the ratio temperature
{
    char comm[2]={'e','k'};//ek
    uint8_t mess[10];//message
    
    //Communicate
    bool retval=getvalue(comm,10,mess);
        
    if(retval)
    {
	//Set the temparature
        onechannel=((double) from(mess,5))/10.0;
        ratio=((double) from(mess+5,5))/10.0;
    }
    
    //Transforming the message to valid output
    return retval;
}

bool UPP::SetResponseTime(const int time)
//Sets the response time: from 0.000 to 10.000s (with discrate steps:0=min; 1=0.01; 2=0.05; 3=0.25; 4=1.0; 5=3.0; 6=10.0)
{
  char comm[2]={'e','z'};//ez
  uint8_t mess;//message
  
  //Checking the correct setup
  if((time<0)||(time>6))
  {
    std::cerr<<"Error! UPP::SetResponseTime: the value \""<<time<<"\" should be between 0 (minimum) and 6 (10s)!\n";
    return false;
  }
  //set the number correctly
  to(time,&mess,1);
  return setvalue(comm, 1, &mess);
}

bool UPP::SetOperationMode(const bool ratioonech)
//Sets the operation mode to 2channel (ratio, true) or 1 channel (false)
{
  char comm[2]={'k','a'};//ka
  uint8_t mess;//message
  ratioonech? to(2,&mess,1) : to(1,&mess,1);
  return setvalue(comm,1,&mess);
}

bool UPP::SwitchLaser(const bool on)
//Switch on(t)/off(f) laser
{
  char comm[2]={'l','a'};//la
  uint8_t mess;//message
  on? to(1,&mess,1) : to(0,&mess,1);
  return setvalue(comm,1,&mess);
}

bool UPP::GetLaserStatus()
//Returns with the status of laser: on(t) / off (f)
{
  char comm[2]={'l','a'};//la
  uint8_t mess;//message
    
  //Communicate
  bool retval=getvalue(comm,1,&mess);

  if(retval) return ((bool) from(&mess,1));
  else return false;
}

bool UPP::ClearMaxStorage()
//Clears the maximum storage (software simulation of external clearance)
{
  char comm[2]={'l','x'};//lx
  return setvalue(comm,0,NULL);
}

bool UPP::ClearPeakMemoryTime(const int time)
//Clear peak memory time. It changes by values (0=OFF; 1=0.01s; 2=0.05s; 3=0.25s; 4=1.0s; 5=5.0s; 6=25.0s; 7=EXTERN; 8=AUTO; 9=HOLD)
{
  char comm[2]={'l','z'};//lz
  uint8_t mess;//message
  
  //Checking the correct setup
  if((time<0)||(time>9))
  {
    std::cerr<<"Error! UPP::ClearPeakMemoryTime: the value \""<<time<<"\" should be between 0 (OFF) and 9 (HOLD)!\n";
    return false;
  }
  //set the number correctly
  to(time,&mess,1);
  return setvalue(comm, 1, &mess);
}

bool UPP::GetSubRange(int &lowlimit, int &upperlimit)
//Returns the lower and upper limits of the subrange in the requested unit
{
  char comm[2]={'m','e'};//me
  uint8_t mess[8];//message
  
  //Communicate
  bool retval=getvalue(comm,8,mess);
  
  if(retval)
  {
      //Set the temparature
      lowlimit=from(mess,4,false);
      upperlimit=from(mess+4,4,false);
  }
  
  //Transforming the message to valid output
  return retval;
}

bool UPP::SetSubRange(const int lowlimit, const int upperlimit)
//Sets the lower and upper limits of the subrange in the requested unit
{
  char comm[2]={'m','1'};//m1
  uint8_t mess[8];//message
  
  //Setting the limits
  to(lowlimit,mess,4,false);
  to(upperlimit,mess+4,4,false);
  return setvalue(comm, 8, mess); //Send request
}

double UPP::GetMeasuredValue()
//Returns with the measured value, -1: communication error, -2: overflow error
{
  char comm[2]={'m','s'};//ms
  uint8_t mess[5];//message
  
  //Communicate
  bool retval=getvalue(comm,5,mess);

  if(!retval) return -1.0;//Comm. error
    
  int value=from(mess,5);
  if (value==88880) return -2.0;//Overflow
  return value/10.0;
}

int UPP::GetSignalStrength()
//Returns with the strength of the signal: 0..1500
{
  char comm[2]={'t','r'};//tr
  uint8_t mess[4];//message
  
  //Communicate
  bool retval=getvalue(comm,4,mess);

  if(!retval) return -1.0;//Comm. error
  else return from(mess,4);
}

int UPP::GetResponseTime()
//Returns with the response time: from 0.000 to 10.000s (with discrate steps:-1=error, 0=min; 1=0.01; 2=0.05; 3=0.25; 4=1.0; 5=3.0; 6=10.0)
{
  //Bulky reading
  double emi, era;//Emissivity, emissivity ratio
  int rtime, pmtime, inttemp, devadd, baud;//Response time, peak-memory time, internal temperature, device address, baud
  bool aout;//analog output
  if (ReadParams(emi, rtime, pmtime, aout, inttemp, devadd, baud, era))
  {
    return rtime;
  }
  else return -1;
}

int UPP::GetPeakMemoryTime()
//Gets the peak memory time. It changes by values (-1=error, 0=OFF; 1=0.01s; 2=0.05s; 3=0.25s; 4=1.0s; 5=5.0s; 6=25.0s; 7=EXTERN; 8=AUTO; 9=HOLD)
{
  //Bulky reading
  double emi, era;//Emissivity, emissivity ratio
  int rtime, pmtime, inttemp, devadd, baud;//Response time, peak-memory time, internal temperature, device address, baud
  bool aout;//analog output
  if (ReadParams(emi, rtime, pmtime, aout, inttemp, devadd, baud, era))
  {
    return pmtime;
  }
  else return -1;
}

//Pyrometer not frequent get and set parameters
void UPP::getRuralSetParameters(bool *temp_unit, int *curr_internal_temp, int* max_internal_temp, int *minint_switchoff_level, double *emissivity, double *window_trans, double *emiss_ratio, bool bulky)
//Gives information about the not so frequently set parameters: temperature unit(C-t, F-f), current and maximal internal temperature, minimum intensity switch-off level, emissivity in onechannel mode, window transmittance, emissivity ratio (epsilon1/epsilon2), analog output settings (t:0..20, f:4..20mA)
{
  if (temp_unit!=NULL) *temp_unit=GetTemperatureUnit();//It returns with the current unit of temperature (C:t, F:f)
  if(max_internal_temp!=NULL) *max_internal_temp=GetInternalTemperature(false);//Returns with the internal temperature, which can be the current one (true) or maximum temperature (false). It is changes between 0..98C or 32..210F; -1: error
  if(minint_switchoff_level!=NULL) *minint_switchoff_level=GetMinimumIntensitySwitchOffLevel();//Gets the minimum intensity switch-off level, -1: error
  
  if(window_trans!=NULL) *window_trans=GetWindowTrans();//Returns with the current setting for the transmittance of the window from 0.05 to 1.0; -1.0: error
  if(bulky)
  {
    //Bulky reading
    double emi, era;//Emissivity, emissivity ratio
    int rtime, pmtime, inttemp, devadd, baud;//Response time, peak-memory time, internal temperature, device address, baud
    bool aout;//analog output
    if (ReadParams(emi, rtime, pmtime, aout, inttemp, devadd, baud, era))
    //Read status information on: emissivity, response time, peak memory time, analog output, internal temperature, device address, baud rate, emissivity ratio
    {
      if(emissivity!=NULL) *emissivity=emi;//Returns with the current setting of emissivity for one channel temperature from 0.05 to 1.0; -1.0: error
      if(curr_internal_temp!=NULL) *curr_internal_temp=inttemp;//Returns with the internal temperature, which can be the current one (true) or maximum temperature (false). It is changes between 0..98C or 32..210F; -1: error
      if(emiss_ratio!=NULL) *emiss_ratio=era;//Returns with the current setting for the emissivity ratio (epsilon1/epsilon2): from 0.8 to 1.2; -1.0: error
    }
  }
  else
  {
    //Individual reading
    if(emissivity!=NULL) *emissivity=GetEmissivity();//Returns with the current setting of emissivity for one channel temperature from 0.05 to 1.0; -1.0: error
    if(curr_internal_temp!=NULL) *curr_internal_temp=GetInternalTemperature(true);//Returns with the internal temperature, which can be the current one (true) or maximum temperature (false). It is changes between 0..98C or 32..210F; -1: error
    if(emiss_ratio!=NULL) *emiss_ratio=GetEmissivityRatio();//Returns with the current setting for the emissivity ratio (epsilon1/epsilon2): from 0.8 to 1.2; -1.0: error
  }
  return;
}

bool UPP::setRuralSetParameters(bool temp_unit, int min_int_switch_off_level, double emissivity, double window_trans, double emiss_ratio, int dirty_window_warn)//Set the values of not frequently set parameters: temperature unit(C-t, F-f), minimum intensity switch-off level, emissivity in onechannel mode, window transmittance, emissivity ratio (epsilon1/epsilon2), "Dirty window" warning level: from 0 to 99%.
{
  return (SetTemperatureUnit(temp_unit) && SetMinimumIntensitySwitchOffLevel(min_int_switch_off_level) && SetEmissivity(emissivity) && SetWindowTrans(window_trans) && SetEmissivityRatio(emiss_ratio) && SetDirtyWindowWarn(dirty_window_warn));
}

bool UPP::getCommunication(int *address, int *baud, bool *analog_output, bool bulky)
//Get communication related parameters: address, baud rate, analog output setting (t: 0..20mA, f:4..20mA)
{
  double emissivity, emrat;
  int response_time, peak_memory_time, int_temp, dev_addr, brate;
  bool aout;
  bool retval=ReadParams(emissivity, response_time, peak_memory_time, aout, int_temp, dev_addr, brate, emrat);//Get information
  if (retval)
  {
    if(baud!=NULL) *baud=brate;
    if(analog_output!=NULL) *analog_output=aout;
    if(address!=NULL)
    {
      if(bulky) *address=dev_addr;
      else *address=GetAddress();
    }
  }
  return retval;
}

bool UPP::setCommunication(int address, int baud_rate, bool analog_output)
//Changes the communication related parameters: device address, baud rate and analog output (t: 0..20mA, f:4..20mA)
{
  return (SetDeviceAddress(address)&&SetBaudRate(baud_rate)&&SetAnalogOutput(analog_output));
}

void UPP::About(std::string *ref_num, std::string *dev_type, std::string *ser_num, std::string *soft_ver, std::string *comm_det, std::string *soft_det, int *basic_range_lo, int *basic_range_hi)
//Gives general information about the device: reference number, device type, detector serial number, device type and software version, communication deatails, software details, and finally with basic range limits
{
    if(ref_num!=NULL) *ref_num=GetReferenceNumber();//Gets the reference number, Error: "ERROR"
    int low=0;//Low
    int high=0;//High limits
    if(dev_type!=NULL) *dev_type=GetDeviceType();//Returns with the device type or with error message
    if(ser_num!=NULL)  *ser_num=GetSerialNumber();//Returns the serial number of the detector
    if(soft_ver!=NULL) *soft_ver=GetDeviceTypeSoftwareVersion();//Returns with the device type (25) and the version of the software (MMYY)
    if(comm_det!=NULL) *comm_det=GetCommunicationDetails();//Returns with the details of the software: DD.MM.YY SOFT.VERSION
    if(soft_det!=NULL) *soft_det=GetSoftwareDetails();//Returns with the details of the software: DD.MM.YY SOFT.VERSION*
    if ((basic_range_lo!=NULL)&&(basic_range_hi!=NULL))
    {
      if (GetBasicRange(low, high))//Returns the lower and upper limits of the basic range in the requested unit
      {
	*basic_range_lo=low;
	*basic_range_hi=high;
      }
      else
      {
	*basic_range_lo=-1000;//Error messages
	*basic_range_hi=-1000;
      }
    }
    return;
}

//private methods: middle level methods

bool UPP::ReadParams(double &emissivity, int &response_time, int &peak_memory_time, bool &analog_output, int &internal_temp, int &device_addr, int &baud, double &emissivity_ratio)
//Read status information on: emissivity, response time, peak memory time, analog output, internal temperature, device address,
//baud rate, emissivity ratio
{
  char comm[2]={'p','a'};//pa
  uint8_t mess[15];//message
  //Communicate
  bool retval=getvalue(comm,15,mess);
  if(!retval) return false;//Comm. error
  //Evaluate datasets
  if (from (mess+10,1)!=0) return false;//Not the correct device
  emissivity= ((double) from(mess,2))/100.0;
  if(emissivity==0.0) emissivity+=1.0;
  response_time=from(mess+2,1);
  peak_memory_time=from(mess+3,1);
  analog_output=(from(mess+4,1)==0);
  internal_temp=from(mess+5,2);
  device_addr=from(mess+7,2);
  switch(from(mess+9,1))
  {
    case 0: baud=1200; break;
    case 1: baud=2400; break;
    case 2: baud=4800; break;
    case 3: baud=9600; break;
    case 4: baud=19200; break;
    case 5: baud=38400; break;
    case 6: baud=57600; break;
    case 8: baud=115200; break;
    default: baud=from(mess+9,1);
  }
  emissivity_ratio=((double)from(mess+11,4))/1000.0;
  return true;
}

//Pyrometer not frequent get and set parameters - start
int  UPP::GetInternalTemperature(const bool currmax)
//Returns with the internal temperature, which can be the current one (true) or maximum temperature (false).
//It is changes between 0..98C or 32..210F; -1: error
{
  char comm[2]={'g','t'};//gt
  if(!currmax)
  {
    comm[0]='t'; comm[1]='m';//tm
  }
  uint8_t mess[3];//message
    
  //Communicate
  bool retval=getvalue(comm,3,mess);

  if(retval) return from(mess,3);
  else return -1;
}

int  UPP::GetMinimumIntensitySwitchOffLevel()
//Gets the minimum intensity switch-off level, -1: error
{
  char comm[2]={'a','w'};//aw
  uint8_t mess[2];//message
    
  //Communicate
  bool retval=getvalue(comm,2,mess);

  if(retval) return from(mess,2);
  else return -1;
}

double UPP::GetWindowTrans()
//Returns with the current setting for the transmittance of the window from 0.05 to 1.0; -1.0: error
{
  char comm[2]={'e','t'};//et
  uint8_t mess[4];//message
    
  //Communicate
  bool retval=getvalue(comm,4,mess);

  if(retval) return from(mess,4)/1000.0;
  else return -1.0;
}

double UPP::GetEmissivity()
//Returns with the current setting of emissivity for one channel temperature from 0.05 to 1.0; -1.0: error
{
  char comm[2]={'e','m'};//em
  uint8_t mess[4];//message
    
  //Communicate
  bool retval=getvalue(comm,4,mess);

  if(retval) return from(mess,4)/1000.0;
  else return -1.0;
}

double UPP::GetEmissivityRatio()
//Returns with the current setting for the emissivity ratio (epsilon1/epsilon2): from 0.8 to 1.2; -1.0: error
{
    char comm[2]={'e','v'};//ev
    uint8_t mess[4];//message
    
    //Communicate
    bool retval=getvalue(comm,4,mess);

    if(retval) return from(mess,4)/1000.0;
    else return -1.0;
}

bool UPP::SetTemperatureUnit(const bool cf)
//Sets the display unit in C (true)/F (false)
{
  char comm[2]={'f','h'};//fh
  uint8_t mess;//message
  cf? to(0,&mess,1) : to(1,&mess,1);
  if (!setvalue(comm,1,&mess)) return false;//Changing the unit on pyrometer
  else  //And in the corresponding class
  {
    cfunit=cf;
    return true;
  }
}

bool UPP::SetMinimumIntensitySwitchOffLevel(const int level)
//It sets the minimum intensity switch-off level: between 2 and 50
{
    char comm[2]={'a','w'};//aw
    uint8_t mess[2];//message
    
    //Check limits
    if((level<2)||(level>50))
    {
	std::cerr<<"ERROR! UPP::SetMinimumIntensitySwitchOffLevel: level is outside valid range: 2..50\n Actual value: "<<level<<std::endl;
	return false;
    }
    //Set values
    to(level,mess,2);
    return setvalue(comm,2, mess);
}

bool UPP::SetEmissivity(const double epsilon)
//Sets the emissivity for one channel temperature: from 0.05 to 1.0
{
  char comm[2]={'e','m'};//em
  uint8_t mess[4];//message
  
  //Checking the correct setup
  if((epsilon<0.050)||(epsilon>1.000))
  {
    std::cerr<<"Error! UPP::SetEmissivity: the value \""<<epsilon<<"\" should be between 0.050 and 1.000!\n";
    return false;
  }
  //set the emissivity correctly
  int r=(int) (1000*epsilon);
  to(r,mess,4);
  return setvalue(comm, 4, mess);
}

bool UPP::SetWindowTrans(const double tau)
//Sets the window transmittance: from 0.05 to 1.0
{
  char comm[2]={'e','t'};//et
  uint8_t mess[4];//message
  
  //Checking the correct setup
  if((tau<0.050)||(tau>1.000))
  {
    std::cerr<<"Error! UPP::SetWindowTrans: the value \""<<tau<<"\" should be between 0.050 and 1.000!\n";
    return false;
  }
  //set the tau correctly
  int r=(int) (1000*tau);
  to(r,mess,4);
  return setvalue(comm, 4, mess);
}

bool UPP::SetEmissivityRatio(const double ratio)
//Sets the emissivity ratio(epsilon1/epsilon2): from 0.8 to 1.2
{
  char comm[2]={'e','v'};//ev
  uint8_t mess[4];//message
  
  //Checking the correct setup
  if((ratio<0.8)||(ratio>1.2))
  {
    std::cerr<<"Error! UPP::SetEmissivityRatio: the value \""<<ratio<<"\" should be between 0.8 and 1.2!\n";
    return false;
  }
  //set the ratio correctly
  int r=(int) (1000*ratio);
  to(r,mess,4);
  return setvalue(comm, 4, mess);
}

bool UPP::SetDirtyWindowWarn(const int level)
//Set the "Dirty window" warning level: from 0 to 99%
{
  char comm[2]={'d','w'};//dw
  uint8_t mess[2];//message
    
  //Check limits
  if((level<0)||(level>99))
  {
    std::cerr<<"ERROR! UPP::SetDirtyWindowWarn: level is outside valid range: 0..99\n Actual value: "<<level<<std::endl;
    return false;
  }
  //Set values
  to(level,mess,2,false);
  return setvalue(comm,2, mess);
}

// not so frequent get and set -end

//Communication
bool UPP::SetAnalogOutput(const bool set)
//Sets the analog output: true 0..20mA, false: 4..20mA
{
    char comm[2]={'a','s'};//as
    uint8_t mess;//message
    set? to(0,&mess,1) : to(1,&mess,1);
    return setvalue(comm,1,&mess);
}

bool UPP::SetBaudRate(const int rate)
//Sets the requested baudrate to communicate: from 1200 to 115200 baud
{
  //Transforming the rate into code number
  int baud;
  switch(rate)
  {
    case   1200: baud=0; break;
    case   2400: baud=1; break;
    case   4800: baud=2; break;
    case   9600: baud=3; break;
    case  19200: baud=4; break;
    case  38400: baud=5; break;
    case  57600: baud=6; break;
    case 115200: baud=8; break;
    default: std::cerr<<"ERROR! UPP::SetBaudRate: This rate (\""<<rate<<"\") is not allowed!\n Only 1.2k,2.4k,4.8k,9.6k, 19.2k, 38.4k, 57.6k, 115.2k are available!"<<std::endl; return false;
  }
  
  //Set the rate
  char comm[2]={'b','r'};//br
  uint8_t mess;//message
  //Set values
  to(baud,&mess,1);
  if(!setvalue(comm,1, &mess,true)) return false;
  
  //Modify the current setting according to the rate
  return ResetCommunication(communicator->getPortName(), rate);
}

bool UPP::SetDeviceAddress(const int address)
//Sets the new device address: 00 to 97 (regular); 98, 99 (global address with response)
{
  char comm[2]={'g','a'};//ga
  uint8_t mess[2];//message
  
  //Checking the correct setup
  if((address<0)||(address>99))
  {
    std::cerr<<"Error! UPP::SetDeviceAddress: the value \""<<address<<"\" should be between 0 and 99!\n";
    return false;
  }
  
  //Check does it need to work?
  int b,c,e,f,dev_addr;
  bool d;
  double a,g;
  if(!ReadParams(a, b, c, d, e, dev_addr,f, g)) return false;
  if (dev_addr==address) return true;
    
  //set the ratio correctly
  to(address,mess,2);
  if (!setvalue(comm, 2, mess,false,true)) return false;
  else
  {
    SetAddress(address);//Set the new address to communicate
    return true;
  }
}

//ABOUT-start
std::string UPP::GetReferenceNumber()
//Gets the reference number, Error: "ERROR"
{
    char comm[2]={'b','n'};//bn
    uint8_t mess[6];//message
    
    //Communicate
    if (!getvalue(comm,6,mess)) return std::string("ERROR");
    
    return tostr(mess,6);
}

std::string UPP::GetDeviceType()
//Returns with the device type or with error message
{
  char comm[2]={'n','a'};//na
  uint8_t mess[16];//message
  
  //Communicate
  bool retval=getvalue(comm,16,mess);

  if(!retval) return std::string("ERROR");//Comm. error
  return tostr(mess,16);
}

std::string UPP::GetSerialNumber()
//Returns the serial number of the detector
{
  char comm[2]={'s','n'};//sn
  uint8_t mess[5];//message
  
  //Communicate
  bool retval=getvalue(comm,5,mess);

  if(!retval) return std::string("ERROR");//Comm. error
  return tostr(mess,5);
}

std::string UPP::GetDeviceTypeSoftwareVersion()
//Returns with the device type (54) and the version of the software (MMYY)
{
  char comm[2]={'v','e'};//ve
  uint8_t mess[6];//message
  
  //Communicate
  bool retval=getvalue(comm,6,mess);

  if(!retval) return std::string("ERROR");//Comm. error
  return tostr(mess,6);
}

std::string UPP::GetCommunicationDetails()
//Returns with the details of the software: DD.MM.YY SOFT.VERSION
{
  char comm[2]={'v','c'};//vc
  uint8_t mess[14];//message
  
  //Communicate
  bool retval=getvalue(comm,14,mess);

  if(!retval) return std::string("ERROR");//Comm. error
  return tostr(mess,14);
}

std::string UPP::GetSoftwareDetails()
//Returns with the details of the software: DD.MM.YY SOFT.VERSION
{
  char comm[2]={'v','s'};//vc
  uint8_t mess[14];//message
  
  //Communicate
  bool retval=getvalue(comm,14,mess);

  if(!retval) return std::string("ERROR");//Comm. error
  return tostr(mess,14);
}

bool UPP::GetBasicRange(int &lowlimit, int &upperlimit)
//Returns the lower and upper limits of the basic range in the requested unit
{
  char comm[2]={'m','b'};//mb
  uint8_t mess[8];//message
  
  //Communicate
  bool retval=getvalue(comm,8,mess);
  
  if(retval)
  {
      //Set the temparature
      lowlimit=from(mess,4,false);
      upperlimit=from(mess+4,4,false);
  }
  
  //Transforming the message to valid output
  return retval;
}

//ABOUT-end

//private low-level methods
bool UPP::setvalue(char *command, const int length, uint8_t *message, const bool onlysend, const bool wait_extra)
//Send the command [2 letters] to addr with the length long message
{
    uint8_t *sbuffer= new uint8_t [5+length];//Sendbuffer
    uint8_t *rbuffer= new uint8_t [3];//Replybuffer
    //Set send buffer datasets
    sbuffer[0]=add[0];//Address 10 digits
    sbuffer[1]=add[1];//Address: last digits
    sbuffer[2]=command[0];//Command letters
    sbuffer[3]=command[1];
    for (int i=0; i<length; i++) sbuffer[4+i]=message[i];//Message
    
    //Sending the message
    bool retval=this->communicate(sbuffer,5+length,rbuffer,3, onlysend, wait_extra);//Returning value
    //Checking agreement: it should be "ok+CR"
    if(retval && (!onlysend))
    {
	retval=((rbuffer[0]=='o')&&(rbuffer[1]=='k')&&(rbuffer[2]==UPP::CR));
    }

    //Delete buffers
    delete [] sbuffer;
    delete [] rbuffer;
    
    return retval;
}

bool UPP::getvalue(char *command, const int length, uint8_t *message, bool limits)
//Get the actual value of command [2 letters] of addr. Its message is length long message. If limits are true, the limits are the question
{
    uint8_t *sbuffer= new uint8_t [(limits? 6 : 5)];//Sendbuffer
    uint8_t *rbuffer= new uint8_t [1+length];//Replybuffer
    //Set the send buffer datasets
    sbuffer[0]=add[0];//Address 10 digits
    sbuffer[1]=add[1];//Address: last digits
    sbuffer[2]=command[0];//Command letters
    sbuffer[3]=command[1];
    if (limits) sbuffer[4]='?';//Get the limits task
    
    //Send message, get the answer
    bool retval=this->communicate(sbuffer,(limits? 6 : 5), rbuffer,1+length);//Returning value
    //Writing the reply into message buffer
    if ((retval)&&(((rbuffer[0]=='n')&&(rbuffer[1]=='o'))||(rbuffer[0]==CR)))
    {
      retval=false;//Set return to false, if not allowed function requested
      usleep((int)(100000.0*TIMEOUT*5));//Wait some time to avoid reading of CR in the next circle
    }
    if (retval) {for(int i=0; i<length; i++) message[i]=rbuffer[i];}
    
    //dispose memory
    delete [] sbuffer;
    delete [] rbuffer;
    
    return retval;
}


bool UPP::communicate(uint8_t *sendbuffer, const int slen, uint8_t *replybuffer, const int rlen, const bool onlysend, const bool wait_extra)
{
    //Adding CR character
    sendbuffer[slen-1]=UPP::CR;
    
    //Writing it to the serial line
    int counterwhole=0;//Whole procedure counter
    int rep;//Local reply
    
    bool repeatfull=false;//Reapeat the full send and receive procedure
    do
    {
	repeatfull=false;
	int counter=0;//Local counter for send and reply
	bool repeatsend=false;//Repeat send procedure
	bool repeatreply=false;//Repeat reply procedure
	
	//Perform error checked writing
	do
	{
#ifdef DEBUG_UPP	    
	    std::cout<<"Sendbuffer:";
	    for(int i=0; i<slen; i++) std::cout<<i<<": "<<sendbuffer[i]<<std::endl;
#endif
	    //Send bytes
	    rep=communicator->Send(sendbuffer,slen);
	    
	    if(rep!=slen)
	    {
		repeatsend=true;
		//Communication error
		std::cerr<<"Sending is unsuccessful(UPP::communicate) on the serial line!\n";
		if(counter++<1) 
		{
		    std::cerr<<"Trying again!"<<std::endl;
		    usleep((int)(100000.0*WAITSUCCESS));//Wait some time until the next trial
		}
		else return false;
	    }
	}while(repeatsend);
	
       if (!onlysend)
       {	 
	if (wait_extra) usleep((int)(500000.0*TIMEOUT));
	//Perform error checked reading
	counter=0;
	do
	{
	    //Receiving the reply
	    rep=communicator->Receive(replybuffer,rlen);
#ifdef DEBUG_UPP	    
	    std::cout<<"Replybuffer:";
	    for(int i=0; i<rlen; i++) std::cout<<i<<": "<<replybuffer[i]<<std::endl;
#endif
	    if (rep<rlen)
	    {
	    
		//Possible reasons: 
		//timeout
		//SERIAL_ERROR --- try it igain
		repeatreply=true;
		usleep((int)(100000.0*WAITSUCCESS));//Wait some time until the next trial
		if(counter++==1)
		{
		    std::cerr<<"Communication error on the serial line (UPP::communicate) during the reading!\n";
		    if (counterwhole++==0) {std::cerr<<"Sending messages again\n"; repeatreply=false; repeatfull=true;} 
		    else return false;
		}
		
	    }
	}while(repeatreply);
	usleep((int)(100000.0*WAITSUCCESS));//Wait some time until the next trial
       }
    }while(repeatfull);
    
    if(onlysend) usleep((int)(1250000.0*TIMEOUT));

    return true;
}

void UPP::to(const int number, uint8_t *buffer, const int blength, const bool dechex)
//Converting int (number) to decimal(t)/hexa(f) digits (into blength long buffer)
{
  int div=1;//Divider
  int multiplier = (dechex ? 10 : 16);//Multiplier
  for(int i=blength-1; i>-1; i--)
  {
    int digit=(number%(multiplier*div))/div;//Decimal count system to hexadecimal
    *(buffer+i)=digit+( (digit<10) ? 48 : 55);//Converting to ascii
    div*=multiplier;
  }
  return;
}

int UPP::from(uint8_t* buffer, const int blength, const bool dechex)
//Converting the decimal(t)/hexa(t) digits containing in blength long buffer to integer
{
  std::stringstream ss;
  for(int i=0; i<blength; i++) ss<<buffer[i];
  int result;
  if (dechex) ss>>result; else ss>>std::hex>>result;
  return result;
}

std::string UPP::tostr(uint8_t* buffer, const int length)
//Converting the digits containing the buffer into string
{
  std::stringstream ss;
  for(int i=0; i<length; i++) ss<<buffer[i];
  return ss.str();
}