#include <algorithm>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <string>
#include <sstream>
#include "HV_CAEN_N1471.h"

template <typename T> std::ostream& operator<< (std::ostream &os, const std::vector<T> &t)
{
    if(!t.empty()) std::copy(t.begin(),t.end(), std::ostream_iterator<T>(os," "));
    return os;
}

void print_help(char *argv)
{
    std::cout<<"This program communicates with CAEN N1471 4 channel high-voltage units"<<std::endl;
    std::cout<<"Usage: "<<argv<<" COMMAND_PATTERN"<<std::endl;
    std::cout<<" where the COMMAND_PATTERN is one of the following (CHN corresponds to channel number: might be 0..3 for specific and 4 for all):"<<std::endl;
    std::cout<<"\tboot\t\t\t: sets basic configuration of the modules (current set 51 micA; 5kV max; +/-50V/s ramp; 10s overcurrent time; RAMP)"<<std::endl;
    std::cout<<"\tsetcfg CHN VOLTAGE\t: sets the operation voltage of channel CH to VOLTAGE"<<std::endl;
    std::cout<<"\tgetcfg CHN\t\t: get the operation voltage(s) already set of channel CH"<<std::endl;
    std::cout<<"\tactual CHN\t\t: get the actual voltage(s) related to channel CH"<<std::endl;
    std::cout<<"\thv CHN on/off\t: switch on/off the high voltage(s) of channel CHN"<<std::endl;
}

int main(int argc, char *argv[])
{
  if(argc==1)
  {
      print_help(argv[0]);
      return 1;
  }
  HV_CAEN_N1471 hc("caen_hv");
  bool generalchk=false;
  
  if(generalchk)
  {
    std::vector<double> mins,maxs;
    std::vector<int> digits;
    std::vector<bool> sw;
    
    //Set voltage related fields
    std::cout<<"Get the set voltage for each channel: "<<(hc.getVoltageSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"Set the HV to 123.456 V for each channel: "<<(hc.setVoltageSet(0,123.456,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the set voltage for each channel: "<<(hc.getVoltageSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"Set the HV to 0V for each channels"<<(hc.setVoltageSet(0,0.0,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the set voltage for each channel: "<<(hc.getVoltageSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"Get the set voltage properties for each channel: "<<(hc.getAboutVoltageSet(0,mins, maxs, digits, 4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMins: "<<mins<<" V"<<std::endl;
    std::cout<<"\tMaxs: "<<maxs<<" V"<<std::endl;
    std::cout<<"\tDigits: "<<digits<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Actual voltage related fields
    std::cout<<"Get the actual voltage for each channel: "<<(hc.getVoltageActual(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Set current related fields
    std::cout<<"Get the set current for each channel: "<<(hc.getCurrentSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" microA"<<std::endl;
    std::cout<<"Set the HV to 123.456 microA for each channel: "<<(hc.setCurrentSet(0,123.456,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the set current for each channel: "<<(hc.getCurrentSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" microA"<<std::endl;
    std::cout<<"Set the HV to 51 microA for each channel: "<<(hc.setCurrentSet(0,51.,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the set current for each channel: "<<(hc.getCurrentSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" microA"<<std::endl;
    std::cout<<"Get the set current properties for each channel: "<<(hc.getAboutCurrentSet(0,mins, maxs, digits, 4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMins: "<<mins<<" microA"<<std::endl;
    std::cout<<"\tMaxs: "<<maxs<<" microA"<<std::endl;
    std::cout<<"\tDigits: "<<digits<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Actual current related fields
    std::cout<<"Get the actual current for each channel: "<<(hc.getCurrentActual(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" microA"<<std::endl;
    std::cout<<"Get the actual current properties for each channel: "<<(hc.getAboutCurrentActual(0,digits, 4) ? "OK" : "ERROR")<<std::endl;
    //std::cout<<"\tHi(1)/low(0) resolution: "<<sw<<std::endl;
    std::cout<<"\tDigits: "<<digits<<std::endl;
    //std::cout<<"Set the actual current properties to high resolution for each channel: "<<(hc.setAboutCurrentActual(0,true, 4) ? "OK" : "ERROR")<<std::endl;
    //std::cout<<"Get the actual current properties for each channel: "<<(hc.getAboutCurrentActual(0,digits, 4) ? "OK" : "ERROR")<<std::endl;
    //std::cout<<"\tHi(1)/low(0) resolution: "<<sw<<std::endl;
    //std::cout<<"\tDigits: "<<digits<<std::endl;
    //std::cout<<"Set the actual current properties to low resolution for each channel: "<<(hc.setAboutCurrentActual(0,false, 4) ? "OK" : "ERROR")<<std::endl;
    //std::cout<<"Get the actual current properties for each channel: "<<(hc.getAboutCurrentActual(0,digits, 4) ? "OK" : "ERROR")<<std::endl;
    //std::cout<<"\tHi(1)/low(0) resolution: "<<sw<<std::endl;
    //std::cout<<"\tDigits: "<<digits<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Maximal voltage related fields
    std::cout<<"Get the maximal voltage for each channel: "<<(hc.getMaxVoltageSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"Set the maximal HV to 123.456 V for each channel: "<<(hc.setMaxVoltageSet(0,123.456,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal voltage for each channel: "<<(hc.getMaxVoltageSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"Set the maximal HV to 5600.0 V for each channel: "<<(hc.setMaxVoltageSet(0,5600.,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal voltage for each channel: "<<(hc.getMaxVoltageSet(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"Get the maximal voltage properties for each channel: "<<(hc.getAboutMaxVoltageSet(0,mins, maxs, digits, 4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMins: "<<mins<<" V"<<std::endl;
    std::cout<<"\tMaxs: "<<maxs<<" V"<<std::endl;
    std::cout<<"\tDigits: "<<digits<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Maximal voltage ramp up speed related fields
    std::cout<<"Get the maximal ramp up voltage speed for each channel: "<<(hc.getMaxVoltageUpramp(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V/s"<<std::endl;
    std::cout<<"Set the maximal ramp up speed to 123.456 V/s for each channel: "<<(hc.setMaxVoltageUpramp(0,123.456,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal ramp up voltage speed for each channel: "<<(hc.getMaxVoltageUpramp(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V/s"<<std::endl;
    std::cout<<"Set the maximal ramp up speed to 50 V/s for each channel: "<<(hc.setMaxVoltageUpramp(0,50.0,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal ramp up voltage speed for each channel: "<<(hc.getMaxVoltageUpramp(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V/s"<<std::endl;
    std::cout<<"Get the maximal ramp up voltage speed properties for each channel: "<<(hc.getAboutMaxVoltageUpramp(0,mins, maxs, digits, 4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMins: "<<mins<<" V/s"<<std::endl;
    std::cout<<"\tMaxs: "<<maxs<<" V/s"<<std::endl;
    std::cout<<"\tDigits: "<<digits<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Maximal voltage ramp down speed related fields
    std::cout<<"Get the maximal ramp down voltage speed for each channel: "<<(hc.getMaxVoltageDownramp(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V/s"<<std::endl;
    std::cout<<"Set the maximal ramp down speed to 123.456 V/s for each channel: "<<(hc.setMaxVoltageDownramp(0,123.456,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal ramp down voltage speed for each channel: "<<(hc.getMaxVoltageDownramp(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V/s"<<std::endl;
    std::cout<<"Set the maximal ramp down speed to 50.0 V/s for each channel: "<<(hc.setMaxVoltageDownramp(0,50.0,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal ramp down voltage speed for each channel: "<<(hc.getMaxVoltageDownramp(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V/s"<<std::endl;
    std::cout<<"Get the maximal ramp down voltage speed properties for each channel: "<<(hc.getAboutMaxVoltageDownramp(0,mins, maxs, digits, 4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMins: "<<mins<<" V/s"<<std::endl;
    std::cout<<"\tMaxs: "<<maxs<<" V/s"<<std::endl;
    std::cout<<"\tDigits: "<<digits<<std::endl;
    std::cout<<"Attempting to set the maximal ramp down speed to 0.0 V/s for each channel (test of undershoot): "<<(hc.setMaxVoltageDownramp(0,0.0,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Attempting to set the maximal ramp down speed to 1000.0 V/s for each channel (test of overshoot): "<<(hc.setMaxVoltageDownramp(0,1000.0,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Maximal overcurrent time related fields
    std::cout<<"Get the maximal overcurrent time for each channel: "<<(hc.getMaxOvercurrentTime(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" s"<<std::endl;
    std::cout<<"Set the maximal overcurrent time to 123.456 s for each channel: "<<(hc.setMaxOvercurrentTime(0,123.456,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal overcurrent time for each channel: "<<(hc.getMaxOvercurrentTime(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" s"<<std::endl;
    std::cout<<"Set the maximal overcurrent time to 10 s for each channel: "<<(hc.setMaxOvercurrentTime(0,10.0,4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the maximal overcurrent time for each channel: "<<(hc.getMaxOvercurrentTime(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" s"<<std::endl;
    std::cout<<"Get the maximal overcurrent time properties for each channel: "<<(hc.getAboutMaxOvercurrentTime(0,mins, maxs, digits, 4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMins: "<<mins<<" s"<<std::endl;
    std::cout<<"\tMaxs: "<<maxs<<" s"<<std::endl;
    std::cout<<"\tDigits: "<<digits<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Power down setup related fields
    std::cout<<"Get the power down setup for each channel: "<<(hc.getPwrDownSetup(0,sw,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tRAMP(1)/KILL(0): "<<sw<<std::endl;
    std::cout<<"Set the power down to KILL for each channel: "<<(hc.setPwrDownSetup(0,false, 4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the power down setup for each channel: "<<(hc.getPwrDownSetup(0,sw,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tRAMP(1)/KILL(0): "<<sw<<std::endl;
    std::cout<<"Set the power down to RAMP for each channel: "<<(hc.setPwrDownSetup(0,true, 4)? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the power down setup for each channel: "<<(hc.getPwrDownSetup(0,sw,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tRAMP(1)/KILL(0): "<<sw<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Polarity related fields
    std::cout<<"Get the polarity for each channel: "<<(hc.getPolarity(0,sw,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t+(1)/-(0): "<<sw<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Status related fields
    std::vector<bool> on,ramp_up,ramp_dwn, curr_ne_set, volt_overshot, volt_undershot, maxv_protect, off_bytrip, over_pwr, overheat, disabled_remote_and_off, kill_bypanel, interlck_bypanel, cal_error;
    std::cout<<"Get the status for each channel: "<<(hc.getStatus(0,on,ramp_up,ramp_dwn, curr_ne_set, volt_overshot, volt_undershot, maxv_protect, off_bytrip, over_pwr, overheat, disabled_remote_and_off, kill_bypanel, interlck_bypanel, cal_error,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tOn: "<<on<<std::endl;
    std::cout<<"\tRamp up: "<<ramp_up<<std::endl;
    std::cout<<"\tRamp down: "<<ramp_dwn<<std::endl;
    std::cout<<"\tActual current !=set: "<<curr_ne_set<<std::endl;
    std::cout<<"\tV>Vset+250V: "<<volt_overshot<<std::endl;
    std::cout<<"\tV<Vset-250V: "<<volt_undershot<<std::endl;
    std::cout<<"\tVmax protection ON: "<<maxv_protect<<std::endl;
    std::cout<<"\tOFF by TRIP: "<<off_bytrip<<std::endl;
    std::cout<<"\tP>1.7W: "<<over_pwr<<std::endl;
    std::cout<<"\tT>105C: "<<overheat<<std::endl;
    std::cout<<"\tDisabled(REMOTE+OFF switch): "<<disabled_remote_and_off<<std::endl;
    std::cout<<"\tKILL (switch): "<<kill_bypanel<<std::endl;
    std::cout<<"\tINTERLOCK (switch): "<<interlck_bypanel<<std::endl;
    std::cout<<"\tCalibration ERROR: "<<cal_error<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Switch on/off HV
    std::cout<<"Set on HV for each channel: "<<(hc.setHVOnOff(0,true,4) ? "OK" : "ERROR")<<std::endl;
    sleep(10);
    std::cout<<"Get the actual voltage for each channel: "<<(hc.getVoltageActual(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\t"<<mins<<" V"<<std::endl;
    std::cout<<"Get the status for each channel: "<<(hc.getStatus(0,on,ramp_up,ramp_dwn, curr_ne_set, volt_overshot, volt_undershot, maxv_protect, off_bytrip, over_pwr, overheat, disabled_remote_and_off, kill_bypanel, interlck_bypanel, cal_error,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tOn: "<<on<<std::endl;
    std::cout<<"\tRamp up: "<<ramp_up<<std::endl;
    std::cout<<"\tRamp down: "<<ramp_dwn<<std::endl;
    std::cout<<"\tActual current !=set: "<<curr_ne_set<<std::endl;
    std::cout<<"\tV>Vset+250V: "<<volt_overshot<<std::endl;
    std::cout<<"\tV<Vset-250V: "<<volt_undershot<<std::endl;
    std::cout<<"\tVmax protection ON: "<<maxv_protect<<std::endl;
    std::cout<<"\tOFF by TRIP: "<<off_bytrip<<std::endl;
    std::cout<<"\tP>1.7W: "<<over_pwr<<std::endl;
    std::cout<<"\tT>105C: "<<overheat<<std::endl;
    std::cout<<"\tDisabled(REMOTE+OFF switch): "<<disabled_remote_and_off<<std::endl;
    std::cout<<"\tKILL (switch): "<<kill_bypanel<<std::endl;
    std::cout<<"\tINTERLOCK (switch): "<<interlck_bypanel<<std::endl;
    std::cout<<"\tCalibration ERROR: "<<cal_error<<std::endl;
    std::cout<<"Switch off HV for each channel: "<<(hc.setHVOnOff(0,false,4) ? "OK" : "ERROR")<<std::endl;
    sleep(10);
    std::cout<<"Get the status for each channel: "<<(hc.getStatus(0,on,ramp_up,ramp_dwn, curr_ne_set, volt_overshot, volt_undershot, maxv_protect, off_bytrip, over_pwr, overheat, disabled_remote_and_off, kill_bypanel, interlck_bypanel, cal_error,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tOn: "<<on<<std::endl;
    std::cout<<"\tRamp up: "<<ramp_up<<std::endl;
    std::cout<<"\tRamp down: "<<ramp_dwn<<std::endl;
    std::cout<<"\tActual current !=set: "<<curr_ne_set<<std::endl;
    std::cout<<"\tV>Vset+250V: "<<volt_overshot<<std::endl;
    std::cout<<"\tV<Vset-250V: "<<volt_undershot<<std::endl;
    std::cout<<"\tVmax protection ON: "<<maxv_protect<<std::endl;
    std::cout<<"\tOFF by TRIP: "<<off_bytrip<<std::endl;
    std::cout<<"\tP>1.7W: "<<over_pwr<<std::endl;
    std::cout<<"\tT>105C: "<<overheat<<std::endl;
    std::cout<<"\tDisabled(REMOTE+OFF switch): "<<disabled_remote_and_off<<std::endl;
    std::cout<<"\tKILL (switch): "<<kill_bypanel<<std::endl;
    std::cout<<"\tINTERLOCK (switch): "<<interlck_bypanel<<std::endl;
    std::cout<<"\tCalibration ERROR: "<<cal_error<<std::endl;
    std::cout<<"Get the actual voltage for each channel: "<<(hc.getVoltageActual(0,mins,4) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    std::cout<<"Module related commands:"<<std::endl; 
    std::string a,b,c;
    int chnls=0,ser=0;
    bool ab,bb,cb;
    //About related fields
    std::cout<<"Get the module about: "<<(hc.getAbout(0,a,chnls,b,ser) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tModule name: "<<a<<std::endl;
    std::cout<<"\tSerial number: "<<ser<<std::endl;
    std::cout<<"\tNumber of channels: "<<chnls<<std::endl;
    std::cout<<"\tFirmware release: "<<b<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Interlock related fields
    std::cout<<"Get the interlock status of the module: "<<(hc.getInterlock(0,ab,bb) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMode "<<(ab ? "OPEN" :"CLOSED")<<std::endl;
    std::cout<<"\tStatus "<<(bb ? "YES" : "NO")<<std::endl;
    std::cout<<"Setting the interlock to OPEN: "<<(hc.setInterlock(0,true) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the interlock status of the module: "<<(hc.getInterlock(0,ab,bb) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMode "<<(ab ? "OPEN:OK" :"CLOSED:ERROR!!!")<<std::endl;
    std::cout<<"Setting the interlock back to CLOSED: "<<(hc.setInterlock(0,false) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"Get the interlock status of the module: "<<(hc.getInterlock(0,ab,bb) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMode "<<(ab ? "OPEN:ERROR!!!!!" :"CLOSED:OK")<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Communication mode related fields
    std::cout<<"Get the communication status of the module: "<<(hc.getCommunication(0,ab,bb) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tMode "<<(ab? "REMOTE":"LOCAL")<<std::endl;
    std::cout<<"\tTermination "<<(bb?"ON":"OFF")<<std::endl;
    std::cout<<"---------------------"<<std::endl;
    //Alarm related fields
    std::cout<<"Get the alarm status of the module: "<<(hc.getAlarms(0,sw,ab,bb,cb) ? "OK" : "ERROR")<<std::endl;
    std::cout<<"\tAlarms of each channel: "<<sw<<std::endl;
    std::cout<<"\tPower fail: "<<ab<<std::endl;
    std::cout<<"\tOver power: "<<bb<<std::endl;
    std::cout<<"\tHV clock fail: "<<cb<<std::endl;
    std::cout<<"Clearing the alarm: "<<(hc.setClrAlarm(0) ? "OK" : "ERROR")<<std::endl;
    //TODO: check minimal and maximal limits
  }
  else
  {
      //Command interpreting
      std::string first(argv[1]);
      std::transform(first.begin(),first.end(),first.begin(),::tolower);
      if(first=="boot")
      {
          //Sets basic configuration of the modules (current set 51 micA; 5kV max; +/-50V/s ramp; 10s overcurrent time; RAMP)
          std::cout<<(hc.setCurrentSet(0,51.0, 4)&&hc.setMaxVoltageSet(0,5000.0, 4)&&hc.setMaxVoltageUpramp(0,50.0, 4)&&hc.setMaxVoltageDownramp(0,50.0, 4)&&hc.setMaxOvercurrentTime(0,10.0, 4)&&hc.setPwrDownSetup(0,true, 4) ? "OK" : "ERROR")<<std::endl;
      }
      else if(first=="setcfg")
      {
          //Sets the operation voltage of channel CH to VOLTAGE
          if(argc<4)
          {
              std::cerr<<"ERROR: This command requires more parameters"<<std::endl;
              print_help(argv[0]);
              return 1;
          }
          std::string second(argv[2]);
          std::string third(argv[3]);
          std::cout<<(hc.setVoltageSet(0,std::stod(std::string(argv[3])), std::stoi(std::string(argv[2]))) ? "OK" : "ERROR")<<std::endl;
      }
      else if(first=="getcfg")
      {
          //Gets the operation voltage(s) already set of channel CH
          if(argc<3)
          {
              std::cerr<<"ERROR: This command requires more parameters"<<std::endl;
              print_help(argv[0]);
              return 1;
          }
          std::vector<double> voltages;
          bool retval=hc.getVoltageSet(0,voltages, std::stoi(std::string(argv[2])));
          if(retval) std::cout<<"OK "<<voltages<<" V"<<std::endl;
          else
          {
              std::cerr<<"ERROR"<<std::endl;
              return 2;
          }
          
      }
      else if(first=="actual")
      {
          //Gets the actual voltage(s) related to channel CH
          if(argc<3)
          {
              std::cerr<<"ERROR: This command requires more parameters"<<std::endl;
              print_help(argv[0]);
              return 1;
          }
          std::vector<double> voltages;
          bool retval=hc.getVoltageActual(0,voltages,std::stoi(std::string(argv[2])));
          if(retval) std::cout<<"OK "<<voltages<<" V"<<std::endl;
          else
          {
              std::cerr<<"ERROR"<<std::endl;
              return 2;
          }
      }
      else if(first=="hv")
      {
          //Switch on/off the high voltage(s) of channel CHN"<<std::endl;
          if(argc<4)
          {
              std::cerr<<"ERROR: This command requires more parameters"<<std::endl;
              print_help(argv[0]);
              return 1;
          }
          std::string second(argv[2]);
          std::string third(argv[3]);
          std::transform(third.begin(),third.end(),third.begin(),::tolower);
          std::cout<<(hc.setHVOnOff(0,third=="on", std::stoi(std::string(argv[2]))) ? "OK" : "ERROR")<<std::endl;
      }
      else
      {
          std::cerr<<"ERROR: The command you provided (\""<<argv[1]<<"\") does not correspond any of the valid commands"<<std::endl;
          print_help(argv[0]);
          return 1;
      }
      return 0;
  }
 return 0;
}

