#include "kd48p.h"

//TESTED in 2013.8.15

//Get actual changing values
bool KD48P::getRegVals(double &processval, double &setpoint, double &actuator)
//Getting actual values of the process, setpoint and actuator signals. It returns with true, if communication error has not been done
{
    uint16_t pv, sp, y;//PV, SP, Y
    bool ret=((mb->getValue(slvno,(uint16_t) 0x80, (uint16_t) 0x01,&pv,true))&&
    (mb->getValue(slvno,(uint16_t) 0x100, (uint16_t) 0x01,&sp,true))&&
    (mb->getValue(slvno,(uint16_t) 0x101, (uint16_t) 0x01,&y,true)));//Get PV,SP and Y
    //Converting
    if(ret)
    {
	processval=u2i(pv);
	setpoint  =u2i(sp);
	actuator  =u2i(y);
    }
    //Returning with the results
    return ret;
}

bool KD48P::getStatus(Vec<int> &alarms, FINISHEDPRG &finishedprogno, int &executedprogno, ECODE &errorcode)
//Getting the status of alarms, the finished and executed program's number, finally the error code. It returns with comm error, if it happens
{
    uint16_t response;
    bool retval=mb->getValue(slvno,(uint16_t) 0x90, (uint16_t) 0x01, &response,true);//Asking the content of the status register
    if(retval)
    {
	//Alarms
	alarms[0]=response&0x0001;//ALARM2
	alarms[1]=response&0x0002;//ALARM3
	alarms[2]=response&0x0004;//ALARM4
	alarms[3]=response&0x0008;//ALARM5
	//Finished programsteps
	uint16_t fin=(response>>4)&0xf;
	finishedprogno=(FINISHEDPRG)fin;
	//Finished program number
	executedprogno=(int)((response>>8)&0xf);
	//Error code
	uint16_t ec=(response>>12)&0xf;
	errorcode=(ECODE) ec;
    }
    return retval;
}

//Time-program controlls
bool KD48P::GetActualProgramNumber(int *prog_no)
//Returns with the actually set program number
{
  uint16_t result;//Result for the asked register
  bool retval=mb->getValue(slvno, 0x102, 0x01, &result);
  if(retval)
  {
    *prog_no=u2i(result);
    return true;
  }
  else return false;
}
bool KD48P::OnOff(bool stop, bool feedback)
//This method switch off (stop=true) or switch on (stop=false) the temperature controller.
//If feedback is true, it returns with success, if the status register changed properly. If false, it is waiting for only the accept.
{
  bool retval=true;//Returning value
  uint16_t result;//Result for the asked register
  uint16_t Tval=(stop ? 0x00 : 0x01);//True value of the requested status
  //Asking the current status
  retval=mb->getValue(slvno, 0x2ff, 0x01, &result, false);
  
  if(retval)
  {
    if(result==Tval) return true;//Actually it has been set -> we should not have to do anything
    else                        //We should set the requested value
    {
      retval=mb->setValue(slvno, 0x2ff, Tval);//Set it to the requested value
      if(retval&&feedback) //If real feedback is needed
      {
	retval=mb->getValue(slvno, 0x2ff, 0x01, &result, false);//Get actual value
	if (retval&&(result==Tval)) return true;//It went into proper state
	else return false;//Not, or communication trouble
      }
      else return retval;
    }
  }
  return false;//Returning communication error
}

bool KD48P::SetProgram(int program_no, int subprogram, int time_before, int heat_speed, int target_temperature, int keeping_time, bool final_term)
//This method sets the control program for the "program_no", and subprogram number, parameters are the time before in min, the heat speed in C/hour units (FREE-0, END (prog) - -1), the target temperature in C, the keeping time in mins, finally, does this the final term of the program?
{
  //Check conformity
  if((program_no<0)||(program_no>MAX_PROG_NO)) return false;
  if((subprogram<0)||(subprogram>MAX_SPRG_NO)) return false;
  if((time_before<0)||(time_before>5999)) return false;
  if((heat_speed<-1999)||(heat_speed>3999)) return false;
  int hilim,lolim;//Get limits for the temperature
  if(!GetInputRangeConfig(lolim,hilim)) return false;
  if((target_temperature<lolim)||(target_temperature>hilim)) return false;
  if((keeping_time<0)||(keeping_time>5999)) return false;
  uint16_t fi_add=0x0;//Final address
  if(final_term)
  {
    if(subprogram!=MAX_SPRG_NO) fi_add=0x202+i2u(program_no)*0x10+i2u(subprogram+1)*0x3;
    else final_term=false;//This is the last subprogram of the last program --- it does not need anything to do
  }
  //Check is OK, rearranging the heat speed
  uint16_t heat_ratio=i2u(heat_speed);
  uint16_t value=0x0;
  bool retval=true;//Returning value
  if((subprogram==0)&&retval)
  {
    mb->setValue(slvno, 0x201+i2u(program_no)*0x10, i2u(time_before));//Set pre-time
    retval=mb->getValue(slvno, 0x201+i2u(program_no)*0x10, 0x1,&value);//Check its value
    if(retval) retval= (value==i2u(time_before));//Set the returning value
  }
  if((final_term)&&retval) 
  {
    mb->setValue(slvno, fi_add, i2u(END));//Set end switch
    retval=mb->getValue(slvno, fi_add, 0x1, &value);//Check its value
    if(retval) retval= (value==i2u(END));//Set the returning value
  }
  if (retval) 
  {
    mb->setValue(slvno, 0x202+i2u(program_no)*0x10+i2u(subprogram)*0x3, heat_ratio);//Set heat ratio
    retval=mb->getValue(slvno, 0x202+i2u(program_no)*0x10+i2u(subprogram)*0x3,0x1, &value);//Check its value
    if(retval) retval= (value==heat_ratio);//Set the returning value
  }
  if (retval) 
  {
    mb->setValue(slvno, 0x203+i2u(program_no)*0x10+i2u(subprogram)*0x3, i2u(target_temperature));//Set target temperature
    retval=mb->getValue(slvno, 0x203+i2u(program_no)*0x10+i2u(subprogram)*0x3,0x1, &value);//Check its value
    if(retval) retval= (value==i2u(target_temperature));//Set the returning value
  }
    if (retval) 
  {
    mb->setValue(slvno, 0x204+i2u(program_no)*0x10+i2u(subprogram)*0x3, i2u(keeping_time));//The controller keep controlled the temperature during this time
    retval=mb->getValue(slvno, 0x204+i2u(program_no)*0x10+i2u(subprogram)*0x3,0x1, &value);//Check its value
    if(retval) retval= (value==i2u(keeping_time));//Set the returning value
  }
  return retval;
}

bool KD48P::GetProgram(int program_no, int subprogram, int *time_before, int *heat_speed, int *target_temperature, int *keeping_time, bool *final_term)
//This method get the control settings for "program_no", and subprogram number, the returning parameters are the time before in min, the heat speed in C/hour units (FREE-0, END (prog) - -0.1), the target temperature in C, the keeping time in mins, finally, does this the final term of the program?
{
  //Check conformity
  if((program_no<0)||(program_no>MAX_PROG_NO)) return false;
  if((subprogram<0)||(subprogram>MAX_SPRG_NO)) return false;
  //Get time before: it will be the same as for all of the "program"
  uint16_t result;//Result for the asked register
  if(!mb->getValue(slvno, 0x201+program_no*0x10, 0x01, &result)) return false;//Get time before
  *time_before=u2i(result);
  if(!mb->getValue(slvno, 0x202+program_no*0x10+subprogram*0x3, 0x01, &result)) return false;//Get heat ratio
  *heat_speed=u2i(result);//Set heat ratio
  if(!mb->getValue(slvno, 0x203+program_no*0x10+subprogram*0x3, 0x01, &result)) return false;//Get target temperature
  *target_temperature=u2i(result);//Set target temperature
  if(!mb->getValue(slvno, 0x204+program_no*0x10+subprogram*0x3, 0x01, &result)) return false;//Get keeping time
  *keeping_time=u2i(result);//Set the keeping time
  if(subprogram==MAX_SPRG_NO) *final_term=true;
  else
  {
    if(!mb->getValue(slvno, 0x202+program_no*0x10+(subprogram+1)*0x3, 0x01, &result)) return false; 
    *final_term= (u2i(result)==-1);
  }
  return true;
}

//Properties

bool KD48P::ResetAlarmProperties()
//Reset the alarm properties
{
  bool retval=true;//Returning value
  for(uint16_t address=0x103; address<0x10b; address++) if(retval) retval=mb->setValue(slvno, address, 0x0);//Sets the "A2", "A2h", "A3", "A3h", "A4", "A4h", "A5", "A5h"
  return retval;
}

//PID 

bool KD48P::SetPID(double gain, int integration_time, int derivation_time, int offset, int deadzone)
//Set the PID parameters: P=100%/gain (0.0~100.0%), integration time(0..1000s), derivation time(0..1000s), offset reset value (+/-1000), hysteresis(0.1000) 
{
  if((gain<0.0)||(gain>100.0)) return false;
  if((integration_time<0)||(integration_time>1000)) return false;
  if((derivation_time<0)||(derivation_time>1000)) return false;
  if((offset<-1000)||(offset>1000)) return false;
  if((deadzone<0)||(deadzone>1000)) return false;
  return (mb->setValue(slvno, 0x110, i2u((int)(gain*10))) && mb->setValue(slvno, 0x111, i2u(integration_time))&& mb->setValue(slvno, 0x112, i2u(derivation_time))&& mb->setValue(slvno, 0x113, i2u(offset))&& mb->setValue(slvno, 0x114, i2u(deadzone)));//Set the "Gain", "Int", "dEr", "mrES", "dzon" registers
}

bool KD48P::GetPID(double *gain, int *integration_time, int *derivation_time, int *offset, int *deadzone)
//Get the PID parameters: P=100%/gain (0.0~100.0%), integration time(0..1000s), derivation time(0..1000s), offset reset value (+/-1000), hysteresis(0.1000)
{
  uint16_t value;//Stores the received data
  bool retval=mb->getValue(slvno, 0x110, 0x01, &value);//get the "Gain" register
  if(retval)
  {
    *gain=((double)u2i(value))/10.0;//gain
    retval=mb->getValue(slvno, 0x111, 0x01, &value);//get the "Int" register
  }
  if(retval)
  {
    *integration_time=u2i(value);//Integration time
    retval=mb->getValue(slvno, 0x112, 0x01, &value);//get the "dEr" register
  }
  if(retval)
  {
    *derivation_time=u2i(value);//Derivation time
    retval=mb->getValue(slvno, 0x113, 0x01, &value);//get the "mrEs" register
  }
  if(retval)
  {
    *offset=u2i(value);//Offset
    retval=mb->getValue(slvno, 0x114, 0x01, &value);//get the "dzon" register
  }
  if(retval)
  {
     *deadzone=u2i(value);//hysteresis
     return true;
  }
  else return false;
}

//Configuration

bool KD48P::SetInputConfig(TC *thermocouple, bool *scalable, bool strong_filter, int decimal_point, int shift, int scalable_lo, int scalable_hi)
//Sets the input configuration parameters: thermocouple || (scalable &&decimal point(0..3)) input, shift(+/-1000), low and high scalable values (-1999..4000)
{
  //First of all, check the range, and get them to be consistent
  if((decimal_point>3)||(decimal_point<0)) return false;
  if((shift>1000)||(shift<-1000)) return false;
  if((scalable_lo<-1999)||(scalable_lo>4000)) return false;
  if((scalable_hi<-1999)||(scalable_hi>4000)) return false;
  if((thermocouple!=NULL)&&(scalable!=NULL)) return false;
  if(scalable_hi<scalable_lo)
  {
    int temp=scalable_lo;
    scalable_lo=scalable_hi;
    scalable_hi=temp;
  }
  
  uint16_t value= strong_filter ? 0x8 : 0x0;//Value to be set
  if(thermocouple!=NULL)
  {
    if(*thermocouple-6<0) value+=(uint16_t) *thermocouple;//Set the correct thermocouple
    else value+=(uint16_t) (*thermocouple-6)+0x10;
  }
  else if(scalable!=NULL)
  {
    value+= ( *scalable ? 0x6 : 0x7);//Set type
    for(int i=0; i<decimal_point; i++) value+=0x10;//Set decimal digits
  }
  //Communicate with controller
  bool retval=mb->setValue(slvno, 0x117, value);//set the "inP" register
  if(retval) retval=mb->setValue(slvno, 0x118, i2u(shift));//set the "SHFt" register
  if(retval) retval=mb->setValue(slvno, 0x119, i2u(scalable_lo));//set the "PvLo" register
  if(retval) retval=mb->setValue(slvno, 0x11a, i2u(scalable_hi));//set the "PvHi" register
  return retval;
}

bool KD48P::GetInputConfig(bool *tcs, TC *thermocouple, bool *scalable, bool *strong_filter, int *decimal_point, int *shift, int *scalable_lo, int *scalable_hi)
//Gets the input and related parameters (inP, SHFt, PvLo, PvHi), parameters are: tcs: t - thermocouple, f- scalable, decimal point will set, the shift, scalable low and high will be shown anyway.in the case of scalable 0..20mA (t) 4..20 mA(f)
{
  //Communicate with controller
  uint16_t value;//Stores the communication value
  bool retval=mb->getValue(slvno, 0x11a, 0x01, &value);//get the "PvHi" register
  if (retval) 
  {
    *scalable_hi=u2i(value);
    retval=mb->getValue(slvno, 0x119, 0x01, &value);//get the "PvLo" register
  }
  if (retval)
  {
    *scalable_lo=u2i(value);
    retval=mb->getValue(slvno, 0x118, 0x01, &value);//get the "SHFt" register
  }
  if (retval) 
  {
    *shift=u2i(value);
    retval=mb->getValue(slvno, 0x117, 0x01, &value);//get the "inP" register
  }
  //Sorting the parameters
  if(retval)
  {
    *strong_filter= (value>>3)&0x1;//Set filter
    if (((value>>1)&0x3 )== 0x3)
    {
      //This is the scalable input
      *tcs=false;
      *scalable= ((value&0x1)==0x0);//Set the scalable input
      *decimal_point=u2i((value>>4)&0x3);//Get decimal point
    }
    else
    {
      //This will be thermocouple
      *tcs=true;
      if((value>>4)&0x1) *thermocouple=(TC) (6+ (int)(value&0x7));//Upper part
      else *thermocouple=(TC)(value&0x7);//Down part
    }
  }
  return retval;
}

bool KD48P::SetInputRangeConfig(int low, int high)
//Sets the input range (-1999..4000)
{
  //Checking the range
  if((low<-1999)||(low>4000)) return false;
  if((high<-1999)||(high>4000)) return false;
  if(low>high) {int temp=low; low=high; high=temp;}
  return (mb->setValue(slvno, 0x11b, i2u(low)) && mb->setValue(slvno, 0x11c, i2u(high)));//set the "SPLo" and "SPHi" registers
}

bool KD48P::GetInputRangeConfig(int &low, int &high)
//Gets the input range
{
  uint16_t value;
  bool retval=mb->getValue(slvno, 0x11b, 0x01, &value);//get the "SPLo" register
  if(retval)
  {
     low=u2i(value);
     retval=mb->getValue(slvno, 0x11c, 0x01, &value);//get the "SPHi" register
     if(retval) high=u2i(value);
  }
  return retval;
}

bool KD48P::SetRegulatorConfig(bool cool_heat, bool str, int y_period, int y_lag, int y_lo, int y_hi)
//Sets the actuator: cooling(t) or heating(f), solid state relay output or regular, time between full off and on state of the output device (0..255), lag in y of the device (0..100), y_lo>y -> y=0 and y_hi<y -> y=100 (0..100 for hi and low)
{
  //Check conformity
  if((y_period<0)||(y_period>255)) return false;
  if((y_lag<0)||(y_lag>100)) return false;
  if((y_lo<0)||(y_lo>100)) return false;
  if((y_hi<0)||(y_hi>100)) return false;
  if(y_lo>y_hi) {int temp=y_lo; y_lo=y_hi; y_hi=temp;}
  //Set values
  uint16_t value= cool_heat ? 0x4 :0x0;//Set cooling or heating
  value+= str ? 0x1 : 0x0;//Solid state relay at analog output
  return (mb->setValue(slvno, 0x11d, value)&&mb->setValue(slvno, 0x11e, 0x0)&&mb->setValue(slvno, 0x11f, i2u(y_period))&&mb->setValue(slvno, 0x120, i2u(y_lag))&&mb->setValue(slvno, 0x121, i2u(y_lo))&&mb->setValue(slvno, 0x122, i2u(y_hi)));//set "out", "EPid", "Yt", "Yd'", "YLo", "YHi" registers
}

bool KD48P::GetRegulatorConfig(bool *cool_heat, bool *str, int *y_period, int *y_lag, int *y_lo, int *y_hi)//Gets the actuator setup: cooling(t) or heating(f), solid state relay output or regular, time between full off and on state of the output device, lag in y of the device, y_lo>y -> y=0 and y_hi<y -> y=100
{
  uint16_t value;//Stores the received data
  bool retval=mb->getValue(slvno, 0x11d, 0x01, &value);//get the "out" register
  if (retval)
  {
    *str=((value&0x3)==0x1);//Solid state relay
    *cool_heat=(((value>>2)&0x1)==0x1);//cooling/heating
    //We do not take care with "EPid" register
    retval=mb->getValue(slvno, 0x11f, 0x01, &value);//get the "Yt" register
  }
  if(retval)
  {
    *y_period=u2i(value);//Time between fully open and close
    retval=mb->getValue(slvno, 0x120, 0x01, &value);//get the "Yd'" register
  }
  if(retval)
  {
    *y_lag=u2i(value);//The small y value, which does not cause any move of the real device
    retval=mb->getValue(slvno, 0x121, 0x01, &value);//get the "YLo" register
  }
  if(retval)
  {
    *y_lo=u2i(value);//The lower setpoint, where y< values should taken as 0
    retval=mb->getValue(slvno, 0x122, 0x01, &value);//get the "YHi" register
  }
  if(retval)
  {
     *y_hi=u2i(value);//The upper setpoint, where y> values should taken as 100
     return true;
  }
  else return false;
  
}

bool KD48P::SetAnalogOutConfig(bool analog_out, int signal_pv_y_sp, int analog_out_lo, int analog_out_hi)
//analog output (0..t 4..f mA), the output signal(ideal signal-1, actuator-2, input signal-3), the corresponding minimum and maximum values (-1999~3999)of the output (true: the number corresponds to the 0 and the 20mA values)
{
  //Check conformity
  if((signal_pv_y_sp<1)||(signal_pv_y_sp>3)) return false;
  if((analog_out_lo<-1999)||(analog_out_lo>3999)) return false;
  if((analog_out_hi<-1999)||(analog_out_hi>3999)) return false;
  if(analog_out_lo>analog_out_hi) {int temp=analog_out_hi; analog_out_hi=analog_out_lo; analog_out_lo=temp;}
  return (mb->setValue(slvno, 0x123, 0x10+0x20*signal_pv_y_sp+ (analog_out ? 0x80 : 0x0))&&mb->setValue(slvno, 0x124,i2u(analog_out_lo))&&mb->setValue(slvno, 0x125,i2u(analog_out_hi)));//Sets the "dLin", "LiLo", "LiHi" registers
}

bool KD48P::GetAnalogOutConfig(bool *analog_out, int *signal_pv_y_sp, int *analog_out_lo, int *analog_out_hi)
//analog output (0..t 4..f mA), the output signal(ideal signal, actuator, input signal), the corresponding minimum and maximum values of the output (true: the number corresponds to the 0 and the 20mA values)
{
  uint16_t value;//Stores the received data
  bool retval=mb->getValue(slvno, 0x123, 0x01, &value);//get the "dLin" register
  if (retval)
  {
    *signal_pv_y_sp=(int)((value>>5)&0x3);//The analog signal type: pv,y,sp
    *analog_out=(((value>>7)&0x1)==0x1);//analog output
    retval=mb->getValue(slvno, 0x124, 0x01, &value);//get the "LiLo" register
  }
  if(retval)
  {
    *analog_out_lo=u2i(value);//The number corresponds to the lowest value of the analog output
    retval=mb->getValue(slvno, 0x125, 0x01, &value);//get the "YHi" register
  }
  if(retval)
  {
     *analog_out_hi=u2i(value);//The number corresponds to the highest value of the analog output
     return true;
  }
  else return false;
}

bool KD48P::ResetAlarmConfig()
//Set the alarm relay -- here, I set only the 0 setting for all of them --- it should be implemented if it needed
{
  bool retval=true;
  uint16_t address=0x126;//Starting point
  for(uint16_t i=0; i<5; i++) if(retval) retval=mb->setValue(slvno, address++,0x0);//Sets the "A2d", "A3d", "A4d", "A5d", "A23L" registers
  return retval;
}

bool KD48P::About(std::string *prog_ver)
//Returns with type and version separated by semicolon -- only KD48P is implemented
{
  uint16_t value;//Stores the received data
  bool retval=mb->getValue(slvno, 0x10, 0x01, &value);//get the "Ontc" register
  if (retval)
  {
    std::stringstream ss;
    if((value>>8)>=0x1a)//KD48P
    {
      ss<<"KD48P-";
      ss<<u2i((value>>8)-0x1a)<<";";//Set regulator type
      ss<<u2i(value&0xff);//Set program version
      ss>>*prog_ver;
      return true;
    }
    else //Other type
    {
      *prog_ver=std::string("ERROR: this is not a KD48P regulator!");
      return false;
    }
    return true;
  }
  else return false;
}


//Bulky configuration reading
bool KD48P::readConfig(std::ostream &os)
//Reads the configuration part of the regulator and writes it into the output stream
{
  
  //Reading the registers
  this->getwrite(os,"Type/version",0x10);//Get the type/version
  this->getwrite(os,"PV",0x80);//Get the actual temperature
  this->getwrite(os,"Status register",0x90);//Get the status register
  this->getwrite(os,"SP",0x100);//Get the SP
  this->getwrite(os,"Y",0x101);//Get the Y
  this->getwrite(os,"PrFL",0x102);//Get the PrFL
  this->getwrite(os,"A2",0x103);//Get the A2
  this->getwrite(os,"A2h",0x104);//Get the A2h
  this->getwrite(os,"A3",0x105);//Get the A3
  this->getwrite(os,"A3h",0x106);//Get the A3h
  this->getwrite(os,"A4",0x107);//Get the A4
  this->getwrite(os,"A4h",0x108);//Get the A4h
  this->getwrite(os,"A5",0x109);//Get the A5
  this->getwrite(os,"A5h",0x10A);//Get the A5h
  this->getwrite(os,"dE-r",0x10B);//Get the dE-r
  this->getwrite(os,"dE-L",0x10C);//Get the dE-L
  this->getwrite(os,"dE-H",0x10D);//Get the dE-H
  //-----------------------------------------------------
  this->getwrite(os,"Gain",0x110);//Get the Gain
  this->getwrite(os,"Int",0x111);//Get the Int
  this->getwrite(os,"dEr",0x112);//Get the dEr
  this->getwrite(os,"mrEs",0x113);//Get the mrEs
  this->getwrite(os,"dzon",0x114);//Get the dzon
  //-------------------------------------------------------
  this->getwrite(os,"inp",0x117,BIN);//Get the inp
  this->getwrite(os,"SHFt",0x118);//Get the SHFt
  this->getwrite(os,"PvLo",0x119);//Get the PvLo
  this->getwrite(os,"PvHi",0x11A);//Get the PvHi
  this->getwrite(os,"SPLo",0x11B);//Get the SPLo
  this->getwrite(os,"SPHi",0x11C);//Get the SPHi
  this->getwrite(os,"out",0x11D,BIN);//Get the out
  this->getwrite(os,"EPid",0x11E,BIN);//Get the EPid
  this->getwrite(os,"Yt",0x11F);//Get the Yt
   this->getwrite(os,"Yd'",0x120);//Ge the Yd'
  this->getwrite(os,"YLo",0x121);//Get the YLo
  this->getwrite(os,"YHi",0x122);//Get the YHi
  this->getwrite(os,"dLin",0x123,BIN);//Get the dLin
  this->getwrite(os,"LiLo",0x124);//Get the LiLo
  this->getwrite(os,"LiHi",0x125);//Get the LiHi
  this->getwrite(os,"A2d",0x126,BIN);//Get the A2d
  this->getwrite(os,"A3d",0x127,BIN);//Get the A3d
  this->getwrite(os,"A4d",0x128,BIN);//Get the A4d
  this->getwrite(os,"A5d",0x129,BIN);//Get the A5d
  this->getwrite(os,"A23L",0x12A,BIN);//Get the A23L
  this->getwrite(os,"con",0x12B,BIN);//Get the con
  this->getwrite(os,"Adr",0x12C);//Get the Adr
  this->getwrite(os,"dPrG",0x12D,BIN);//Get the dPrG
//  this->getwrite(os,"Ontc",0x12E);//Get the Ontc
//  this->getwrite(os,"d.FrE",0x12F);//Get the d.FrE

//----------------------------------------
  this->getwrite(os,"PreTime",0x201);//Get the PreTime
  this->getwrite(os,"1ramp",0x202);//Get the 1st ramp
  this->getwrite(os,"1SP",0x203);//Get the 1st SP
  this->getwrite(os,"1soak",0x204);//Get the 1st soak
  this->getwrite(os,"2ramp",0x205);//Get the 2nd ramp
  this->getwrite(os,"2SP",0x206);//Get the 2nd SP
  this->getwrite(os,"2soak",0x207);//Get the 2nd soak
  this->getwrite(os,"3ramp",0x208);//Get the 3rd ramp
  this->getwrite(os,"3SP",0x209);//Get the 3rd SP
  this->getwrite(os,"3soak",0x20A);//Get the 3rd soak
  this->getwrite(os,"4ramp",0x20B);//Get the 4th ramp
  this->getwrite(os,"4SP",0x20C);//Get the 4th SP
  this->getwrite(os,"4soak",0x20D);//Get the 4th soak
//----------------------------------------
  this->getwrite(os,"PreTime",0x211);//Get the PreTime
  this->getwrite(os,"1ramp",0x212);//Get the 1st ramp
  this->getwrite(os,"1SP",0x213);//Get the 1st SP
  this->getwrite(os,"1soak",0x214);//Get the 1st soak
  this->getwrite(os,"2ramp",0x215);//Get the 2nd ramp
  this->getwrite(os,"2SP",0x216);//Get the 2nd SP
  this->getwrite(os,"2soak",0x217);//Get the 2nd soak
  this->getwrite(os,"3ramp",0x218);//Get the 3rd ramp
  this->getwrite(os,"3SP",0x219);//Get the 3rd SP
  this->getwrite(os,"3soak",0x21A);//Get the 3rd soak
  this->getwrite(os,"4ramp",0x21B);//Get the 4th ramp
  this->getwrite(os,"4SP",0x21C);//Get the 4th SP
  this->getwrite(os,"4soak",0x21D);//Get the 4th soak
  
  return true;
}

bool KD48P::getwrite(std::ostream &os, std::string label, uint16_t regtoread, FORMAT frm)
//This method read the content of the register and writes into stream 
{
  bool retval=true;//Returning value
  uint16_t result;//Result for the asked register
  int counter=0;//Loop counter
  do
  {
    retval=mb->getValue(slvno, regtoread, 0x01, &result, false);//Reading the requested value
  }while((counter++<4)&&(!retval));
  os<<label<<": "<<(retval? "OK " : "Trouble ");
  if (retval)
  {
    switch(frm)
    {
      case(DECIMAL): os<<std::dec<<u2i(result)<<std::endl; break;//Decimal
      case(HEXA):    os<<std::hex<<result<<std::dec<<std::endl; break;//hexa
      case(BIN):     os<<((result>>7)&0x1)<<((result>>6)&0x1)<<((result>>5)&0x1)<<((result>>4)&0x1)<<((result>>3)&0x1)<<\
	((result>>2)&0x1)<<((result>>1)&0x1)<<(result&0x1)<<std::endl;
    }
    return true;
  }
  else return false;
}

//    bool writeConfig();//Writes the configuration part of the regulator
