#include <sstream>
#include "Thermometer.h"
#include "Modbus.h"

#ifndef KD48P_H
#define KD48P_H

//This class is tested successfully in 2013.08.15

enum FINISHEDPRG { OFF, PRE, FIRST_RAMP, FIRST_FREE, FIRST_SOAK, SEC_RAMP, SEC_FREE, SEC_SOAK,THI_RAMP, THI_FREE, THI_SOAK, FOUR_RAMP, FOUR_FREE, FOUR_SOAK, PROG_OFF };//Program status

enum ECODE { NO_ERROR, E_AD, E_HL, UNDERFLOW_E, OVERFLOW_E};//Error codes

enum FORMAT {DECIMAL, HEXA, BIN};//Formats

enum TC {T, J, K, S, PT, KTY, E, L, N, R, JPT, NI};//thermocouple types

class KD48P: public Thermometer
{
  public:
    KD48P(Modbus *mbus, int slavenumber=1): Thermometer(TEMP_PARAMETERS)
    //Constructor: modbus object, slave number
    {
	mb=mbus;
	slvno=(uint8_t) slavenumber;
    }
   //Set slave numbers
    int getSlaveNumber(){ return ((int) slvno);}//Get the actual slave number
    void setSlaveNumber(int sl){ slvno=(uint8_t) sl;}//Set the slave number
   //Get regulator related actual properties
    bool getRegVals(double &processval, double &setpoint, double &actuator);//Getting actual values of the process, setpoint and actuator signals. It returns with true, if communication error has not been done
    bool getStatus(Vec<int> &alarms, FINISHEDPRG &finishedprogno, int &executedprogno, ECODE &errorcode);//Getting the status of alarms, the finished and executed program's number, finally the error code. It returns with comm error, if it happens
   //Time-Program controll
    bool SetActualProgramNumber(int prog_no){if ((prog_no<0)||(prog_no>MAX_PROG_NO)) return false; mb->setValue(slvno, 0x102,i2u(prog_no)); int num; if(GetActualProgramNumber(&num) && (num==prog_no)) return true; else return false;};//Sets the actual program number (0..MAX_PROG_NO)
    bool GetActualProgramNumber(int *prog_no);//Returns with the actually set program number
    bool switchOn(bool feedback=true){ return this->OnOff(false,feedback);}//Start the execution of the program --- returns with success of the operation (fedback=false), or checked by the status registers (feedback=true)
    bool switchOff(bool feedback=true){ return this->OnOff(true,feedback);}//Stop the execution of the program --- returns with success of the operation (fedback=false), or checked by the status registers (feedback=true)
    bool goToNext(){return mb->setValue(slvno, 0x2fe,0x1);}//Go to the next program step --- returns with its status
    bool SetProgram(int program_no, int subprogram, int time_before, int heat_speed, int target_temperature, int keeping_time, bool final_term);//This method sets the control program for the "program_no" (0..MAX_PROG_NO), and subprogram number (0..MAX_SPRG_NO), parameters are the time before in min (0..5999), the heat speed in C/hour units (FREE-0.0, END (prog) - -0.1, -199.9~399.9), the target temperature in C (input_range), the keeping time in mins (0..5999), finally, does this the final term of the program?
    bool GetProgram(int program_no, int subprogram, int *time_before, int *heat_speed, int *target_temperature, int *keeping_time, bool *final_term);//This method get the control settings for "program_no" (0..MAX_PROG_NO), and subprogram number (0..MAX_SPRG_NO), the returning parameters are the time before in min (0..5999), the heat speed in C/hour units (FREE-0.0, END (prog) - -0.1, -199.9~399.9), the target temperature in C (input_range), the keeping time in mins (0..5999), finally, does this the final term of the program?
    bool ResetAllPrograms(){return (switchOff(true) && mb->setValue(slvno, 0x123,0x9));}//Reset all of the control program
   //Reset alarm properties
   //ResetAll
    bool ResetAll_SolidStateRelay(int inp_hi=500, int inp_lo=10, TC thermocouple=TC::K){return (ResetAlarmProperties()&&SetPID()&&SetInputConfig(&thermocouple,NULL)&&SetInputRangeConfig(inp_lo, inp_hi)&&SetRegulatorConfig()&&SetAnalogOutConfig()&&ResetAlarmConfig()&&SetTimeBaselineConfig()&&SetActualProgramNumber(0));};//Resets the parameters according to the solid state relay: inputs are the type of the thermocouple, low and high part of the range
    //bool ResetFactoryConfig(){return (switchOff(true) && mb->setValue(slvno, 0x123,0xd));}//Reset the config to the factory setting -- it modifies the communication, as well -- it works, but commented out to avoid troubles
    bool ResetAlarmProperties();//Reset the alarm properties --- this reset only, it should be implemented, if it need
   //PID setting
    bool SetPID(double gain=20.3, int integration_time=129, int derivation_time=28, int offset=0, int deadzone=0);//Set the PID parameters: P=100%/gain (0.0~100.0%), integration time(0..1000s), derivation time(0..1000s), offset reset value (+/-1000), hysteresis(0.1000) 
    bool GetPID(double *gain, int *integration_time, int *derivation_time, int *offset, int *deadzone);//Get the PID parameters: P=100%/gain (0.0~100.0%), integration time(0..1000s), derivation time(0..1000s), offset reset value (+/-1000), hysteresis(0.1000) 
   //Configuration settings and reading
    bool SetInputConfig(TC *thermocouple=NULL, bool *scalable=NULL, bool strong_filter=false, int decimal_point=0, int shift=0, int scalable_lo=0, int scalable_hi=0);//Sets the input configuration parameters: thermocouple || (scalable &&decimal point(0..3)) input, shift(+/-1000), low and high scalable values (-1999..4000)
    bool GetInputConfig(bool *tsc, TC *thermocouple, bool *scalable, bool *strong_filter, int *decimal_point, int *shift, int *scalable_lo, int *scalable_hi);//Gets the input and related parameters (inP, SHFt, PvLo, PvHi), parameters are: tcs: t - thermocouple, f- scalable, decimal point will set, the shift, scalable low and high will be shown anyway.in the case of scalable 0..20mA (t) 4..20 mA(f)
    bool SetInputRangeConfig(int low=0, int high=0);//Sets the input range (-1999..4000)
    bool GetInputRangeConfig(int &low, int &high);//Gets the input range
    bool SetRegulatorConfig(bool cool_heat=false, bool str=true, int y_period=1, int y_lag=0, int y_lo=0, int y_hi=100);//Sets the actuator: cooling(t) or heating(f), solid state relay output or regular, time between full off and on state of the output device (0..255), lag in y of the device (0..100), y_lo>y -> y=0 and y_hi<y -> y=100 (0..100 for hi and low)
    bool GetRegulatorConfig(bool *cool_heat, bool *str, int *y_period, int *y_lag, int *y_lo, int *y_hi);//Gets the actuator setup: cooling(t) or heating(f), solid state relay output or regular, time between full off and on state of the output device (0..255), lag in y of the device (0..100), y_lo>y -> y=0 and y_hi<y -> y=100
    bool SetAnalogOutConfig(bool analog_out=true, int signal_pv_y_sp=2, int analog_out_lo=0, int analog_out_hi=100);//analog output (0..t 4..f mA), the output signal(ideal signal-1, actuator-2, input signal-3), the corresponding minimum and maximum values (-1999~3999)of the output (true: the number corresponds to the 0 and the 20mA values)
    bool GetAnalogOutConfig(bool *analog_out, int *signal_pv_y_sp, int *analog_out_lo, int *analog_out_hi);//analog output (0..t 4..f mA), the output signal(ideal signal, actuator, input signal), the corresponding minimum and maximum values of the output (true: the number corresponds to the 0 and the 20mA values)
    bool ResetAlarmConfig();//Set the alarm relay -- here, I set only the 0 setting for all of them --- it should be implemented if it needed
    //Get version does not exist due to the lack of need of this setup --- not implemented
    //For safety reason, the communication nor the setting, nor the reading are implemented: to read/set, please use manual setting
    bool SetTimeBaselineConfig() {return mb->setValue(slvno, 0x12d,0x85);}//Set the time-baseline program (14 piece of 4 step progs, unit C/hours, heat& cooling are accepted)
    //Get version: does not necessary to implement
    //Get operation hours: it seems to me, that this configuration does not exist
    bool About(std::string *prog_ver);//Returns with type and version separated by semicolon -- only KD48P is implemented
    //Bulky reading
    bool readConfig(std::ostream &os);//Reads the configuration part of the regulator and writes it into the output stream
    //static constants
    static const int END=-1;//Constant should be set to the next ramp, if it finishes
    static const int FREE=0;//Free heating/cooling set for ramp
    static const int INFTY=0;//Keeping soak the requested temperature for infinite times
  private:
    //Constant
    static const int TEMP_PARAMETERS=3;//Number of temperature parameters to provide information on processes
    //Private members
    uint8_t slvno;//Slave number
    Modbus* mb;//Modbus communication object
    //Constatnts
    static const int MAX_PROG_NO=1;
    static const int MAX_SPRG_NO=3;
    //Private methods
    bool OnOff(bool stop=true, bool feedback=true);//This method switch off (stop=true) or switch on (stop=false) the temperature controller.If feedback is true, it returns with success, if the status register changed properly. If false, it is waiting for only the accept. --- WORKS
    bool getwrite(std::ostream &os, std::string label, uint16_t regtoread, FORMAT frm=DECIMAL);//This method read the content of the register and writes into stream 
    uint16_t i2u(int input){return ((input<0) ? ((0xffff-(uint16_t)abs(input))+0x1) : (uint16_t) input);}//Transforming int to 16 bit unsigned
    int u2i(uint16_t input){return ((input>0x7fff) ? -((int)(0xffff-input)+1) : (int) input);}//Transforming 16 bit unsigned int to int
    //Inherited method
    void getTempHW(double *results){double res[TEMP_PARAMETERS]; if(getRegVals(res[0], res[1], res[2])) for(int i=0; i<TEMP_PARAMETERS; i++) results[i]=res[i]; else for(int i=0; i<TEMP_PARAMETERS; i++) results[i]=-5000.0; return;};//Returns with the temperatures: -5000: error, should be implemented in child class
};
#endif
